if (!Array.prototype.unique)
    Array.prototype.unique = function () {
        const seen = {};
        return this.filter(item => seen.hasOwnProperty(item) ? false : (seen[item] = true));
    };

Object.equals = function(x, y, exclusions) {
    // if both x and y are null or undefined and exactly the same
    if (x === y)
        return true;

    // if they are not strictly equal, they both need to be Objects
    if (!(x instanceof Object) || !(y instanceof Object))
        return false;

    // they must have the exact same prototype chain, the closest we can do is
    // test their constructor.
    if (x.constructor !== y.constructor)
        return false;

    for (var p in x) {
        if (exclusions.includes(p)) continue;

        // other properties were tested using x.constructor === y.constructor
        if (!x.hasOwnProperty(p)) continue;

        // allows to compare x[ p ] and y[ p ] when set to undefined
        if (!y.hasOwnProperty(p))
            return false;

        // if they have the same strict value or identity then they are equal
        if (x[p] === y[p]) continue;

        // Numbers, Strings, Functions, Booleans must be strictly equal
        if (typeof(x[p]) !== "object")
            return false;

        // Objects and Arrays must be tested recursively
        if (!Object.equals(x[p], y[p], exclusions))
            return false;
    }

    for (p in y) {
        if (exclusions.includes(p)) continue;

        // allows x[p] to be set to undefined
        if (y.hasOwnProperty(p) && y[p] !== undefined && !x.hasOwnProperty(p))
            return false;
    }
    return true;
};

Object.differences = function(x, y, exclusions, name, results, indent) {
    results = results || [];
    indent  = indent  || "";

    // if both x and y are null or undefined and exactly the same
    if (x === y)
        return results;

    // if they are not strictly equal, they both need to be Objects
    if (!(x instanceof Object) || !(y instanceof Object)) {
        results.push(indent + name + " was changed from '" + JSON.stringify(x) + "' to '" + JSON.stringify(y) + "'");
        return results;
    }
    // they must have the exact same prototype chain, the closest we can do is
    // test their constructor.
    if (x.constructor !== y.constructor) {
        results.push(indent + name + " seems to have changed to a different type");
        return results;
    }

    for (var p in x) {
        if (exclusions.includes(p)) continue;

        // other properties were tested using x.constructor === y.constructor
        if (!x.hasOwnProperty(p)) continue;

        // allows to compare x[ p ] and y[ p ] when set to undefined
        if (!y.hasOwnProperty(p)) {
            results.push(indent + name + " lost property '" + p + "' with value '" + JSON.stringify(x[p]) + "'");
            continue;
        }

        // if they have the same strict value or identity then they are equal
        if (x[p] === y[p]) continue;

        // Numbers, Strings, Functions, Booleans must be strictly equal
        if (typeof(x[p]) !== "object") {
            results.push(indent + name + " has property '" + p + "' that was changed from '" + JSON.stringify(x[p]) + "' to '" + JSON.stringify(y[p]) + "'");
            continue;
        }

        // Objects and Arrays must be tested recursively
        Object.differences(x[p], y[p], exclusions, name + " > " + p, results, indent + "    ");
    }

    for (p in y) {
        if (exclusions.includes(p)) continue;

        // allows x[p] to be set to undefined
        if (y.hasOwnProperty(p) && y[p] !== undefined && !x.hasOwnProperty(p)) {
            results.push(indent + name + " has a new property '" + p + "' with value '" + JSON.stringify(y[p]) + "'");
        }
    }
    return results;
};

String.prototype.trim = function (char) {
    if (!char) return this.replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g, '');
    return this.replace(RegExp("^[\\" + char + "]+|[\\" + char + "]$", 'g'), '');
};

const friendly = function (name) {
    return name.toLowerCase().replace(/_/g, " ");
};

JobEditor = function (options) {
    var self = this;

    var fullList  = [];
    var controls  = {};

    this.config   = {};
    this.options  = options;
    this.events   = {};

    var emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const event = self.events[eventName] || [];
        event.forEach(fn => { fn.call(null, data); });
    };

    var subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);

        // Return an 'unsubscribe' function for this subscription.
        return () => {
            self.events[eventName] = self.events[eventName].filter(eventFn => fn !== eventFn);
        };
    };

    var onConfigChange = function (userDidSomething) {
        // TODO: Show how many filter or how many entities.
        if (controls.jobModeControl.checked) {
            controls.entityCounter.innerHTML = controls.selectedFilters.items.length + " group(s) of criteria";
            controls.saveButton.classList.remove("Disabled");
        }
        else {
            controls.entityCounter.innerHTML = controls.selectedList.items.length + " " + friendly(controls.jobTypeControl.value) + "s";
            controls.saveButton.classList.remove("Disabled");
        }
    }

    var hasConfigChanged = function (msg) {
        let currentJobData = unrenderJob();

        const hasChanged = originalJobData && !Object.equals(originalJobData, currentJobData, ["key", "id", "timestamp", "userID"]);
        if (hasChanged) {
            const changes = Object.differences(originalJobData, currentJobData, ["key", "id", "timestamp", "userID"], "Job");
            if (msg) changes.push(msg);
            window.alert("The following changes were detected in " + currentJobData.name + ":\n" + changes.join("\n"));
        }
        return hasChanged;
    }

    let comparators = [
        { icon: '&#x2208',    value: '=',       name: "Equals or in",     minValues: "1" },
        { icon: '&#x2209;',   value: '<>',      name: "Not equals or in", minValues: "1" },
        { icon: '=&#x2205',   value: 'IS NULL', name: "Is NULL",          minValues: "0", maxValues: "0" },
        { icon: '<span style="position:relative; top:1px;font-size:90%;padding-right:1px">&#8800;</span>&#x2205;',
                              value: 'IS NOT NULL', name: "Is not NULL",  minValues: "0", maxValues: "0" },
        { icon: '&#x227B;',   value: '>',       name: "Larger than",      minValues: "1", maxValues: "1" },
        { icon: '&#x227A;',   value: '<',       name: "Smaller than",     minValues: "1", maxValues: "1" },
        { icon: '[&#x2217;]', value: 'BETWEEN', name: "Between",          minValues: "2", maxValues: "2" },
        { icon: '&#x39e;',    value: 'LIKE',     name: "Like",            minValues: "1" },
        { icon: '&#x3a0;',    value: 'NOT LIKE', name: "Not Like",        minValues: "1" }
    ];

    let itemHash   = {}
    let knownItems = [];
    var addItem = function (item, silent) {
        if (itemHash.hasOwnProperty(item.entityId)) {
            if (!silent)
                console.error("Duplicate item in list: " + item.displayName + " (" + item.entityId + ")");
        }
        else {
            itemHash[item.entityId] = item;
            knownItems.push(item);
            return true;
        }
        return false;
    }
    var forgetItems = function () {
        itemHash   = {}
        knownItems = [];
    }

    var toTechnologyString = function (technologies) {
        technologies = technologies || [];

        return technologies
            .map(technology => {
                return technology.type +
                      (technology.version ? (" "   + technology.version) : "") +
                      (technology.edition ? (" - " + technology.edition) : "");
                })
            .join(", ");
    }

    var originalJobData = null;

    var isReadyToSwitch = function () {
        if (!originalJobData) return true;

        if (!controls.targetControl.value) {
            return true;    // We can always move away from a job that will not be un-rendered anyways.
        }

        let jobName = controls.jobNameControl.value.trim();
        if (!jobName)
            return window.alert("Please specify a name for the job.") || false;

        // We check if the config has changed unless we don't have job data right now at all.
        if (hasConfigChanged() &&
            !window.confirm("Do you want to discard the changes you made to this job?"))
            return false;

        return true;
    }

    var renderJob = function (jobData) {
        if (!self.config.Tenants[jobData.source]) {
            return;
        }
        entitiesSampleSplit.collapse(1);    // Collapse test area.

        controls.mgtZoneControl.removeAttribute("prevValue");

        if (!jobData) {
            originalJobData = null;
            return;
        }

        // This here is our new 'original' job data (ie. before editing).
        originalJobData = jobData;

        jobData.template    = jobData.template    || controls.templateControl.querySelectorAll("option")[0].value;
        jobData.destination = jobData.destination || "None";
        jobData.type        = jobData.type        || "HOST";
        jobData.subtype     = jobData.subType     || "Problems",
        jobData.mz          = jobData.mz          || ".";
        jobData.optimize    = jobData.optimize    || "true";

        controls.jobNameControl.value  = jobData.name;
        controls.targetControl.value   = jobData.destination;
        controls.templateControl.value = jobData.template;
        controls.outputControl.value   = jobData.subtype;

        // Populate the other controls with values that pertain to this source.
        onSourceChange(jobData.source);

        controls.jobTypeControl.value     = jobData.type;     // This should trigger an 'onTypeChange' event.
        controls.mgtZoneControl.value     = jobData.mz;
        controls.granularityControl.value = jobData.granularity;
        controls.jobModeControl.checked   = jobData.mode !== "explicit";
        controls.relatedControl.value     = jobData.related || "none";

        // We set 'prevValue' because we may display a warning when the user changes it, and then we can undo.
        controls.mgtZoneControl.setAttribute("prevValue", controls.mgtZoneControl.value);
        controls.jobModeControl.setAttribute("prevValue", controls.jobModeControl.checked);

        if (jobData.query) {
            jobData.groups = jobData.groups || [];
            jobData.sorts  = jobData.sorts  || [];

            controls.whereControl.options   = self.usqlCriteria[jobData.type];
            controls.whereControl.value     = jobData.where;

            controls.fieldControl.options   = self.usqlFields[jobData.type];
            controls.fieldControl.value     = jobData.fields;

            controls.groupByControl.options = self.usqlGroups[jobData.type];
            controls.groupByControl.value   = jobData.groups;

            controls.orderByControl.options = self.usqlSorts[jobData.type];
            controls.orderByControl.value   = jobData.sorts;

            controls.metricControl.options  = self.usqlMetrics[jobData.type];
            controls.metricControl.value    = jobData.metrics;

            controls.availableList.items   = [];
            controls.selectedList.items    = [];
            controls.selectedFilters.value = [];
            forgetItems();
        }
        else {
            controls.metricControl.options  = self.entityMetrics[jobData.type];
            controls.metricControl.value    = jobData.metrics.map(metric => {
                let parts = metric.split("#");

                return {
                    type:   parts[0].split(":")[0],
                    value:  parts[0],
                    values: parts[1] ? parts[1].split(",") : []
                }
            });
            // If we have entities in the job, it's an 'explicit' job.
            controls.selectedList.items = jobData.entities;
            controls.selectedList.tooltipTemplate  = tooltipTemplates[controls.jobTypeControl.value];

            forgetItems();
            controls.selectedList.items.forEach(addItem);

            // If we have criteria in the job, it's an 'implicit' job.
            controls.filterControl.options   = filterOptions[jobData.source] ? filterOptions[jobData.source][jobData.type] || [] : [];
            controls.selectedFilters.options = filterOptions[jobData.source] ? filterOptions[jobData.source][jobData.type] || [] : [];

            // This is where we need a bit of translation to make job definition compatible with filter controls.
            controls.selectedFilters.items = jobData.criteria ? jobData.criteria.map(criteriaGroup => {
                const filterGroup = [];
                if (!Array.isArray(criteriaGroup))
                    criteriaGroup = [criteriaGroup];

                criteriaGroup.forEach(filter => {
                    if (!Array.isArray(filter.values))
                        filter.values = [filter.values];

                    if (!filter.context)    // Was already translated?
                        filterGroup.push(filter);
                    else
                        filterGroup.push({  // Translate.
                            type:   filter.context,
                            name:   filter.key,
                            value:  filter.key,
                            values: filter.values.map(value =>
                                value == "."
                                ? "<i>(exists)</i>"
                                : value == "*"
                                ? "<i>(has a value)</i>"
                                : value
                            ),
                            comparator: filter.operator.charAt(0) == '!' ? filter.operator.substring(1) : filter.operator,
                            negated:    filter.operator.charAt(0) == '!'
                        });
                });
                return filterGroup;
            }) : [];
        }
        // Set visibility of filter or entity list.
        onModeChange(null, !jobData.entities || jobData.entities.length === 0);

        currJobId = jobData.id;
        onTypeChange(null, true);

        if (!controls.relatedControl2) {
            controls.relatedControl2 = document.getElementById("includeWhat2");
            if (controls.relatedControl2) {
                controls.relatedControl2.value = controls.relatedControl.value;
                controls.relatedControl.addEventListener("change", function () {
                    controls.relatedControl2.value = this.value;
                });
            }
        }
        else {
            controls.relatedControl2.value = controls.relatedControl.value;
        }
        hasConfigChanged("It is possible that certain options are not available anymore.");
    }

    var unrenderJob = function () {
        function quoted(value) {
            return Number(value) == value ||
                   value === "FALSE" ||
                   value === "TRUE"
                 ? value
                 : ("'" + value + "'");
        }

        function toSELECT(fields, metrics) {
            let clause = [];

            fields.forEach(field => {
                if (!field.values || field.values.length == 0)
                    clause.push(
                        (field.type == "UNIQUE" ? "DISTINCT " : "")
                      +  field.value
                      + (field.alias ? " AS " + field.alias : "")
                   );
                else
                    field.values.forEach(value => {
                        var valueParts = value.split("=");
                        var valueAspects = valueParts[0].split("|");
                        if (valueAspects[0] == "DATETIME" || valueAspects[0] == "TOP")
                            clause.push((field.type == "UNIQUE" ? "DISTINCT " : "")
                                     +  valueAspects[0] + "(" + field.value
                                     + ", " + valueAspects.slice(1)
                                                          .map(aspect => quoted(aspect))
                                                          .join(", ")
                                     + ")"
                                     + (valueParts[1] ? " AS " + valueParts[1] : ""));
                        else
                            clause.push((field.type == "UNIQUE" ? "DISTINCT " : "")
                                     +  valueAspects[0] + "(" + field.value + ")"
                                     + (valueParts[1] ? " AS " + valueParts[1] : ""));
                    });
            });

            metrics.forEach(metric => {
                if (!metric.values || metric.values.length == 0)
                    clause.push(metric.value + (metric.alias ? " AS " + metric.alias : ""));
                else
                    metric.values.forEach(value => {
                        var valueParts = value.split("=");
                        if (valueParts[0] == "COUNT DISTINCT")
                            clause.push("COUNT(DISTINCT "   + metric.value + ")" + (valueParts[1] ? " AS " + valueParts[1] : ""));
                        else
                            clause.push(valueParts[0] + "(" + metric.value + ")" + (valueParts[1] ? " AS " + valueParts[1] : ""));
                    });
            });
            return clause.join(", ");
        }

        function toWHERE(conditions) {
            if (conditions.length === 0) return "";

            let clause = [];

            conditions.forEach(condition => {
                switch (condition.comparator) {
                    case "IS NULL":
                        clause.push(condition.value + " IS NULL");
                        break;

                    case "IS NOT NULL":
                        clause.push(condition.value + " IS NOT NULL");
                        break;

                    case "<>":
                        condition.values.forEach(value => {
                            clause.push(condition.value + " <> " + quoted(value));  // Multiple 'AND' conditions
                        });
                        break;

                    case ">":
                        clause.push(condition.value + " > " + quoted(condition.values[0]));
                        break;

                    case "<":
                        clause.push(condition.value + " < " + quoted(condition.values[0]));
                        break;

                    case "BETWEEN":
                        clause.push(condition.value + " BETWEEN " + quoted(condition.values[0]) + " AND " + quoted(condition.values[1]));
                        break;
                        
                    case "LIKE":
                            clause.push(condition.value + " LIKE " + quoted(condition.values[0]));
                            break;
                    case "NOT LIKE":
                        clause.push(condition.value + " NOT LIKE " + quoted(condition.values[0]));
                        break;
                    case "=":
                    default:
                        if (condition.values.length == 0) {
                            clause.push(condition.value + " = ''");
                        }
                        else if (condition.values.length == 1) {
                            clause.push(condition.value + " = " + quoted(condition.values[0]));
                        }
                        else {
                            clause.push(condition.value + " IN(" + condition.values.map(value => quoted(value)).join(", ") + ")");
                        }
                        break;
                }
            });

            return " WHERE " + clause.join(" AND ");
        }

        function toGROUPBY(fields) {
            if (fields.length === 0) return "";

            let clause = fields.map(field => field.value);

            return " GROUP BY " + clause.join(", ");
        }

        function toORDERBY(fields) {
            if (fields.length === 0) return "";

            let clause = fields.map(field => field.value + (field.values[0] ? " " + field.values[0] : ""));

            return " ORDER BY " + clause.join(", ");
        }

        function toLIMIT(limit) {
            if (!limit) return "";

            return " LIMIT " + limit;
        }

        var jobData = {};

        jobData.name = controls.jobNameControl.value = controls.jobNameControl.value.trim();
        if (!jobData.name)
            return window.alert("Please specify a name for the job.");

        jobData.source      = originalJobData.source;
        jobData.timestamp   = (new Date()).getTime(),
        jobData.template    = controls.templateControl.value;
        jobData.destination = controls.targetControl.value;
        jobData.type        = controls.jobTypeControl.value;
        jobData.subtype     = controls.outputControl.value;
        jobData.granularity = controls.granularityControl.value;
        jobData.optimize = originalJobData.optimize;

        if (jobData.type == "PROBLEM") {
            // TODO - Only filters are allowed, and current problems may or may not show up
            // in the UI.
            jobData.problems = "problems";
            jobData.mode = "implicit";
            jobData.mz = "none";
            jobData.mzName = "";
            jobData.metrics = [];
            jobData.criteria = controls.selectedFilters.items;

            jobData.criteria = jobData.criteria.map(filterGroup => {
                const criteriaGroup = [];
                filterGroup.forEach(filter => {
                    criteriaGroup.push({
                        context: filter.type,
                        key: filter.value,
                        operator: filter.negated ? ("!" + filter.comparator) : filter.comparator,
                        values: filter.values.map(value =>
                            value == "<i>(exists)</i>"      ? "." :
                            value == "<i>(has a value)</i>" ? "*" :
                            value
                        )
                    });
                });
                return criteriaGroup;
            });
        }
        else if (jobData.type == "usersession" || jobData.type == "useraction") {
            jobData.mode    = "implicit";
            jobData.mz      = "none";
            jobData.mzName  = "";
            jobData.limit   = controls.limitControl.value = Number(controls.limitControl.value) || "";
            jobData.where   = controls.whereControl.value;
            jobData.fields  = controls.fieldControl.value;
            jobData.metrics = controls.metricControl.value;
            jobData.groups  = controls.groupByControl.value;
            jobData.sorts   = controls.orderByControl.value;
            jobData.query   = "SELECT " + toSELECT(jobData.fields, jobData.metrics)
                            + " FROM "  + jobData.type
                            + toWHERE(jobData.where)
                            + toGROUPBY(jobData.groups)
                            + toORDERBY(jobData.sorts)
                            + toLIMIT(jobData.limit);
        }
        else {
            jobData.mode    = controls.jobModeControl.checked ? "implicit" : "explicit";
            jobData.related = controls.relatedControl.value;
            jobData.mz      = controls.mgtZoneControl.value;
            jobData.mzName  = self.zoneKeys[jobData.mz];
            jobData.metrics = controls.metricControl.value.map(metric => {
                return metric.value + "#" + metric.values.join(",");
            });

            if (controls.jobModeControl.checked) {
                jobData.entities = null;
                jobData.criteria = controls.selectedFilters.items;

                // Make filter definitions compatible with Library query format.
                jobData.criteria = jobData.criteria.map(filterGroup => {
                    const criteriaGroup = [];
                    filterGroup.forEach(filter => {
                        criteriaGroup.push({
                            context:  filter.type,
                            key:      filter.value,
                            operator: filter.negated ? ("!" + filter.comparator) : filter.comparator,
                            values:   filter.values.map(value =>
                                value == "<i>(exists)</i>"
                                ? "."
                                : value == "<i>(has a value)</i>"
                                ? "*"
                                : value
                            )
                        });
                    });
                    return criteriaGroup;
                });
            }
            else {
                jobData.entities = controls.selectedList.items.map(entity => {
                    return {
                        entityId:    entity.entityId,
                        displayName: entity.displayName
                        // Yeah - now we don't have properties and so on to show in the UI.
                        // It seems that the best way to solve that is to retrievwe the full
                        // definitions of thse entities when we're showing them in the UI.
                    };
                });
                jobData.criteria = null;
            }
        }

        return jobData;
    }

    var onSourceChange = function (source) {
        if (!self.config.Tenants[source]) {
            return;
        }
        var definitions = self.config.Tenants[source].definitions;

        self.zoneKeys      = {};
        self.metricKeys    = {};
        self.entityMetrics = {};
        self.entityTypes   = {
            usersession: {
                value: "usersession",
                label: "User Sessions"
            },
            useraction: {
                value: "useraction",
                label: "User Actions"
            },
            problem: {
                value: "PROBLEM",
                label: "Problems"
            }
        };
        for (let type in apiEndPoints) {
            if (apiEndPoints[type])                                   // If we have an endpoint for this type
                self.entityTypes[type] = self.entityTypes[type] || {  // then add it as an available type.
                    value: type,
                    label: toUserReadableName(type)
                };
        }

        self.usqlMetrics   = {};
        self.usqlFields    = {};
        self.usqlCriteria  = {};
        self.usqlGroups    = {};
        self.usqlSorts     = {};

        function createMetrics(props, scope) {
        /*  Some things are integers, and thus automatically quality as metrics.
                They get ["SUM", "MAX", "MIN", "AVG", "MEDIAN", "COUNT DISTINCT"] options. Default is "AVG".
            Some things are strings: they get added as metrics with the option ["COUNT", "COUNT DISTINCT"].
                Default is "COUNT".
            Other things don't qualify as metrics.
         */
            return Object
                .keys(props)
                .filter(prop => props[prop].type == 'integer' ||
                                props[prop].type == 'number'  ||
                                props[prop].type == 'string')
                .map(prop => {
                    let entry = {
                        value:  scope ? (scope + "." + prop) : prop,
                        name:   scope ? ("<i>" + scope + "</i>." + prop) : prop,
                        descr:  props[prop].description
                    };
                    if (props[prop].type == "integer") {
                        entry.values  = ["SUM", "MAX", "MIN", "AVG", "MEDIAN", "COUNT DISTINCT"];
                        entry.default = "AVG";
                    }
                    if (props[prop].type == "string") {
                        entry.values  = ["COUNT", "COUNT DISTINCT"];
                        entry.default = "COUNT";
                    }

                    self.metricKeys[entry.value] = entry.name;
                    return entry;
                });
        }

        function createFields(props, scope) {
            /*  Some things are integers: they are not fields.
                All other things qualify as fields.
                We account for "DISTINCT" by simply adding flavors: ALL field and DISTINCT field.
            */
            return Object
                .keys(props)
                .filter(prop => props[prop].format == "int64" ||    // That's an epoch datetime.
                                props[prop].type != 'integer' &&
                                props[prop].type != 'number')
                .reduce((arr, prop) => {
                    let values = props[prop].format == "int64"
                               ? ["YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "DATETIME|Format|Granularity"]
                               : ["TOP|Number"];

                    arr.push({
                        value:   scope ? (scope + "." + prop) : prop,
                        name:    scope ? ("<i>" + scope + "</i>." + prop) : prop,
                        type:    "ALL",
                        descr:   props[prop].description,
                        values:  values,
                        default: null
                    });

                    if (props[prop].type != 'boolean') {
                        arr.push({
                            value:   scope ? (scope + "." + prop) : prop,
                            name:    scope ? ("<i>" + scope + "</i>." + prop) : prop,
                            type:    "UNIQUE",
                            descr:   props[prop].description,
                            values:  values,
                            default: null
                        });
                    }
                    // For dates: User has options to extract something from a date
                    return arr;
                },[]);
        }

        function createCriteria(props, scope) {
            return Object
                .keys(props)
                .filter(prop => props[prop].type != 'object' &&
                                props[prop].type != 'array')
                .map(prop => {
                    return {
                        value:  scope ? (scope + "." + prop) : prop,
                        name:   scope ? ("<i>" + scope + "</i>." + prop) : prop,
                        descr:  props[prop].description,
                        values: props[prop].enum
                              ? props[prop].enum
                              : props[prop].type == "boolean"
                              ? ["TRUE", "FALSE"]
                              : null
                    };
                });
        }

        function createGroups(props, scope) {
            return Object
                .keys(props)
                .filter(prop => props[prop].type != 'object' &&
                                props[prop].type != 'array')
                .map(prop => {
                    return {
                        value:  scope ? (scope + "." + prop) : prop,
                        name:   scope ? ("<i>" + scope + "</i>." + prop) : prop,
                        descr:  props[prop].description
                    };
                });
        }

        function createSorts(props, scope) {
            return Object
                .keys(props)
                .filter(prop => props[prop].type != 'object' &&
                                props[prop].type != 'array')
                .map(prop => {
                    return {
                        value:  scope ? (scope + "." + prop) : prop,
                        name:   scope ? ("<i>" + scope + "</i>." + prop) : prop,
                        descr:  props[prop].description,
                        values: ["ASC", "DESC"]
                    };
                });
        }

        /*
            FOR THE PURPOSES OF METRICS AND FIELDS TO BE SELECTED (THUS NOT AS FILTERS)

            We handle 'AS' by making the column headers of the sample result editable.

            myfield: TOP(n)  -> then myfield should really be grouped by. Must, really.
            For TOP(n) we can have values, 1, 5, 10, 20, 50, 100

            starttime: DATETIME(‘5m’). Format is 'yyyy-MM-dd HH:mm'.
            We can have a predefined list of options for scenarios like this:
            DATETIME(1m), DATETIME(5m), DATETIME(10m), DATETIME(15m), DATETIME(30m), same for 's', h' and 'd'
            We can let this be driven by the polling interval setting: Values go up to the selected granularity.
         */
        if (!definitions.usql) {
            window.alert("Warning: " + source + " does not contain USQL definitions.");
        }
        else {
            const userSessionProps = definitions.usql.UserSession.properties;
            const userActionProps  = definitions.usql.UserSessionUserAction.properties;

            self.usqlCriteria.usersession = createCriteria(userSessionProps).concat(createCriteria(userActionProps, "useraction"));
            self.usqlMetrics.usersession  = createMetrics(userSessionProps).concat(createMetrics(userActionProps, "useraction"));
            self.usqlFields.usersession   = createFields(userSessionProps).concat(createFields(userActionProps, "useraction"));
            self.usqlGroups.usersession   = createGroups(userSessionProps).concat(createGroups(userActionProps, "useraction"));
            self.usqlSorts.usersession    = createSorts(userSessionProps).concat(createSorts(userActionProps, "useraction"));

            self.usqlCriteria.useraction  = createCriteria(userActionProps).concat(createCriteria(userSessionProps, "usersession"));
            self.usqlMetrics.useraction   = createMetrics(userActionProps).concat(createMetrics(userSessionProps, "usersession"));
            self.usqlFields.useraction    = createFields(userActionProps).concat(createFields(userSessionProps, "usersession"));
            self.usqlGroups.useraction   = createGroups(userActionProps).concat(createGroups(userSessionProps, "usersession"));
            self.usqlSorts.useraction    = createSorts(userActionProps).concat(createSorts(userSessionProps, "usersession"));
        }

        // Available entity types into dropdown.
        populateSelector(controls.jobTypeControl, Object.values(self.entityTypes));

        // Finally, the MZ control:
        var zones = definitions.zones ? definitions.zones.values || [] : [];
        if (!Array.isArray(zones))
            zones = [];

        zones.forEach(zone => {
            zone.val   = zone.id;
            zone.label = zone.name;

            self.zoneKeys[zone.id] = zone.name;  // Harmonize with metric keys naming convention
        })
        zones.sort((a, b) => {
            return a.name < b.name ? -1 : a.name > b.name ? 1 : 0;
        });

        [   { id: "empty", name: "Everything not in any Management Zone" },
            { id: "*",     name: "Everything in a Management Zone" },
            { id: ".",     name: "Everything" }
        ].forEach(zone => {
            zone.val   = zone.id;
            zone.label = zone.name;

            self.zoneKeys[zone.id] = zone.name;  // Harmonize with metric keys naming convention
            zones.unshift(zone);
        });

        // Available zones into dropdown.
        populateSelector(controls.mgtZoneControl, zones, ".", true);

        // Now we need to clear EVERYTHING ELSE in the UI.
        // We can probably leave the job type in place. The MZ is unselected.
        // Let's see how it plays out.
        // We should have a default type selected. We need to populate that with something.
        controls.jobTypeControl.value = controls.jobTypeControl.value || "HOST";
        onTypeChange();
    }

    var onTypeChange = function (evt, isNewlySelectedJob) {
        var source = originalJobData.source;
        var definitions = self.config.Tenants[source].definitions;

        var currentTarget = evt ? evt.currentTarget : controls.jobTypeControl;
        var prevValue = currentTarget.getAttribute("prevValue");
        var currValue = currentTarget.value;

        // TODO: Depending on the mode, the test and the warning is different!
        var accept  = controls.jobModeControl.checked
                    ? (controls.selectedList.items.length > 0 && evt
                        ? window.confirm("This will remove all " + friendly(controls.jobTypeControl.value) + "s from this job.")
                        : true)
                    : (controls.selectedFilters.items.length > 0 && evt
                        ? window.confirm("This will remove all filters from this job.")
                        : true);
        if (accept) {
            currentTarget.setAttribute("prevValue", currValue);

            // Hide/show the proper controls and initialize.
            if (currValue == "PROBLEM") {
                document.body.classList.remove("Entities");
                document.body.classList.remove("sessions");
                document.body.classList.add("Problems");

                // TODO: For problems, we probably have no metrics at all.
                controls.metricControl.options   = self.entityMetrics[currValue];
                controls.filterControl.options   = filterOptions[source][currValue];
                controls.selectedFilters.options = filterOptions[source][currValue];
            }
            else if (currValue == "usersession" || currValue == "useraction") {
                document.body.classList.remove("Entities");
                document.body.classList.remove("Problems");
                document.body.classList.add("Sessions");

                // Collapse the entity selector;
                options.splitter.collapse(0);

                controls.whereControl.options   = self.usqlCriteria[currValue];
                controls.metricControl.options  = self.usqlMetrics[currValue];
                controls.fieldControl.options   = self.usqlFields[currValue];
                controls.groupByControl.options = self.usqlGroups[currValue];
                controls.orderByControl.options = self.usqlSorts[currValue];

                setInterval(function () {
                    var jobData = jobEditor.unrender();
                    controls.sqlPreviewContainer.innerHTML = jobData.query
                                                           ? "<div class='PickerToggle'><span>USQL Statement:</span></div><code>" + jobData.query + "</code>"
                                                           : "";
                }, 1000);
            }
            else {
                document.body.classList.remove("Sessions");
                document.body.classList.remove("Problems");
                document.body.classList.add("Entities");

                controls.metricControl.options   = self.entityMetrics[currValue];
                controls.filterControl.options   = filterOptions[source][currValue];
                controls.selectedFilters.options = filterOptions[source][currValue];

                // We need to get new data.
                getData();
            }

            // Populate metric control values:
            var metrics = definitions.metrics && Array.isArray(definitions.metrics)
                        ? definitions.metrics
                        : [];

            if (metrics.length == 0) {
                window.alert("Warning: " + source + " does not contain metric definitions.");
            }
            else {
                metrics.forEach(metric => {
                    metric.value  = metric.timeseriesId;
                    metric.name   = metric.displayName;
                    metric.entity = metric.dimensions[0];
                    metric.type   = metric.timeseriesId.split(":")[0];
                    metric.values = metric.aggregationTypes;

                    if (!metric.values || metric.values.length === 0)
                        metric.values = ["NUM"];

                    var index = metric.values.indexOf("PERCENTILE");
                    if (index !== -1)
                        metric.values[index] = "PERCENTILE|Number";

                    self.metricKeys[metric.value]   = metric.name;
                    self.entityMetrics[metric.entity] = self.entityMetrics[metric.entity] || [];

                    // If we can retrieve this type of entity, then prepare for it.
                    if (apiEndPoints[metric.entity]) {
                        // Check if metric is already in the list before adding it (by timeseries ID!).
                        const exists = self.entityMetrics[metric.entity].some(m => m.value === metric.value);
                        if (!exists) self.entityMetrics[metric.entity].push(metric);

                        // Put it also in the accounting of entity types we can retrieve metrics for.
                        self.entityTypes[metric.entity] = self.entityTypes[metric.entity] || {
                            value: metric.entity,
                            label: toUserReadableName(metric.entity)
                        };
                    }
                });
            }

            if (!isNewlySelectedJob) {
                controls.selectedList.items    = [];
                controls.selectedFilters.items = [];
                forgetItems();
            }
            controls.filterMatches.innerHTML = "";
            controls.availableList.items = [];

            onConfigChange(!!evt);
        }
        else {
            currentTarget.value = prevValue;
        }
    }

    var onModeChange = function (evt, override) {
        if (override === true) {
            controls.jobModeControl.checked = true;
        }
        else if (override === false) {
            controls.jobModeControl.checked = false;
        }
        else {
            // It happened interactively... Inform user of consequences.
            var prevValue = controls.jobModeControl.getAttribute("prevValue");

            var accept  = controls.jobModeControl.checked
                        ? (controls.selectedList.items.length > 0 && evt
                            ? window.confirm("This will remove all " + friendly(controls.jobTypeControl.value) + "s from this job.")
                            : true)
                        : (controls.selectedFilters.items.length > 0 && evt
                            ? window.confirm("This will remove all filters from this job.")
                            : true);
            if (!accept) {
                controls.jobModeControl.checked = prevValue != "false";
                return;
            }
        }
        onConfigChange();

        controls.jobModeControl.setAttribute("prevValue", controls.jobModeControl.checked);

        var selectorBox = document.getElementById("availableCount").parentElement;

        if (controls.jobModeControl.checked) {
            options.selectedList.style.display = "none";
            options.selectedList.previousElementSibling.style.display = "none";
            options.selectedFilters.style.display = "";
            selectorBox.style.display = "none";
        }
        else {
            options.selectedList.style.display = "";
            options.selectedList.previousElementSibling.style.display = "";
            options.selectedFilters.style.display = "none";
            selectorBox.style.display = "";
        }
        getData();
    }

    var onZoneChange = function (evt) {
        var prevValue = evt.currentTarget.getAttribute("prevValue");
        var currValue = evt.currentTarget.value;

        if (prevValue && currValue == "none") {
            window.alert("Please select a Management Zone");
            evt.currentTarget.value = prevValue;
            return;
        }

        var accept  = controls.jobModeControl.checked
                    ? (controls.selectedList.items.length > 0 && prevValue
                        ? window.confirm("This will remove all " + friendly(controls.jobTypeControl.value) + "s from this job.")
                        : true)
                    : (controls.selectedFilters.items.length > 0 && prevValue
                        ? window.confirm("This will remove all filters from this job.")
                        : true);
        if (accept) {
            evt.currentTarget.setAttribute("prevValue", currValue);

            controls.selectedList.items = [];

            getData();  // Get a new list of entities (if applicable).

            onConfigChange(!!prevValue);   // Take note of change in UI.
        }
        else {
            evt.currentTarget.value = prevValue;
        }
    }

    var addItems = function (items) {
        items.forEach(item => addItem(item, true));

        controls.selectedList.items = knownItems;
        controls.selectedList.selectedItems = items.map(item => item.entityId);

        controls.availableList.specials = controls.selectedList.items.map(item => item.entityId);
        controls.availableList.items    = filter(fullList);

        onConfigChange(true);
    }

    var removeItems = function (items) {
        var selectedIds = items && Array.isArray(items)
                        ? items.map(item => item.entityId)
                        : controls.selectedList.selectedItems.map(item => item.entityId);

        controls.selectedList.items = controls.selectedList.items.filter(item => !selectedIds.includes(item.entityId));

        forgetItems();
        controls.selectedList.items.forEach(addItem);

        controls.availableList.specials = controls.selectedList.items.map(item => item.entityId);
        controls.availableList.items    = filter(fullList);

        controls.availableList.selectedItems = selectedIds;

        onConfigChange(true);
    }

    function filter(fullList) {
        // We don't filter in-place if we're selecting by criteria instead of entities.
        // Actually, we do! Why not? It's a good aid to preview the result when you add
        // them to the job!
        // if (controls.jobModeControl.checked) return fullList;

        // TODO: If we are filtering by entities, then there's a convenient filter at
        // the top-right to narrow down the data you want to select from.
        // That's what this thing implements. But we do need the tags then as well.
        var filters = controls.filterControl.value || [];

        // Index tag values for quick access.
        const filteredList = filters.length === 0
                           ? fullList
                           : fullList.filter(function (item) {
            for (var i=0; i < filters.length; i++) {
                var filter = filters[i];
                var isIncluded = filter.type === "PROPERTY"
                               ? ( filter.values.length == 0
                                   ? item.hasOwnProperty(filter.name)
                                   : filter.values.includes(item[filter.name])
                                 )
                               : ( filter.values.length == 0
                                   ? item.tagValuesByKey.hasOwnProperty(filter.type + ":" + filter.name)
                                   : filter.values.includes(item.tagValuesByKey[filter.type + ":" + filter.name])
                                 );

                if (( isIncluded &&  filter.negated) ||
                    (!isIncluded && !filter.negated))
                    return false;
            }
            return true;
        });

        if (controls.jobModeControl.checked) {
            if (filters.length > 0)
                controls.filterMatches.innerHTML = filteredList.length + " sub-matches among " + fullList.length + " matching " + friendly(controls.jobTypeControl.value) + "s";
            else
                controls.filterMatches.innerHTML = filteredList.length + " matching " + friendly(controls.jobTypeControl.value) + "s";
        }
        else {
            controls.filterMatches.innerHTML = filteredList.length + " matches among " + fullList.length + " available " + friendly(controls.jobTypeControl.value) + "s";
        }
        return filteredList;
    }

    function analyze(filteredList) {
        var toLocaleString = function (time) {
            var date = new Date();
            date.setTime(time);
            return date.toLocaleString();
        }

        /*  Based on the available fields and tag values, we will
            create an inventory that helps us generate dropdowns easily.
         */
        let tags   = {};
        let fields = {};
        var now    = (new Date()).getTime();
        var hour   = 1000*60*60;

        const endPointInfo = self.config.Tenants[originalJobData.source];
        const entityURL = endPointInfo.baseURL + uiEndPoints[controls.jobTypeControl.value];

        // TODO: MAke normal iteration
        filteredList.forEach(function (item, i) {
            item.id = item.entityId;

            item.entityURL       = entityURL + item.entityId;

            var firstSeenSince   = Math.round((now - item.firstSeenTimestamp) / hour);
            var lastSeenSince    = Math.round((now - item.lastSeenTimestamp)  / hour);

            item.firstSeenTime   = toLocaleString(item.firstSeenTimestamp)
                                 + (firstSeenSince <=72 ? (" <span style='font-style:italic;font-size:85%;'> (" + firstSeenSince + " hours ago)</span>") : "");
            item.lastSeenTime    = toLocaleString(item.lastSeenTimestamp)
                                 + (lastSeenSince  <=72 ? (" <span style='font-style:italic;font-size:85%;'> (" + lastSeenSince  + " hours ago)</span>") : "");
            item.managementZones = item.managementZones || [];
            item.tags            = item.tags || [];
            item.tagValuesByKey  = {};
            item.zoneValuesByKey = {};

            // Collect tag values.
            item.tags.forEach(function (tag) {
                tag.value = tag.value || "(no value)";

                item.tagValuesByKey[tag.context + ":" + tag.key] = tag.value;

                tags[tag.context] = tags[tag.context] || {};

                if (tags[tag.context].hasOwnProperty(tag.key)) {
                    if (!tags[tag.context][tag.key].includes(tag.value))
                        tags[tag.context][tag.key].push(tag.value);
                }
                else {
                    tags[tag.context][tag.key] = [tag.value];
                }
            });

            // Collect field values (except common ones).
            for (var prop in item) {
                if (typeof item[prop] === "string" &&
                    !excludedFields.includes(prop)) {

                    if (fields.hasOwnProperty(prop)) {
                        if (!fields[prop].includes(item[prop]))
                            fields[prop].push(item[prop]);
                    }
                    else {
                        fields[prop] = [item[prop]];
                    }
                }
                else {
                    // We don't use fields that are not contain simple strings.
                    // And we also (per the above) exclude common fields.
                }
            }
        });

        // For the filter control:
        let filterOptions = [];

        for (var name in fields) {
            filterOptions.push({
                type:   "PROPERTY",
                name:   name,
                values: fields[name]
            });
        }
        for (var context in tags) {
            for (var name in tags[context]) {
                filterOptions.push({
                    type:   context,
                    name:   name,
                    values: tags[context][name]
                });
            }
        }
        // This is the 'play' filter.
        controls.filterControl.options = filterOptions;
    }

    let dataRequest = null;
    function getData(evt) {
        // Clear out old data - visually, too.
        fullList  = [];

        // Do we have a valid management zone selection?
        if (!apiEndPoints[controls.jobTypeControl.value] ||
            controls.mgtZoneControl.value == "none")
            return;

        // Abort any outstanding request - the user wants something else now.
        if (dataRequest) {
            dataRequest.abort();
            dataRequest = null;
        }

        // Preserve the selections, but clear the actual list.
        controls.availableList.specials = controls.selectedList.items.map(item => item.entityId);
        controls.availableList.items    = [];

        try {
            controls.spinner.style.visibility = "visible";

            if (controls.jobModeControl.checked) {
                dataRequest = new JSONHttpRequest({
                    address:      "/library/" + originalJobData.source + "/" + controls.jobTypeControl.value + "?" // select=entityId,displayName,tags"
                                  + '&ZONE:id=' + controls.mgtZoneControl.value,
                    method:       "POST",
                    onComplete:   function (data) { dataRequest = null; processData(data);  },
                    onError:      function (err)  { dataRequest = null; controls.spinner.style.visibility = "hidden"; },
                    progressElem: controls.filterMatches
                });

                evt = evt || controls.selectedFilters.selectedItem || { value: [] };

                dataRequest.send(JSON.stringify(evt.value.map(val => {
                    let values = val.values.map(value => value.replace("<i>(exists)</i>", ".").replace("<i>(has a value)</i>", "*"));

                    return {
                        context:  val.type,
                        key:      val.value,
                        values:   values,
                        operator: val.negated ? ("!" + val.comparator) : val.comparator
                    }
                })));
            }
            else {
                dataRequest = new JSONHttpRequest({
                    address:      "/library/" + originalJobData.source + "/" + controls.jobTypeControl.value + "?"
                                  + '&ZONE:id=' + controls.mgtZoneControl.value,
                    onComplete:   function (data) { dataRequest = null; processData(data); },
                    onError:      function (err)  { dataRequest = null; controls.spinner.style.visibility = "hidden"; },
                    progressElem: controls.filterMatches
                });

                dataRequest.send();
            }
        }
        catch (ex) {
            alert("The Dynatrace API could not be called: " + ex.message);
        }
    }

    function processData(jsonData) {
        controls.spinner.style.visibility = "hidden";

        fullList  = jsonData.sort(function (a, b) {
            if (!a.sortName)
                a.sortName = a.displayName.trim().toLowerCase();
            if (!b.sortName)
                b.sortName = b.displayName.trim().toLowerCase();

            return a.sortName < b.sortName
                 ? -1
                 : a.sortName > b.sortName
                 ? 1
                 : 0;
        });

        // We update the filterOptions control in 'analyze()'.
        controls.filterMatches.innerHTML = jsonData.length + " matches";

        analyze(fullList);

        controls.availableList.specials = controls.selectedList.items.map(item => item.entityId);
        controls.availableList.items    = filter(fullList);

        controls.selectedList.tooltipTemplate  = tooltipTemplates[controls.jobTypeControl.value];
        controls.availableList.tooltipTemplate = tooltipTemplates[controls.jobTypeControl.value];
    }

    function searchSelected(value) {
        return controls.selectedList.search(value);
    }
    function searchAvailable(value) {
        return controls.availableList.search(value);
    }

    function init(config) {
        self.config = config;

        var sources = Object.keys(config.Tenants);
        var destinations = Object.keys(config.Destinations).concat("None");

        // We're going to preload everything for all sources,
        // so that we can quicky and seamlessly switch between them.
        let pendingSources = [].concat(sources);
        sources.forEach(source => {
            emit("progress", "Connecting to source " + source + "...");

            loadDefinitions(source, config.Tenants[source], (err) => {
                let idx = pendingSources.indexOf(source);
                if (idx != -1) {
                    pendingSources.splice(idx, 1);

                    if (err) {
                        // We found it, and so we will find it in 'sources', too.
                        sources.splice(sources.indexOf(source), 1);
                        emit("progress", "WARNING: Source " + source + " will be omitted. " + err);
                    }
                    else {
                        emit("progress", "Successfully connected to source " + source);
                    }
                }

                if (pendingSources.length == 0) {
                    if (sources.length === 0) {
                        emit("progress", "There are no sources available. The editor cannot work.");
                    }
                    const templates = Object.keys(config.Templates);
                    populateSelector(controls.templateControl, templates);
                    populateSelector(controls.targetControl, destinations);

                    emit("ready", {});
                }
            });
        });
    }

    const filterOptions = {};

    function loadDefinitions(tenant, endPointInfo, onReady) {
        filterOptions[tenant]  = {
            // TODO: For these 3 default failters, the maxValue is 1. Only one value is allowed.
            PROBLEM: [
                { type: "PROBLEM", name: "status", values: [ "OPEN", "CLOSED" ]},
                { type: "PROBLEM", name: "impactLevel", values: [ "APPLICATION", "ENVRIONMENT", "INFRASTRUCTURE", "SERVICE" ]},
                { type: "PROBLEM", name: "severityLevel", values: [ "AVAILABILITY", "CUSTOM_ALERT", "ERROR", "MONITORING_UNAVAILABLE", "PERFORMANCE", "RESOURCE_CONTENTION" ]},
                // And then we add all the tags we are aware of, across the board.
            ]
        };

        // Retrieve the definitions.
        endPointInfo.definitions = {};
        endPointInfo.baseURL  = "https://"
                     + (endPointInfo.host || endPointInfo.hosts[0])
                     + endPointInfo.url;

        try {
        // First, get the contexts for this tenant. After that we will be making a bunch of
        // paralllel requests to get definitions and index values.
        var xhr = new JSONHttpRequest({
            address:    "/library/" + tenant + "/CONTEXT",
            onComplete: function (contexts) {
                // Prepare all requests we are going to make from here.
                // Track all requests we'll be making as an array of Promises.
                var requests = [];

                requests.push((function() {
                    return new Promise((resolve, reject) => {
                        var xhr = new JSONHttpRequest({
                            address: "/library/" + tenant + "/usql",
                            onComplete: function (usql) {
                                endPointInfo.definitions.usql = usql.definitions || usql.components.schemas;
                                resolve();
                            },
                            onError: function (error) {
                                // Wut?? Wel, then use our offline copy.
                                console.log("Request for USQL metadata failed. Getting local copy instead.");
                                var xhralt = new JSONHttpRequest({
                                    address:    "/spec2.json",
                                    onComplete: function (usql) {
                                        endPointInfo.definitions.usql = usql.definitions;
                                        resolve();
                                    },
                                    onError: function (err) {
                                        reject(error.message);
                                    }
                                });
                                xhralt.send();
                            }
                        });
                        xhr.send();
                    });
                })()); // Call it to obtain the promise.

                requests.push((function() {
                    return new Promise((resolve, reject) => {
                        var xhr = new JSONHttpRequest({
                            address: "/library/" + tenant + "/metrics",
                            onComplete: function (metrics) {
                                endPointInfo.definitions.metrics = metrics;
                                resolve();
                            },
                            onError: function (error) {
                                // Wut?? Wel, then use our offline copy.
                                console.log("Request for timeseries metadata failed. Getting local copy instead.");
                                var xhralt = new JSONHttpRequest({
                                    address:    "/timeseries.json",
                                    onComplete: function (metrics) {
                                        endPointInfo.definitions.metrics = metrics;
                                        resolve();
                                    },
                                    onError: function (err) {
                                        reject(error.message);
                                    }
                                });
                                xhralt.send();
                            }
                        });
                        xhr.send();
                    });
                })()); // Call it to obtain the promise.

                // Get contexts that have '.values'. Ignore the rest.
                // Special cases: ZONE and PROP
                // Zones need to be mapped to id and name
                // Props need to do into 'fields'
                // Other contexts: Do /librart/(tentant)/(CONTEXT)?select=displayName,values
                // Wait until they are all done.
                // Do this last
                contexts.forEach(context => {
                    if (!context.values) return;    // These are the entity types. We _could_ use them, but we have them defined already.

                    if (context.displayName === "ZONE") {
                        endPointInfo.definitions.zones = {
                            values: context.values.map(value => {
                                return {
                                    id:   value.entityId,
                                    name: value.displayName
                                };
                            })
                        };
                    }
                    else {
                        // Its a tag context or "PROP". We are going to have the retrieve all distinct tags/props and their values.
                        requests.push((function() {
                            return new Promise((resolve, reject) => {
                                var xhr = new JSONHttpRequest({
                                    address:    "/library/" + tenant + "/" + context.displayName,
                                    onComplete: function (tags) {
                                        // We get an array with objects. Each object:
                                        // - the displayName is the name of the tag. We should the contextValue in front of it.
                                        // - the 'values' list is the overal, consolidated list of values for the tag.
                                        tags.forEach(tag => {
                                            let tagName = tag.displayName;
                                            Object.keys(tag).forEach(key => {

                                                if (key === "displayName" || key === "values") return;

                                                const type = key;
                                                filterOptions[tenant][type] = filterOptions[tenant][type] || [];

                                                const filter = {
                                                    type: context.displayName,
                                                    name: tagName,
                                                    values: tag[key].concat(["<i>(exists)</i>", "<i>(has a value)</i>"])
                                                };
                                                filterOptions[tenant][type].push(filter);

                                                // All tags (but not property values) are relevant for problems, too...
                                                if (filter.type !== "PROPERTY") {
                                                    let found = false;
                                                    for (let i = 0; !found && i < filterOptions[tenant]["PROBLEM"].length; i++) {
                                                        if (filterOptions[tenant]["PROBLEM"][i].type === filter.type &&
                                                            filterOptions[tenant]["PROBLEM"][i].name === filter.name) {
                                                            // Known tag filter - merge the values.
                                                            filterOptions[tenant]["PROBLEM"][i].values =
                                                            filterOptions[tenant]["PROBLEM"][i].values.concat(filter.values).unique();

                                                            found = true;
                                                        }
                                                    }
                                                    if (!found) {
                                                        // New tag filter - add as new option.
                                                        filterOptions[tenant]["PROBLEM"].push(filter);
                                                    }
                                                }
                                            })
                                        });
                                        resolve();
                                    },
                                    onError: function (error) {
                                        reject("The Dynatrace API returns: " + error.message);
                                    }
                                });
                                xhr.send();
                            });
                        })()); // Call it to obtain the promise.
                    }
                });

                Promise
                .all(requests)
                .then(() => {
                    onReady();
                })
                .catch(error => {
                    alert("The Dynatrace API could not be called: " + error);
                    onReady();
                });
            },
            onError: function (error) {
                onReady("The Library API returns: " + error.message);
            }
        });
        xhr.send();
        }
        catch (ex) {
            alert("The Dynatrace API could not be called: " + ex.message);
        }
    }

    // Now we create the actual UI.
    options.container.innerHTML = "";

    var topRightArea = document.createElement("span");
    topRightArea.id = "topRightArea";

    var deleteButton = document.createElement("input");
    deleteButton.type = "button";
    deleteButton.value = "DELETE";
    deleteButton.style.background  = "red";
    deleteButton.style.borderColor = "red";
    deleteButton.addEventListener("click", function (evt) {
        emit("delete", null);
    });
    topRightArea.appendChild(deleteButton);

    var saveButton = document.createElement("input");
    saveButton.type = "button";
    saveButton.value = "SAVE";
    saveButton.addEventListener("click", function (evt) {
        emit("save", null);
    });
    topRightArea.appendChild(saveButton);

    options.container.appendChild(topRightArea);

    var jobNameContainer = document.createElement("div");
    jobNameContainer.id = "jobNameContainer";

    var jobNameControl   = document.createElement("input");
    jobNameControl.id = "jobName";
    jobNameControl.type  = "text";
    jobNameControl.value = "";
    jobNameControl.placeholder = "Job name";
    jobNameControl.style.width = "calc(100% - 200px)";

    jobNameContainer.appendChild(jobNameControl);
    options.container.appendChild(jobNameContainer);

    controls.jobNameControl = jobNameControl;

    var jobDefinition = document.createElement("div");
    jobDefinition.style.display = "inline-block";
    jobDefinition.style.width   = "calc(50% - 10px)";
    jobDefinition.style.verticalAlign = "top";
    jobDefinition.style.padding = "10px 0px";
    options.container.appendChild(jobDefinition);

    var playArea = document.createElement("div");
    playArea.style.display = "inline-block";
    playArea.style.width   = "calc(50% - 10px)";
    jobDefinition.style.verticalAlign = "top";
    playArea.style.padding = "10px 0px";
    options.container.appendChild(playArea);

    var controlsContainer = document.createElement("div");
    controlsContainer.style.padding = "0px";

    var jobFromToContainer = document.createElement("div");
    jobFromToContainer.id = "jobFromToContainer";
    jobFromToContainer.style.paddingBottom = "16px";

    var templateLabel = document.createElement("span");
    templateLabel.style.display = "inline-block";
    templateLabel.style.minWidth = "50px";
    templateLabel.innerHTML = "Format: ";

    var templateControl = document.createElement("select");
    templateControl.id = "templateControl";

    jobFromToContainer.appendChild(templateLabel);
    jobFromToContainer.appendChild(templateControl);

    var targetLabel = document.createElement("span");
    targetLabel.style.display = "inline-block";
    targetLabel.style.padding = "0px 6px 0px 12px";
    targetLabel.innerHTML = "Destination: ";

    var targetControl = document.createElement("select");
    targetControl.id = "targetControl";

    jobFromToContainer.appendChild(targetLabel);
    jobFromToContainer.appendChild(targetControl);

    var jobTypeContainer = document.createElement("span");

    var jobTypeLabel = document.createElement("span");
    jobTypeLabel.style.display = "inline-block";
    jobTypeLabel.style.minWidth = "50px";
    jobTypeLabel.innerHTML = "Type: ";

    var jobTypeControl = document.createElement("select");
    jobTypeControl.style.color = "#000000";
    jobTypeControl.addEventListener("change", onTypeChange);

    jobTypeContainer.appendChild(jobTypeLabel);
    jobTypeContainer.appendChild(jobTypeControl);

    var limitLabel = document.createElement("span");
    limitLabel.className = "Sessions";
    limitLabel.style.padding = "0px 6px 0px 12px";
    limitLabel.innerHTML = "Limit: ";

    var limitControl = document.createElement("input");
    limitControl.id = "limitControl";
    limitControl.className = "Sessions";
    limitControl.type = "text";
    limitControl.style.width = "50px";
    limitControl.placeholder = "1000";

    jobTypeContainer.appendChild(limitLabel);
    jobTypeContainer.appendChild(limitControl);

    controlsContainer.appendChild(jobTypeContainer);

    var mgtZoneContainer = document.createElement("div");
    mgtZoneContainer.className = "Entities";

    var mgtZoneLabel = document.createElement("span");
    mgtZoneLabel.style.display = "inline-block";
    mgtZoneLabel.style.minWidth = "50px";
    mgtZoneLabel.innerHTML = "Zone: ";

    var mgtZoneControl = document.createElement("select");
    mgtZoneControl.id = "mgtZoneControl";
    mgtZoneControl.addEventListener("change", onZoneChange);

    mgtZoneContainer.appendChild(mgtZoneLabel);
    mgtZoneContainer.appendChild(mgtZoneControl);
    controlsContainer.appendChild(mgtZoneContainer);

    var spinner = document.createElement("div");
    spinner.id = "spinner";
    spinner.className = "lds-ellipsis";
    spinner.style.visibility = "hidden";
    spinner.innerHTML = "<div></div><div></div><div></div><div></div>";

    mgtZoneContainer.appendChild(spinner);

    var outputContainer = document.createElement("div");
    outputContainer.className = "Problems";

    var outputLabel = document.createElement("span");
    outputLabel.style.display = "inline-block";
    outputLabel.style.minWidth = "50px";
    outputLabel.innerHTML = "Output: ";

    var outputControl = document.createElement("select");
    outputControl.id = "outputControl";
    outputControl.innerHTML = '<option value="Problems">Problems with affected entities</option><option value="Entities">Entities affected by problems</option>';

    outputContainer.appendChild(outputLabel);
    outputContainer.appendChild(outputControl);
    controlsContainer.appendChild(outputContainer);

    jobDefinition.appendChild(controlsContainer);
    jobDefinition.appendChild(jobFromToContainer);

    // The field picker
    var fieldToggle = document.createElement("div");
    fieldToggle.className = "PickerToggle Sessions";
    fieldToggle.innerHTML = '<span id="fieldToggleControl">Use the fields below to create an ID for each result: </span>';

    jobDefinition.appendChild(fieldToggle);

    var fieldContainer = document.createElement("div");
    fieldContainer.className = "fieldArea Sessions";

    var fieldBoxes = document.createElement("div");
    fieldBoxes.id = "fieldBoxes";
    fieldBoxes.onchange = onConfigChange;
    fieldContainer.appendChild(fieldBoxes);

    jobDefinition.appendChild(fieldContainer);

    // The metric picker
    var metricToggle = document.createElement("div");
    metricToggle.className = "";
    metricToggle.innerHTML = '<span id="metricToggleControl">Collect for each </span>';
    var granularityControl = document.createElement("select");
    granularityControl.id = "granularity";
    populateSelector(granularityControl, granularities, "_min", true);
    granularityControl.onchange = onConfigChange;
    metricToggle.appendChild(granularityControl);
    metricToggle.insertAdjacentHTML('beforeend', " period: ");
    // About the above: I am proud of the ugliest approach ever.

    jobDefinition.appendChild(metricToggle);

    var metricContainer = document.createElement("div");
    metricContainer.className = "PickerArea";

    var metricBoxes = document.createElement("div");
    metricBoxes.metricBoxes = "metricBoxes";
    metricBoxes.style.paddingBottom = "10px";
    metricBoxes.onchange = onConfigChange;
    metricContainer.appendChild(metricBoxes);

    jobDefinition.appendChild(metricContainer);

    // The filter picker (for metrics only)
    // Only for "EXPLICIT" "ENTITIES" - and goes into PLAY AREA
    // TODO!
    var filterToggle = document.createElement("div");
    filterToggle.className = "PickerToggle Entities";
    filterToggle.innerHTML = '<span id="filterToggleControl" style="display:block; position:relative;">Filters: <i style="float:right; padding-right:6px;">(<b>not</b> part of the job definition)</i></span>';

    playArea.appendChild(filterToggle);

    var filterContainer = document.createElement("div");
    filterContainer.style.paddingTop = "16px";
    filterContainer.className = "PickerArea Entities Problems";

    var filterBoxes = document.createElement("div");
    filterBoxes.id = "filterBoxes";
    filterBoxes.onchange = onConfigChange;
    filterContainer.appendChild(filterBoxes);

    playArea.appendChild(filterContainer);

    var sqlPreviewContainer = document.createElement("div");
    playArea.appendChild(sqlPreviewContainer);

    // Back to JOBDEFINITION area
    // The where picker (for usql only)
    var whereToggle = document.createElement("div");
    whereToggle.className = "PickerToggle Sessions";
    whereToggle.innerHTML = '<span id="whereToggleControl">Filters: </span>';

    jobDefinition.appendChild(whereToggle);

    var whereContainer = document.createElement("div");
    whereContainer.className = "PickerArea Sessions";

    var whereBoxes = document.createElement("div");
    whereBoxes.id = "whereBoxes";
    whereContainer.appendChild(whereBoxes);

    jobDefinition.appendChild(whereContainer);

    // The Group By control
    var groupByToggle = document.createElement("div");
    groupByToggle.className = "PickerToggle Sessions";
    groupByToggle.innerHTML = '<span id="groupByToggleControl">Group by: </span>';

    jobDefinition.appendChild(groupByToggle);

    var groupByContainer = document.createElement("div");
    groupByContainer.className = "PickerArea Sessions";

    var groupByBoxes = document.createElement("div");
    groupByBoxes.id = "groupByBoxes";
    groupByContainer.appendChild(groupByBoxes);

    jobDefinition.appendChild(groupByContainer);

    // The Order By Control
    var orderByToggle = document.createElement("div");
    orderByToggle.className = "PickerToggle Sessions";
    orderByToggle.innerHTML = '<span id="orderByToggleControl">Order by: </span>';

    jobDefinition.appendChild(orderByToggle);

    var orderByContainer = document.createElement("div");
    orderByContainer.className = "PickerArea Sessions";

    var orderByBoxes = document.createElement("div");
    orderByBoxes.id = "orderByBoxes";
    orderByContainer.appendChild(orderByBoxes);

    jobDefinition.appendChild(orderByContainer);

    var filterCriteriaLabel = document.createElement("div");
    filterCriteriaLabel.className = "PickerToggle Problems";
    filterCriteriaLabel.style.paddingBottom = "12px";
    filterCriteriaLabel.style.marginTop = "0px";
    filterCriteriaLabel.innerHTML = "Filter criteria:";
    options.container.appendChild(filterCriteriaLabel);

    var entitiesToggle = document.createElement("div");
    entitiesToggle.className = "PickerToggle Entities";
    entitiesToggle.style.width = "calc(100% - 8px)";

    var matchCountSpan = document.createElement("span");
    matchCountSpan.id = "matchCount";
    matchCountSpan.innerHTML = ""; // No entities in selected MZ
    entitiesToggle.appendChild(matchCountSpan);

    var jobModeControl = document.createElement("span");
    jobModeControl.className = "onoffswitch";
    var jobModeCheckbox = document.createElement("input");
    jobModeCheckbox.type = "checkbox";
    jobModeCheckbox.id   = "jobModeToggle";
    jobModeCheckbox.name = "onoffswitch";
    jobModeCheckbox.className = "onoffswitch-checkbox";
    jobModeCheckbox.checked = false;
    jobModeCheckbox.addEventListener("change", onModeChange);
    jobModeControl.appendChild(jobModeCheckbox);
    jobModeControl.insertAdjacentHTML("beforeend", '<label class="onoffswitch-label" for="jobModeToggle"><span class="onoffswitch-inner"></span><span class="onoffswitch-switch"></span></label>');

    var entityCounter = document.createElement("span");
    jobModeControl.appendChild(entityCounter);
    entitiesToggle.appendChild(jobModeControl);

    options.container.appendChild(entitiesToggle);

    Split([jobDefinition, playArea], {
        sizes: [70, 30],
        minSize: [400, 200],
        direction: 'horizontal'
    });

    // Consolidate key controls for quick access later.
    controls = {
        spinner:             spinner,
        templateControl:     templateControl,
        targetControl:       targetControl,
        entityCounter:       entityCounter,
        filterMatches:       document.getElementById("matchCount"),   // TODO: This must be created above.
        jobNameControl:      jobNameControl,
        jobTypeControl:      jobTypeControl,
        jobModeControl:      jobModeCheckbox,
        outputControl:       outputControl,
        relatedControl:      document.getElementById("includeWhat"),
        relatedControl2:     document.getElementById("includeWhat2"),
        mgtZoneControl:      mgtZoneControl,
        limitControl:        limitControl,
        granularityControl:  granularityControl,
        fieldToggle:         fieldToggle,
        metricToggle:        metricToggle,
        filterToggle:        filterToggle,
        whereToggle:         whereToggle,
        fieldControl:        new FilterControl(fieldBoxes,   { negatable: false, emptytext: "", allowAlias: true, editable: false }),
        metricControl:       new FilterControl(metricBoxes,  { negatable: false, emptytext: "", allowAlias: true, minValues: 1, tooltip:true }),
        filterControl:       new FilterControl(filterBoxes,  { negatable: true,  emptytext: "(exists)", minValues: 1 }),
        whereControl:        new FilterControl(whereBoxes,   { comparators: comparators, editable: true }),
        groupByControl:      new FilterControl(groupByBoxes, { negatable: false, emptytext: "", maxValues: 0 }),
        orderByControl:      new FilterControl(orderByBoxes, { negatable: false, emptytext: "", maxValues: 1 }),
        selectedFilters:     new FilterList(options.selectedFilters,  {}, getData),
        saveButton:          saveButton,
        addButton:           options.addButton,
        removeButton:        options.removeButton,
        sqlPreviewContainer: sqlPreviewContainer
    };
    controls.availableList = new ListTable(options.availableList, { selectionControls: options.availableCount }),
    controls.selectedList  = new ListTable(options.selectedList,  { selectionControls: options.selectedCount, helper: controls.availableList }),

    controls.filterControl.addEventListener("change", function (evt) {
        controls.availableList.specials = controls.selectedList.items.map(item => item.entityId);
        controls.availableList.items    = filter(fullList);
    });

    controls.addButton.addEventListener("click", function (evt) {
        addItems(controls.availableList.selectedItems);
    });

    controls.removeButton.addEventListener("click", function (evt) {
        removeItems(controls.selectedList.selectedItems);
    });

    return {
        init:      init,
        on:        subscribe,
        get sources() {
            return Object.keys(self.config.Tenants);
        },
        get emptyJob() {
            return {
                id: null,
                source: null,
                type: "HOST",
                name: "(New Job)",
                mz: ".",
                mzName: "Everything",
                granularity: "_min",
                metrics: [],
                mode: "implicit",
                entities: null,
                related: "none",
                criteria: [],
                optimize: "true"
            };
        },
        render:    renderJob,
        unrender:  unrenderJob,
        clear:     function () {
            originalJobData = unrenderJob();
        },
        add:       addItems,
        remove:    removeItems,
        isReadyToSwitch: isReadyToSwitch,
        searchSelected:  searchSelected,
        searchAvailable: searchAvailable
    }
}