JobNavigator = function (container, jobEditor) {
    var self = this;

    this.container = container;
    this.jobEditor = jobEditor;

    this.events  = {};

    var jobs = null;
    var selectedSource   = null;
    var selectedJobIndex = 0;
    var selectedJobPanel = null;
    var jobIdGenerator = 0;

    var createJob = function (jobData, doSelect) {
        var jobControl = document.createElement("div");
        jobControl.className = "JobItem";
        jobControl.setAttribute('data-i', jobData.id);
        if (jobEditor.sources.includes(jobData.source)) {
            jobControl.addEventListener("click", selectJob);
        }

        var controlsCtl =  document.createElement("div");
        controlsCtl.className = "Controls";

        var destGraphic = document.createElement("span");
        destGraphic.className = "icon-arrow-right2";
        destGraphic.innerHTML = '<span style="font-family: BerninaSansWeb, Arial, sans-serif; font-size: 11pt; vertical-align: text-bottom; padding-left: 4px;">' + jobData.destination + '</span>';
        controlsCtl.appendChild(destGraphic);

        jobControl.appendChild(controlsCtl);

        var titleCtl =  document.createElement("div");
        titleCtl.className = "Title";
        titleCtl.innerHTML = jobData.name || "(New job)";

        jobControl.appendChild(titleCtl);

        var contentsCtl =  document.createElement("div");
        contentsCtl.className = "Contents";
        contentsCtl.title = jobData.metrics.join(", ");

        if (jobData.query) {
            contentsCtl.innerHTML = jobData.timestamp
                                  ? jobData.metrics.length + " metrics, "
                                  + jobData.fields.length  + " fields, "
                                  + jobData.where.length + " criteria from "
                                  + toUserReadableName(jobData.type) + "."
                                  : "This job is currently empty";
        }
        else if (jobData.entities) {
            contentsCtl.innerHTML = jobData.timestamp
                                  ? jobData.metrics.length + " metrics for "
                                  + jobData.entities.length + " " + toUserReadableName(jobData.type)
                                  + " from " + jobData.mzName + "."
                                  : "This job is currently empty";
        }
        else if (jobData.problems) {
            contentsCtl.innerHTML = jobData.timestamp
                                  ? (jobData.subtype === "Problems" ? "Problems with affected entities" : "Entities affected by problems")
                                  +  " matching any of " + jobData.criteria.length + " groups(s) of criteria."
                                  : "This job is currently empty";
        }
        else if (jobData.criteria) {
            contentsCtl.innerHTML = jobData.timestamp ?
                jobData.metrics.length + " metrics for " +
                toUserReadableName(jobData.type) +
                " from " + jobData.mzName +
                " matching any of " + jobData.criteria.length + " groups(s) of criteria." :
                "This job is currently empty";
        } else {
            contentsCtl.innerHTML = jobData.timestamp
                                  ? jobData.metrics.length + " metrics for "
                                  + toUserReadableName(jobData.type) +
                                  + " (no entities or filters specified) "
                                  : "This job is currently empty";

        }
        jobControl.appendChild(contentsCtl);

        var metaDataCtl =  document.createElement("div");
        metaDataCtl.className = "Metadata";
        metaDataCtl.innerHTML = jobData.timestamp
                              ? "Last saved" + (jobData.userID ? " by " + jobData.userID : "") + " on " + (new Date(jobData.timestamp)).toLocaleString()
                              : "Not yet saved";
        metaDataCtl.title = metaDataCtl.innerHTML;

        jobControl.appendChild(metaDataCtl);

        // Add to the proper tab.
        let tabContent = self.scrollList.querySelector("#tab_" + jobData.source + " .tab-content");
        if (!tabContent) {
            self.scrollList.appendChild(createTab(jobData.source, false));
            tabContent = self.scrollList.querySelector("#tab_" + jobData.source + " .tab-content");
        }

        tabContent.appendChild(jobControl);

        if (doSelect)
            selectJob({ currentTarget: jobControl });

        return jobControl;
    };

    var deleteJob = function (ev) {
        // The job is already selected - otherwise the "Delete" button
        // would not be visible. So, delete the current job and select
        // and render another one.
        if (!window.confirm("Do you want to delete this job?")) return;

        // Remove from the proper tab.
        const tab = selectedJobPanel.parentElement;
        tab.removeChild(selectedJobPanel);
        selectedJobPanel = null;

        jobs.splice(selectedJobIndex, 1);
        selectedJobIndex = -1;

        saveJobs(true);

        // Show empty page.
        jobEditor.render();
        document.getElementById("emptyPanel").style.display = "block";

        if (ev) {
            // This may be overkill, but at least one of these is necessary...
            // This is for when the 'delete' button on the little panel is clicked.
            ev.cancelBubble = true;
            ev.preventDefault = true;
            ev.stopImmediatePropagation();
            ev.stopPropagation();
        }
    };

    var selectJob = function (evt) {
        // Ensure that 'empty page is hidden.
        document.getElementById("emptyPanel").style.display = "none";

        var newSelectedPanel = evt.currentTarget;
        if (newSelectedPanel === selectedJobPanel) return; // Ignore click on selected job.

        if (jobEditor.isReadyToSwitch()) {
            if (selectedJobPanel) selectedJobPanel.classList.remove("Selected");
            newSelectedPanel.classList.add("Selected");

            // Ensure that the tab is expanded.
            newSelectedPanel.parentElement.parentElement.querySelector("input[type=checkbox").checked = true;

            selectedJobPanel = newSelectedPanel;
            selectedJobIndex = Number(newSelectedPanel.getAttribute("data-i"));;
            if (isNaN(selectedJobIndex)) console.error("Job has no index!");

            jobEditor.render(jobs[selectedJobIndex]);
        }
    }

    var loadJobs = function (isFirstLoad) {
        const scrollTop = emptyScrollList();

        var xhr = new JSONHttpRequest({
            address:      "jobs.json",
            onComplete:   function (data) {
                jobs = data;
                jobIdGenerator = jobs.length;
                jobs.forEach(function (jobData, i) {
                    jobData.id = i;
                    createJob(jobData, i === selectedJobIndex);
                });
                // Go back to last scroll position.
                self.scrollList.scrollTop = scrollTop;

                // Show empty page if necessary.
                if (!isFirstLoad)
                    document.getElementById("emptyPanel").style.display = jobs.length === 0 ? "block" : "none";
                else if (jobs.length === 0)
                    document.getElementById("emptyPanel").classList.add("Visible");
            }
        });
        xhr.send();
    };

    var saveJobs = function (skipUnrender) {
        if (!skipUnrender) {
            var job = jobEditor.unrender();
            if (!job) return;   // Can happen if config validation fails.

            job.userID = userID;
            jobs[selectedJobIndex] = job;
            jobEditor.clear();  // Resets its 'dirty' flag.
        }

        var xhr = new JSONHttpRequest({
            method:     "POST",
            address:    "jobs.json",
            onComplete:  loadJobs,
            onError: function (ex) {
                window.alert("Could not save jobs: " + ex.message);
            }
        });
        xhr.send(JSON.stringify(jobs, null, 4));
    }

    function displayProgress(message) {
        var panel = document.getElementById("initProgress");

        if (message)
            panel.insertAdjacentHTML("beforeend", "<p>" + message + "</p>");
        else
            panel.parentElement.insertAdjacentHTML("beforeend", "<button id='continueButton' role='button' type='submit' onclick='showEditor(event)'>Continue</button>");
    }

    var emptyScrollList = function () {
        self.scrollList = self.scrollList || createScrollList(self.container);
        const scrollTop = self.scrollList.scrollTop;
        const tabContentSections = self.scrollList.querySelectorAll(".tab-content");
        tabContentSections.forEach(section => {
            section.innerHTML = "";
        });
        return scrollTop;
    };

    var createTab = function (source, isKnownSource) {
        const tab = document.createElement("div");
        tab.id = "tab_" + source;
        tab.className = "tab";
        if (!isKnownSource) {
            tab.style.display = "none";
        }

        const toggle = document.createElement("input");
        toggle.id = "toggle_" + source;
        toggle.type = "checkbox";
        tab.appendChild(toggle);

        const label = document.createElement("label");
        label.className = "tab-label";
        label.setAttribute("for", "toggle_" + source);
        label.innerHTML = source;
        tab.appendChild(label);

        if (isKnownSource) {
            const button = document.createElement("input");
            button.type = "button";
            button.value = "New Job";
            button.addEventListener("click", evt => {
                if (jobEditor.isReadyToSwitch()) {
                    const jobData = jobEditor.emptyJob; // Ask for an empty job.
                    jobData.id = jobIdGenerator++;
                    jobData.source = source;

                    jobs.push(jobData);

                    createJob(jobData, true);
                    setTimeout(() => {
                        // TODO: Ideally, we scroll to to of the tab.
                        self.scrollList.scrollTop = self.scrollList.scrollHeight;
                    }, 10);
                }
            });
            tab.appendChild(button);
        }
        const content = document.createElement("div");
        content.id = "content_" + source;
        content.className = "tab-content";
        tab.appendChild(content);

        return tab;
    }

    var createScrollList = function (container) {
        container.innerHTML = "";

        var scrollList = document.createElement("div");
        scrollList.className = "ScrollList Scrollbar tabs";
        scrollList.style.overflowY = "auto";
        scrollList.style.height = "calc(100% - 16px)";
        scrollList.style.width = "100%";

        jobEditor.sources.forEach(source => {
            scrollList.appendChild(createTab(source, true));
        });

        container.appendChild(scrollList);

        return scrollList;
    }

    jobEditor.on("progress", displayProgress);
    jobEditor.on("ready",    () => {
        displayProgress();  // Show 'Click to continue'.
        this.jobList = createScrollList(this.container);
        loadJobs(true);     // Load, but don't hide messages yet.
    });
    jobEditor.on("save",     saveJobs);
    jobEditor.on("delete",   deleteJob);

    return {
        load:   loadJobs,
        save:   saveJobs
    }
}