JobTester = function (container, config) {
    const self = this;

    let gridId = container.id;

    this.config = config;
    this.selectionControls  = config.selectionControls 
                            ? ( typeof config.selectionControls === "string"
                                ? document.getElementById(config.selectionControls)
                                : config.selectionControls
                            )
                            : { innerHTML: null };

    this.events  = {};

    const emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const event = self.events[eventName] || [];
        event.forEach(fn => { fn.call(null, data); });
    };

    const subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);

        // Return an 'unsubscribe' function for this subscription.
        return () => {
            self.events[eventName] = self.events[eventName].filter(eventFn => fn !== eventFn);
        };
    };
        
    let grid = null;
    
    function createColumns(sampleRow) {
        var columns = [
            {
                id: "entityId",
                name: "Entity ID",
                field: "entityId",
                width: 100,
            }, {
                id: "displayName",
                name: "Name",
                field: "displayName",
                width: 100,          
                sortable: true,
            }, {
                id: "lastSeen",
                name: "Last Seen",
                field: "lastSeen",
                width: 100,      
                formatter: (row, cell, value) => new Date(value).toISOString(),
                headerCssClass: "Right",
                cssClass: "Right",   
                sortable: true,
            }
        ];

        for (var col in sampleRow) {
            if (col === "displayName" || 
                col === "entityId" || 
                col === "lastSeen") continue;

            columns.push({
                id: col,
                name: col,
                field: col,
                width: 100,
                headerCssClass: "Right",
                cssClass: "Right",                
                sortable: true
            });
        }

        return columns;
    }

    function processData(jsonData) {
        if (!jsonData || jsonData.length === 0) {
            grid.setColumns([]);
            grid.setData([]);
            grid.invalidate();
            return;
        }
            
        var columns = createColumns(jsonData[0]);
        grid.setColumns(columns);

        var data = jsonData.map(rowData => {
            let row = {};
            columns.forEach(column => {
                row[column.id]  = column.colType
                                ? rowData[column.colType][column.id]
                                : rowData[column.id];
            });
            return row;
        })
        grid.setData(data);
        grid.invalidate();
    }

    function selectAllItems() {
        // Create array with indices: [0 .. N-1] where N is the length of the data.
        grid.setSelectedRows([...Array(grid.getData().length).keys()]);
    }

    function invertSelection() {
        var availableIndices = grid.getData().map((row, i) => i);
        var selectedIndices  = grid.getSelectedRows();

        grid.setSelectedRows(availableIndices.filter((row, i) => !selectedIndices.includes(i)));
    }

    function selectNoItems() {
        grid.setSelectedRows([]);
    }

    // Create the UI, selection controls.
    self.selectionControls.innerHTML = "";

    var selectAllControl = document.createElement("span");
    selectAllControl.className = "icon-radio-checked";
    selectAllControl.addEventListener("click", selectAllItems);
    selectAllControl.innerHTML = "&nbsp;";

    self.selectionControls.appendChild(selectAllControl);

    var invertSelControl = document.createElement("span");
    invertSelControl.className = "icon-contrast";
    invertSelControl.addEventListener("click", invertSelection);
    invertSelControl.innerHTML = "&nbsp;";
    
    self.selectionControls.appendChild(invertSelControl);

    var selectNoneControl = document.createElement("span");
    selectNoneControl.className = "icon-radio-unchecked";
    selectNoneControl.addEventListener("click", selectNoItems);
    selectNoneControl.innerHTML = "&nbsp;";
    
    self.selectionControls.appendChild(selectNoneControl);
        
    grid = new Slick.Grid("#" + gridId, [], [], {
        enableCellNavigation: true,
        enableColumnReorder: true,
        forceFitColumns: true,
        multiColumnSort: true,
        multiSelect: true
    });
    grid.setSelectionModel(new Slick.RowSelectionModel());
    grid.registerPlugin(new Slick.CellExternalCopyManager({
        readOnlyMode: true,
        includeHeaderWhenCopying: true,
    }));

    grid.onSort.subscribe(function (e, args) {
        let data = grid.getData();
        let cols = args.sortCols;

        data.sort(function (dataRow1, dataRow2) {
            for (var i = 0, l = cols.length; i < l; i++) {
                var field  = cols[i].sortCol.field;
                var sign   = cols[i].sortAsc ? 1 : -1;
                var value1 = dataRow1[field], value2 = dataRow2[field];
                var result = (value1 == value2 ? 0 : (value1 > value2 ? 1 : -1)) * sign;
                if (result != 0) return result;
            }
            return 0;
        });
        grid.invalidate();
        grid.render();
    });

    // Return the interface.
    return {
        on: subscribe,
        resize: function(height, width) {
            $("#" + gridId).height(height);
            $("#" + gridId).width(width);
            
            grid.resizeCanvas();
        },
        set data(jsonData) {
            processData(jsonData);
        }
    };
    
}
