const TimelineManager = function (workerSorter) {

    const players = {};
    const knowns  = Array.isArray(workerSorter)
                  ? workerSorter
                  : null;
    const sorter  = Array.isArray(workerSorter)
                  ? (a, b) => workerSorter.indexOf(a) - workerSorter.indexOf(b)
                  : workerSorter;

    function toReadableEntity(worker) {
        return worker.charAt(0) + worker.substring(1).toLowerCase().replace(/_/g," ");
    }

    function getStartOfMinute(ts) {
        let timestamp = new Date();
        if (ts) timestamp.setTime(ts);
        timestamp.setSeconds(0, 0);
        return timestamp.getTime();
    }

    function insertSorted(parent, childrenClassName, child, attribute) {
        if (!parent) return;

        const children = parent.querySelectorAll(childrenClassName);
        for (var i=0; i < children.length && child.parentElement !== parent; i++) {
            let thatChildValue = attribute ? children[i].getAttribute(attribute) : children[i].id;
            let thisChildValue = attribute ? child.getAttribute(attribute)       : child.id;

            if (thatChildValue > thisChildValue)
                parent.insertBefore(child, children[i]);
        }
        if (child.parentElement !== parent)
            parent.appendChild(child);
    }

    function addPlayer(playerName, playerAdmin) {
        if (!players[playerName]) {
            players[playerName] = playerAdmin;
            players[playerName].mins = {};
            players[playerName].workers = {};
        }

        return players[playerName];
    }

    function hasPlayer(playerName) {
        return players[playerName] && players[playerName].timeline;
    }

    function getPlayer(playerName) {
        return players[playerName];
    }

    function getPlayers() {
        return Object.keys(players);
    }

    function addWorker(update, time, drawer, titleGenerator) {
        const player = players[update.player];
        if (!player) return null;

        let workerAdmin = players[player.name].workers[update.worker.name];
        if (!workerAdmin) {
            workerAdmin = players[player.name].workers[update.worker.name] = {
                id:   update.worker.id,
                name: update.worker.name,
                type: update.worker.type,
                next: time,
                mins: {},
                elem: null,
                drawer: drawer,
                titleGenerator: titleGenerator || function (player, worker, type) {
                    return '<span>' + toReadableEntity(worker) + '</span>';
                }
            };

            // Makes sure that the next execution is in the admin so that that minute gets drawn.
            if (workerAdmin.next) {
                // Draw the first known execution for this worker.
                const nextMinute = getStartOfMinute(workerAdmin.next);

                if (!workerAdmin.mins[nextMinute]) {
                    workerAdmin.mins[nextMinute] = { cycles: {} };
                }

                if (!workerAdmin.mins[nextMinute].cycles[workerAdmin.next]) {
                    workerAdmin.mins[nextMinute].cycles[workerAdmin.next] = {
                        cycle:  nextMinute,
                        requested: nextMinute,  // TODO: This is for drawLibraryRequest. Parameterize!
                        tenant: player.name,  // Backwards compatibility
                        player: player.name,
                        worker:   workerAdmin.name,
                        progress: "plan",
                        log: { planned: workerAdmin.next }
                    };

                    createMinute(player.name, nextMinute);
                    setTimeout( workerAdmin.drawer, 10,
                                workerAdmin.mins[nextMinute].cycles[workerAdmin.next],
                                workerAdmin.mins[nextMinute]);
                }
            }
        }
        return workerAdmin;
    }

    function createPipelinePanel(player, workerAdmin, segment) {
        let created = false;

        let pipelinePanel = document.getElementById(player.name + "_" + workerAdmin.name + "_" + segment);
        if (!pipelinePanel) {
            pipelinePanel = document.createElement("div");
            pipelinePanel.id           = player.name + "_" + workerAdmin.name + "_" + segment;
            pipelinePanel.className    = "PipelinePanel";
            pipelinePanel.style.height = "12px";
            pipelinePanel.style.width  = "100%";
            pipelinePanel.setAttribute("minute", segment);
            pipelinePanel.setAttribute("player", player.name);
            pipelinePanel.setAttribute("worker", workerAdmin.name);

            workerAdmin.mins[segment] = workerAdmin.mins[segment] || { cycles: {} };
            workerAdmin.mins[segment].panel = pipelinePanel;
            workerAdmin.mins[segment].renderer = SVG(pipelinePanel).nested();

            created = true;
        }
        return {
            pipelinePanel: pipelinePanel,
            created:       created
        };
    }

    function drawWorker(player, workerAdmin, time) {
        let panelWasCreated = false;

        let openingMinute  = document.getElementById(player.name + "_opening");
        let timelineMinute = document.getElementById(player.name + "_" + time);

        let openingPanel = document.getElementById(player.name + "_" + workerAdmin.name + "_opening");
        if (!openingPanel) {
            openingPanel = document.createElement("div");
            openingPanel.id           = player.name + "_" + workerAdmin.name + "_opening";
            openingPanel.className    = "PipelinePanel";
            openingPanel.style.height = "12px";
            openingPanel.style.width  = "100%";
            openingPanel.setAttribute("player", player.name);
            openingPanel.setAttribute("worker", workerAdmin.name);

            openingMinute.appendChild(openingPanel);

            // Draw the worker name in the opening block, plus any controls.
            openingPanel.innerHTML = workerAdmin.titleGenerator(player.name, workerAdmin.name, workerAdmin.type);
            panelWasCreated = true;
        }

        // Ensure that segments exist for this worker for all known future minutes.
        // NOTE: We also have to look at newly added future minutes, which a new worker may bring!
        let nextTimeSegment = timelineMinute;
        while (nextTimeSegment) {
            let segment = nextTimeSegment.getAttribute("minute");
            let { pipelinePanel, created} = createPipelinePanel(player, workerAdmin, segment);

            if (created) nextTimeSegment.appendChild(pipelinePanel);

            workerAdmin.mins[segment].cycles[segment] = workerAdmin.mins[segment].cycles[segment] || {
                cycle:  segment,
                tenant: player.name,  // Backwards compatibility
                player: player.name,
                worker:   workerAdmin.name,
                progress: "plan",
                log: { planned: workerAdmin.next }
            };

            nextTimeSegment = nextTimeSegment.nextSibling;
        }

        return panelWasCreated;
    }

    function invalidateWorker(player, worker, hideControls) {
        let openingPanel = document.getElementById(player + "_" + worker + "_opening");
        if (!openingPanel) return;

        openingPanel.style.color = "red";
        if (hideControls) {
            let controls = openingPanel.querySelector(".FloatControl");
            controls.style.display = "none";
        }
    }

    function revalidateWorker(player, worker) {
        let openingPanel = document.getElementById(player + "_" + worker + "_opening");
        if (!openingPanel) return;

        openingPanel.style.color = "";

        let controls = openingPanel.querySelector(".FloatControl");
        controls.style.display = "";
    }

    function getWorker(playerName, worker) {
        return players[playerName].workers[worker.name || worker];
    }

    function setCycle(worker, time, update) {
        if (!worker || !worker.mins)
            return null;

        let minute = getStartOfMinute(time);

        // Check if the future minute already has a timeline. If not, create it.
        if (!worker.mins[minute]) {
            worker.mins[minute] = {
                panel:    null,
                renderer: null,
                cycles:   {}
            };
            worker.mins[minute].cycles[time] = update || worker.mins[minute].cycles[time];

            createMinute(null, minute);
        }
        else {
            worker.mins[minute].cycles[time] = update || worker.mins[minute].cycles[time];
        }

        return worker.mins[minute];
    }

    function updateCycle(worker, time, update) {
        const minAdmin = setCycle(worker, time, update);
        if (!minAdmin)
            return null;

        return minAdmin.cycles[time];
    }

    function destroyAncientMinutes(arrearMinutes) {
        const ancientMinute = getStartOfMinute() - arrearMinutes * 60 * 1000;

        const timeAxis = document.getElementById("timeAxis");
        const timelineTickers = timeAxis.querySelectorAll(".TimelineTicker");

        for (let i=0; i < timelineTickers.length; i++) {
            const minute = Number(timelineTickers[i].getAttribute("minute"));
            if (minute && minute < ancientMinute) {
                // Remove the timelineTicker from the DOM.
                timelineTickers[i].parentElement.removeChild(timelineTickers[i]);
            }
        }

        const timeBar = document.querySelector(".TimelineBar");
        const timelineMinutes = timeBar.querySelectorAll(".TimelineMinute");

        for (let i=0; i < timelineMinutes.length; i++) {
            const minute = Number(timelineMinutes[i].getAttribute("minute"));
            if (minute && minute < ancientMinute) {
                let playerName   = timelineMinutes[i].parentElement.id;
                let workerPanels = timelineMinutes[i].querySelectorAll(".PipelinePanel");
                let player = players[playerName];

                if (!player) continue;

                let minAdmin = player.mins;
                if (minAdmin) delete minAdmin[minute];

                for (let j=0; j < workerPanels.length; j++) {
                    const workerName = workerPanels[j].getAttribute("worker");

                    let minAdmin = player.workers[workerName]
                                 ? player.workers[workerName].mins
                                 : null;
                    if (minAdmin) delete minAdmin[minute];
                }
                // Remove the whole timelineMinute from the DOM.
                timelineMinutes[i].parentElement.removeChild(timelineMinutes[i]);
            }
        }
    }

    function createMinute(playerName, time) {
        if (!playerName) {
            for (playerName in players)
                createMinute(playerName, time);
            return;
        }

        const player = players[playerName];
        if (!player) return;

        // Destroy everything older than 10 minutes ago.
        destroyAncientMinutes(10);

        let minuteWasCreated = false;
        let workerWasCreated = false;

        // We may have a timeline for this minute, but we may not have a line for every worker we now know of.

        let openingMinute  = document.getElementById(player.name + "_opening");
        if (!openingMinute) {
            openingMinute = document.createElement("div");
            openingMinute.id = player.name + "_opening";
            openingMinute.className   = "TimelineMinute Opening";
            openingMinute.innerHTML = '<span class="icon-stats-bars2" style="position:absolute; font-size:12px;" onclick="showPlayerHealth(event)" data-player="' + player.name + '" data-worker="' + player.engine + '"> </spann>';
            openingMinute.style.width = "";

            openingMinute.setAttribute("engine", player.engine);   // For visual identification of row.
            openingMinute.setAttribute("worker", player.name);     // For visual identification of row.

            insertSorted(player.opening, ".TimelineMinute", openingMinute, "minute");

            player.mins["opening"] = openingMinute;                // To prove that we create this minute.
        }

        let timelineMinute = document.getElementById(player.name + "_" + time);
        if (time && !timelineMinute) {
            // First, create a minute segment at the tope that shows the time.
            let timelineTicker = document.getElementById("Time_" + time);
            if (time && !timelineTicker) {
                timelineTicker = document.createElement("span");
                timelineTicker.id = "Time_" + time;
                timelineTicker.className = "TimelineTicker";
                timelineTicker.innerHTML = "<span>" + new Date(time).toLocaleTimeString()+ "</span>";
                timelineTicker.setAttribute("minute", time);

                let timeAxis = document.getElementById("timeAxis");
                insertSorted(timeAxis, ".TimelineTicker", timelineTicker, "minute");

                const timeline = document.getElementById("timeline");
                timeAxis.style.width = timeline.offsetWidth + "px";
                timelineTicker.scrollIntoView({ behavior: "smooth", inline: "end" });
            }

            // Next, create the timeline that will host the notes.
            timelineMinute = document.createElement("div");
            timelineMinute.id = player.name + "_" + time;
            timelineMinute.className   = "TimelineMinute";
            timelineMinute.style.width = fullWidth + "px";

            timelineMinute.setAttribute("minute", time);            // For proper insertion on timeline.

            insertSorted(player.timeline, ".TimelineMinute", timelineMinute, "minute");

            player.mins[time] = timelineMinute;                  // To prove that we create this minute.
            minuteWasCreated  = true;
        }

        Object
        .keys(player.workers)
        .sort(sorter)
        .forEach(workerName => {
            const worker = player.workers[workerName];

            // Create the bar for the current time and track if new ones were added.
            workerWasCreated = drawWorker(player, worker, time) || workerWasCreated;
        });

        // Let the page know we need more space if new morkers were created.
        if (workerWasCreated)
            window.parent.postMessage(
                {   purpose: "size",
                    name:    player.engine,
                    height:  document.body.scrollHeight + 52
                },  "*");

        // Let the page know that we added a minute to the timeline, if we did.
        if (minuteWasCreated)
            window.parent.postMessage(
                {   purpose: "minute",
                    name:    player.engine,
                    minute:  time
                },  "*");

        return timelineMinute;
    }

    function createOpeningRow(firstWorker) {

        let openingBar = document.querySelector(".OpeningBar");
        if (!openingBar) {
            openingBar = document.createElement("div");
            openingBar.className = "OpeningBar";
            document.getElementById("opening").appendChild(openingBar);
        }

        let timelineBar = document.querySelector(".TimelineBar");
        if (!timelineBar) {
            timelineBar = document.createElement("div");
            timelineBar.className = "TimelineBar";
            document.getElementById("timeline").appendChild(timelineBar);
        }

        for (let playerName in players) {
            let player = players[playerName];

            if (knowns && Object.keys(player.workers).length !== knowns.length)
                console.log("Pipelines for some workers are missing ");

            let playerOpeningRow = document.getElementById(player.name);
            if (!playerOpeningRow) {
                playerOpeningRow = document.createElement("div");
                playerOpeningRow.id = player.name;
                playerOpeningRow.className = "PlayerRow Opening";

                openingBar.appendChild(playerOpeningRow);
                player.opening = playerOpeningRow;

                let playerTimelineRow = document.createElement("div");
                playerTimelineRow.id = player.name;
                playerTimelineRow.className = "PlayerRow";
                player.timeline = playerTimelineRow;

                timelineBar.appendChild(playerTimelineRow);

                /*
                const initMinuteWorkers = firstWorker
                                        ? [ player.workers[firstWorker] ]
                                        : Object.values(player.workers);

                initMinuteWorkers.forEach(worker => {
                    createMinute(player, worker.next, worker.titleGenerator);
                }); */
            }
        }
    }

    function getDrawPanel(panel, type, time, tooltipGenerator) {
        if (!panel[type]) {
            panel[type] = {};
        }

        let drawPanel = panel[type][time];
        if (!drawPanel) {
            drawPanel = panel[type][time] = panel.renderer.group();

            drawPanel.node.setAttribute("cycle", time);

            drawPanel.node.onmouseover = function (evt) {
                if (evt.currentTarget === drawPanel.node)
                    showTooltip(evt, tooltipGenerator);
            };
            drawPanel.node.onmouseout = function (evt) {
                if (evt.currentTarget === drawPanel.node)
                    hideTooltip(evt);
            };
        }
        return drawPanel;
    }

    function hideTooltip(evt) {
        if (evt.currentTarget.lastElementChild &&
            evt.currentTarget.lastElementChild.tagName !== "circle")
            evt.currentTarget.lastElementChild.setAttribute("fill-opacity", "0.01");

        var panel  = evt.currentTarget.parentElement.parentElement.parentElement;
        var player = panel.getAttribute("tenant") || panel.getAttribute("player");

        window.parent.postMessage(
            {   purpose: "hideTooltip",
                name: players[player].engine
            },  "*");
    }

    function showTooltip(evt, htmlGenerator) {
        if (evt.currentTarget.lastElementChild &&
            evt.currentTarget.lastElementChild.tagName !== "circle")
            evt.currentTarget.lastElementChild.setAttribute("fill-opacity", "0.25");

        var panel  = evt.currentTarget.parentElement.parentElement.parentElement;
        var cycle  = evt.currentTarget.getAttribute("cycle");
        var minute = panel.getAttribute("minute");
        var player = panel.getAttribute("tenant") || panel.getAttribute("player");
        var worker = panel.getAttribute("worker");

        if (!minute || !player || !cycle) return;

        let cycles = players[player].workers[worker].mins[minute].cycles;
        let pipeline = cycles[cycle];
        if (!pipeline) return;

        const rect = evt.target.getBoundingClientRect();

        window.parent.postMessage(
            {   purpose: "showTooltip",
                name: players[player].engine,
                element: {
                    x : rect.left + window.scrollX,
                    y : rect.top  + window.scrollY,
                    h : rect.height,
                    w : rect.width
                },
                offset: {
                    x : -12,
                    y : 10
                },
                max: {
                    h : 400,
                    w : 400
                },
                position: "center-below",
                contents: htmlGenerator(pipeline),
                notch:    true
            },  "*");
    }

    return {
        getStartOfMinute: getStartOfMinute,
        addPlayer: addPlayer,
        hasPlayer: hasPlayer,
        getPlayer: getPlayer,
        getPlayers: getPlayers,
        invalidateWorker: invalidateWorker,
        revalidateWorker: revalidateWorker,
        addWorker: addWorker,
        getWorker: getWorker,
        setCycle: setCycle,
        updateCycle: updateCycle,
        createMinute: createMinute,
        createOpeningRow: createOpeningRow,
        getDrawPanel: getDrawPanel,
        showTooltip: showTooltip,
        hideTooltip: hideTooltip
    };
}