# Source: Ticker.js



----------
```javascript
'use strict';

const { Readable } = require('stream');

 /**
  * The Ticker that drives the frequency ofpipeline executons.
  * @private
  */
class ReadableTicker {

    constructor(options) {
        this.curator   = options.curator;
        this.consumer  = options.consumer;
        this.logger    = options.logger;
        this.tracker   = options.tracker || { progress: () => {} };  // For live diagnostic messages.
        this.trigger   = null;
        this.lastRun   = null;
        this.nextRun   = null;
        this.isPaused  = false;
        this.isFlowing = false;

        this.plannedRunTime = (new Date()).getTime();
    }

    /**
     * Our AsyncIterator allows us to use this in Readable.from() to turn this into
     * a Readable that NodeJS's pipeline can use.
     */
    [Symbol.asyncIterator]() {
        const self = this;

        return {
            next() {
                return new Promise(resolve => {
                    // We'll call 'resolve' at a time of _our_ choosing...
                    self.scheduleNext(resolve);
                });
            }
        };
    }
    scheduleNext(trigger) {
        const self = this;
        const minDelay = 100;

        if (this.isPaused) {
            // This covers a weird edge case where we are called while paused.
            this.trigger = this.trigger || trigger;
            return;
        }
        else {
            this.trigger = trigger;
        }

        /*  We trigger the next execution in the pipeline at the planned time, by calling 'resolve' in
            a scheduled run. The planned time is (plannedRunTime - now) milliseconds from now - or in
            1ms if we missed it. We keep advancing at our intended rate until the next execution is in
            the future.
         */
        const now         = (new Date()).getTime();
        const prevRunTime = this.plannedRunTime || now;
        const runDelay    = this.plannedRunTime &lt;= now
                          ? 1 // ms
                          : this.plannedRunTime  - now;
        const runInterval = Math.ceil(this.curator.cycleTime / (this.consumer.runCount() || 1));

        do {    // Schedule the next run, skipping intervals we missed already.
            this.plannedRunTime += runInterval;
        } while (this.plannedRunTime &lt; (now + minDelay));

        this.tracker.progress('ticker', {
            stage:     "plan",
            cycle:     this.plannedRunTime,
            timestamp: this.plannedRunTime,
            tenant:    this.consumer.tenant,
            type:      this.consumer.type,
            data:      null
        });

        this.nextRun = setTimeout(() => {
            this.nextRun = null;
            if (!this.trigger) return;

            this.tracker.progress('ticker', {
                stage:     "fire",
                cycle:     prevRunTime,
                timestamp: (new Date()).getTime(),
                tenant:    this.consumer.tenant,
                type:      this.consumer.type,
                data:      null
            });

            this.trigger({
                done:  false,
                value: this.consumer.getNext(),
                cycle: prevRunTime
            });
            this.trigger  = null;           // Never call this twice!
        }, Math.max(runDelay, minDelay));   // Breathing space for others.
    }
    pause() {
        if (this.isPaused) return;

        // Cancel the next execution. We reschedule on 'resume'.
        clearTimeout(this.nextRun);
        this.producer.pause();
        this.isPaused = true;
        this.pausedAt = (new Date()).getTime();

        this.tracker.progress('ticker', {
            stage:     "pause",
            cycle:     this.pausedAt,
            tenant:    this.consumer.tenant,
            type:      this.consumer.type,
            data:      null
        });
    }
    resume() {
        if (!this.isPaused) return;

        clearTimeout(this.nextRun); // Just to be sure...

        this.isPaused = false;
        this.producer.resume();

        // To get started immediately, we state that we should have run already.
        // We use the still-unused 'resolve' from the last iteration.
        this.plannedRunTime = (new Date()).getTime();
        this.scheduleNext(this.trigger);

        this.tracker.progress('ticker', {
            stage:     "resume",
            resumed:   (new Date()).getTime(),
            cycle:     this.pausedAt,
            tenant:    this.consumer.tenant,
            type:      this.consumer.type,
            data:      null
        });
    }
    async start(delay) {
        // Once started, never start again - not even after stopping.
        if (this.producer) throw new Error("A stopped pipeline cannot be restarted");

        // Wait for the specified amount of time.
        await new Promise(resolve => setTimeout(resolve, delay || 1));

        // Starting will happen automatically as soon as we construct the pipeline.
        this.plannedRunTime = (new Date()).getTime();    // Fire on the first iteration.

        // The below is supported in NodeJS 12.4+.
        const self = this;
        const iterator = this[Symbol.asyncIterator]();
        this.producer = Readable.from(
            async function * () {
                let nextRun = await iterator.next();
                self.lastRun = new Date();
                yield nextRun;
            }(), {
                objectMode: true,
                highWaterMark: 1
            }
        );
        // This is our NodeJS 10.x backfill.
        /* this.producer  = new Readable({
            objectMode: true,
            highWaterMark: 1,
            read: async function () {
                try {
                    // Wait for the next trigger to fire.
                    let nextRun = await iterator.next();
                    self.lastRun = new Date();

                    // Push next run info to the output.
                    this.push(nextRun);
                    // Streaming logic will call us again.
                }
                catch (ex) {
                    self.logger.warn("Write after end by Ticker");
                }
            }
        }); */

        this.consumer.start(this);          // First let downstream pipeline connect up.
        this.producer.pipe(this.consumer);  // Now we connect to the downstream pipe.

        this.tracker.progress('ticker', {
            stage:     "start",
            timestamp: (new Date()).getTime(),
            tenant:    this.consumer.tenant,
            type:      this.consumer.type,
            data:      null
        });

        this.isFlowing = true;
        return true;
    }
    stop() {
        if (!this.isFlowing) return;

        if (this.producer) {
            clearTimeout(this.nextRun); // Cancel next scheduled production.
            this.producer.pause();      // Prevent asking for any more data.
            this.producer.unpipe();     // Disconnect the consumer.
            this.consumer.stop();       // Will cleanup and call 'end' on itself.
            this.producer.destroy();    // Now we're totally done with this pipe.
        }

        this.tracker.progress('ticker', {
            stage:     "stop",
            timestamp: (new Date()).getTime(),
            tenant:    this.consumer.tenant,
            type:      this.consumer.type,
            data:      null
        });

        this.isFlowing = false;
    }
    getDefinition() {
        return {
            tenant:    this.consumer.tenant,
            type:      this.consumer.type,
            state:     this.isPaused ? "paused" : this.consumer &amp;&amp; this.consumer.consumer ? "running" : "stopped",
            nextRun:   this.plannedRunTime,
            interval:  Math.ceil(this.curator.cycleTime / (this.consumer.runCount() || 1))
        };
    }
    getStatistics() {      
        return {
            tenant:     this.consumer.tenant,
            type:       this.consumer.type,
            count:      this.consumer.getEntities().size,
            state:      this.isPaused ? "paused" : this.consumer &amp;&amp; this.consumer.consumer ? "running" : "stopped",
            strategy:   this.consumer.constructor.name,
            chunkType:  this.consumer.crit || this.consumer.chunking,
            chunkCount: this.consumer.values ? this.consumer.values.size : 1,
            fullCycle:  Math.round(this.consumer.getCycleTime() / 1000),
            avgChunkTime: this.consumer.avgChunkTime,
            avgChunkSize: this.consumer.avgChunkSize, 
            failedChunks: this.consumer.failedChunks,
            totalChunks:  this.consumer.totalChunks,
            lastRun:    this.lastRun ? this.lastRun.toLocaleTimeString() : "(N/A)",
            lastResult: this.consumer.lastResult,
            nextRun:    Math.round((this.plannedRunTime - (new Date()).getTime()) / 1000),
            isBusy:     this.producer &amp;&amp; this.producer.readableLength !== 0
        }
    }
    isAlive() {
        /*  When are we still alive?
            1)  We've started flowing and that hasn't stopped yet AND any of the following is true:
            2a) We're paused.
            2b) A next run is scheduled. It will fire! But a next run will not be scheduled until the
                current one completes. So, it may be null if we're still processing the current run.
            2c) If the planned run is more recent than 120 secs in the past, we'll allow for the
                possibility that the current run is still busy but not actually blocked (yet...).
         */
        return  this.isFlowing &amp;&amp;
                (   this.isPaused ||
                    !!this.nextRun ||
                    this.plannedRunTime > (new Date()).getTime() - 120000
                );
    }
};

module.exports = {
    ReadableTicker:  ReadableTicker
}
```


