'use strict';

const { Transform } = require('stream');
const Ring = require('./Ring.js');

Math.sum = arr => arr.reduce((a,b) => a + b, 0);
Math.avg = arr => arr.reduce((a,b) => a + b, 0) / arr.length;

/**
 * The Collector class is an abstract class that other collectors must extend.
 * @private
 */
class Collector extends Transform {
    constructor(options) {
        super({
            readableObjectMode: true,
            writableObjectMode: true,
            highWaterMark: 1
        });

        this.apiTimeout   = (options.config.maxQueueTime || 30) * 1000,
        this.pageSize     = options.pageSize || 1000;
        this.curator      = options.curator;
        this.scheduler    = options.scheduler;
        this.tenant       = options.tenant;
        this.type         = options.type;
        this.friendlyType = options.friendlyType;
        this.dynatraceAPI = options.dynatraceAPI;
        this.endPoint     = options.endPoint;
        this.consumer     = options.consumer;
        this.config       = options.config  || {};  // For miscellaneous config options.
        this.tracker      = options.tracker || { progress: () => {} };  // For live diagnostic messages.
        this.logger       = options.logger  || {};  // Default logger. MAYBE...
        this.pipeline     = null;                   // Will be populated in 'start()'.
        this.recentResults = [];
        this.lastResult    = "(N/A)";
    }

    /*  Because of backpressure, our actual speed will vary. So, we need to keep track
        of the rate at which we get called, and estimate our effective full cycle time
        from that. The 'getCycleTime()' method uses at least the most recent 10 executions
        (up to a full cycle) calculate the average elapsed time between executions, and 
        multiplies that with the number of (expected) executions for a full cycle to 
        calculate the estimate.
     */
    addExecResult(startTime, chunkSize, hasFailed, cycleTime) {
        const now = new Date().getTime();
        this.recentResults.push({ 
            time:      startTime, 
            duration:  now - startTime,
            size:      chunkSize,
            hasFailed: hasFailed
        });

        // Discard everything older than the cycle time but keep at least 10.
        const discardTime = now - cycleTime;
        while (this.recentResults.length > 10 && 
            this.recentResults[0].time < discardTime)
            this.recentResults.shift();
    }

    get avgChunkTime() {
        return Math.sum(this.recentResults.filter(r => !r.hasFailed).map(r => r.duration)) / (this.runCount() || 1);
    }
    get avgChunkSize() {
        return Math.sum(this.recentResults.filter(r => !r.hasFailed).map(r => r.size)) / (this.runCount() || 1);
    } 
    get failedChunks() {
        return this.recentResults.filter(r => r.hasFailed).length;
    }
    get totalChunks() {
        if (this.runCount() != this.recentResults.length)
            this.logger.warn("Planned number of " + this.friendlyType + " chunks in " + this.tenant + " per cycle: " + this.runCount() + ", actually retrieved: " + this.recentResults.length);

        return this.recentResults.length;
    }
    
    getCycleTime() {
        const totDistance = this.recentResults.length > 0
                          ? this.recentResults[this.recentResults.length-1].time - this.recentResults[0].time
                          : 0;
        const avgDistance = totDistance / (this.recentResults.length || 1); 
        const cycleTime   = Math.min(Math.max(avgDistance * (this.runCount() || 1), this.curator.cycleTime), 24*60*60*1000);

        return cycleTime;
    }
    getEntities() {
        return this.curator.getEntities(this.tenant, this.type);
    }

    getDiscovered() {
        return this.curator.getDiscovered(this.tenant, this.type);
    }

    getZones(startsWithValue) {
        return this.curator.getZones(this.tenant, this.type, startsWithValue);
    }

    getZoneId(name) {
        return this.curator.getZoneId(this.tenant, this.type, name);
    }

    /**
     * Retrieve a subset of entities and publish them to consumers.
     *
     * @param {Object} queryData
     * Specification of what subset of data to collect, in the form of a query string.
     *
     * @param {String} unused
     * Ignored.
     *
     * @param {Function} onDone
     * The callback function to be called when the dataset has been retrieved and published.
     */
    async _transform(runData, unused, onDone) {
        /*  The assumption is that the dynatraceAPI will always accept our requests. If it has
            to wait, it waits. That's OK. We will not call it again until the next call to
            "_transform", which can only happen after we've called 'onDone()' - with or without
            data to publish. And we hold off doing that until the request has completed and its
            results successfully published.
            Because we want the Collectors to immediately slow down when downstream components
            in the pipe cannnot keep up, we set our writeableHighWaterMark to 1. This has the
            effect of immediately building up backpressure here, which slows down successive
            calls to '_transform' to a sustainable level.
            So, the dynatraceAPI will not exert backpressure by itself. Rather, the fact that
            we may have to wait a long time for it to complete and return results, builds up
            backpressure right here. And _that_ is felt by the previous step in the pipeline.

            Due to influence of backpressure, we continuously track how long we probably need
            to make a full cycle, at this point. That gives us the appropriate relative time -
            the point in time (and after that - hence, 'since') for which the API reports
            entities. What we pass is a Dynatrace API-supported value, ceiled if necessary.
         */
        if (!runData || !runData.value) {
            this.tracker.progress('collector', {
                stage:     "warn",
                cycle:     runData ? runData.cycle : new Date().getTime(),
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                warning:   "Syncing " + this.friendlyType + "s in " + this.tenant + ": no data for this run",
                data:      null
            });

            onDone();    // Nothing to produce.
            return;
        }

        const queryData  = runData.value;
        const cycle      = runData.cycle;

        try {
            const startTime = (new Date()).getTime();

            this.tracker.progress('collector', {
                stage:     "begin",
                cycle:     cycle,
                timestamp: startTime,
                data:      queryData
            });

            const cycleTime    = this.getCycleTime();
            const relativeTime = this.scheduler
                                     .getGranularityValue(Math.max(cycleTime, 60*60000), true)
                                     .substring(1);  // Remove leading '_'.

            const queryString  = (this.pageSize ? 'pageSize='     + this.pageSize + '&' : '') +
                                 (relativeTime  ? 'relativeTime=' + relativeTime  + '&' : '') +
                                 (queryData     ? queryData : '');

            this.activeRequest = this.dynatraceAPI.request(
                {   tenant:    this.tenant,
                    endPoint:  '/api/v1/' + this.endPoint,
                    queryData: queryString,
                    noReject:  true,
                    timeout:   this.apiTimeout
                },
                (error, entities) => {
                    entities = entities || [];

                    if (error) {
                        error = "Error while syncing " + this.friendlyType + "s in " + this.tenant + ": "
                              + (error.status ? error.status + " - " + error.message : error) + ". Query: "
                              + queryString;

                        this.logger.warn(error);
                        this.tracker.progress('collector', {
                            stage:     "error" ,
                            cycle:     cycle,
                            timestamp: (new Date()).getTime(),
                            pipeline:  this.pipeline,
                            data:      entities.length,
                            error:     error
                        });

                        this.lastResult = error;                        
                    }
                    else {
                        const info = "Synced " + (entities ? entities.filter(entity => entity.entityId.indexOf(this.type) === 0).length : 0) + 
                            " " + this.friendlyType + "s in " + this.tenant;

                        this.tracker.progress('collector', {
                            stage:     "end",
                            cycle:     cycle,
                            timestamp: (new Date()).getTime(),
                            pipeline:  this.pipeline,
                            data:      entities.length,
                            info:      info
                        });

                        this.lastResult = info + " at " + (new Date()).toLocaleTimeString();                        
                    }
    
                    // We consumed the data and this is what we produced in response.
                    onDone(null, {
                        cycle: cycle,
                        data: entities
                    });     

                    this.addExecResult(startTime, entities.length, !!error, cycleTime);
                }
            );

            this.activeRequest.on("error", error => {
                this.tracker.progress('collector', {
                    stage:     "error",
                    cycle:     cycle,
                    timestamp: (new Date()).getTime(),
                    pipeline:  this.pipeline,
                    error:    "Error while retrieving " + this.friendlyType + "s in " + this.tenant + ": " +
                        (error.status ? error.status + " - " + error.message : error) + ". Query: " +
                        queryString
                });
            });

            this.activeRequest.on("progress", stage => {
                this.tracker.progress('collector', {
                    stage:     stage,
                    cycle:     cycle,
                    timestamp: (new Date()).getTime(),
                    pipeline:  this.pipeline
                });
            });
        }
        catch (ex) {
            const error = this.constructor.name + " failed: " + ex.message;

            this.tracker.progress('collector', {
                stage:     "error",
                cycle:     cycle,
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                error:     error
            });

            onDone(); // The show must go on.
        }
    }

    start(pipeline) {
        this.pipeline = pipeline;           // Track the pipeline we are part of.
        if (!this.consumer) return;         // If we have no consumer, then never mind.
        this.consumer.start(pipeline);      // First let downstream pipeline connect up.
        this.pipe(this.consumer);           // Now we connect to the downstream pipe.

        this.tracker.progress('collector', {
            stage:     "init",
            timestamp: (new Date()).getTime(),
            data:      null
        });
    }

    stop() {
        if (this.activeRequest)
            this.activeRequest = this.activeRequest.cancel();

        if (!this.consumer) return;         // If we have no consumer, then never mind.
        this.consumer.stop();               // Tell it to stop.
        this.unpipe(this.consumer);         // Disconnect it from the downstream pipe.
        this.end();
        this.consumer.destroy();
        this.consumer = null;
    }
}


/**
 * The LibraryCollector syncs with another Library instance.
 *
 *   In order to hit the ground running when we take over, we will need to receive
 *   continuous updates from the primary instance's embedded library (which is a
 *   copy of ourselves) with respect to all the changes that are happening in it.
 *
 *   The approach is that is that we ask for a list of known entities of a type,
 *   with checksums to efficiently check if entities (by entityId) that we know of
 *   have changed. We can then ask for fresh, full copies of the things that have
 *   changed or that appear new, and delete entities the primary library did not
 *   report. Then on an ongoing basis, we can ask for updates and deletes since
 *   some timestamp.
 * @private
 */
class LibraryCollector extends Collector {
    constructor(options) {
        super(options);

        this.primary = options.primary;

        /*  The ReadableTicker uses the curator.cycleTime/consumer.runCount() to
            determine the next execution. We want to poll for updates/deletes
            every X mins (whatever is configured). So, we will report our runCount
            as "cycleTime/Xmins".
         */
        this.requiredRunCount = Math.round(this.curator.cycleTime / (options.pollTime*60*1000));
    }

    /*  We override the super class' "_transform()" method be cause we won't be
        reaching out to the Dynatrace API. Instead, we will connect to another
        Librray instance with a different set of requests.
     */
    async _transform(runData, unused, onDone) {
        /*  We assume that this Library instance has already loaded all entities from its
            index files or from the database, so it starts with a reasonably up-to-date admin.

            1)  First-time run since this library started (first call to 'getNext()'):
                a) Ask for a list of all host IDs and their checksums that the primary instance
                    knows of (using '&select=entityId,checksum). Keep the timestamp of this first
                    request for use in the next request.
                b) For things that are new to us or that have different checksums: Ask for full
                    entities wrapped in proxies (using '&select=entityId,checksum,lastUpdated,entity').
                c) Add to the 'deletions' list the IDs of entities we know of but that weren't included.
            2)  On an ongoing basis (subsequent calls to 'getNext()'):
                a) Ask for FULL entities that have changed or that are new (using '&since...').
                    We use '&metadata=entityId,checksum,lastUpdated,entity' to the request to
                    receive entites wrapped in a proxy. In that way, we get the checksum and the
                    lastUpdate info as well.
                b) Ask for recent deletions (using '&since...&metadata=entityId') at a distinct endpoint.
            3)  All calls to 'getNext()':
                a) For 'updated' entities, check against the entities the curator has.
                    If we have it already and the checksum we received matches the one we
                    have for it, we can ignore it.
                    Otherwise, add it to the list of things to return for this cycle.
                b) Upon receipt of everything, we return an object with updates and deletes.
                    The updates are unindexed proxies that contain full entity definitions.
                    For deleted entities, we simply include an ID list (if we have it).
         */
        if (!runData || !runData.value) {
            this.tracker.progress('collector', {
                stage:     "warning",
                cycle:     runData ? runData.cycle : new Date().getTime(),
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                warning:   this.constructor.name + ": no data for this run",
                data:      0
            });

            onDone();
            return;
        }

        const cycleTime = this.curator.cycleTime;
        const queryData = runData.value;
        const cycle     = runData.cycle;    // This is the next 'since' value.

        try {
            const now = (new Date()).getTime();

            this.tracker.progress('collector', {
                stage:     "begin",
                cycle:     cycle,
                timestamp: now,
                data:      queryData
            });

            const relativeTime = this.lastTimestamp;
            const queryString  = (relativeTime  ? 'since=' + relativeTime  + '&' : '') +
                                 (queryData     ? queryData : '');

            // The dynatraceAPI we get here is actually an instance of the 'LibraryAPI' - it has the same interface.
            this.activeRequest = this.dynatraceAPI.request(
                {   tenant:    this.primary,
                    endPoint:  this.endPoint,
                    queryData: queryString,
                    noReject:  true,
                    timeout:   this.apiTimeout
                },
                (error, entities) => {
                    if (!error)
                        this.tracker.progress('collector', {
                            stage:     "end",
                            cycle:     cycle,
                            timestamp: (new Date()).getTime(),
                            pipeline:  this.pipeline,
                            data:      entities ? entities.length : null
                        });

                    if (!relativeTime) {
                        /*  First-time run since this library started (first call to 'getNext()'):
                            a) We asked for a list of all host IDs and their checksums that the primary instance
                                knows of (using '&select=entityId,checksum). Keep the timestamp of this first
                                request for use in the next request.
                            b) For things that are new to us or that have different checksums: Ask for full
                                entities wrapped in proxies (using '&select=entityId,checksum,lastUpdated,entity').
                            c) Add to the 'deletions' list the IDs of entities we know of but that weren't included.
                         */
                        const knownEntities = this.getEntities();
                        const unconfirmeds  = knownEntities.ids;
                        const newOrUpdated  = [];

                        for (let i=0; i < (entities || []).length; i++) {
                            const entity = knownEntities.get(entities[i].entityId);

                            if (entity) {
                                unconfirmeds.remove(entities[i].entityId);
                            }
                            if (!entity || entity.checksum !== entities[i].checksum) {
                                newOrUpdated.push(entities[i].entityId);
                            }
                        }

                        if (newOrUpdated.length === 0) {
                            onDone(null, {
                                cycle: cycle,
                                data:  { updates: [], deletes: unconfirmeds }
                            });

                            this.addExecResult(startTime, 0, false, cycleTime);
                        }
                        else
                            this.dynatraceAPI.request(
                            {   tenant:    this.primary,
                                endPoint:  this.endPoint,
                                queryData: null,
                                postData:  JSON.stringify([{
                                    context:  "VALUE",
                                    key:      "entityId",
                                    values:   newOrUpdated.join(","),
                                    operator: "="
                                }]),
                                noReject:  true,
                                timeout:   this.apiTimeout
                            },
                            (error, entities) => {
                                entities = entities || [];

                                onDone(error, {
                                    cycle: cycle,
                                    data:  { updates: entities, deletes: unconfirmeds }
                                });

                                this.addExecResult(startTime, entities.length, !!error, cycleTime);
                            });
                    }
                    else {
                        /*  On an ongoing basis (subsequent calls to 'getNext()'):
                            a) We asked for FULL entities that have changed or that are new (using '&since...').
                                We added '&select=entityId,checksum,lastUpdated,entity' to the request to
                                receive entites wrapped in a proxy. In that way, we get the checksum and the
                                lastUpdate info as well.
                            b) We also received recent deletions (using '&since...').
                         */
                        onDone(null, {
                            cycle: cycle,
                            data:  entities
                        });

                        this.addExecResult(startTime, entities.length, false, cycleTime);
                    }
                }
            );

            this.activeRequest.on("error", error => {
                error = "Error while retrieving " + this.friendlyType + "s in " + this.tenant + ": "
                      + (error && error.status ? error.status + " - " + error.message : error || "N/A") 
                      + ". Query: " + queryString;

                this.tracker.progress('collector', {
                    stage:     "error",
                    cycle:     cycle,
                    timestamp: (new Date()).getTime(),
                    pipeline:  this.pipeline,
                    error:     error
                });

                onDone(error);

                this.addExecResult(startTime, 0, true, cycleTime);
            });

            this.activeRequest.on("progress", stage => {
                this.tracker.progress('collector', {
                    stage:     stage,
                    cycle:     cycle,
                    timestamp: (new Date()).getTime(),
                    pipeline:  this.pipeline
                });
            });
        }
        catch (ex) {
            const error = this.constructor.name + " failed: " + ex.message;

            this.tracker.progress('collector', {
                stage:     "error",
                cycle:     cycle,
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                error:     error
            });

            onDone(error); // The show must go on.
        }

        // Update timestamp for the next cycle.
        this.lastTimestamp = cycle;
    }

    getNext() {
        return this.lastTimestamp
            ? "select=entityId,checksum,lastUpdated,entity"
            : "select=entityId,checksum";
    }

    runCount() {
        return this.requiredRunCount;
    }
}

/**
 * The ChunkedCollector retrieves entities of a certain type per a specified chunking criterion.
 *
 * @param {object} options - The following properties are expected:
 * - tenant
 * - type
 * - endPoint
 * - chunking
 *
 * EX: chunking: { HOST: "[ENVIRONMENT]Cloud=prod-cdc01,prod-cdc02,..." }
 * EX: chunking: { HOST: "[CONTEXTLESS]Cloud=prod-cdc01,prod-cdc02,..." }
 * EX: chunking: { HOST: "ZONE=mz1,mz2,mz3"}
 * EX: chunking: { HOST: "HOSTGROUP=mz1,mz2,mz3" } (for hosts only)
 * @private
 */
class ChunkedCollector extends Collector {
    constructor(options) {
        super(options);

        this.chunking = options.chunking || "unknown=novalue";

        let parts = this.chunking.split("=");
        this.crit = parts[0];
        if (this.crit === "ZONE" && !parts[1])
            this.values = new Ring(this.getZones())
        else if (this.crit === "ZONE" && parts[1].contains("*"))    // For now, only: "ZONE=SomePrefix*" is supported.
            this.values = new Ring(this.getZones(parts[1].replace("*", "")));
        else
            this.values = new Ring(parts[1].split(","));
    }

    /**
     * Returns the query string to obtain the next iteration of this ChunckedCollector.
     *
     */
    getNext() {
        const nextValue = this.values.next();
        if (!nextValue)
            return null;

        // console.log("Retrieving " + this.friendlyType + "s in " + this.tenant + " where " + this.crit + "= " + nextValue + " - total known: " + this.getEntities().size);

        return this.crit === "ZONE"
             ? ("managementZone=" + (this.getZoneId(nextValue) || "unavailable"))
             : this.crit === "HOSTGROUP"
             ? ("hostGroupName=" + nextValue)
             : ("tag=" + this.crit + ":" + nextValue);
    }

    runCount() {
        return this.values.length;
    }
}

/**
 * The SimpleCollector retrieves all entities of a certain type at once.
 * @private
 */
class SimpleCollector extends Collector {
    constructor(options) {
        super(options);

        this.chunking       = options.chunking || "ALL";
        this.includeDetails = options.chunking !== "LIST";
    }

    /**
     * Runs the next iteration of this SimpleCollector.
     *
     * @returns {boolean} Whether the Collector is ready for the next call to this method.
     * If it returns 'false', callers must immediately listen for the 'drain' event and
     * wait until it is received before calling this method again.
     */
    getNext() {
        return "includeDetails=" + this.includeDetails;
    }

    runCount() {
        return 1;   // Once paging is supported, we can modify this based on the response from the API.
    }
}

/**
 * The DiscoverCollector retrieves a group of entities, prioritizing newly discovered ones.
 * @private
 */
class DiscoverCollector extends Collector {
    constructor(options, idListLimit) {
        super(options);

        this.chunking  = Math.min(Number(idListLimit), Number(options.chunking));
        this.batchSize = this.chunking;
    }

    getNext() {
        /*  Retrieve entities of this type as found on the 'discovered' list.
            Those have priority. Then, add as many as will fit in the remaining
            max batch size from the already discovered list (to refresh those).
            Every entity will be (re-)appended to the entity list as they are
            (re-)loaded. Thus, we can always use 'top()' to get the stalest ones.
         */
        let newlyDiscovered = this.getDiscovered().slice(0, this.batchSize);
        let knownEntities   = this.getEntities().top(this.batchSize - newlyDiscovered.length);

        // console.log("Retrieving next " + this.batchSize + " " + this.friendlyType + "s in " + this.tenant + " - total known: " + this.getEntities().size);

        return newlyDiscovered.concat(knownEntities).map(id => "entity=" + id).join("&");
    }

    runCount() {
        /*  Here we can do something smart. If we are still discovering new items,
            we should go as fast as possible, ignoring the cycle time, until we
            stabilize. Then, we can go as fast (or slow, rather) as the rest to
            adhere to the same refresh cycle.

            The approach is this: If we are at the stable level with our entities,
            we go along with the rest, speed-wise. But if the growth rate of the
            entities we build up is larger than a certain percentage, then we will
            report a larger number of required runs. To make this determination,
            we will have to periodically check the entity size. The percentage-
            difference between previous and next will be compared against a
            threshold, the outcome if which can be used as a boolean. If set to
            'false' (unstable), we report that we want, say, 1M runs...
         */

        // This is the 'normal' number.
        return Math.ceil(this.getEntities().size / (this.batchSize || 1));
    }
}

/**
 * The PerEntityCollector retrieves a group of entities per an individual related entity.
 * @private
 */
class PerEntityCollector extends Collector {
    constructor(options) {
        super(options);

        this.chunking   = options.chunking;
        this.entityType = this.chunking.toLowerCase();  // Good enough for now. Only 'host' is supported by the DT API.

        const list  = this.curator.getEntities(this.tenant, this.chunking);
        this.values = list.getIterator();
    }

    getNext() {
        const nextValue = this.values.next();
        if (!nextValue)
            return null;

        // console.log("Retrieving " + this.friendlyType + "s in " + this.tenant + " for " + nextValue + " - total known: " + this.getEntities().size);

        return this.entityType + "=" + nextValue;
    }

    runCount() {
        return this.values.size;
    }
}


/**
 * The NoCollector is for debug purposes. It allows us to exclude retrieval of a certain class of entities.
 * @private
 */
class NoCollector extends Collector {
    constructor(options) {
        super(options);
    }

    getNext() {
        return null;
    }

    runCount() {
        return 0;
    }
}

module.exports = {
    ChunkedCollector:   ChunkedCollector,
    SimpleCollector:    SimpleCollector,
    DiscoverCollector:  DiscoverCollector,
    PerEntityCollector: PerEntityCollector,
    LibraryCollector:   LibraryCollector,
    NoCollector:        NoCollector
};