'use strict';
const fs = require('fs');
/* FUTURE
const yaml = require('js-yaml');

const YAML = {
    parse:     data => yaml.safeLoad(data),
    serialize: data => yaml.safeDump(data)
}; */

const ConfigReader = function (fileName = "config.json", rootDir = "../config") {
    // Assumption is that we are in '/server' by default.
    const configPath = rootDir + "/";
    const configFile = fileName;

    const watchers = {};
    const events = {};

    const emit = function (eventName, section, data) {
        // Call the subscribed event handlers.
        const event = events[eventName] || [];
        event.forEach(handler => handler.call(null, section, data));
    };

    const subscribe = function (eventName, handler) {
        // Add the subscription (event handler) for this event.
        events[eventName] = events[eventName] || [];
        events[eventName].push(handler);
    };

    let config = null;

    function readFileOrDefault(fileName) {
        let fileContent = null;

        try {
            // TODO: Require only works for JSON files!
            // We want to support other types, too.
            fileContent = require(fileName);
            if (!fileContent) throw new Error("FILE_NOT_FOUND");
        }
        catch (ex) {
            // Generate '<name>.json' from '<name>.default.json'.
            const parts = fileName.split(".");
            const ext = parts.pop();
            parts.push("default");
            parts.push(ext);

            fileContent = require(parts.join("."));
            fs.writeFileSync(fileName, JSON.stringify(fileContent, undefined, 4));
        }
        return fileContent;
    }

    function loadFile(property, config = {}) {
        const fileName = configPath + (config[property] || property);
        const fileContent = readFileOrDefault(fileName);
      
        config[property] = fileContent[property] || fileContent;

        if (!watchers[fileName]) {
            watchers[fileName] = watchFile(fileName, content => {
                try {
                    // Parse before assigning in case there are errors.
                    const parsedContent = JSON.parse(content);
                    config[property] = parsedContent;

                    emit("changed", property, config[property]);
                }
                catch (ex) {
                    emit("warning", fileName + " could not be parsed", ex.message);
                }
            });
        }

        return config[property];
    }

    function watchFile(fileName, onChanged) {
        let previousContent = null;

        return fs.watch(fileName, 'utf8', event => {
            const currentContent = fs.readFileSync(fileName, 'utf8');
            if (currentContent !== "" && currentContent !== previousContent) {
               previousContent = currentContent;
               onChanged(currentContent);
            }
        });
    }

    function loadSection(config, section) {
        if (typeof config[section] === "string") {
            loadFile(section, config);
        }  

        if (Array.isArray(config[section])) {
            let configAsObject = {};
            config[section].forEach(item => configAsObject[item.name] = item);
            config[section] = configAsObject;
        }
    }

    function loadConfig(content) {
        try {
            if (content) {
                // This happens when the main file has changed.
                config = JSON.parse(content);
            }
            else {
                const fileName = configPath + configFile;
                config = readFileOrDefault(fileName);

                watchFile(fileName, content => loadConfig(content));
            }

            // Backwards compatibility.
            config.Libraries = config.Libraries || config.library;  

            loadSection(config, "Tenants");
            loadSection(config, "Templates");
            loadSection(config, "Destinations");
            loadSection(config, "Libraries");
            loadSection(config, "Authorizations");

            if (content)
                emit("change", "config", config);
            else
                emit("loaded", "config", config);

            return config;  // For those that called us explicitly.
        }
        catch (ex) {
            emit("error", ex.message);
            return null;
        }
    }

    return {
        load: function (fileName) {
            if (fileName)
                return loadFile(fileName);
            else 
                return loadConfig();
        },
        on: function (eventName, handler) {
            subscribe(eventName, handler);
            return this;    // For chaining.
        }
    };
};

module.exports = ConfigReader;