const fs      = require('fs');
const express = require('express');
const webui   = express();
const http    = require('http').Server(webui);

const ConfigReader = require('./ConfigReader.js');
const DynatraceUI  = require('./DynatraceUI.js');
const OpenAPIHelper = require('./OpenAPIHelper.js');

const contentType = 'application/json';
let dynatraceUI   = null;
let openAPIHelper = null;
let server        = null;

let config = {
    configPort: 82,
    HTTPRequests: require("../config/customRequests.json")
};

let configReader = new ConfigReader();
configReader.on("loaded", newConfig => {
    config = newConfig;

    if (!config.Tenants.hasOwnProperty("INTERNAL")) {
        config.Tenants.INTERNAL = {
            host: "localhost:" + config.configPort,
            protocol: "http://",
            url:  "",
            account: "user|pass",
        }
    }

    if (dynatraceUI)    // TODO: This is old. We now have "changed" events.
        dynatraceUI.config = config;
    else
        dynatraceUI = new DynatraceUI(config);

    openAPIHelper = new OpenAPIHelper(config);

    // And finally we'll start listening for requests.
    if (server) {
        server.close();
        server.listen(config.configPort, () => {
            console.log("ConfigService is listening on port " + config.configPort);
        });
    }
    else {
        server = http.listen(config.configPort, () => {
            console.log("ConfigService is listening on port " + config.configPort);
        });
        server.on("error", err => {
            console.error("ConfigService cannot listen on port " + config.configPort + ": " + err.message);
            server.close();
        });
    }
});
configReader.on("changed", (section, config) => {
    console.log("Configuration change detected for " + section);
});
configReader.on("warning", warning => {
    console.warn("Configuration change warning: " + warning);
});
configReader.on("error", error => {
        console.error("Configuration could not be applied: " + error);
        process.exit(1);
});

webui.use('/doc/libraryapi.json', (req, res) => openAPIHelper.getLibraryAPI(req, res));
webui.use('/doc/configapi.json', (req, res) => openAPIHelper.getConfigAPI(req, res));
webui.use('/doc', express.static("../doc")); // The documentation.

webui.use("/test/:requestName", function (request, response) {
    let answer = "OOPS";
    let statusCode = 200;

    switch (request.params.requestName) {
        case "first":
            answer = {
                age:      request.query.age,
                userName: request.query.name,
                extras: { first: "Bart", last:  "Oostlander" }
            };
            break;
        case "second":
            if (request.query.pwd === "B") {
                answer = { message: "You are logged in"  }
            }
            else {
                answer = { message: "Sorry - the passowrd is 'B'." };
                statusCode = 401;
            }
            break;
        case "third":
            answer = {
                answer: request.query.greeting + " at " + request.query.age + ", " + request.query.fullName
            };
            break;
        default:
            statusCode = 404;
            break;
    }

    response.statusCode = statusCode;
    response.setHeader('Content-Type', contentType);
    response.write(JSON.stringify(answer));
    response.end();
});

// Access to Dynatrace UI URLs is exposed at /config and /custom.
// Optionally, user may call /login/Tenant?user=...&pwd=... first to use own credentials.
webui.use("/login/:tenantKey", function (request, response) {
    dynatraceUI
    .login(request.params.tenantKey, request.query.user, request.query.pwd)
    .then(confirmation => {
        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(confirmation));
        response.end();
    })
    .catch(err => {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(err));
        response.end();
    });
});

webui.use("/logout/:tenantKey", function (request, response) {
    dynatraceUI
        .logout(request.params.tenantKey)
        .then(confirmation => {
            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(confirmation));
            response.end();
        })
        .catch(err => {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(err));
            response.end();
        });
});

webui.use("/custom/:tenantKey/:requestName", function (request, response) {
    if (request.params.requestName === "login") {
        dynatraceUI
        .login(request.params.tenantKey, request.query.user, request.query.pwd)
        .then(confirmation => {
            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(confirmation));
            response.end();
        })
        .catch(err => {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(err));
            response.end();
        });
    }
    else {
        dynatraceUI
        .login(request.params.tenantKey, request.query.user, request.query.pwd)
        .then(res => {
            // Get the URL associated with the request name.
            dynatraceUI.scriptedRequest(
                request.params.tenantKey,
                config.HTTPScripts[request.params.requestName],
                request.url.split("?")[1],
                true) // skip login - already done.
            .then(result => {
                response.statusCode = 200;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify(result));
                response.end();
            })
            .catch(error => {
                // Copy select things from the error. Not all of its properties are useful.
                const sanitizedError = {
                    message:  error.message,
                    request:  error.config,
                    response: error.response ? {
                        status:     error.response.status,
                        statusText: error.response.statusText,
                        headers:    error.response.headers,
                        data:       error.response.data
                    } : "N/A",
                    next: error.config.next
                }
                delete sanitizedError.request.next;
                delete sanitizedError.request.transformRequest;
                delete sanitizedError.request.transformResponse;

                response.statusCode = 500;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify(sanitizedError, undefined, 4));
                response.end();
            });
        })
        .catch(err => {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({ error: "Could not log in: " + err }));
            response.end();
        });
    }
});

webui.use("/config/:tenantKey/:entityType/:setting", function (request, response) {
    if (request.params.entityType === "login") {
        dynatraceUI
        .login(request.params.tenantKey, request.query.user, request.query.pwd)
        .then(confirmation => {
            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(confirmation));
            response.end();
        })
        .catch(err => {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(err));
            response.end();
        });
    }
    else {
        if (request.query.id) {
            dynatraceUI
            .login(request.params.tenantKey, request.query.user, request.query.pwd)
            .then(res => {
                // Take the 'id' parameter from the query string and remove it from there.
                // The request will be submitted to the indicated Dynatrace REST endpoint.
                let idList = request.query.id.split(",");
                delete request.query.id;

                Promise
                .all(idList.map(id =>
					dynatraceUI.request(
						request.params.tenantKey,
						request.params.entityType.toUpperCase(),
						request.params.setting,
						id,
						request.query,
						true) // skip login - already done.
					.then(result => new Promise((resolve, reject) => resolve(result)))
                    .catch(error => new Promise((resolve, reject) => resolve(error)))
                ))
                .then(result => {
                    response.statusCode = 200;
                    response.setHeader('Content-Type', contentType);
                    response.write(JSON.stringify(result));
                    response.end();
                })
                .catch(error => {
					// This will never happen because all individual errors are caught
					// and all Promises wil always 'resolve'.
                    response.statusCode = 500;
                    response.setHeader('Content-Type', contentType);
                    response.write(JSON.stringify(error, undefined, 4));
                    response.end();
                });
            })
            .catch(err => {
                response.statusCode = 500;
                response.setHeader('Content-Type', contentType);
                response.write("Could not log in: " + err);
                response.end();
            });
        }
        else {
            response.statusCode = 400;
            response.setHeader('Content-Type', contentType);
            response.write("Missing 'id' parameter");
            response.end();
        }
    }
});
