'use strict';

const async         = require('async');
const WorkerProxy   = require('./WorkerProxy.js');
const Scheduler     = require('./Scheduler.js');
const ServiceBroker = require('./ServiceBroker.js');
const Diagnostics   = require('./Diagnostics.js');

Math.avg = (...args) => args.reduce((a,b) => a + b, 0) / args.length;

/*  NOTES

    Users can define jobs as a management zone plus tags, or as a 
    specific list of entities (IDs). The first option is not implemented
    yet. We will deal with that in the following manner:

    We will maintain an Entity Cache process. It analyzes all jobs, and 
    for each it retrieves the properties for the entities it contains.
    If a job contains an entity list, it is easy. We have all relevant
    properties for each entity already (includng tags). We should still
    periodically refresh though. It would be ideal if, for each entity,
    the criteria by which it was brought in would be captured. That makes
    retrieving in bulk easier.

    If a job contains tags, use the MZ (a mandatory part of such jobs)
    to retrieve the list of matching entities. We will do our own tag
    filtering, with the 'OR' and 'NOT' capabilities we have. 
    NOTE: 
    Host groups are a separate query parameter. We may be able to use that.
    Also note that each entity has a 'Management Zone' property, which
    indicates in which MZs the entity occurs.

    Every job as assigned to a Worker will consist of an entity list. That 
    allows us to split and distribute the work better. The Workers will ask 
    the Entity Cache process for entity properties (including tags!). It  
    will do so, inevitably, in an asynchronous manner. That also allows the 
    Entity Cache to update its cache if necessary.

    The Entity Cache periodically refreshes itself. The Controller always
    asks the Entity Cache for an entity-list-oriented accounting of the
    jobs. It will then proceed as ususal.
 */

if (!Array.prototype.getById)
    Array.prototype.getById = function (id) {
        for (var i=0; i < this.length; i++)
            if (this[i].id == id)
                return this[i];
        return null;
    };

if (!Array.prototype.removeById)
    Array.prototype.removeById = function (id) {
        for (var i=0; i < this.length; i++)
            if (this[i].id == id)
                return this.splice(i, 1)[0];
        return null;
    };

const Controller = function (config, socket, logger, runOption) {
    const self = this;

    let diagnostics     = null;
    let scheduler       = null; // Determines what jobs need to be performed.
    let libraryProxy    = null; // Bootstrap object to start a Library process.
    let rosteredWorkers = [];   // Workers available for work - busy or not.
    let retiringWorkers = [];   // Workers still doing some work before dying.
    let lastMinute      = 0;    // The minute-value of the last cycle.
    let previousTick    = 0;    // Timestamp of last kicked off cycle.
    let pendingJobList  = null; // Holds joblist until the scheduler is ready.

    this.events  = {};

    const subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);
    };

    const emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const eventHandlers = self.events[eventName] || [];
        eventHandlers.forEach(fn => { fn.call(null, data); });
    };

    function getHealthMetrics(request, response) {
        return new Promise((resolve, reject) => {
            const now = new Date();
            now.setSeconds(0, 0);

            const healthMetrics = { 
                timestamp: now,
                events:  [],
                metrics: []
            };

            // Get cycle metrics for previous minute.
            const cycleInfo = scheduler.getLastCycle(now.getTime() - 60*1000);
            if (cycleInfo) {
                Object.keys(cycleInfo.metrics).forEach(metric => {
                    healthMetrics.metrics.push({
                        metricId: metric, 
                        dimensions: null,
                        value: cycleInfo.metrics[metric]
                    });
                });
            }
        
            // Get metrics from the libraries.
            Promise.all(libraryProxy.getHealthMetrics())
            .then(results => {
                results.forEach(result => {
                    healthMetrics.events  = healthMetrics.events.concat(result.events);
                    healthMetrics.metrics = healthMetrics.metrics.concat(result.metrics);
                })
                resolve(healthMetrics);
            })
            .catch(err => {
                reject(err);
            });
        });
    }

    function onWorkerDied(id) {
        rosteredWorkers.removeById(id);
        retiringWorkers.removeById(id);
    };

    function createWorker(i, onWorkerReady) {
        var worker = new WorkerProxy(config, scheduler.tuning, socket, logger);
    
        worker.on("ready", worker => onWorkerReady(worker ? null : "error during initialization", worker))
              .on("done", onJobDone)
              .on("died", onWorkerDied);
    };
    
    function removeWorkers(workers) {
        workers.forEach(worker => {
            rosteredWorkers.removeById(worker.id); 
        });
    }

    function retireWorkers(retirees) {
        /*  We will let these workers finish any work they may be doing,
            but once they are done with that, they will die. If we get
            a number, we'll take the oldest X workers.
         */
        if (typeof retirees == "number") {
            if (retirees > 0)
                logger.info("We could retire " + retirees + " workers, but we we'll keep them on standby");
            return 0;
            retirees = retirees > 0 
                     ? rosteredWorkers.slice(0, retirees)
                     : [];
        }
        retirees.forEach(worker => {
            var retiree = rosteredWorkers.removeById(worker.id);
            if (retiree) {
                logger.info("Retiring worker " + retiree.id);
                retiringWorkers.push(retiree);
                retiree.retire();
            }
        });
        return retirees.length;
    }
    
    function retireAllWorkers() {
        rosteredWorkers.forEach(worker => {
            retiringWorkers.push(worker);
            worker.retire();
        });
        rosteredWorkers = [];
    }

    function replenishWorkers(suggestedCapacity) {
        // Create new workers if required, but not beyond the maximum allowed.
        let workersToCreate = suggestedCapacity > rosteredWorkers.length 
                            ? Math.min(suggestedCapacity - rosteredWorkers.length,   // EG: req: 10 - 2 avb = 8 workers
                                       config.maxWorkers - rosteredWorkers.length)   // EG  max:  8 - 2 avb = 6 workers
                            : 0;

        logger.info(rosteredWorkers.length + " workers are available. Creating " + workersToCreate 
                    + " new workers (max is " + config.maxWorkers + ")");

        return new Promise((resolve, reject) => {
            if (workersToCreate == 0) {
                setTimeout(resolve, 1);
            }
            else {
                async.times(            
                    workersToCreate,   
                    createWorker, 
                    function(err, readyWorkers) {
                        if (err)
                            logger.error("    An error occurred while creating workers: " + err);
                        else     
                            rosteredWorkers = rosteredWorkers.concat(readyWorkers.filter(worker => !!worker));
                        
                        resolve();
                    }
                );
            }
        });
    }

    async function startNextCycle(tick) {
        let cycleInfo = null;

        try {
            // Assess and remediate the damage incurred in the previous cycle.
            let { retirees, fallen } = scheduler.evaluateCycle(previousTick);

            retireWorkers(retirees); 
            removeWorkers(fallen);

            // Create a new cycle and initialize its administration.
            cycleInfo = scheduler.createNewCycle(tick);
            if (!cycleInfo)
                throw new Error("Administration for cycle " + tick + " could not be created");

            // Tell any browsers that are listening.
            socket.emit('cycle', cycleInfo);

            // Collect the jobs we need to execute for this cycle.
            let { jobs, capacity, errors, sizes } = await scheduler.collectJobs(cycleInfo);

            if (errors) {
                // Add the errors to the error log.
                cycleInfo.log.errors.splice(cycleInfo.log.errors.length, 0, ...errors); 
            }

            // Spin up or down workers to match required capacity.
            if (capacity > rosteredWorkers.length)
                await replenishWorkers(capacity);
            else 
                retireWorkers(rosteredWorkers.length - capacity);  // EG: Have 6 workers, need 2 -> retire 4.

            // Assign jobs to as many workers as we need. The rest can remain on standby.
            let { assignments, workers } = assignJobs(jobs, rosteredWorkers.slice(0, capacity), tick);

            cycleInfo.assignments = assignments;
            cycleInfo.workers     = workers;
                
            cycleInfo.log.workers     = workers.map(worker => worker.id);
            cycleInfo.log.assignments = workers.map(worker => ({
                worker: worker.identity.name,
                jobs:   assignments[worker.id] ? assignments[worker.id].jobs.map(job => job.name) : []
            }));
            cycleInfo.log.jobs = jobs && jobs.length > 0 ? jobs.map(job => ({
                name: job.name,
                type: job.type,
                size: job.entities ? job.entities.length : job.limit || config.maxJobSize
            })) : null;

            cycleInfo.progress = "started";
            cycleInfo.log.started = (new Date()).getTime();

            cycleInfo.metrics.failedWorkers  = workers.length;
            cycleInfo.metrics.totalWorkers   = workers.length;
            cycleInfo.metrics.failedJobs     = 0; 
            cycleInfo.metrics.totalJobs      = jobs.length;
            cycleInfo.metrics.minJobSize     = sizes.length > 0 ? Math.min(...sizes) : 0;
            cycleInfo.metrics.maxJobSize     = sizes.length > 0 ? Math.max(...sizes) : 0;
            cycleInfo.metrics.avgJobSize     = sizes.length > 0 ? Math.avg(...sizes) : 0;
            cycleInfo.metrics.minJobDuration = 0;
            cycleInfo.metrics.maxJobDuration = 0;
            cycleInfo.metrics.avgJobDuration = 0;
            cycleInfo.metrics.totalDuration  = 0;
        }
        catch (ex) {
            if (cycleInfo) {
                cycleInfo.progress = "failed";

                cycleInfo.log.errors.push(ex.message);
                cycleInfo.log.failed = (new Date()).getTime(); 
            }
            else {
                cycleInfo = scheduler.getDummyCycle(tick, "failed", ex.message);
            }
        }
        finally {
            // Advance the hand.
            previousTick = tick;
                
            if (cycleInfo && cycleInfo.log.jobs)
                logger.info("Awaiting completions for cycle " + tick);
            else 
                logger.info("No jobs available for cycle " + tick);

            socket.emit('cycle', cycleInfo);  
        }
    }
        
    function assignJobs(jobs, workers, tick) {
        if (!jobs || jobs.length === 0 || !workers || workers.length === 0) return {
            assignments: {},
            workers: []
        };

        /*  Allocate jobs to workers. Assumption: jobs are sorted in order
            of decreasing size (number of entities).
            For each job, we assign it to the worker with the smallest 
            assignment so far. Since we keep the assignments sorted, that
            means that we should always assign it to the first worker in 
            the list.
         */
        var allocations = workers.map(worker => {
            return { worker: worker, jobs: [], size: 0, cycle: tick };
        });

        let assignTime = (new Date()).getTime();

        jobs.forEach(job => {
            const jobSize = job.entities ? job.entities.length : job.limit || config.maxJobSize;

            job.assigned = assignTime++;    // To keep them unique in the browser.

            allocations[0].jobs = allocations[0].jobs.concat(job);
            allocations[0].size += jobSize;

            allocations.sort((workA, workB) => workA.size - workB.size);
        });

        /*  Now all jobs have been allocated to workers. Tell the workers, 
            and they will start processing their assigned work immediately.
            Note that some workers might not have gotten an assignment for
            this cycle. We must exclude them from the completion tracking.
         */
        // Don't send assignments to workers that have no jobs. Can happen!
        allocations = allocations.filter( allocation => allocation.jobs.length > 0);

        logger.info("Worker assignments for cycle " + tick + ":");
        allocations.forEach(function(allocation) {
            logger.info("    Worker " + allocation.worker.id + ":\t" + allocation.jobs.map(job => job.name).join(", ") + " (" + allocation.size + " entities total)" );

            allocation.worker.assignment = allocation;
        });

        // Return allocations as an assignment hash map to track completion.
        return {
            assignments: allocations.reduce((map, allocation) => {
                map[allocation.worker.id] = allocation;
                return map;
            }, {}),
            workers: allocations.map(allocation => allocation.worker)
        };
    }
 
    function onJobDone(jobStatus) {
        const workerID = jobStatus.worker.id;
        logger.info("Worker " + workerID + " produced " + jobStatus.sent + " messages for " + jobStatus.size + " entities for '" + jobStatus.name + "' in cycle " + jobStatus.cycle);

        let cycleInfo = scheduler.getCycleInfo(jobStatus.cycle);
        if (!cycleInfo) return;

        try {
            cycleInfo.log.results.push(
                "Worker " + workerID + " finished '" + jobStatus.name + "' at " + new Date().toLocaleTimeString()
            );

            // Were there any errors or messages?
            if (jobStatus.log && jobStatus.log.summary) {
                cycleInfo.log.durations.push(jobStatus.log.summary.duration);

                if (jobStatus.log.summary.errors && jobStatus.log.summary.errors.length > 0) {
                    cycleInfo.log.errors.push("Worker " + workerID + " errors:");
                    cycleInfo.metrics.failedJobs++; 

                    logger.warn("Worker " + workerID + " errors:");

                    jobStatus.log.summary.errors.forEach(err => {
                        logger.warn("    " + err);
                        cycleInfo.log.errors.push("    " + err);
                    });

                    scheduler.dingJob(jobStatus.id, cycleInfo);
                } 
                if (jobStatus.log.messages && jobStatus.log.messages.length > 0) {
                    cycleInfo.log.results.push("Worker " + workerID + " results:");
                    jobStatus.log.messages.forEach(msg => cycleInfo.log.results.push("    " + msg));
                } 
                if (jobStatus.log.feedback == "split") {
                    cycleInfo.log.results.push(
                        "    Worker recommends that '" + jobStatus.name + "' be split"
                    );
                    scheduler.splitJob(jobStatus.id);
                }
            }

            if (!cycleInfo.assignments || !cycleInfo.workers) {
                cycleInfo.progress = "done";

                cycleInfo.log.finished = (new Date()).getTime();
                cycleInfo.log.duration = cycleInfo.log.finished - cycleInfo.log.started;
                cycleInfo.elapsedTime  = cycleInfo.log.duration;
                cycleInfo.metrics.totalDuration = cycleInfo.elapsedTime;

                cycleInfo.log.results.push(
                    "This cycle does not appear to have assignments or workers"
                );

                socket.emit('cycle', cycleInfo);  
            }
            else {
                // Remmove job from jobs assigned to this worker. When empty, worker is done for this cycle.
                let assignment = cycleInfo.assignments[workerID];    
                if (assignment && assignment.jobs) {
                    assignment.jobs.removeById(jobStatus.id);
                    // This is where we track each job time.
                }

                if (!assignment || !assignment.jobs || assignment.jobs.length === 0) {
                    // No more jobs - clear the assignment list and remove the worker from the 'busy' list.
                    cycleInfo.assignments[workerID] = null;
                    cycleInfo.workers.removeById(workerID);
                    cycleInfo.metrics.failedWorkers--;

                    cycleInfo.log.results.push(
                        "Worker " + workerID + " completed all assignments - " + cycleInfo.workers.length + " workers still busy"
                    );

                    if (cycleInfo.workers.length === 0) {
                        logger.info("All workers completed their assigned jobs for cycle " + cycleInfo.cycle);

                        cycleInfo.progress = "done";

                        cycleInfo.log.finished = (new Date()).getTime();
                        cycleInfo.log.duration = cycleInfo.log.finished - cycleInfo.log.started;
                        cycleInfo.elapsedTime  = cycleInfo.log.duration;

                        cycleInfo.metrics.totalDuration  = cycleInfo.elapsedTime;
                        cycleInfo.metrics.minJobDuration = Math.min(...cycleInfo.log.durations);
                        cycleInfo.metrics.maxJobDuration = Math.max(...cycleInfo.log.durations);
                        cycleInfo.metrics.avgJobDuration = Math.avg(...cycleInfo.log.durations);

                        socket.emit('cycle', cycleInfo);
                    }
                }
            }
        }
        catch (ex) {
            logger.info(
                "Error while processing completion of '" + jobStatus.name + "' for cycle " + jobStatus.cycle + " by worker " + workerID + ": " + ex.message
            );

            cycleInfo.progress = "warning";

            cycleInfo.timestamp = (new Date()).getTime();
            cycleInfo.log.errors.push(
                "Error while processing completion of '" + jobStatus.name + "' for cycle " + jobStatus.cycle + " by worker " + workerID + ": " + ex.message
            );

            socket.emit('cycle', cycleInfo);  
        }
    }

    function heartbeat(forcedRun) {
        /*  We tick every second and check if we crossed a minute boundary. On startup that
            is right away, which would lead to the immediate creation of a new cycle. Thus, 
            the number of seconds in the minute should be less than 10 to start a new cycle.
         */
        var now = new Date();
        var minutes = now.getMinutes();
        var seconds = now.getSeconds();

        if (!forcedRun && (minutes === lastMinute || seconds > 10))
            return; 

        // Now we're good for a new cycle. Align the timestamp to the exact minute boundary.
        lastMinute = minutes;
        now.setSeconds(0, 0);

        logger.info(" ");
        logger.info("Starting new cycle at " + now.toLocaleTimeString() + " (epoch " + now.getTime() + ")");

        startNextCycle(now.getTime()); 
    }

    // Our heartbeat...
    let timer = null;
    const startTimer = function () {
        /*  If the 'runOption' says '1', it means: run once.
            Else we set an interval to check each second if it's time for the next cycle.
         */
        if (runOption === 1) {
            setTimeout(heartbeat, 1, true);
        }
        else {
            timer = timer || setInterval(heartbeat, 1000);
        }
    };
    const stopTimer = function () {
        if (timer) {
            clearInterval(timer);
            retireAllWorkers();     // We don't need our workers anymore.
        }
        timer = null;
    };

    logger.info("    Launching or connecting to library processes");

    libraryProxy = new ServiceBroker("LibraryService", config, socket, logger, true); // Autorestart
    libraryProxy.once("ready", libraryIF => {
        // We needed to wait until all libraries have loaded their entities.
        // Now the clock can start ticking and jobs can get executed.
        logger.info("    All library processes are available");

        // Replace the initial interface with the full interface.
        libraryProxy = libraryIF;

        // Start the diagnostics to track memory.
        diagnostics = new Diagnostics(config, libraryProxy, logger);
        diagnostics.start();

        // Create the scheduler. We will assign it jobs later, unless we have them already.
        // If the library becomes unavailable and then available again, we don't recreate.
        scheduler = scheduler || new Scheduler(config, libraryProxy, logger, pendingJobList, tuningInfo => {
            // Whenever the scheduler updates its tuning info, tell all workers about it. 
            // Note that we also set the tuning info when a worker is created.
            rosteredWorkers.forEach(worker => worker.tuning = tuningInfo);
        });

        switch (runOption) {
            case 0:   
                emit("ready", "Jobs will be executed per their schedules until stopped")              
                break;
            case 1:
                emit("ready", "Jobs will be executed once per manual request to run");
                break;
            default:
                emit("ready", "This failover instance is ready to take over when the primary instance goes down");
                break;
        }  
    });

    return {
        on:    subscribe,
        start: startTimer,
        stop:  stopTimer,
        end: function(onEnd) {
            stopTimer();
            libraryProxy.retire(onEnd);
        },
        killWorker: function(workerID) {
            return retireWorkers([{ id: workerID }]);
        },
        get scheduler() {
            return scheduler;
        },
        get config() {
            return config;
        },
        get workers() {
            return rosteredWorkers.map(worker => worker.identity);
        },
        get libraries() {
            return scheduler.libraries.map(library => library.identity);
        },
        get schedules() {
            return scheduler.schedules;
        },
        get tuningInfo() {
            return scheduler.tuning;
        },
        set tuningInfo(info) {
            scheduler.tuning = info;
        },
        get isRunning() {
            return !!timer;
        },
        set config (newConfig) {
            logger.info("Reloading configuration...");
            config = newConfig;

            retireAllWorkers();
            if (scheduler)
                scheduler.config = config;
        },
        set jobs (jobList) {
            if (scheduler)
                scheduler.jobs = jobList;
            else
                pendingJobList = jobList;
        },
        get jobs () {
            return scheduler ? scheduler.jobs : pendingJobList;
        },
        getHealthMetrics: getHealthMetrics
    };
};

module.exports = Controller;
