'use strict';

const IndexedSet      = require('./IndexedSet.js');
const Janitor         = require("./Janitor.js");
const PipelineManager = require('./PipelineManager.js');

class Curator {
    #pipelineManager = null;
    #started    = false;
    #entities   = {};
    #discovered = {};
    #deletions  = {};
    #janitor    = null;

    cycleTime   = 60*60*1000;

    get indexer() {
        return this.globals.indexer;
    }
    get dataAccess() {
        return this.globals.dataAccess;
    }    
    get entityAccess() {
        return this.globals.entityAccess;
    }     

    constructor(globals) {
        this.globals       = globals;
        this.config        = globals.config;
        this.friendly      = globals.friendly;
        this.scheduler     = globals.scheduler;
        this.logger        = globals.logger;
        this.socket        = globals.socket;
        this.relationships = globals.relationships;
        this.tenants       = globals.tenants;
        this.types         = globals.types;
        this.dynatraceAPI  = globals.dynatraceAPI;
        this.libraryAPI    = globals.libraryAPI;
        this.apiEndPoints  = globals.apiEndPoints;

        try {
            const maxSpeed = this.scheduler.getGranularityTime(this.config.library.refreshRate);
            this.cycleTime = maxSpeed;     // We start here, and slow down as needed.

            this.logger.info("    Maximum cycle time is " + this.cycleTime + "ms (" + this.config.library.refreshRate + ")");

            this.tenants.forEach(tenant => {
                this.#entities[tenant]   = {};
                this.#discovered[tenant] = {};
                this.#deletions[tenant]  = {};
         
                this.types.forEach(type => {
                    this.#entities[tenant][type]   = new IndexedSet([], "entityId", tenant+"_"+type);
                    this.#discovered[tenant][type] = [];
                    this.#deletions[tenant][type]  = [];
                });
            });    

            // Also initialize the janitor. We will start it when we start.
            this.logger.info("    Creating the janitor");
            this.#janitor = new Janitor(globals, this);

            this.logger.info("    Creating the pipeline manager");
            this.#pipelineManager = new PipelineManager(globals);

            this.logger.info("    The curator initialized successfully");
        }
        catch (ex) {
            this.logger.error("   Curator failed to initialize: " + ex.message);
        }    
    }

    loadManagementZones() {
        if (!this.dynatraceAPI) return;

        const self = this;

        return new Promise((resolve, reject) => {
            this.logger.info("Retrieving all management zones from Dynatrace for " + this.tenants.join(", "));

            let promises = this.tenants.map(tenant => {
                // Load the zones for each tenant so that we can chunk our entities by them.
                return new Promise((resolve, reject) => {
                    /*  Retrieve all known management zones in this tenant. Our chunking
                        options include chunking by a list of MZs, but we need to know about
                        their names and IDs first. This is a one-tinme thing.
                    */
                    this.dynatraceAPI.request({
                        tenant:    tenant,
                        endPoint:  '/api/config/v1/managementZones',
                        queryData: ''
                    },
                    function (err, zones) {
                        if (err) {
                            self.logger.warn("    Could not retrieve management zones in " + tenant + ": " + err.message);
                            self.logger.warn("    Chunking by management zone will not be supported for " + tenant);
                            self.logger.warn("    Management zones for " + tenant + " will be rebuilt from entity data ");

                            zones = [];
                        }
                        else {
                            zones = !zones || !Array.isArray(zones.values) ? [] : zones.values;

                            self.logger.info("    Found " + zones.length + " management zones in " + tenant);
                        }

                        self.indexer.indexZones(tenant, zones);

                        resolve();
                    });
                });
            });

            Promise.all(promises)
            .then(async () => {
                this.logger.info("    Management zones for all tenants have been loaded");
                resolve();
            })
            .catch(ex => {
                this.logger.warn("    Loading management zones for some tenants failed: " + (ex.message ? ex.message : ex));
                reject(ex);
            });
        });
    }

    getEntities(tenant, type) {
        if (!tenant) 
            return this.#entities;

        if (!this.#entities.hasOwnProperty(tenant))
            this.#entities[tenant] = {};

        if (!type) 
            return this.#entities[tenant];
        
        if (!this.#entities[tenant].hasOwnProperty(type))
            this.#entities[tenant][type] = new IndexedSet([], "entityId", tenant+"_"+type);

        return this.#entities[tenant][type];
    }

    setEntities(tenant, type, list) {
        this.#entities[tenant][type] = new IndexedSet(list, "entityId", tenant+"_"+type);

        return this.#entities[tenant][type];
    }

    addEntity(tenant, type, entity, staleTime, checksum, lastUpdated, lastUsed) {
        const entityProxy = {
            entityId:    entity.entityId,
            displayName: entity.displayName,
            isUnknown:   entity.isUnknown,
            type:        type,
            tenant:      tenant,
            checksum:    checksum,
            lastUpdated: lastUpdated,
            lastSeen:    Number(entity.lastSeenTimestamp) || (new Date()).getTime(),
            lastUsed:    lastUsed,
            referrers:   new Set(),
            flows:       { IN: [], OUT: [] },
            stack:       { IN: [], OUT: [] },
            group:       { IN: [], OUT: [] },
            entity:      null
        }
        // Add to the proxy if the entity was recently used in a query.
        if (staleTime && entityProxy.lastUsed > staleTime) {
            entityProxy.entity = entity;
        }

        // Now that we know of this entity, it's no longer unknown.
        this.#entities[tenant][type].add(entityProxy);
        this.#discovered[tenant][type].remove(entity.entityId);

        return entityProxy;
    }

    deleteEntity(tenant, type, entity) {
        if (!entity.deleted) {
            delete entity.entity;  // Clear from cache.
            entity.deleted = (new Date()).getTime();
            this.#deletions[tenant][type].push(entity);
            this.#discovered[tenant][type].remove(entity.entityId);
        }
        return this.#entities[tenant][type].delete(entity);
    }

    getDeletedEntities(tenant, type = "ALL", since = 0) {
        if (!this.#deletions[tenant])
            return null;

        if (type === "ALL")
            return [].concat(...types.map(t => this.getDeletedEntities(tenant, t, since)));

        if (!this.#deletions[tenant][type])
            return null;

        return this.#deletions[tenant][type].filter(entityProxy => entityProxy.deleted >= since);
    }    

    getEntity(tenant, type, entity) {
        return this.#entities[tenant][type].get(entity.entityId);
    }

    getZones(tenant, type, startsWithValue) {
        return this.indexer.getZones(tenant, type, startsWithValue);
    }

    getZoneId(tenant, type, name) {
        return this.indexer.getZoneId(tenant, type, name);
    }

    getPipelines(tenants, types) {
        return this.#pipelineManager.getPipelines(tenants, types);
    }

    async getPipelinesInfo() {
        if (!this.#started)
            throw new Error("Curator is not ready yet");

        return this.#pipelineManager.getPipelinesInfo();
    }

    async getPipelineStats(tenant, type) {
        return this.#pipelineManager.getPipelineStats(tenant, type);
    }

    getHealthMetrics() {
        const healthMetrics = [];

        this.tenants.forEach(tenant => {
            this.types.forEach(type => {
                healthMetrics.push({
                    metricId: "numberOfEntities", 
                    dimensions: [ tenant, type ],
                    value: this.#entities[tenant][type].length
                });
            })
        });

        return healthMetrics.concat(this.#pipelineManager.getHealthMetrics());
    }

    getPipelineHealth() {
        const health = this.#pipelineManager.getPipelineHealth();
        health.push(this.#janitor.getPurgerHealth());
        return health;
    }

    async controlPipeline(tenant, type, action) {
        if (!this.#started)
            throw new Error("Curator is not ready yet");

        return this.#pipelineManager.controlPipeline(tenant, type, action);
    }

    getDiscovered(tenant, type) {
        return this.#discovered[tenant][type] || [];
    }

    addDiscovered(tenant, type, entityId) {
        const discovered = this.#discovered[tenant][type] || [];

        if (!discovered.includes(entityId))
            discovered.push(entityId);
    }

    clearDiscovered(tenant, type) {
        this.#discovered[tenant][type] = [];
    }

    start(isOnStandby = false) {
        if (this.#started)
            this.stop(true);

        this.#pipelineManager.start(isOnStandby)       
        this.#janitor.start();
        this.#started = true;
    }

    stop(quiet) {
        this.#pipelineManager.stop(quiet);
        this.#janitor.stop();
        this.#started = false;
    }

    clear() {
        this.stop();
    }

    get isRunning() {
        return this.#started;
    }
}

module.exports = Curator;
