'use strict';

/*******
 *  Interactions with the database.
 *  - createTable:      Creates the table we will be working with in the database.
 *  - loadEntity:       Load a specific entity (by ID and tenant) from the database.
 *  - loadEntities:     Load all entities stored in the database and update our administration.
 *  - countEntities:    Counts how many entities of a given type in a given tenant are in the database.
 *  - cacheEntities:    Load the full entities for the specified proxies from the database and cache them.
 *  - refreshEntities:  Store all entities received from a Dynatrace API call and update our administration.
 *  - retireEntities:   Remove all entities of a given type for a specified tenant from the database and
 *                      from our administration that have not been seen since the specified timestamp.
 */

const crypto = require('crypto');

const maxDisplayNameSize  = 512;
const maxStringValueSize  = 256;

const executeSerially = function (work, resolve, reject, results) {
    results = results || { errors: [], data: [] };

    const task = work.shift();
    if (task) {
        task(
            data => {
                if (data) results.data.push(data);
                setTimeout(executeSerially, 1, work, resolve, reject, results);
            },
            error => {
                if (error) results.errors.push(error);
                setTimeout(executeSerially, 1, work, resolve, reject, results);
            }
        );
    }
    else {
        if (results.errors.length > 0)
            reject(results);
        else
            resolve(results);
    }
};

class DataAccess {
 
    get curator() {
        return this.globals.curator;
    }
    get indexer() {
        return this.globals.indexer;
    }    
    get databasePool() {
        return this.globals.databasePool;
    }      
    get dataAccess() {
        return this.globals.dataAccess;
    }    
    get entityAccess() {
        return this.globals.entityAccess;
    }        
    get dynatraceAPI() {
        return this.globals.dynatraceAPI;
    }        
    constructor(globals, cacheCycle) {
        this.globals      = globals;
        this.logger       = globals.logger;
        this.friendly     = globals.friendly;
        this.tenants      = globals.tenants;
        this.types        = globals.types;
        this.apiEndPoints = globals.apiEndPoints;
        this.cacheCycle   = cacheCycle;
    }
 
    createTable() {}
 
    loadEntity(entityId, tenant) {
        const self = this;

        return new Promise(function (resolve, reject) {
            const taskName = "loadEntity_" + tenant + "_" + type;

            self.databasePool.connect(tenant, taskName)
            .then(connection => {
                const statement = "SELECT id, entity FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' AND id='" + entityId + "'";

                connection
                    .query(statement)
                    .then(rows => {
                        connection.release(() => {
                        if (rows.length !== 1)
                            reject("WARNING: When trying to load " + entityId + " " + rows.length + " were returned for " + tenant);
                        else
                            resolve(JSON.parse(rows[0].entity || "{}"));
                        });
                    })
                    .catch(error => {
                        connection.release((err) => {
                            reject(error);
                        });
                    });
            })
            .catch(error => {
                reject("Error connecting to database: " + error.message)
            });
        });
    }

    loadEntities(tenants) {
        const self = this;

        return new Promise(function (resolve, reject) {
            const workList = [];

            tenants.forEach(tenant => {
                self.types.forEach(type => {
                    workList.push(function (resolve, reject) {
                        const taskName = "loadEntities_" + tenant + "_" + type;

                        self.databasePool.connect(tenant, taskName)
                        .then(connection => {
                            self.indexer.clear(tenant, type);

                            const statement  = "SELECT * FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' AND entityType='" + type + "'";

                            connection
                            .query(statement)
                            .then(rows => {
                                let failed = false;
                                try {
                                    rows = rows || [];
                                    for (var r=0; r < rows.length; r++) {
                                        self.entityAccess.updateEntity(tenant, type, JSON.parse(rows[r].entity || "{}"), undefined, rows[r].checksum, rows[r].lastUpdated);
                                    }
                                    self.logger.info("    Found " + rows.length + " " + self.friendly(type) + "s in " + tenant);
                                }
                                catch (ex) {
                                    failed = true;
                                    self.logger.error("    Could not load " + self.friendly(type) + "s in " + tenant + ": " + ex.message);
                                }
                                finally {
                                    connection.release(() => {
                                        resolve(failed ? tenant : null);
                                    });
                                }
                            })
                            .catch(err => {
                                self.logger.error("    Could not load " + self.friendly(type) + "s in " + tenant + ": " + err);

                                connection.release(() => {
                                    resolve(tenant);
                                });
                            });
                        })
                        .catch(err => {
                            // No connection!
                            resolve(tenant);
                        });
                    });
                });
            });

            // Execute the tasks one at a time.
            executeSerially(workList,
                ({errors, data}) => {
                    if (data.length === 0)
                        resolve();
                    else
                        reject(data.unique().join(", "));
                },
                ({errors, data}) => {
                    reject(data.unique().join(", "));
                }
            );
        });
    }

    getEntity(tenant, type, entityId, refreshEntities, noReject) {
        const self = this;

        return new Promise(function (resolve, reject) {
            self.dynatraceAPI.request({
                tenant:    tenant,
                endPoint:  '/api/v1/' + apiEndPoints[type] + '/' + entityId,
                noReject:  true
            },
            function (err, entity) {
                const logicalErrors = [404, 404, 500];

                if (err && logicalErrors.includes(err.status)) {
                    const now = (new Date()).getTime();

                    logger.warn("Dynatrace could not find " + entityId + " in " + tenant + ". Creating placeholder: 'Missing " + friendly(type) + " " + entityId + "'");

                    entity = {
                        entityId:    entityId,
                        displayName: "Missing " + friendly(type) + " " + entityId,
                        tenant:      tenant,
                        type:        type,
                        lastSeen:    now,
                        firstSeen:   now,
                        isUnknown:   true       // This will end up in the proxy.
                    };
                }

                if (err && !logicalErrors.includes(err.status)) {
                    if (noReject)
                        resolve();
                    else
                        reject("DynatraceAPI could not get " + friendly(type) + " " + entityId + " in " + tenant + ": " + err.message);
                }
                else if (refreshEntities)
                    self.refreshEntities(tenant, type, [ entity ])
                        .then(() => resolve())
                        .catch(err => reject(err));
                else
                    resolve(entity);
            });
        });
    }

    getEntities(tenant, type, entityIds, refreshEntities) {
        const self = this;

        if (!entityIds || entityIds.length === 0)
            return new Promise(function (resolve, reject) { resolve(); });

        if (entityIds.length === 1)
            return self.getEntity(tenant, type, entityIds[0], refreshEntities);

        return new Promise(function (resolve, reject) {
            // Make sure that the query string doesn't get too long. We're using a max of 2048 chars here.
            const maxEntities = Math.floor(2048 / ("&entity=" + entityIds[0]).length);  // Ex.: 2048/25 chars = max 81 entities per request
            const noOfChunks  = Math.ceil(entityIds.length / maxEntities);              // If 100 entities requested, chunks = ceil(100/81) = 2
            const chunkSize   = Math.ceil(entityIds.length / noOfChunks);               // Therefore the chunk size = 100/2 = 50

            if (noOfChunks > 1) {
                const entityIdChunks = [];
                while (entityIds.length > 0) {
                    entityIdChunks.push(entityIds.splice(0, chunkSize));
                }

                const promises = entityIdChunks.map(chunk => self.getEntities(tenant, type, chunk));
                Promise.all(promises)
                .then(results => {
                    const entities = results.flat().filter(result => !!result);

                    if (refreshEntities)
                        self.refreshEntities(tenant, type, entities)
                            .then(() => resolve())
                            .catch(error => reject("Could not process received " + friendly(type) + "s in " + tenant + ": " + error));
                    else
                        resolve(entities);
                })
                .catch(err => reject("Could not get all " + friendly(type) + "s in " + tenant + ": " + err));
            }
            else {
                // We have an acceptable amount of entity IDs to ask the Dynatrace API for.
                self.dynatraceAPI.request({
                    tenant:    tenant,
                    endPoint:  '/api/v1/' + self.apiEndPoints[type],
                    queryData: entityIds.map(id => "entity=" + id).join("&"),
                    noReject:  true
                },
                function (err, entities) {
                    if (err || !entities)
                        logger.warn("DynatraceAPI could not get " + self.friendly(type) + "s in " + tenant + ": " + (err.message || "no results"));

                    if (entities.length < entityIds.length) {
                        /*  Sometimes not all requested entities are found by the Dynatrace API. That really should
                            NEVER happen, but it does. We remediate in two ways: we ask Dynatrace in a different way,
                            and if then still not successful, we create a proxy marked "unknown" (in 'getEntity()').
                        */
                        let promises = [ entities ];    // Our already-retrieved entities. Mixing this with promises is allowed!

                        let foundIds = entities.map(entity => entity.entityId);
                        for (let i=0; i < entityIds.length; i++) {
                            if (!foundIds.includes(entityIds[i])) {
                                promises.push(self.getEntity(tenant, type, entityIds[i], null, true));
                            }
                        }

                        Promise.all(promises)
                        .then(results => {
                            entities = results.flat().filter(result => !!result);

                            if (refreshEntities)
                                self.refreshEntities(tenant, type, entities)
                                    .then(() => resolve())
                                    .catch(error => reject("Could not process received " + self.friendly(type) + "s in " + tenant + ": " + error));
                            else
                                resolve(entities);
                        })
                        .catch(err => reject(err));
                    }
                    else {
                        if (refreshEntities)
                            self.refreshEntities(tenant, type, entities)
                                .then(() => resolve())
                                .catch(error => reject("Could not process received " + self.friendly(type) + "s in " + tenant + ": " + error));
                        else
                            resolve(entities);
                    }
                });
            }
        });
    }

    countEntities(tenant, type) {
        const self = this;

        return new Promise(function (resolve, reject) {
            const taskName = "countEntities_" + tenant + "_" + type;

            self.databasePool.connect(tenant, taskName)
            .then(connection => {
                const statement  = "SELECT count(id) FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' AND entityType='" + type + "'";

                connection
                .query(statement)
                .then(result => {
                    connection.release(() => {
                        if (result.length === 1)
                            resolve(result[0]["count(id)"]);
                        else
                            reject("Database returned: " + JSON.stringify(result));
                    });
                })
                .catch(error => {
                    connection.release(() => {
                        reject(error);
                    });
                });
            })
            .catch(error => {
                reject(error);
            });
        });
    }

    async cacheEntities(tenant, type, entityProxies) {
        const self = this;

        // If there is no tenant, we'll figure out what tenants are in the data,
        // and call ourselves in recurse for each tenant found.
        if (!tenant) {
            return new Promise (function (resolve, reject) {
                if (entityProxies.length === 0)
                    return resolve();

                const proxiesPerTenant = {};
                for (let i=0; i < entityProxies.length; i++) {
                    let proxy = entityProxies[i];
                    proxiesPerTenant[proxy.tenant] = proxiesPerTenant[proxy.tenant] || [];
                    proxiesPerTenant[proxy.tenant].push(proxy);
                }
                const promises = Object.keys(proxiesPerTenant).map(tenant =>
                    cacheEntities(tenant, type, proxiesPerTenant[tenant])
                );
                Promise.all(promises).then(resolve).catch(reject);
            });
        }

        return new Promise(function (resolve, reject) {
            const now = (new Date()).getTime();

            entityProxies = entityProxies.filter(proxy => {
                // In memory or not, we're using it now: push flushing from cache further into future.
                proxy.lastUsed = now;

                // If we have the entity cached, we're good. No need to read it from database again.
                if (proxy.entity) return false;

                // Let's create a dummy entity in case we can't find anything for it in the database.
                proxy.entity = self.entityAccess.createEntity(proxy);

                // If the entity the proxy refers to is officially unknown, don't try the database.
                if (proxy.isUnknown) return false;

                // Entity is not in the cache yet: go query the database for it.
                return true;
            });

            if (entityProxies.length === 0)
                return resolve();   // Nothing to load from database.

            const taskName = "cacheEntities_" + tenant + "_" + Math.random();

            self.databasePool.connect(tenant, taskName)
            .then(connection => {
                // This is to be able to quickly find the entity associated with a specific row in the resultset.
                const proxyIndex = {};
                for (let i=0; i < entityProxies.length; i++) {
                    proxyIndex[entityProxies[i].entityId] = entityProxies[i];
                }

                let statement = "SELECT id, entity FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' and id IN ('"
                            + entityProxies.map(entity => entity.entityId).join("','") + "')";

                connection
                .query(statement)
                .then(rows => {
                    try {
                        let extras  = [];
                        let missing = [];
                        let proxy = null;

                        for (let r=0; r < rows.length; r++) {
                            proxy = proxyIndex[rows[r].id];
                            if (proxy)
                                proxy.entity = JSON.parse(rows[r].entity || "{}");
                            else
                                extras.push(rows[r].id);    // TODO: if we ever have extras, we should consider adding them to the internal admin.
                        }

                        if (rows.length < entityProxies.length) {
                            for (let i=0; i < entityProxies.length; i++) {
                                let proxy = entityProxies[i];
                                if (!proxy.entity) {
                                    missing.push(proxy.entityId);
                                    self.entityAccess.removeEntity(tenant, type, proxy);    // Not in database? Remove from admin. Better would be to ask Dynatrace.
                                }
                            }
                        }

                        connection.release(() => {
                            self.logger.info("Cached " + rows.length + " entities in " + tenant);

                            if (extras.length > 0)
                                reject("Database returned the following extra " + extras.length + " entities in " + tenant + ": " + extras.join(", "));
                            else if (missing.length > 0)
                                reject("The following " + missing.length + " entities in " + tenant + " were not found in the database: " + missing.join(", "));
                            else
                                resolve();
                        });
                    }
                    catch (ex) {
                        connection.release(() => {
                            reject("Caching " + self.friendly(type) + "s in " + tenant + " yields: " + ex.message)
                        });
                    }
                })
                .catch(error => {
                    connection.release(() => {
                        reject(error);
                    });
                });
            })
            .catch(error => {
                reject("Error connecting to database: " + error.message)
            });
        });
    }

    streamEntities(tenant, type, entityProxies) {
        const self = this;

        return new Promise((resolve, reject) => {
            // If there is no tenant, we'll figure out what tenants are in the data,
            // and call ourselves in recurse for each tenant found.
            const emitter = {
                events: [],
                on: function (eventName, handler) {
                    this.events[eventName] = this.events[eventName] || [];
                    this.events[eventName].push(handler);
                },
                emit: function (eventName, data) {
                    const eventHandlers = this.events[eventName] || [];
                    eventHandlers.forEach(handler => { handler.call(null, data); });
                }
            };

            if (entityProxies.length === 0) {
                setTimeout(() => {
                    emitter.emit("end");
                }, 10);

                return resolve(emitter);
            }

            if (!tenant) {
                const proxiesPerTenant = {};
                for (let i=0; i < entityProxies.length; i++) {
                    let proxy = entityProxies[i];
                    proxiesPerTenant[proxy.tenant] = proxiesPerTenant[proxy.tenant] || [];
                    proxiesPerTenant[proxy.tenant].push(proxy);
                }
                const promises = Object.keys(proxiesPerTenant).map(tenant =>
                    streamEntities(tenant, type, proxiesPerTenant[tenant])
                );

                Promise.all(promises)
                .then(streams => {
                    let metasReceived = 0;
                    let endsReceived  = 0;

                    streams.forEach(stream => stream
                        .on("fields", meta => {
                            if (metasReceived === 0)
                                emitter.emit("fields", meta);
                        })
                        .on("data", row => {
                            emitter.emit("data", row);
                        })
                        .on("error", err => {
                            emitter.emit("error", err);
                        })
                        .on("end", () => {
                            endsReceived++;
                            if (endsReceived === streams.length)
                                emitter.emit("end");
                        })
                    );
                })
                .catch(error => {
                    self.logger.error("Could not stream resultset to client: " + error);
                    emitter.emit("error", error);
                });

                return resolve(emitter);
            }

            const taskName = "cacheEntities_" + tenant + "_" + Math.random();

            self.databasePool.connect(tenant, taskName)
            .then(connection => {
                const statement = "SELECT id, entity FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' and id IN ('"
                                + entityProxies.map(entity => entity.entityId).join("','") + "')";

                connection
                .stream(statement)
                .on("end", () => {
                    connection.release(() => {});
                })
                .on("error", error => {
                    connection.release(() => {});
                });

                resolve(connection);
            })
            .catch(error => {
                setTimeout(() => {
                    emitter.emit("error", error);
                }, 10);

                resolve(emitter);
            });
        });
    }

    refreshEntities(tenant, type, entityList, retryCount, errors) {
        const self = this;

        const createChecksum = function (entity) {
            // Because JSON.stringify doesn't always return the same string for an object
            // that in terms of properties and values is the same, we sort the resulting
            // string internally before creating the hash from it.
            return crypto
                .createHash('md5')
                .update(JSON
                    .stringify(entity)  // We should remove all dates from here - at least 'lasts
                    .split('')
                    .sort()
                    .join(''),
                    'utf8')
                .digest('hex');
        }

        retryCount = retryCount === undefined ? self.databasePool.retryLimit : retryCount;
        errors = errors || [];

        const staleTime  = (new Date()).getTime() - self.cacheCycle;

        // First of all, mark the proxies of all entities that Dynatrace could not find.
        // We won't be putting them in the database.
        if (entityList) {
            for (let i=0; i < entityList.length; i++) {
                if (entityList[i].isUnknown) {
                    self.entityAccess.updateEntity(tenant, type, entityList[i], staleTime, "N/A", (new Date()).getTime());
                }
            }
        }

        return new Promise(function (resolve, reject) {
            const taskName = "refreshEntities_" + tenant + "_" + type;

            /*  If entityList is 'false', we're restarting a pipeline. In that scenario we
                need to release any database connection(s) we may have open for this task.
            */
            if (entityList === false) {
                self.databasePool.release(taskName);
                return resolve();
            }

            /*  Multiple types may leverage the same endpoint to the API. In effect, we are
                retrieving from the same endpoints more than once, for different types. So,
                we have to filter the entity list so that only the ones of the intended type
                remain. That is only true for applications right now - and we actually only
                support 'APPLICATION'.
            */
            entityList = (entityList || []).filter(entity => entity.entityId.indexOf(type) === 0);
            if (entityList.length === 0) {
                return resolve();
            }

            function retry(error, insult) {
                // Let's not repeat any errors we already saw.
                if (error && !errors.includes(error))
                    errors.push(error.message || error);

                // Insult on injury: an error happend WHILE we were dealing with an error!
                if (insult && !errors.includes(insult))
                    errors.push(insult.message || insult);

                if (retryCount > 0 && self.databasePool.isRetryable(error)) {
                    // Retries left: Call ourselves after 100 ms, and call our 'resolve()' when that resolves...
                    setTimeout((function (resolve, reject, retriesLeft, errors) {
                        return () => self.refreshEntities(tenant, type, entityList, retriesLeft, errors)
                                        .then(() => resolve())
                                        .catch(err => reject(err));
                    })(resolve, reject, retryCount - 1, errors), 100);
                }
                else {
                    // No retries left: We're going to have to give up and call 'reject()'.... :-(
                    reject("Could not update " + self.friendly(type) + "s in the database: " + errors.join(". "));
                }
            }
            
            self.databasePool.connect(tenant, taskName)
            .then(async function (connection) {
                // Track how many entities we haven't processed yet.
                let pending = entityList.length;

                async function onEntityProcessed(err) {
                    pending--;      // We've procesed one.

                    if (!connection) return;

                    if (err) {
                        // A query reports an error. Rollback the transaction. Retry it.
                        connection.rollback(
                            () => retry(err),
                            (error) => retry(err, error)
                        );
                        connection = null;  // Other queries may report the same error!
                    }
                    else if (pending <= 0) {
                        // No pending queries anymore. Commit the transaction. Retry if that fails.
                        connection.commit(
                            () => resolve(),
                            (error) => retry(error)
                        );
                        connection = null;  // Just to be safe...
                    }
                    else {
                        // We still have pending queries. Don't do anything yet.
                    }
                }

                connection.begin()
                .then(() => {
                    const statement  = "REPLACE INTO " + self.databasePool.tableName + " (id, tenant, entityType, lastSeen, lastUpdated, displayName, entity, checksum) VALUES (?, ?, ?, ?, ?, ?, ?, ?)";

                    /*  We have discovered that '.forEach()' has horrendous performance and memory
                        characteristics. But avoiding it comes at a cost... We 'll have to fix the
                        current value of 'entity' so that inside the 'query().then()' we are still
                        referring to the same one. So, we just bind the whole body to a dedicated
                        scope using an anonymous function.
                    */
                    for (let i=0; i < entityList.length; i++) {
                        (function (entity) {
                            // Check if this is a new entity or if it has changed and needs to be updated.
                            const lastUpdated = (new Date()).getTime();
                            const lastSeen    = !entity.lastSeenTimestamp || entity.lastSeenTimestamp == -1
                                            ? (new Date()).getTime()
                                            : Number(entity.lastSeenTimestamp);
                            const checksum    = createChecksum(entity);

                            let needsUpdate = true;
                            try {
                                // No need to do anything if this entity is either unknown by Dynatrace,
                                // or is known to us and it's checksum is still the same.
                                const curatedEntity = self.curator.getEntity(tenant, type, entity);
                                needsUpdate = !(entity.isUnknown || (curatedEntity && curatedEntity.checksum === checksum));
                            }
                            catch (ex) {
                                // The entity may not exist yet - then it needs to be updated, obviously.
                            }

                            if (!needsUpdate)
                                setImmediate(onEntityProcessed);
                            else
                                connection
                                .query(statement, [
                                    entity.entityId,
                                    tenant,
                                    type,
                                    lastSeen,
                                    lastUpdated,
                                    entity.displayName.substring(0, maxDisplayNameSize),
                                    JSON.stringify(entity, (key, value) => {
                                        if (typeof value === "string") {
                                            value = value                           // Protect the database:
                                                .replace(/\0/g, '')                 // Remove null characters.
                                                .substring(0, maxStringValueSize);  // Truncate overly long values.
                                        }
                                        return value;
                                    }),
                                    checksum
                                ])
                                .then(result => {
                                    try {
                                        if (result.affectedRows === 0)
                                            throw new Error("Could not insert/update '" + entity.displayName + "' in database");

                                        self.entityAccess.updateEntity(tenant, type, entity, staleTime, checksum, lastUpdated);
                                        onEntityProcessed();
                                    } 
                                    catch (err) {
                                        /*  We could not successfully execute a query in the transaction. We need to
                                            roll back the whole transaction and retry - and at some point give up.
                                        */
                                        onEntityProcessed(err.message);
                                    }
                                })
                                .catch(err => {
                                    /*  We could not successfully execute a query in the transaction. We need to roll
                                        back the whole transaction and retry a couple of times before giving up.
                                    */
                                    onEntityProcessed(err);
                                });
                        })(entityList[i]); // Call and pass current entity to anonymous function to get a private scope!
                    }
                })
                .catch(err => {
                    // We could not successfully start a transaction. Let's retry - or ultimately reject our promise.
                    connection.release(() => retry(err));   // We did get a connection, so we do need to close it first.
                });
            })
            .catch(err => {
                // We could not obtain a connection. Let's retry - or ultimately reject our promise.
                retry(err.message || err);
            });
        });
    }

    getRetirableEntities(tenant, type, cutOffTime) {
        const self = this;

        return new Promise(function (resolve, reject) {
            const taskName = "retireEntities_" + tenant + "_" + type;
            let statement = "SELECT id FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' AND entityType='" + type + "' AND lastSeen<" + cutOffTime;

            self.databasePool.connect(tenant, taskName)
            .then(connection => {
                connection.query(statement)
                .then(entities => {
                    connection.release(() => resolve(entities.map(e => e.id)));
                })
                .catch(err     => {
                    self.logger.error(err);
                    connection.release(() => reject(err));
                });
            })
            .catch(err => reject(err));
        });
    }

    async retireEntities(tenant, type, cutoffDate) {
        const self = this;

        // With this shared task name, each tenant/type combination has to wait
        // until the previous one finished. This avoids deadlocks...
        const taskName = "retireEntities";
        const cutOffTime = cutoffDate ? cutoffDate : (new Date()).getTime();

        let deleteCount = 0;

        // Update our internal administration first.
        const proxies = self.curator.getEntities(tenant, type);
        try {
            const entityIds = await self.getRetirableEntities(tenant, type, cutOffTime) || [];

            entityIds.forEach(entityId => {
                let entityProxy = proxies.get(entityId);
                if (entityProxy) {
                    self.entityAccess.removeEntity(tenant, type, entityProxy);
                    deleteCount++;
                }
            });

            self.curator.clearDiscovered(tenant, type);
        }
        catch (ex) {
            self.logger.error("Cannot get retireable entities from database: " + ex);
        }

        // Next, update the database.
        self.databasePool.connect(tenant, taskName)
        .then(connection => {
            let statement = "DELETE FROM " + self.databasePool.tableName + " WHERE tenant='" + tenant + "' AND entityType='" + type + "' AND lastSeen<" + cutOffTime;

            connection.begin()
            .then(() => {
                connection.query(statement)
                .then(result => {
                    self.logger.info("Deleted " + result.affectedRows + " " + self.friendly(type) + "s in " + tenant + " older than " + (new Date(cutOffTime)).toLocaleString() + " (removed " + deleteCount + ", " + proxies.length + " remaining)");

                    connection.commit();
                })
                .catch(err => {
                    self.logger.error("Could not purge " + self.friendly(type) + "s in " + tenant + ": " + (err.message || err));

                    connection.rollback();
                });
            })
            .catch(err => {
                self.logger.error("Could not start transaction in database: " + (err.message || err));

                connection.release();
            });
        })
        .catch(err => {
            self.logger.error("Could not connect to database: " + (err.message || err));
        });
    }
}

class SQLite extends DataAccess {

    createTable() {
        const self = this;

        return new Promise((resolve, reject) => {                      
            const statement = "CREATE TABLE `" + self.databasePool.tableName + "` ( " 
                + " `id` varchar(255) NOT NULL, "
                + " `tenant` nvarchar(100) NOT NULL, "
                + " `entityType`varchar(100) NOT NULL, "
                + " `lastSeen` UNSIGNED BIG INT NOT NULL, "
                + " `lastUpdated` UNSIGNED BIG INT NOT NULL, "
                + " `displayName` nvarchar(" + maxDisplayNameSize + ") NOT NULL, "
                + " `checksum` binary(32) NULL DEFAULT NULL, "
                + " `entity` longtext NOT NULL, "
                + " PRIMARY KEY (`id`,`tenant`) "
                + ")";
            
            const promises = self.tenants.map(tenant => new Promise((resolve, reject) => {
                self.logger.info("    Looking for entities table for " + tenant);

                self.databasePool.connect(tenant, "createTable")
                .then(connection => {
                    connection
                    .query(statement)
                    .then(result => {
                        connection.release(() => {
                            self.logger.info("    Created entities table for " + tenant);
                            resolve();
                        });
                    })
                    .catch(error => {
                        connection.release(() => {
                            if (error.code === "SQLITE_ERROR" && error.message.includes("already exists")) {
                                self.logger.info("    Found entities table for " + tenant);
                                resolve();
                            }
                            else {
                                self.logger.error("    Could not create entities table for " + tenant + ": " + (error.message || error));
                                reject(error);
                            }
                        });
                    });
                })
                .catch(error => {
                    reject(error);
                });
            }));

            Promise.all(promises).then(resolve).catch(error => {
                self.logger.error("Could not connect to database: " + (error.message || error));
                reject("Could not connect to database: " + (error.message || error));
            });
        });
    }
}

class InnoDB extends DataAccess {

    createTable() {
        const self = this;

        return new Promise((resolve, reject) => {
            const statement = "CREATE TABLE `" + self.databasePool.tableName + "` ( "
                + " `id` varchar(255) NOT NULL, "
                + " `tenant` nvarchar(100) NOT NULL, "
                + " `entityType` enum('HOST','SERVICE','PROCESS_GROUP','PROCESS_GROUP_INSTANCE','APPLICATION','MOBILE_APPLICATION','CUSTOM_APPLICATION') NOT NULL, "
                + " `lastSeen` bigint(20) unsigned NOT NULL, "
                + " `lastUpdated` bigint(20) unsigned NOT NULL, "
                + " `displayName` nvarchar(" + maxDisplayNameSize + ") NOT NULL, "
                + " `checksum` binary(32) NULL DEFAULT NULL, "
                + " `entity` longtext NOT NULL, "
                + " PRIMARY KEY (`id`,`tenant`) "
                + ") ENGINE=InnoDB DEFAULT CHARSET=utf8";
            
            const promises = self.tenants.map(tenant => new Promise((resolve, reject) => {
                self.logger.info("    Looking for entities table for " + tenant);

                self.databasePool.connect(tenant, "createTable")
                .then(connection => {
                    connection
                    .query(statement)
                    .then(result => {
                        connection.release(() => {
                            self.logger.info("    Created entities table for " + tenant);
                            resolve();
                        });
                    })
                    .catch(error => {
                        connection.release(() => {
                            if (error.code === "ER_TABLE_EXISTS_ERROR") {
                                self.logger.info("    Found entities table for " + tenant);
                                resolve();
                            }
                            else {
                                self.logger.error("    Could not create entities table for " + tenant + ": " + (error.message || error));
                                reject(error);
                            }
                        });
                    });
                })
                .catch(error => {
                    reject(error);
                });
            }));

            Promise.all(promises).then(resolve).catch(error => {
                self.logger.error("Could not connect to database: " + (error.message || error));
                reject();
            });                
        });
    }
}

module.exports = { InnoDB, SQLite }