'use strict';
 /**
  * This is the interface that database connectors must implement.
  * @interface
  */
class DatabaseAPI {
        /**
         * Connect to the database using the options specified in the 'config'
         * object, and initialize a connection pool.
         * @param {Object} config - The configuration settings for this adapter.
         * @param {Object} logger - The logger at which can be logged.
         */   
        constructor(config, logger) {}

        /**
         * Creates and returns a {@link DatabaseConnection} object, initialized with the parameters below.
         * The object's {@link DatabaseConnection#state} value is initialized to "Queued".
         * @param {string} name - The name of the connection. This name will be used by the DatabaseTracker to serialize requests for a database connection by the same task.
         * @param {function} resolve - (Private) Function to be called when a connection has been acquired.
         * @param {function} reject - (Private) Function to be called when a connection cannot be acquired.
         * @param {function} cleanup - (Private) Function to be called when a connection has been released.
         * @returns {DatabaseConnection}
         */
        createConnection(name, resolve, reject, cleanup) {}

        /**
         * Acquire a connection to the database for the provided {@link DatabaseConnection}.
         *
         * First set the {@link DatabaseConnection#state} value to "Connecting".
         * Next, acquire a connection ('conn') from the pool.
         * Once available, call {@link DatabaseConnection#initialize} with that connection (i.e. 'conn').
         * After that, call {@link DatabaseConnection#resolve} with the provided {@link DatabaseConnection}.
         * Finally, call 'onAcquired()' (if specified).
         *
         * If a connection cannot be acquired,
         * call {@link DatabaseConnection#cleanup} with the provided {@link DatabaseConnection}, and
         * call {@link DatabaseConnection#reject} with the error.
         * @param {DatabaseConnection} connection - The connection object created by {@link DatabaseAPI#createConnection}.
         * @param {function} [onAcquired] - The function to be called once the connection has been acquired.
         * @async
         */
        acquire(connection, onAcquired) {}

        /**
         * Return the number of currently active connections.
         */
        get activeConnections() {}

        /**
         * Return the maximum number of available connections.
         */
        get totalConnections() {}

        /**
         * Close the connection pool. Call resolve() afterwards.
         * @param {function} resolve
         * @async
         */
        close(resolve) {}
}

 /**
  * A connection object that exposes standardized methods to interact with the database.
  * An instance of this class can only be created by {@link DatabaseAPI#createConnection}.
  * The {@link DatabaseConnection#state} value is initialized to "Queued".
  * 
  * This is the interface that database connections must implement.
  * @interface
  */
class DatabaseConnection {
    /**
     * @param {Object} adaptor - An adaptor that inherits from {@link DatabaseAPI}.
     * @param {Object} config  - The configuration settings for this adapter.
     * @param {Object} logger  - The logger at which can be logged.
     */   
    constructor(adaptor, config, logger) {}    

    /** @member {string} - The name of the connection. This name will be used by
     * the DatabaseTracker to serialize requests for a database connection by the same task.
     */
    get name() {}
    /** @member {string} - The state of the connection. Valid values are "Queued",
     * "Connecting" and "Connected".
     */
    get state() {}
    /** @member {function} - Function to be called when a connection has been acquired. */
    get resolve() {}
    /** @member {function} - Function to be called when a connection cannot be acquired. */
    get reject() {}
    /** @member {function} - Function to be called when a connection has been released. */
    get cleanup() {}

    /**
     * This method should only be called by {@link DatabaseAPI#acquire}. The 'conn'
     * argument is the value passed by 'acquire()', and for which the methods below
     * will be called. it is recommended to save this value as 'this.conn'.
     * Upon completion of this method the {@link DatabaseConnection#state} value must
     * be set to "Connected". Note that the other methods only need to be available
     * after this method has been called.
     * @param {object} conn
     */
    initialize(conn) {}

    /**
     * Start a new transaction. The 'commit()' or 'rollback()' method will be called later.
     * @async
     */
    begin() {}

    /**
     * Return a promise to execute the statement, using the values if specified.
     * @param {string} statement - SQL statement, optionally with '?' placeholders to be
     * resolved against the provided values.
     * @param {string[]} [values] - Values for the placeholders.
     * @async
     */
    query(statement, values) {}

    /**
     * Execute the statement, using the values if specified, and stream the result back.
     * @param {string} statement - SQL statement, optionally with '?' placeholders to be
     * resolved against the provided values.
     * @param {string[]} [values] - Values for the placeholders.
     * @async
     */
    stream(statement, values) {}

    /**
     * Commit transaction and release the connection. Call 'onDone()' or 'onError(err)' when done.
     * @param {function} onDone
     * @param {function} onError
     * @async
     */
    commit(onDone, onError) {}

    /**
     * Rollback transaction and release the connection. Call 'onDone()' or 'onError(err)' when done.
     * @param {function} onDone
     * @param {function} onError
     * @async
     */
    rollback(onDone, onError) {}

    /**
     * Release the connection. Call 'cleanup()' if noCleanup not 'true'. Call onReleased() when done.
     * @param {function} onReleased
     * @param {boolean} [noCleanup=false]
     * @async
     */
    release(onReleased, noCleanup) {}
}

const DatabaseManager = function (adaptor, config, logger) {
    const tableName = "__Entities__";

    const connectTimeout = config.connectTimeout;
    // We leave this here as the adapters use it as well.

    const retryLimit = config.retryLimit || 1;
    delete config.retryLimit;    // The database adaptor itself doesn't need this property.

    const maxQueueTime = (config.maxQueueTime || 30) * 1000;
    delete config.maxQueueTime;  // The database adaptor itself doesn't need this property.

    const connectionTTL = (config.connectionTTL || 30) * 1000;
    delete config.connectionTTL; // The database adaptor itself doesn't need this property.

    const database = new adaptor(config, logger, tableName);
    const tasks    = {};
    const queue    = [];   // Overflow queue for requests that exceed the connection limit.

    function healthReport(callback, extendedVersion) {
        const now = (new Date()).getTime();

        let report = null;

        if (extendedVersion) {
            report = [];

            for (let taskName in tasks) {
                if (tasks.hasOwnProperty(taskName)) {
                    let taskReport = {
                        name: taskName,
                        queued: [],
                        connecting: [],
                        connected: []
                    };

                    for (let i = 0; i < tasks[taskName].connections.length; i++) {
                        let connData = tasks[taskName].connections[i];

                        switch (connData.state) {
                            case "Queued":
                                taskReport.queued.push({
                                    waiting: Math.round((now - connData.queueTime) / 1000),
                                    timeout: Math.round(maxQueueTime / 1000)
                                });
                                break;
                            case "Connecting":
                                taskReport.connecting.push({
                                    waiting: Math.round((now - connData.queueTime) / 1000),
                                    timeout: Math.round(connectTimeout / 1000)
                                });
                                break;
                            case "Connected":
                                taskReport.connected.push({
                                    waiting: Math.round((now - connData.queueTime) / 1000),
                                    timeout: Math.round(connectionTTL / 1000)
                                });
                                break;
                            default:
                                logger.error("Database connection with unknown state '" + connData.state + "'found");
                        }
                    }
                    report.push(taskReport);
                }
            }
        }
        else {
            let pending = 0;
            let active = 0;

            for (let taskName in tasks) {
                if (tasks.hasOwnProperty(taskName)) {
                    if (tasks[taskName].connections.length > 0) {
                        active += 1;
                        pending += tasks[taskName].connections.length - 1;
                    }
                }
            }
            report = {
                taskActive:  active,
                taskPending: pending,
                queued: queue.length,
                active: database.activeConnections,
                total:  database.totalConnections
            };
        }

        if (callback)
            callback(report);
        else
            return report;
    }

    function acquireConnection(connection) {
        // We're handling this request now, so clear the autoCancel timer.
        clearTimeout(connection.autoCancel);

        database.acquire(connection, () => {
            // Set a timer to auto-release the connection if that doesn't happen automatically.
            connection.queueTime = (new Date()).getTime();
            setTimeout(connection.release.bind(connection), connectionTTL);
        });
    }

    function cancelConnection(connection) {
        cleanupConnection(connection);

        connection.reject("Connection was not issued within " + Math.round(maxQueueTime/1000) + "s");
    }

    function cleanupConnection(connection) {
        if (!tasks[connection.name] ||
            !tasks[connection.name].connections ||
            !tasks[connection.name].connections.includes(connection))
            return;

        tasks[connection.name].connections.remove(connection);

        let nextConnection = null;
        if (tasks[connection.name].connections.length === 0) {
            delete tasks[connection.name];      // Don't keep old tasks around.
            nextConnection = queue.shift();     // From the start - oldest first.
        }
        else {                                  // From the start - oldest first.
            nextConnection = tasks[connection.name].connections[0];
            const index = queue.indexOf(nextConnection);
            if (index >= 0) queue.splice(index, 1);
        }

        if (!nextConnection) return;            // Nobody's waiting for anything.

        // We have a connection from the overflow queue or one that was serialized.
        acquireConnection(nextConnection);
    }

    return {
        connect: function (tenant, taskName) {
            // Ensure that the tenant name does not contain characters that are
            // invalid in table names.
            tenant = tenant.replace(/[^\w]/g, "_");
            
            return new Promise((resolve, reject) => {
                /*  Get and/or create the task for which we're creating a connection.
                    Note that creating a connection is not the same thing as acquiring
                    a connection to the database. We'll call 'acquireConnection()' when
                    the time is right. At that point our promise will get resolved or
                    rejected.
                 */
                tasks[taskName] = tasks[taskName]  || { connections: [] };

                const connection = database.createConnection(tenant, taskName, resolve, reject, cleanupConnection);
                connection.queueTime  = (new Date()).getTime();     // Track its age.
                connection.autoCancel = setTimeout(cancelConnection, maxQueueTime, connection);

                tasks[taskName].connections.push(connection);

                if (database.activeConnections >= config.connectionLimit) {
                    /*  If we are at the connection limit, we need to put this request
                        into an overflow queue, to be handled as previous connections get
                        released. This queue is first-in-first-out, obviously.
                     */
                    queue.push(connection);
                    return;
                }

                if (tasks[taskName].connections.length > 1) {
                    /*  If this task already had a connection, this one will have to wait.
                        It will be handled when the task releases its previous connection.
                     */
                    return;
                }

                // The coast is clear. Let's acquire an actual database connection.
                acquireConnection(connection);
            });
        },
        release: function (taskName) {
            logger.info("    Releasing database connections for " + taskName);

            if (!tasks[taskName] || !tasks[taskName].connections) return;

            tasks[taskName].connections.forEach(connection => connection.release(undefined, true));
            delete tasks[taskName];
        },
        close: function (onDone) {
            database.close(onDone);
        },
        connections: healthReport,
        getHealthMetrics: function (frequency, callback) {
            if (frequency && callback)
                return setInterval(healthReport, frequency, callback, true);
            else
                return healthReport();
        },
        get tableName() {
            return tableName;
        },        
        get retryLimit() {
            return retryLimit;
        },
        isRetryable: database.isRetryable
    }
}

module.exports = DatabaseManager;