const http  = require('http');
const https = require('https');
const EventEmitter = require('events');
const Ring  = require('./Ring.js');

if (!Array.prototype.remove)
    Array.prototype.remove = function (elem) {
        for (let i=0; i < this.length; i++)
            if (this[i] === elem)
                return this.splice(i, 1);
        return null;
    };
if (!Array.prototype.removeAll)
    Array.prototype.removeAll = function () {
        return this.splice(0, this.length);
    };

/**
 * @typedef RequestOptions
 * @property {string} tenant    - Name of the tenant. The tenant's definition in config.json
 *                                provides the base URL, port and token information.
 * @property {string} endPoint  - The URL of the request. The URL is relative to the base URL for
 *                                the specified tenant.
 * @property {string} [queryData=""]    - The query parameters to be appended to the request.
 * @property {string} [postData=null]   - The data to be posted as part of the request.
 * @property {string} [method=GET|POST] - The HTTP request method. The default is 'GET'
 *                                if the 'postData' parameter is empty, and 'POST' otherwise.
 * @property {boolean} [stream=false]   - Whether the response should be streamed back to the caller.
 * WHAT HAPPENS HERE? More details. The postData is piped to the reuqest,
 * amd the response if streamed back to the caller. How? See PROCESS_GROUP_INSTANCE implementation.
 * @property {number} [timeout=maxQueueTime] - The number of seconds before the request will time out.
 * @property {number} [retryLimit=tenantRetryLimit] - The number of times the request can be retried.
 *        The default is the number of retries specified in the tenant definition in config.json.
 */

/**
 * @typedef RequestError
 * @property {number} code - The HTTP status code.
 * @property {string} message - The error message.
 */

/**
 * @callback RequestCallback
 * @param {RequestError} error - The error if the request could not be fulfilled.
 * @param {object} response - The Dynatrace API's response, as a parsed JSON object,
 * if the request was succesful.
 */

 /**
  * A cancellable event emitter at which the progress of a request
  * can be monitored and at which the request can be canceled. See the NodeJS
  * [EventEmitter]{@link https://nodejs.org/api/events.html#events_class_eventemitter}
  * documentation for more information.
  * @extends EventEmitter
  */
class CancellableEventEmitter extends EventEmitter {
    /**
     * Creates an instance.
     * @param {object} request - An object that should expose an 'abort()' method
     * that accepts a {@link RequestError}.
     */
    constructor(request) {
        super();
        this.request = request || { abort: () => {} };
        this.on('error', () => {});     // One MUST be registered. Otherwise... BOOM!
    }

    /**
     * Cancel the request. It will either (1) cause the {@link RequestCallback} provided to
     * the {@link DynatraceAPI#request} method to be called, or (2) the  {@link CancellablePromise}
     * returned by {@link DynatraceAPI#request} to reject. In both cases a {@link RequestError}
     * will be passed as the first argument, with code 500 and this reason as its message.
     * @param {string} reason - The reason for the cancellation.
     */
    cancel(reason) {
        try {
            this.request.abort({
                status: 500,
                message: reason || "The request was aborted"
            });
        } catch (ex) {
            /* NOP */ }
        return null;
    }
}

/**
 * A cancellable Promise. This is technically not a real 'Promise', but rather an emulation that
 * also exposes a 'cancel(reason)' method.
 * See the [Promise]{@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise}
 * documentation for more information. Note: this class is documented for informational purposes
 * only. You cannot create your own instance of this class.
 * @extends Promise
 */
class CancellablePromise {
    /**
     * Cancel the promise. It will cause the promise to reject with a {@link RequestError}
     * with code 500 and this reason as its message. This method should only be
     * called after 'then()' and 'catch()' have been called.
     * @param {string} reason - The reason for the cancellation.
     */
    cancel(reason) {}
}

/**
 */

/**
 * Creates a pool of connections to the Dynatrace API for multiple tenants and clusters.
 * @constructor
 * @param {object} config - The configuration for the tenants.
 * @param {object} limits - The rate limits, timeouts and queue sizes.
 * @classdesc Provides load balancing and rate-limited access to a pool of hosts that
 * expose the Dynatrace API.
 */
function DynatraceAPI(config, limits) {}

/**
 * Submit a request to the Dynatrace API. The DynatraceAPI will find the best
 * host to handle the request, or queue the request if necessary.
 * @param {RequestOptions} requestOptions - The parameters for the request.
 * @param {RequestCallback} callback - (Optional) The callback that handles the response. If specified, this method returns a {@link CancellableEventEmitter}. If not specified, this method returns a {@link CancellablePromise}.
 * @returns {CancellableEventEmitter|CancellablePromise} Either (1) an object at
 * which the progress of the request can be monitored and at which the request can
 * be canceled, or (2) a Promise(-like) object at which the request can be canceled.
 */
DynatraceAPI.prototype.request = function (requestOptions, callback) {}

/**
 * Cancel all running and/or queued requests and reset the connection pool.
 */
DynatraceAPI.prototype.restart = function () {}

/**
 * Cancel all running and/or queued requests and delete the connection pool.
 */
DynatraceAPI.prototype.shutdown = function () {}

/**
 * Report the health of the connection pool. If no parameters are specified, the health
 * report will be returned. If both parameters are specified, teh callback will be called
 * with the health report as its only parameter, per the specified frequency.
 * @param {number} [frequency] - The requency with which the health of the
 * connection pool should be reported.
 * @param {function} [callback] - The function that consumes the health report.
 */
DynatraceAPI.prototype.getHealth = function (frequency, callback) {}

const DynatraceAPIFactory = function (config, limits) {
    const self = this;

    config = config || {};
    limits = limits || {};

    let   reqRateLimit  = limits.reqRateLimit  || 1200;  // Maximum request rate in requests per minute.
    const requestLimit  = limits.requestLimit  || 10;    // Upper limit to number of outstanding requests.
    const maxQueueTime  = (limits.maxQueueTime || 10) * 1000;    // Maximum time for a request to stay queued (ms).
    const retryLimit    = limits.retryLimit    || 1;     // Maximum number of retries before reporting an error.
    const queueLimit    = limits.maxQueueSize  || 10;    // Maximum length of the local queues (ie. highWaterMark).

    const Queue = function (queueLimit = Number.POSITIVE_INFINITY) {
        const queue = [];

        function abort(error, delay) {
            // Note: 'this' is assumed to be bound to the 'reqData' object.
            this.attempts = 0;  // Prevent retries.

            // We'll abort the actual request (if it still exists) for good measure.
            if (this._request) {
                this._request.destroy(); // Will trigger 'onError' but that's OK.

                release(this, error || {
                    status:  408,
                    message: "Request did not return" + (delay ? " within " + delay + "s" : "")
                });
            } else {
                release(this, error  || {
                    status:  429,
                    message: "Request was not issued" + (delay ? " within " + delay + "s" : "")
                });
            }
        }

        function place(reqData) {
            let timeout = reqData.timeout || maxQueueTime;

            reqData.queueTime  = (new Date()).getTime();
            reqData.abort      = abort.bind(reqData);
            reqData.autoCancel = setTimeout(reqData.abort, timeout, null, Math.round(timeout / 1000));

            queue.push(reqData);
        }

        function take(i) {
            if (queue.length === 0) return undefined;

            const reqData = i > 0 ? queue.splice(i, 1)[0] : queue.shift();
            clearTimeout(reqData.autoCancel);

            return reqData;
        }

        function release(reqData, error, data) {
            reqData = queue.remove(reqData);
            if (!reqData) return;
            delete reqData._request;

            clearTimeout(reqData.autoCancel);

            if (error)
                reqData.emitter.emit("error", error.status + " - " + error.message);
            else
                reqData.emitter.emit("progress", "end");

            // The callback will call the requestor only once, ever, and will catch any exceptions.
            setImmediate(() => reqData.callback(error, data));
        }

        function clear(i) {
            let reqData = null;

            if (i === undefined) {
                while (reqData = queue.shift()) {
                    clearTimeout(reqData.autoCancel);

                    reqData.abort({
                        status: 500,
                        message: "The request queue was cleared"
                    });
                }
            }
            else if (reqData = take(i)) {
                clearTimeout(reqData.autoCancel);

                reqData.abort({
                    status: 500,
                    message: "The request was cancelled"
                });
            }
            else {
                // NOP
            }
        }

        return {
            place:   place,
            take:    take,
            release: release,
            clear:   clear,
            peek:    (i) => queue[i || 0],
            get isFull() { return queue.length >= queueLimit; },
            get length() { return queue.length; }
        };
    };

    const Host = function (hostName) {
        const me = hostName;
        const issuedList = new Queue(requestLimit);
        const localQueue = new Queue(queueLimit);
        const bucket  = {                           // This is the leaky bucket we use to throttle requests.
            capacity: reqRateLimit,                 // Rate in max requests per minute (equals bucket size).
            dripRate: 60000 / reqRateLimit,         // If reqRateLimit=30 (req/min), add 1 drop every 2000ms.
            content:  0,
            updated:  0
        };
        let hasConnectedOnce = false;

        function reset() {
            hasConnectedOnce = false;
            localQueue.clear();
            issuedList.clear();
            refill(true);

            // Ensure that we keep emptying the global queue.
            setTimeout(acceptNext, bucket.dripRate);
        }

        function refill(completely) {
            // Refill the bucket proportional to the time elapsed since last refill.
            const now      = (new Date()).getTime();
            const refill   = completely
                           ? bucket.capacity
                           : Math.floor((now - bucket.updated) / bucket.dripRate);        // How many drops should be added?
            bucket.content = Math.min(bucket.capacity, bucket.content + refill);          // Don't exceed bucket capacity.
            bucket.updated = Math.min(now, bucket.updated + (refill * bucket.dripRate));  // Update timestamp for refill.
        };

        function getAPIResetDelay(headers) {
            if (!headers)
                return 0;

            let rateLimitRemaining = Number(headers["x-ratelimit-remaining"]);
            if (isNaN(rateLimitRemaining) || rateLimitRemaining > 0)
                return 0;

            // Abide by the request rate limit the API tells us.
            let apiRateLimit = Number(headers["x-ratetime-limit"]);
            if (!isNaN(apiRateLimit) && apiRateLimit > 0 && apiRateLimit < reqRateLimit) {
                reqRateLimit = apiRateLimit;
                bucket.capacity = reqRateLimit;
                bucket.dripRate = 60000 / reqRateLimit;
            }

            // Calculate delay until the API resets its usage limit
            let resetTimeEpoch = Number(headers["x-ratelimit-reset"]);
            if (resetTimeEpoch > 9999999999999)  // 13 digits - ms
                resetTimeEpoch = Math.ceil(resetTimeEpoch / 1000);

            const resetDelay =  resetTimeEpoch - (new Date()).getTime();
            return resetDelay < 0 || resetDelay > (2*60*60*1000) ? 0 : resetDelay;
        }

        async function issueRequest(reqData, delay, nextPageKey, allData) {
            if (delay)  // Wait for the specified amount of time.
                await new Promise(resolve => setTimeout(resolve, delay));

            allData = allData || [];

            try {
                if (nextPageKey) {
                    // Keep the original path so that we can set the nextPageKey again and again easily.
                    reqData._path = reqData._path || reqData.options.path;
                    reqData.options.path = reqData._path +"&nextPageKey=" + encodeURIComponent(nextPageKey);
                }
                else {
                    // Tell the original requestor that we're handling its request now.
                    reqData.emitter.emit("progress", "start");
                }

                // Keep the actual request handle in the request data so that we can cancel it later.
                const request = reqData._request = reqData.protocol.request(reqData.options, result => {
                    result.setEncoding('utf8');

                    // Check if the response contains headers telling us to slow down.
                    let waitTime = getAPIResetDelay(result.headers);
                    if (waitTime > 0 && nextPageKey) {
                        /*  If we are overloading the host but we are retrieving a paged dataset,
                            then we must wait for as long as we're told... and continue collecting
                            the rest of the data.
                        */
                        result.on('data', chunk => { /* Must read all data to avoid memory leak. */ });

                        // Defensive contruction to avoid closure contributing to heap buildup.
                        setTimeout(function(reqData, delay, nextPageKey, allData) {
                            return function() {
                                issueRequest(reqData, delay, nextPageKey, allData);
                            }
                        }(reqData, 100, nextPageKey, allData), waitTime);

                        return;
                    }

                    if (reqData.stream) {
                        // If the caller wants the output stream, just stream result back.
                        onComplete(reqData, result);
                        return;
                    }

                    var data = [];

                    result.on('data', chunk => data.push(chunk));
                    result.on('end', () => {
                        try {
                            var parsedData = JSON.parse(data.join(""));        // Parse concatenated chunks.
                            if (!parsedData.error) {                           // If data is not an error...
                                allData = allData.concat(parsedData);          // Add to previously collected.

                                if (result.headers["next-page-key"]) {         // If there is a next page...
                                    // Defensive construction to avoid closure contributing to heap buildup.
                                    setTimeout(function(reqData, delay, nextPageKey, allData) {
                                        return function() {
                                            issueRequest(reqData, delay, nextPageKey, allData);
                                        }
                                    }(reqData, null, result.headers["next-page-key"], allData), 1);

                                    // PREVIOUSLY:
                                    //   setTimeout(issueRequest, 1, reqData, null, // Get next set asynchronously.
                                    //       result.headers["next-page-key"], allData   // Recurse with extra args.
                                    //   );
                                }
                                else {    // We've received everything. No paging? Just return the parsed data.
                                    onComplete(reqData, nextPageKey ? allData : parsedData);
                                }
                            }
                            else {                                              // Data contains an error.
                                onError(reqData, result.headers, {              // HTTP 429: Too many requests.
                                    status:  parsedData.error.code || (waitTime > 0 ? 429 : 500),
                                    message: "The API returned: " + parsedData.error.message
                                });
                            }
                        }
                        catch (ex) {                                            // Data could not be parsed.
                            onError(reqData, result.headers, {
                                status:  waitTime > 0 ? 429 : 500,              // HTTP 429: Too many requests.
                                message: "Error parsing API response: " + ex.message +
                                        ". Response: " + data.join("").substring(0, 100)
                            });
                        }
                    });
                    result.on('aborted', () => {
                        onError(reqData, null, {
                            status:  502,
                            message: "The response was aborted"
                        });
                    })
                });
                // Ensure that the request doesn't timeout before we will.
                request.setTimeout(reqData.timeout || maxQueueTime);

                request.on('error', err => {
                    delete reqData._request;
                    onError(reqData, null, {
                        status:  err.code || 500,
                        message: err.message || err
                    });
                });
                request.on('abort', err => {
                    delete reqData._request;
                    onError(reqData, null, {
                        status:  500,
                        message: "The request was aborted"
                    });
                });
                request.on('timeout', err => {
                    // We need to do the aborting since nodeJS won't...
                    reqData.abort({
                        status:  408,
                        message: "The request timed out"
                    });
                });

                if (reqData.postData) {
                    if (reqData.stream)
                        request.pipe(reqData.postData);
                    else
                        request.write(reqData.postData);
                }

                request.end();  // Required to send the request.

                return request;
            }
            catch (ex) {
                delete reqData._request;
                ex.status = 500;
                onError(reqData, null, ex);
                return null;
            }
        }

        function onError(reqData, headers, err) {
            err = err || { status: 500, message: "Reason unknown"};

            let now      = (new Date()).getTime();
            let waitTime = getAPIResetDelay(headers) + 10;
            let timeout  = reqData.timeout || maxQueueTime;
            let timeLeft = (reqData.timestamp + timeout) > (now + waitTime);

            if (err.status === 404) {
                issuedList.release(reqData, err);
            }
            else if (err.status === 429) {
                // Downstream service says "Too many requests". Wait and retry as long as we have time.
                if (timeLeft)
                    return setTimeout(issueRequest, waitTime, reqData);
                else
                    issuedList.release(reqData, {   // We ran out of time. Report the error.
                        status: 503,
                        message: "Due to API usage limiting, the request could not be serviced within " + timeout + "ms"
                    });
            }
            else {
                // Some other error. Retry as many times as we're allowed for this request.
                if (timeLeft && reqData.attempts-- > 0)
                    return setTimeout(issueRequest, waitTime, reqData);   // Retry (almost) right away.
                else
                    issuedList.release(reqData, {   // We ran out of retries. Report the error.
                        status: 500,
                        message: "Request to API failed: " + err.message
                    });
            }
            // Ensure that we keep emptying the global queue.
            setTimeout(acceptNext, bucket.dripRate);
        };

        function onComplete(reqData, data) {
            // We had at least one recent successful connection to this host.
            hasConnectedOnce = true;

            // This request can be removed from the list of still-running requests.
            // This will also do the callback to the original requester.
            issuedList.release(reqData, null, data);

            // Ensure that we keep emptying the global queue.
            setTimeout(acceptNext, bucket.dripRate);
        };

        const exposedInterface = {
            get name() {
                return me;
            },
            get health() {
                return "Host " + me + (hasConnectedOnce ? " (alive)" : " (dead)") + ": issued=" + issuedList.length + ", local queue=" + localQueue.length;
            },
            get localQueue() {
                return localQueue;
            },
            get issuedList() {
                return issuedList;
            },
            get canConnect() {
                return hasConnectedOnce;
                // TODO: If we are down, trigger a 'ping' here, too. Maybe. Anyway, if something
                // goes wrong, issue periodic pings until the host is connected again.
            },
            accept: function (reqData, forced) {
                // First, schedule a recursive call to empty the local and global queues.
                if (localQueue.length > 0 || self.globalQueue.length > 0)
                    setTimeout(acceptNext, bucket.dripRate);

                // If we're full and not forced to accept, politely decline.
                if (localQueue.isFull && !forced)
                    return false;

                // Add the request (if any) to the end of the queue.
                if (reqData) {
                    localQueue.place(reqData);
                    reqData = null;    // We're going to take the oldest (first) request first.
                }

                // Make sure were all caught up with the dripping...
                refill();

                // If the bucket is empty, or if there are too many requests still running,
                // we schedule a recursive call to empty the local and global queues.
                if (bucket.content <= 0 || issuedList.isFull) {
                    setTimeout(acceptNext, bucket.dripRate);
                    return true;   // We queued the request.
                }

                // Take the first request waiting in line, if any, but if the global queue has an
                // older one for us, take that one. The first request in our queue us our oldest.
                const oldestTime = localQueue.length > 0 ? localQueue.peek().queueTime : null;

                reqData = self.globalQueue.acceptNext(me, oldestTime) || localQueue.take();
                if (!reqData)
                    return false;   // No more requests waiting that we should or could handle.

                // We accept this request and we'll issue it right now. Update the admin and do it.
                bucket.content--;
                issuedList.place(reqData);

                reqData.options.hostname = me;
                reqData.attempts = reqData.retryLimit;

                setTimeout(issueRequest, 1, reqData);

                return true;    // Accepted - and on its way!
            },
            get availability() {
                // If our waiting list is full, we are infinitely unavailable!
                if (localQueue.isFull)
                    return  Number.NEGATIVE_INFINITY;

                // Our unavailability is proportional to the length of our waiting list.
                if (localQueue.length > 0)
                    return -1 * localQueue.length;

                refill();  // Update the bucket.

                // Our availability is proportional to the fullness of the bucket. If we have outstanding
                // requests, with a limit to them, then that proportionally reduces our bucket fullness.
                return issuedList.length === 0
                     ? bucket.content
                     : bucket.content * (requestLimit - issuedList.length) / requestLimit;
            },
            reset: reset
        };
        const acceptNext = exposedInterface.accept.bind(exposedInterface);

        return exposedInterface;
    };

    this.hosts   = {};
    this.tenants = {};
    this.globalQueue = new Queue();
    this.globalQueue.acceptNext = function (host, olderThan) {
        olderThan = olderThan || (new Date()).getTime();

        let reqData = null;

        for (let i=0; !reqData && i < this.length; i++) {
            const queuedReq = this.peek(i) || {};

            let tenant = self.tenants[queuedReq.tenant];
            let suitableHosts = tenant && tenant.hosts ? tenant.hosts.map(host => host.name) : null;
            if (!suitableHosts)
                this.clear(i), i--;        // This request will never be handled. Clear it.
            else if (suitableHosts.includes(host) && queuedReq.queueTime < olderThan)
                reqData = this.take(i);    // The calling host can handle this one. Take it.
        }

        return reqData;
    };

    /*  We use a single, consolidated accounting of all hosts so that we can properly
        distribute requests accross them, and throttle the requests to each them.
        Note that the tenants' host lists point into our consolidated host object.
     */

    // Collect the tenants in the configuration and initialize our administration.
    Object.keys(config).forEach(tenant => {
        let tenantHosts = [].concat(config[tenant].host);  // Make array.

        // Add the hosts to consolidated hosts object (no duplicates).
        tenantHosts.forEach(host => {
            self.hosts[host] = self.hosts[host] || new Host(host);
        });

        self.tenants[tenant] = {
            hosts:   tenantHosts.map(host => self.hosts[host]),
            port:    config[tenant].port  || 443,
            noSSL:   config[tenant].noSSL || false,
            url:     config[tenant].url   || "",
            token:   config[tenant].token || "No token specified",
            retryLimit: config[tenant].retryLimit || retryLimit || 1
        };
    });

    // Convert the hosts object to a Ring for easy round-robining.
    this.hosts = new Ring(Object.values(this.hosts));

    function reset(forced) {
        if (this.hosts) this.hosts.forEach(host => host.reset(forced));
    }

    /*  This is the main function. Requests as submitted here and shopped across
        the available hosts until we find one that can best handle the request.
        The 'onDone()' callback is called with the result of executing the request,
        or with the error if something went wrong.
     */
    function request(requestOptions, onDone) {
        // Apparently, under error conditions there are multiple
        // callbacks to 'onDone()'. We can either make that a 'callOnce'
        // thing or we can get to the root of things.
        if (self.hosts.length === 0) {
            setTimeout(onDone, 1, { status: 404, message: "No hosts in pool" });
            return new CancellableEventEmitter();
        }

        const tenantOptions = self.tenants[requestOptions.tenant];
        if (!tenantOptions) {
            setTimeout(onDone, 1, { status: 404, message: "Tenant " + requestOptions.tenant + " not found" });
            return new CancellableEventEmitter();
        }

        const timeout   = requestOptions.timeout && requestOptions.timeout < 1000
                        ? requestOptions.timeout * 1000
                        : requestOptions.timeout || maxQueueTime;
        const queryData = requestOptions.queryData ? "?" + requestOptions.queryData : "";
        const method    = requestOptions.method || (requestOptions.postData ? "POST" : "GET");

        // Prepare everything we need to send this request on its way.
        const reqData = {
            options:    {
                port:   tenantOptions.port,
                path:   tenantOptions.url + requestOptions.endPoint + queryData,
                method: method,
                headers: {
                    'Authorization': 'Api-Token ' + tenantOptions.token,
                    'Content-Type' : 'application/json',
                    'Accept'       : 'application/json'
                },
                timeout: timeout,
                rejectUnauthorized: false   // We may bypass load balancers and stuff, so certs may mismatch.
            },
            protocol:   tenantOptions.noSSL ? http : https,
            tenant:     requestOptions.tenant,
            stream:     requestOptions.stream,
            postData:   requestOptions.postData,
            timeout:    requestOptions.timeout,
            timestamp:  (new Date()).getTime(),
            retryLimit: requestOptions.retryLimit || tenantOptions.retryLimit,
            callback:   (err, data) => {
                try {
                    onDone(err, data);
                }
                catch (ex) {
                    reqData.emitter.emit("error", "Error delivering data to requestor: " + ex.message);
                }
            }
        };
        reqData.emitter = new CancellableEventEmitter(reqData);

        let i = self.hosts.length;

        if (requestOptions.noQueue) {
            /*  STRATEGY "noQueue": Don't queue this request because the caller is
                able to wait and retry if it runs into the throttle limit.

                Find a host that is able to accept this request NOW. Stop looking
                for one once we've asked all available hosts (self.hosts.length).
                If nobody is available, we've run out of capacity.
             */
            while (i > 0) {
                let host = self.hosts.next();  // Round-robin over available hosts.

                // If this host serves this tenant, ask it to accept our request.
                if (tenantOptions.hosts.includes(host) && host.accept(reqData))
                    return reqData.emitter;

                i--;    // Count down number of hosts we tried, and keep trying.
            }
            // Nobody accepted! We'll return an error below.
        }
        else if (requestOptions.noReject) {
            /*  STRATEGY "noReject": If there are no hosts that can handle the
                request immediately, put it on a global queue until somebody gets
                to it. Return backpressure if we're getting overloaded.
             */
            while (i > 0) {
                let host = self.hosts.next();  // Round-robin over available hosts.

                // If this host serves this tenant, ask it to accept our request.
                if (tenantOptions.hosts.includes(host) && host.accept(reqData))
                    return reqData.emitter;

                i--;    // Count down number of hosts we tried, and keep trying.
            }
            /*  Nobody accepted! So we put it in the global queue. If everybody is
                busy, requests will complete in the future, and the requests in
                this queue will be consumed.
             */
            self.globalQueue.place(reqData);
            return reqData.emitter;
        }
        else {
            /*  STRATEGY "allowTimeout": It's OK to queue the request, as long as
                it gets done in a reasonable time (ie. up to maxQueueTime).

                Find the most available host and tell the winner to accept our
                request. Queueing is allowed up a maximum. Only if all queues are
                full have we run out of capacity and will we reject the request.
             */
            let bestHost = { host: null, availability: Number.NEGATIVE_INFINITY };

            while (i > 0) {
                let host = self.hosts.next();  // Iterate over available hosts.

                // If this host serves this tenant, is it more available than the best one so far?
                if (tenantOptions.hosts.includes(host) && host.availability > bestHost.availability)
                    bestHost = { host: host, availability: host.availability };

                i--;    // Count down number of hosts we evaluated and compared.
            }
            if (bestHost.host && bestHost.host.accept(reqData, true))
                return reqData.emitter;
        }

        /*  Before we report that everybody is busy and the caller better back off,
            let's see if we aren't dealing with some network issue ('unreachable').
         */
        if (self.hosts.some(host => host.canConnect)) {
            setImmediate(() => onDone({
                status:  429,
                message: "All endpoints for " + requestOptions.tenant + " are busy"
            }));
        }
        else {
            setImmediate(() => onDone({
                status:  503,
                message: "All endpoints for " + requestOptions.tenant + " are unreachable"
            }));
            reset(true);  // If all hosts are unreachable, cancel all waiting requests.
        }

        return reqData.emitter;
    }

    function healthReport(callback) {
        const now = (new Date()).getTime();

        const report = [];

        let hostReport = {
            name:   "Dynatrace API global queue",
            queued: []
        };

        for (let i = 0; i < self.globalQueue.length; i++) {
            let reqData = self.globalQueue.peek(i);

            hostReport.queued.push({
                url:     reqData.options.path,
                waiting: Math.round((now - reqData.queueTime) / 1000),
                timeout: Math.round(reqData.timeout / 1000)
            });
        }
        report.push(hostReport);

        for (let h = 0; h < self.hosts.length; h++) {
            let host = self.hosts.at(h);

            hostReport = {
                name:   host.name,
                queued: [],
                issued: []
            };

            for (let i = 0; i < host.localQueue.length; i++) {
                let reqData = host.localQueue.peek(i);

                hostReport.queued.push({
                    url:     reqData.options.path,
                    waiting: Math.round((now - reqData.queueTime) / 1000),
                    timeout: Math.round(reqData.timeout / 1000)
                });
            }

            for (let i = 0; i < host.issuedList.length; i++) {
                let reqData = host.issuedList.peek(i);

                hostReport.issued.push({
                    url:     reqData.options.url,
                    waiting: Math.round((now - reqData.queueTime) / 1000),
                    timeout: Math.round(reqData.timeout / 1000)
                });
            }

            report.push(hostReport);
        }

        if (callback)
            callback(report);
        else
            return report;
    }

    // Return our public interface.
    return {
        request: function (requestOptions, onDone) {
            /*  If there's a callback, return a CancellableEventEmitter.
                If there is no callback, return a CancellablePromise.

                Both objects expose a 'cancel(reason)' method that may
                be called later for whatever reason.
             */
            return onDone
                ? request(requestOptions, onDone) // Returns a CancellableEventEmitter.
                : {                // And this here is our CancellablePromise instance.
                    _cancellable: request(requestOptions, (err, data) => {
                        /*  This is our own 'onDone'. It is called asynchronously,
                            obviously, so the caller has time to first call 'then()'
                            and 'catch()'.
                         */
                        if (err)
                            this._reject  && this._reject(err);
                        else
                            this._resolve && this._resolve(data);

                        this._reject = this._resolve = null;    // Call only once!
                    }),
                    then:  resolveFunction => {
                        this._resolve = resolveFunction;
                    },
                    catch: rejectFunction => {
                        this._reject  = rejectFunction;
                    },
                    cancel: reason => {
                        // The 'cancellable' is a CancellableEventEmitter.
                        this._cancellable && this._cancellable.cancel(reason);
                    },
                }
        },
        restart: function () {
            reset(true);        // Cancel all waiting requests.
        },
        shutdown: function () {
            reset(true);        // Cancel all waiting requests.
            self.hosts = [];    // No new requests accepted.
        },
        getHealth: function (frequency, callback) {
            if (frequency && callback)
                return setInterval(healthReport, frequency, callback);
            else
                return healthReport();
        }
    };
};

module.exports = DynatraceAPIFactory;