const http  = require('http');
const https = require('https');
const axios = require('axios').default;

const Resolver = require('./Resolver.js');

const DynatraceUI = function(config) {
    const self     = this;
    const resolver = new Resolver();
    const protocol = "https://";

    const uiEndPoints = {
        APPLICATION: "/#uemapplications/uemappmetrics;uemapplicationId=",
        SERVICE: "/#services/serviceOverview;id=",
        PROCESS_GROUP: "/#processgroupdetails;id=",
        PROCESS: "/#processdetails;id=",
        HOST: "/#hosts/hostdetails;id="
    };
    const restEndpoints = {
        SERVICE: "/rest/configuration/deployment/entities/SERVICE",
        PROCESS_GROUP: "/rest/configuration/deployment/entities/PROCESS_GROUP",
        PROCESS: "/rest/configuration/deployment/entities/PROCESS",
        HOST: "/rest/configuration/deployment/entities/HOST",
        PING: "/rest/about/status?gtf=l_2_HOURS"
    };

    self.config = config;

    let tenants = {};
    function loadTenants() {
        tenants = {};
        Object.keys(self.config.Tenants).forEach(tenant => {
            tenants[tenant] = {
                host:     self.config.Tenants[tenant].host, // "MyCluster.dynatrace-managed.com",
                protocol: self.config.Tenants[tenant].protocol || protocol,
                url:      self.config.Tenants[tenant].url,  // "/e/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxx",
                account:  self.config.Tenants[tenant].account
                        ? self.config.Tenants[tenant].account.split("|")
                        : [ "no_uid", "no_pwd" ],
                session: null
            };
        });
    }

    function getTenantName(nameOrId) {
        if (tenants[nameOrId])
            return nameOrId;

        try {
            return Object.keys(tenants).filter(tenant => tenants[tenant].url === nameOrId)[0];
        }
        catch (ex) {
            return null;
        }
    }

    /*  The service account has two purposes:
        1) To log in to a Managed cluster tenant using predefined credentials.
        2) To avoid logging in because we can't. Examples: SaaS clusters and the "INTERNAL" tenant.
            2a) In this case, it serves as a way to force checks on credentials anyway.
            2b) We should not log in again and again. Just once.
            2c) We should keep credentials per client IP address, ideally.
    */
    function checkServiceAccountCredentials(tenant, uid, pwd) {
        return new Promise(function (resolve, reject) {
            if (tenants[tenant].session) // Already logged in.
                return resolve("OK");

            const account = tenants[tenant].account || [];

            if (uid === account[0] && pwd === account[1]) {
                tenants[tenant].session = { token: "Validated" };
                resolve("OK");
            } else {
                reject("Incorrect credentials");
            }
        });
    }

    function createHeaders(protocol, host, refererURL, token, cookies, extras) {
        let headers = {
            'Origin': protocol + host,
            'Host': Array.isArray(host) ? host[0] : host,
            'Referer': refererURL ? (protocol + host + refererURL) : "",
            'Upgrade-Insecure-Requests': 1,
            'DNT': 1,
            'Tab-Id': 'tab-streamer',
            'Accept': "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
        //  'Accept-Encoding': "gzip, deflate, br",
            'Accept-Encoding': "utf8",
            'Accept-Language': "en-US,en;q=0.9",
            'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36",
        };

        if (token) {
            headers['X-CSRFToken'] = token;
        }

        if (cookies) {
            headers['Cookie'] = Object
                .entries(cookies)
                .map(cookieParts => cookieParts[0] + "=" + cookieParts[1])
                .join("; ");
        }

        return extras ? Object.assign(headers, extras) : headers;
    }

    function getRequestBody(request, callback) {
        if (request.method !== 'POST' && request.method !== 'PUT')
            return setImmediate(callback);

        const data = [];

        request.on('data', (chunk) => data.push(chunk));
        request.on('end',  () => callback(request.body = data.join("")));
    }

    function getResponseBody(response, callback) {
        const data = [];

        response.setEncoding('utf8');  // We should set it, too. And do we really need to?
        response.on('data', (chunk) => data.push(chunk));
        response.on('end', () => callback(response.body = data.join("")));
    }

    function areWeLoggedIn(tenant) {
        return new Promise(function (resolve, reject) {
            if (!tenants[tenant].session)
                return reject("No session");

            const host  = Array.isArray(tenants[tenant].host)
                        ? tenants[tenant].host[0]
                        : tenants[tenant].host;
            const requestOptions = {
                rejectUnauthorized: false,   // We may bypass load balancers and stuff, so certs may mismatch.
                hostname: host,
                path:     tenants[tenant].url + restEndpoints["PING"],
                method:   "GET",
                strictSSL: false,
                headers:  createHeaders(
                    tenants[tenant].protocol,
                    host,
                    tenants[tenant].url + "/login",
                    tenants[tenant].session.token,
                    tenants[tenant].session.cookies,
                    { 'Content-Type': "application/json" })
            };

            let pingRequest = https.request(requestOptions, pingResponse => {
                getResponseBody(pingResponse, data => {
                    if (pingResponse.statusCode === 200 )
                        resolve();
                    else
                        reject(pingResponse.statusMessage);
                });
            });
            pingRequest.end();
        });
    }

    function gotoLoginPage(tenant) {
        // Start with an empty session.
        tenants[tenant].session = {
            headers: {},
            cookies: { dtSa: "-", dtLatC: "1" },
            token:   null
        };

        return new Promise(function (resolve, reject) {
            const host  = Array.isArray(tenants[tenant].host)
                        ? tenants[tenant].host[0]
                        : tenants[tenant].host;

            const requestOptions = {
                rejectUnauthorized: false,   // We may bypass load balancers and stuff, so certs may mismatch.
                hostname: host,
                path:     tenants[tenant].url + "/login",
                method:   "GET",
                headers:  createHeaders(tenants[tenant].protocol, host)
            };

            let pageRequest = https.request(requestOptions, pageResponse => {
                getResponseBody(pageResponse, loginPage => {
                    /*  We need to extract the X-CSRFToken from the login page HTML:
                            <input type="hidden" name="X-CSRFToken" id="csrf_token" value="your|token"/>
                     */
                    let xsrfTagStart = Math.max(
                        loginPage.indexOf('name="X-CSRFToken"'),
                        loginPage.indexOf('id="csrf_token"')
                    );
                    if (xsrfTagStart === -1)
                        return reject("X-CSRF token not found in login page");

                    loginPage = loginPage.substring(xsrfTagStart);
                    xsrfTagStart = loginPage.indexOf('value="');
                    loginPage = loginPage.substring(xsrfTagStart + 7);

                    let xcsrfToken = loginPage.substring(0, loginPage.indexOf('"'));
                    tenants[tenant].session.token = xcsrfToken;

                    // Collect additional cookies, if any.
                    if (pageResponse.headers["set-cookie"])
                        pageResponse.headers["set-cookie"].forEach(cookie => {
                            let cookieParts = cookie.split(";")[0].split("=");
                            tenants[tenant].session.cookies[cookieParts.shift()] = cookieParts.join("=");
                        });

                    resolve();
                });
            });
            pageRequest.end();
        });
    }

    function submitCredentials(tenant, uid, pwd) {
        return new Promise(function (resolve, reject) {
            const creds = "user="  + (uid || tenants[tenant].account[0])
                        + "&pass=" + (pwd || tenants[tenant].account[1])
                        + "&backUri="
                        + "&X-CSRFToken=" + encodeURIComponent(tenants[tenant].session.token);
            const host  = Array.isArray(tenants[tenant].host)
                        ? tenants[tenant].host[0]
                        : tenants[tenant].host;

            const requestOptions = {
                rejectUnauthorized: false,   // We may bypass load balancers and stuff, so certs may mismatch.
                hostname: host,
                path:     tenants[tenant].url ? (tenants[tenant].url + "/login") : "",
                method:   "POST",
                headers:  createHeaders(
                    tenants[tenant].protocol,
                    host,
                    tenants[tenant].url + "/login",
                    tenants[tenant].session.token,
                    tenants[tenant].session.cookies,
                    { 'Content-Type': "application/x-www-form-urlencoded" })
            };

            let loginRequest = https.request(requestOptions, loginResponse => {
                const loginPath = "/login";

                getResponseBody(loginResponse, data => {
                    let redirect = data && data.includes("Could not authenticate user")
                                 ? loginPath
                                 : loginResponse.statusCode == 200
                                 ? ""
                                 : loginResponse.statusCode == 302
                                 ? loginResponse.headers["location"] || ""
                                 : loginPath;

                    if (redirect.includes(loginPath))
                        return reject("Login rejected - redirected to: " + redirect + " (HTTP status " + loginResponse.statusCode + ")");

                     // Collect the cookies.
                    loginResponse.headers["set-cookie"].forEach(cookie => {
                        let cookieParts = cookie.split(";")[0].split("=");
                        if (cookieParts[0] === "dtCookie")
                            cookieParts = cookieParts.concat(
                                "app:cluster1.dynatrace-managed.com=1=app:cluster2.dynatrace-managed.com=1".split("=")
                            );  // NodeJS is losing this last part. Check browser cookies for actual values.

                        tenants[tenant].session.cookies[cookieParts.shift()] = cookieParts.join("=");
                    });

                    // We have to follow the redirect to obtain the rest of the headers and cookies we need.
                    resolve(redirect);
                });
            });
            loginRequest.on('error', ex => reject(ex.message));
            loginRequest.write(creds);
            loginRequest.end();
        });
    }

    function gotoLandingPage(tenant, landingPage) {
        // Note: we are actually going to a page of our choosing
        // instead of the login response-suggested landing page.
        return new Promise(function (resolve, reject) {
            const host  = Array.isArray(tenants[tenant].host)
                        ? tenants[tenant].host[0]
                        : tenants[tenant].host;
            const requestOptions = {
                rejectUnauthorized: false,   // We may bypass load balancers and stuff, so certs may mismatch.
                hostname: host,
                path:     tenants[tenant].url + "/#settings",
                method:   "GET",
                headers:  createHeaders(
                    tenants[tenant].protocol,
                    host,
                    tenants[tenant].url + "/login",
                    tenants[tenant].session.token,
                    tenants[tenant].session.cookies,
                    { 'Content-Type': "application/json; charset=UTF-8" }
                )
            };

            let redirectRequest = https.request(requestOptions, redirectResponse => {
                getResponseBody(redirectResponse, landingPage => {
                    // Can we stay here? If so, we're logged in, presumably.
                    if (redirectResponse.statusCode != 200) {
                        console.log("Response after redirect: " + redirectResponse.statusCode);
                        if (redirectResponse.statusCode == 302)
                            console.log("Re-redirected to: " + redirectResponse.headers["location"]);

                        return reject("Login unsuccessful - status: " + redirectResponse.statusCode);
                    }

                    /*  The XCSRFToken we have to use from now on can be found in the HTML of the landing page:
                            <script type="text/javascript" nonce='f43832db-a507-4a91-aca8-23d73e87e400' >
			                    var csrf_token = '844846cb-4526-4591-83bc-23a354c3d4a1|3|aa539dc3-4a6f-4457-ac1c-5e2896b296d4';
                                var csrf_header_name = 'X-CSRFToken';
                     */
                    let xsrfTagStart = landingPage.indexOf("var csrf_token = '");
                    if (xsrfTagStart === -1)
                        return reject("X-CSRF token not found in landing page");

                    landingPage = landingPage.substring(xsrfTagStart + "var csrf_token = '".length);
                    let xcsrfToken = landingPage.substring(0, landingPage.indexOf("'"));
                    tenants[tenant].session.token = xcsrfToken;

                    // Collect additional cookies, if any.
                    if (redirectResponse.headers["set-cookie"])
                        redirectResponse.headers["set-cookie"].forEach(cookie => {
                            let cookieParts = cookie.split(";")[0].split("=");
                            tenants[tenant].session.cookies[cookieParts.shift()] = cookieParts.join("=");
                        });

                    resolve("OK");    // Return to original requester.
                });
            });
            redirectRequest.on('error', ex => reject(ex.message));
            redirectRequest.end();
        });
    }

    function scriptedRequest(tenant, request, collected, requester, retryCount = 0) {
        if (!request) 
            return new Promise((resolve, reject) => reject("Unknown request"));

        const host = Array.isArray(tenants[tenant].host)
                   ? tenants[tenant].host[0]
                   : tenants[tenant].host;

        collected = collected || { request: {}, response: {} };
        requester = requester || axios.create({
            baseURL: tenants[tenant].protocol + host + tenants[tenant].url,
            timeout: 10000,
            headers: createHeaders(
                tenants[tenant].protocol,
                host,
                tenants[tenant].url + "/",
                tenants[tenant].session.token,
                tenants[tenant].session.cookies,
                {   'Connection': "keep-alive",
                    'Content-Type': "application/json; charset=UTF-8",
                    'Accept': "*/*",
                    'X-Last-Action': (new Date()).getTime(),
                }
            )
        });

        return new Promise((resolve, reject) => {
            if (retryCount > 360) {
                reject({
                    message: "Too many retries",
                    config: request
                });
            }
            request.method = request.method || (request.data ? "POST" : "GET");
            request.params = resolver.resolveObject(request.params, collected);
            request.data   = resolver.resolveObject(request.data,   collected);
            request.url    = resolver.resolveObject(request.url,    collected);

            // Decode params because Axios will encode them.
            for (let qp in request.params)
                request.params[qp] = decodeURIComponent(request.params[qp]);

            requester.request(request)
            .then(response => {
                if (request.transformer) {
                    const transformer = require('./' + request.transformer);
                    collected.transformed = transformer.transform(response.data);
                }

                // Collect the response so that we can use it for the next request in the chain.
                Object.assign(collected.request, request.params, request.data);
                Object.assign(collected.response, response.data);

                // Any "onXXX" property can instruct us how to deal with specific
                // non-error status codes. For now, we only support "repeat".
                // EX: If we get a 202, then repeat until we get something else.
                const respPolicy  = request["on" + response.status];
                const nextRequest = typeof respPolicy === "object"
                                  ? respPolicy
                                  : request.next;

                if (respPolicy === "repeat") {
                    setTimeout(() => {
                        scriptedRequest(tenant, request, collected, requester, retryCount+1)
                            .then(resolve)
                            .catch(reject);
                    }, 1000);
                }
                else if (nextRequest) {
                    scriptedRequest(tenant, nextRequest, collected, requester)
                        .then(resolve)
                        .catch(reject);
                }
                else {
                    resolve(response.data);
                }
            })
            .catch(reject);
        });
    }

    function setMonitoringSetting(tenant, type, entityId, settingName, settingValue) {
        return new Promise(function(resolve, reject) {
            const host  = Array.isArray(tenants[tenant].host)
                        ? tenants[tenant].host[0]
                        : tenants[tenant].host;
            const requestOptions = {
                rejectUnauthorized: false,   // We may bypass load balancers and stuff, so certs may mismatch.
                hostname: host,
                path:     tenants[tenant].url + restEndpoints[type] + "/" + entityId + "/" + settingName,
                method:   settingValue ? "PUT" : "GET", // 'truthy': any actual value is seen as 'true'. 'falsy': null, undefined, 0, "" are seen as false.
                headers:  createHeaders(
                    tenants[tenant].protocol,
                    host,
                    tenants[tenant].url + "/",
                    tenants[tenant].session.token,
                    tenants[tenant].session.cookies,
                    {   'Connection': "keep-alive",
                        'Content-Type': "application/json; charset=UTF-8",
                        'Accept': "*/*",
                        'X-Last-Action': (new Date()).getTime()
                    })
            };

            let request = https.request(requestOptions, response => {
                getResponseBody(response, data => {
                    if (response.statusCode < 200 || response.statusCode > 299)
                        reject("Cannot execute request: " + response.statusMessage);
                    else
                        resolve(data || settingValue);
                });
            });
            request.on('error', ex => reject(ex.message));
            request.write(settingValue);
            request.end();
        });
    }

    function setMonitoringMode(tenant, type, entityId, monitoringMode) {
        return new Promise(function(resolve, reject) {
            const host  = Array.isArray(tenants[tenant].host)
                        ? tenants[tenant].host[0]
                        : tenants[tenant].host;

            const requestOptions = {
                rejectUnauthorized: false,   // We may bypass load balancers and stuff, so certs may mismatch.
                hostname: host,
                path:     null,
                method:   "GET",
                headers:  createHeaders(
                    tenants[tenant].protocol,
                    host,
                    tenants[tenant].url + "/",
                    tenants[tenant].session.token,
                    tenants[tenant].session.cookies,
                    {   'Connection': "keep-alive",
                        'Content-Type': "application/json; charset=UTF-8",
                        'Accept': "*/*",
                        'X-Last-Action': (new Date()).getTime(),
                    })
            };

            /*  We first need to know what the current mode is, because we may have to submit
                several requests to set it according to the specified value of 'monitoringMode'.
             */
            requestOptions.path   = tenants[tenant].url + restEndpoints[type] + "/" + entityId + "/monitoringModeHistory";
            requestOptions.method = "GET";

            let request = https.request(requestOptions, response => {
                getResponseBody(response, data => {
                    if (response.statusCode < 200 || response.statusCode > 299)
                        return reject("Cannot get monitoring mode: " + response.statusMessage);

                    let modeData = JSON.parse(data);
                    let lastChange = modeData.records.pop();
                    let mode = lastChange.mode;

                    // If the mode already is what the request wants us to make it, we're done.
                    if (mode === monitoringMode || mode === "ON" && monitoringMode !== "OFF")
                        return resolve({ entityId: entityId, monitoringMode: mode });

                    // Figure out what the calls are we need to make so set the requested mode.
                    let steps = [];
                    switch (monitoringMode) {
                        case "OFF":
                            steps.push({ endpoint: "/monitoringState", body: '{"isMonitoringEnabled":false}' });
                        break;
                        case "ON":
                            steps.push({ endpoint: "/monitoringState", body: '{"isMonitoringEnabled":true}' });
                        break;
                        case "INFRASTRUCTURE":
                            steps.push({ endpoint: "/monitoringState", body: '{"isMonitoringEnabled":true}' });
                            steps.push({ endpoint: "/monitoringMode",  body: 'false' });
                        break;
                        case "FULL_STACK":
                            steps.push({ endpoint: "/monitoringState", body: '{"isMonitoringEnabled":true}' });
                            steps.push({ endpoint: "/monitoringMode",  body: 'true' });
                        break;
                        default:
                            // Just return the mode we just found.
                            return resolve({ entityId: entityId, monitoringMode: mode });
                    }

                    // Now we know what we need to do - let's do it.
                    let promises = steps.map(step => {
                        return new Promise(function (resolve, reject) {
                            requestOptions.path   = tenants[tenant].url + restEndpoints[type] + "/" + entityId + step.endpoint,
                            requestOptions.method = "PUT";

                            let request = https.request(requestOptions, response => {
                                getResponseBody(response, data => {
                                    if (response.statusCode < 200 || response.statusCode > 299)
                                        return reject("Cannot set monitoring mode: " + response.statusMessage);

                                    resolve();
                                });
                            });
                            request.on('error', ex => reject(ex.message));
                            request.write(step.body);
                            request.end();
                        });
                    });
                    Promise
                    .all(promises)
                    .then(result => resolve({ entityId: entityId, monitoringMode: monitoringMode }))
                    .catch(error => reject({  entityId: entityId, error: error }));
                });
            });
            request.on('error', ex => reject(ex.message));
            request.end();
        });
    }

    function runIdPLogin(configuration) {
        const adaptor = require(`./sso-adaptors/${configuration.provider}.js`);
        return adaptor(configuration);
    }

    // OK. Everything is defined. Now lets load the tenants and then return our public interface.
    loadTenants();

    return {
        login: function (tenantNameOrUrl, uid, pwd) {
            return new Promise(function (weAreLoggedIn, weAreLoggedOut) {
                const tenant = getTenantName(tenantNameOrUrl);
                if (!tenant || !tenants[tenant] || !tenants[tenant].host)
                    return weAreLoggedOut("Tenant " + tenantNameOrUrl + " is unknown");

                const host  = Array.isArray(tenants[tenant].host)
                            ? tenants[tenant].host[0]
                            : tenants[tenant].host;

                if (tenants[tenant].ssoConfiguration) {
                    const idpConfig = tenants[tenant].idp;
                    idpConfig.userEmail    = uid || idpConfig.userEmail;
                    idpConfig.userPassword = pwd || idpConfig.userPassword;
    
                    runIdPLogin(idpConfig)
                    .then(weAreLoggedIn)
                    .catch(weAreLoggedOut);
                }
                else if (/^localhost|^127\.0\.0\.1/.test(host))
                    checkServiceAccountCredentials(tenant, uid, pwd)
                    .then(weAreLoggedIn)
                    .catch(weAreLoggedOut);
                else
                    areWeLoggedIn(tenant)
                    .then(() => weAreLoggedIn("OK"))
                    .catch(notLoggedIn => {
                        gotoLoginPage(tenant)
                        .then(() => {
                            submitCredentials(tenant, uid, pwd)
                            .then(landingPage => {
                                gotoLandingPage(tenant, landingPage)
                                .then(confirmation => weAreLoggedIn(confirmation))
                                .catch(error => weAreLoggedOut(error));
                            })
                            .catch(error => weAreLoggedOut(error));
                        })
                        .catch(error => weAreLoggedOut(error));
                    });
            });
        },
        logout: function (tenantNameOrUrl) {
            return new Promise(function (resolve, reject) {
                const tenant = getTenantName(tenantNameOrUrl);
                if (!tenant || !tenants[tenant] || !tenants[tenant].host)
                    return reject("Tenant " + tenantNameOrUrl + " is unknown");

                tenants[tenant].session = null;
                resolve("OUT");
            });
        },
        request: async function (tenantNameOrUrl, type, setting, entityId, query, skipLogin) {
            const tenant = getTenantName(tenantNameOrUrl);

            if (!tenant || !tenants[tenant])
                return new Promise(function (resolve, reject) {
                    reject("Unknown tenant " + tenant);
                });

            // Make sure we're logged in!
            if (!skipLogin)
                await this.login(tenant); // TODO: What about the credentials??

            if (setting === "monitoringMode") {
                return setMonitoringMode(tenant, type, entityId, query.value);
            }
            else {
                return setMonitoringSetting(tenant, type, entityId, setting, query.value);
            }
        },
        scriptedRequest: async function (tenantNameOrUrl, request, query = "", skipLogin) {
            const tenant = getTenantName(tenantNameOrUrl);

            if (!tenant || !tenants[tenant])
                return new Promise(function (resolve, reject) {
                    reject("Unknown tenant " + tenant);
                });

            // Get the query parameters.
            const params = typeof query === "string"
                         ? query.split("&").reduce((params, param) => {
                                keyValue = param.split("=");
                                params[keyValue.shift()] = keyValue.join("=");
                                return params;
                           }, {})
                         : query;

            // Make sure we're logged in!
            if (!skipLogin)
                await this.login(tenant, params.user, params.pwd);

            if (typeof request === "string") {
                return scriptedRequest(tenant, { url: request, params: params });
            }
            else {
                const script = JSON.parse(JSON.stringify(request));  // Make a copy!
                return scriptedRequest(tenant, script, { request: params, response: {} });
            }
        },
        set config (value) {
            self.config = value;
            loadTenants();
            console.log("DynatraceUI updated after configuration change");
        },
        get config() {
            return self.config;
        }
    };
};

module.exports = DynatraceUI;