'use strict';

class EntityAccess {

    get curator() {
        return this.globals.curator;
    }
    get indexer() {
        return this.globals.indexer;
    }    

    constructor(globals) {
        this.globals  = globals;
        this.logger   = globals.logger;
        this.friendly = globals.friendly;
    }
    
    /*******
     *  Updating and validating a specific entity.
     *  - createEntity:     Create an entity (Dynatrace-style) from a proxy.
     *  - updateEntity:     Add a given entity of a given type to the specified tenant administration.
     *  - removeEntity:     Delete a given entity of a given type to the specified tenant administration.
     */

    createEntity(proxy) {
        return {
            entityId:           proxy.entityId,
            displayName:        proxy.displayName,
            customizedName:     proxy.displayName,
            discoveredName:     proxy.displayName,
            firstSeenTimestamp: proxy.firstSeen,
            lastSeenTimestamp:  proxy.lastSeen,
            tags: [{    // Add a tag of our own...
                context: "STREAMINGAPI",
                key: "isUnknown",
                value: "true"
            }],
            fromRelationships: {},
            toRelationships:   {},
            managementZones:   []
        };
    }

    removeEntity(tenant, type, entityProxy) {
        if (!entityProxy) return;

        try {
            // Delete this entity from the entity set.
            this.curator.deleteEntity(tenant, type, entityProxy);

            this.indexer.forgetEntity(entityProxy);
        }
        catch (ex) {
            this.logger.warn(
                "Could not completely remove " + this.friendly(type) + " " + entityProxy.entityId +
                " (" + entityProxy.displayName + ") in " + tenant + ": " + ex.message
            );
        }
    }

    updateEntity(tenant, type, entity, staleTime, checksum, lastUpdated) {
        const oldEntity = this.curator.getEntity(tenant, type, entity);
        if (oldEntity) {
            // There's no need to do anything if this entity was reported by
            // Dynatrace as unknown. We mark the proxy as such and carry on.
            if (entity.isUnknown) {
                oldEntity.isUnknown = true;
                return;
            }

            // There's no need to do anything if we already have this entity
            // and its checksum is the same as the provided one.
            if (checksum && oldEntity.checksum === checksum)
                return;

            // Delete the old version of this entity from the entity set.
            this.curator.deleteEntity(tenant, type, oldEntity);

            // Also remove this old version from all sets that referenced it.
            this.indexer.removeReferences(oldEntity);
        }

        // Add the new version of this entity.
        const entityProxy = this.curator.addEntity(tenant, type, entity, staleTime, checksum, lastUpdated, oldEntity ? oldEntity.lastUsed : 0);

        // Index the entity.
        this.indexer.analyze(tenant, type, entityProxy, entity);
    }
}

module.exports = EntityAccess;