// eslint.enable

const cleanup = require('node-cleanup');
const express = require('express');
const webui   = express();
const http    = require('http');
const server  = http.Server(webui);
const socket  = require('socket.io')(server);
const winston = require('winston');

const Controller    = require('./Controller.js');
const OpenAPIHelper = require('./OpenAPIHelper.js');
const ConfigReader  = require('./ConfigReader.js');
const SFQueryServer = require('./SFQueryServer.js');

const loggerTransports = {
    console: new winston.transports.Console({ level: 'silly' }),
    file:    new winston.transports.File({ filename: '../log/StreamerService.log', maxsize: 1048576, maxFiles: 10 }),
    silent:  new winston.transports.Console({ level: 'error', silent: true })
};

const controllerLogger = winston.createLogger({
    format: winston.format.combine(
        winston.format.label({ label: 'Controller' }),
        winston.format.timestamp({ format: 'YYYY-MM-DD HH:mm:ss.SSS' }),
        winston.format.printf(({ level, message, label, timestamp }) => {
            return `${timestamp} [${label}] ${level}: ${message}`;
        })
    ),
    transports: [ loggerTransports.silent ]  // We may add transports later, as the config specifies.
});

const contentType  = 'application/json';
const jobsFileName = "../config/jobs.json";

let openAPIHelper  = null;
let queryServer    = null;
let controller     = null;
let listener       = null;
let config         = null;

try {
    // Everything that follows requires a config, so we do this synchronously.
    const configReader = new ConfigReader();
    configReader.on("error", error => {
        console.error("Error while loading config: " + error);
    });
    configReader.on("changed", (section, config) => {
        console.warn("Configuration change detected - please restart to take effect");
    });

    config = configReader.load();
    // Tell the controller to use a different worker than the default "Worker.js".
    config.workerName = "ServiceFlowWorker.js";

    openAPIHelper = new OpenAPIHelper(config);

    // Set up where we will be logging to.
    (config.logging || []).forEach(transport => {
        if (loggerTransports[transport])
            controllerLogger.add(loggerTransports[transport]);
    });

    controllerLogger.info("ExtractionService starting...");

    queryServer = new SFQueryServer(config, controllerLogger);
    queryServer.on("ready", message => {
        // Start listening for HTTP requests.
        listener = server.listen(config.extractionPort, () => {
            controllerLogger.info("ExtractionService is listening on port " + config.extractionPort);
        });
        listener.on("error", err => {
            controllerLogger.info("ExtractionService cannot listen on port " + config.extractionPort + ": " + err.message);
            listener.close();
        });
        listener.setTimeout(120000);
    });
    queryServer.on("error", error => {
        controllerLogger.error("ExtractionService could not start: " + error);
        process.exit(1);
    });

    controller = new Controller(config, socket, controllerLogger);
    controller.on("ready", message => {
        // The Controller is ready to run.
        controller.start();

        try {
            // Read the jobs file and give it to the Controller (only the extraction jobs).
            controller.jobs = require(jobsFileName).filter(job => job.subtype === "Extraction");

            controllerLogger.info("ExtractionService is ready. " + message);
        }
        catch (ex) {
            controllerLogger.error("Error reading jobs: " + ex.message);
        }
    });
    controller.on("error", error => {
        controllerLogger.error("ExtractionService could not start: " + error);
        process.exit(1);
    });
}
catch (ex) {
    controllerLogger.error("ExtractionService could not start: " + ex.message);
    process.exit(1);
}

webui.use('/doc/serviceflowapi.json', (req, res) => openAPIHelper.getServiceFlowAPI(req, res));
webui.use('/doc', express.static("../doc")); // The documentation.

webui.get("/applications/:tenantKey", function (request, response) {
    const tenant = request.params.tenantKey;
    const appId  = request.query.id;

    queryServer.getApplications(tenant, appId)
    .then(list => {
        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(list, undefined, 4));
        response.end();
    })
    .catch(error => {
        response.statusCode = isNaN(error) ? 500 : error;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify({ error: isNaN(error) ? error : "Not found" }));
        response.end();
    });
});

webui.get("/serviceflows/:tenantKey", function (request, response) {
    const tenant = request.params.tenantKey;
    const { appMethodId, startTime, endTime, serviceId, async } = request.query;

    queryServer.getServiceFlows(tenant, appMethodId, startTime, endTime, serviceId, async)
    .then(tree => {
        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(tree, undefined, 4));
        response.end();
    })
    .catch(error => {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify({ error: error }));
        response.end();
    });
});

webui.get("/controller/kill/:workerID",  function (request, response) {
    try {
        const killed = controller.killWorker(request.params.workerID);

        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(killed));
        response.end();
    }
    catch (ex) {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(ex));
        response.end();
    }
});

cleanup(function (exitCode, signal) {
    controllerLogger.info("ExtractionService is exiting...");

    if (listener)
        listener.close();     // Don't leave HTTP port open.

    if (controller)
        controller.end(function () {
            controllerLogger.info("The Controller has been ended");
            controllerLogger.info("Terminating self...   Goodbye");
            controllerLogger.info("----------- :-( -------------");

            process.kill(process.pid, signal);
        });

    cleanup.uninstall();    // Don't call cleanup handler again.
    return false;
});
