'use strict';

const yaml = require('js-yaml');
const fs   = require('fs');
const YAML = {
    parse:     data => yaml.safeLoad(data),
    serialize: data => yaml.safeDump(data)
};

const FormatterFactory = function (templates) {
    templates = templates || {};
    templates.test = {  // For testing in the editor.
        default: {
            "entityId": "[prop.entityId]",
            "displayName":"[prop.displayName]",
            "lastSeen":"[prop.lastSeenTimestamp]",
            "*": "[metric.*]"
        }
    };

    const dots   = /\./g;
    const hashes = /#/g;
    const lastBracket = /\][^|]/;
    const entityTypes = [ "CUSTOM_APPLICATION", "MOBILE_APPLICATION", "APPLICATION", "SERVICE", "PROCESS_GROUP", "PROCESS_GROUP_INSTANCE", "HOST" ];
    const entityChain = ["HOST", "PROCESS_GROUP_INSTANCE", "PROCESS_GROUP", "SERVICE", "APPLICATION"];

    function join(values) {
        return values
            .map(value => typeof value === "object" ? JSON.stringify(value) : value)
            .join(",");
    }

    function followPropertyPath(source, path, wildcardSeen) {
        const getField = function (source, field, iterate) {
            /*  If we have no field, we should take every property in the source we can
                find, and copy it over to the replacement.
                If we have a field and we are told to iterate over the props, we resolve
                each sub-property individually, per the indicated field.
                If we have no field and we have never seen a wildcard, we simply return
                the value of the indicated field.
             */
            if (typeof source !== "object" || !source) {
                return source;
            }
            else if (!field || iterate) {
                const props = Object.keys(source);
                // If all keys are numbers, replacement is an array, otherwise an object.
                const replacement = props.every(p => Number(p) == p) ? [] : {};

                for (let prop of props)
                    replacement[prop] = field && source[prop].hasOwnProperty(field)
                                      ? source[prop][field]
                                      : source[prop];

                return replacement;
            }
            else  {
                return source[field];
            }
        };

        let i = 0;
        while (source && i < path.length) {
            /*  As soon as we first encounter a "*" field, we copy over all (matching - TODO!)
                properties from the current replacement object to a new replacement object.
                For non-"*" fields: If we had a "*" in the past, we go through all properties in the
                replacement object and 'resolve' each per the property chain. If we haven't seen a
                "*" yet, we simply replace the replacement with the value of the indicated field.
             */
            let isWildCard = path[i] === "*";
            source = getField(source, isWildCard ? null : path[i], wildcardSeen);
            wildcardSeen = wildcardSeen || isWildCard;
            i++;
        }
        return source;
    }

    function getGenericProp(constants, propPath) {
        // NOTE: We must not touch 'propPath'. Other invocations need it intact.
        let propName   = propPath[0];
        let isWildCard = propName ? propName.startsWith("*") : false;

        let replacement = isWildCard || !propName
                        ? constants.locals
                        : propName === "zone"
                        ? constants.locals.mzName
                        : constants.locals[propName]
                        ? constants.locals[propName]
                        : propName && constants.globals
                        ? constants.globals[propName]
                        : undefined;

        replacement = followPropertyPath(replacement, propPath.slice(1), isWildCard);
        return replacement; // || "(N/A)";
    }

    function getEntityProp(entity, constants, propPath) {
        // NOTE: We must not touch 'propPath'. Other invocations need it intact.
        let propName   = propPath[0];
        let isWildCard = propName ? propName.startsWith("*") : false;

        let replacement = isWildCard || !propName
                        ? entity
                        : entity[propName]
                        ? entity[propName]
                        : propName && constants.globals
                        ? constants.globals[propName]
                        : undefined;

        replacement = followPropertyPath(replacement, propPath.slice(1), isWildCard);
        return replacement; // || "(N/A)";
    }

    function getEntityTag(entity, constants, tagPath) {
        // Collect the tags in an easily harvestable manner for this and subsequent replacements.
        if (!entity.contexts) {
            entity.contexts = {};
            entity.tags = entity.tags || [];
            for (let i=0; i < entity.tags.length; i++) {
                const tag = entity.tags[i];
                entity.contexts[tag.context] = entity.contexts[tag.context] || {};
                entity.contexts[tag.context][tag.key] = entity.contexts[tag.context].hasOwnProperty(tag.key)
                                                      ? entity.contexts[tag.context][tag.key] + ", " + tag.value
                                                      : tag.value;
            }
        }

        // Request may be for tag.* or tag.CONTEXTLESS.*  or tag.CONTEXTLESS.MyTag
        // The category is 'tag' - we're past that at this point.
        let replacement = null;

        // NOTE: We must not touch 'tagPath'. Other invocations need it intact.
        let context = tagPath[0];
        if (!context.startsWith("*")) {
            replacement = entity.contexts[context]
                        ? entity.contexts[context]
                        : constants.globals
                        ? constants.globals[context]
                        : undefined;

            replacement = followPropertyPath(replacement, tagPath.slice(1));
        }
        else {
            replacement = {};
            for (let context in entity.contexts) {
                if (entity.contexts.hasOwnProperty(context)) {
                    for (let tag in entity.contexts[context]) {
                        if (entity.contexts[context].hasOwnProperty(tag)) {
                            replacement[context + ":" + tag] = entity.contexts[context][tag];
                        }
                    }
                }
            }
        }

        return replacement; // || "(N/A)";
    }

    function getEntityMetric(metrics, metricName, mappedField, metadata) {
        /*  If we don't have the metric as such, but we have dimension values for that metric,
            then we aggregate to produce a single value from them. Since each metric has some
            kind of aggregation (#AVG), a dimension agregation is the one after that (optional).
            eg. metric.com.dynatrace.builtin:host.disk.availablespace#AVG#MAX (defaults to AVG)
         */
        if (metricName.startsWith("*")) {
            let replacement = null;

            if (mappedField) {
                /*  The macro has a mapping ("=>") and the value is an object. That is the convention for
                    resolving the value into a list of values, not an object we need to embed or assign.
                 */
                replacement = Object.entries(metrics)               // Get metric names and values.
                .sort((a,b) => a[0]<b[0] ? -1 : a[0]>b[0] ? 1 : 0)  // Sort by name for consistency.
                .map(entry => {
                    const name = entry[0].replace(dots, "-").replace(hashes, "_");
                    const meta = metadata[name] || metadata[name.split("_")[0]];

                    return mappedField === "id"                     // Map to the metric's id,
                         ? entry[0]
                         : mappedField === "name"                   // Or to its name,
                         ? name                                     // Temporary thing - We have to get the real from somewhewre
                         : mappedField === "value"                  // Or to its value,
                         ? entry[1]
                         : meta                                     // Or, if we have meta data
                         ? meta[mappedField]                        // Then map to specified field.
                         : "(N/A)"                                  // Otherwise not available
                });
            }
            else {
                /*  No mapping to lists, but possibly a mapping of the metric property names themselves.
                    And if not, them we'll still translate metric names (dots and hashes) because some
                    destinations don't like "." or "#".
                 */
                replacement = {};
                const mappedField = metadata && metricName.split("=>")[1];
                const metricNames = Object.keys(metrics).sort();   // Sort by name for consistency.

                for (let metric of metricNames) {
                    const name = metric.replace(dots, "-").replace(hashes, "_");
                    const meta = mappedField
                               ? metadata[name] || metadata[name.split("_")[0]] || {}
                               : null;
                    const prop = meta ? meta[mappedField] || name : name;
                    replacement[prop] = metrics[metric];
                }
            }

            return replacement;
        }

        if (metrics.hasOwnProperty(metricName)) {
            if (mappedField) {
                const name = metricName.replace(dots, "-").replace(hashes, "_");
                const meta = metadata[name] || metadata[name.split("_")[0]];

                return mappedField === "id"                     // Map to the metric's id,
                     ? metricName
                     : mappedField === "name"                   // Or to its name (TODO: We have to get that from somewhewre!!),
                     ? name
                     : mappedField === "value"                  // Or to its value,
                     ? metrics[metricName]
                     : meta                                     // Or, if we have meta data
                     ? meta[mappedField]                        // Then map to specified field.
                     : "(N/A)";                                 // Otherwise we'll use the value.
            }
            else {
                return metrics[metricName];
            }
         }

        // Metric not found? See if we have the dimensions for it, instead.
        // First, remove any dimension aggregation, if present. We'll use it later.
        let metricNameParts = metricName.split("#");
        let aggregation     = metricNameParts.length > 2 ? metricNameParts.pop() : "AVG";
        let cleanMetricName = metricNameParts.join("#");

        let values = [];
        for (let metric in metrics) {
            if (metric.startsWith(cleanMetricName))
                values.push(metrics[metric]);
        }
        if (values.length > 0) {
            switch (aggregation.toUpperCase()) {
                case "AVG": return values.reduce((a,b) => a + b, 0) / values.length;
                case "SUM": return values.reduce((a,b) => a + b, 0);
                case "MIN": return Math.min(...values);
                case "MAX": return Math.max(...values);
                case "NUM": return values.length;
            }
        }
        return undefined;
    }

    function getRelatedBranches(entity, entities, linkages, visited, direction) {
        const getEntityById = function (id) {
            for (let i = 0; i < entities.length; i++)
                if (entities[i].id === id)
                    return entities[i];
        }

        // Avoid following circular references - to be safe.
        visited = visited || {};
        visited[entity.id] = true;

        const seen = entity => {
            if (visited[entity.id]) return true;
            return !(visited[entity.id] = true);
        };

        // We want our little section of the whole graph: only those entities and linkages that matter to us.
        if (!direction) {
            const incoming = getRelatedBranches(entity, entities, linkages, visited, "IN");
            const outgoing = getRelatedBranches(entity, entities, linkages, visited, "OUT");

            return {
                related:   [ ...incoming.related,   ...outgoing.related   ],
                inventory: { ...incoming.inventory, ...outgoing.inventory },
                relations: [ ...incoming.relations, ...outgoing.relations ]
            };
        }

        const relations = linkages
            .filter(linkage => linkage[direction === "IN" ? "target" : "source"] === entity.id);
        const related = relations
            .map(linkage => getEntityById(linkage[direction === "IN" ? "source" : "target"]));
        const inventory = related.reduce((entities, entity) => {
            entities[entity.id] = entity;
            return entities;
        }, {});

        const all = related.reduce((all, entity) => {
            if (seen(entity)) return all;

            const {
                related,
                inventory,
                relations
            } = getRelatedBranches(entity, entities, linkages, visited, direction);

            Object.assign(all.inventory, inventory);    // Is key-value pairs (ID:entity).
            all.related.push(...related);               // Is array of relevant entities.
            all.relations.push(...relations);           // Is array of relevant linkages.

            return all;
        }, {
            related: related,
            inventory: inventory,
            relations: relations
        });

        return all;
    }

    function getRelatedEntities(entity, relatedType, constants) {
        /*  Starting at the given entity, find all related entities of the specified type.
            The starting point may not be the 'main' entity for which we are creating a
            message. Instead, it may be an entity that is related to the 'main' entity
            if a switch has occurred previously.
         */

        const traverseRelations = function (entity, path, from, to) {
            /*  Recurse through the linkages, starting from 'entity', and get all entities
                of the last type in the path (that's our destination so to speak).

                A tenant name preceeds each entity ID in a graph, with an "_" in between.
                Note that we want to treat "CUSTOM_APPLICATION", "MOBILE_APPLICATION" and
                "APPLICATION" the same, and that will in fact, coincidentally, be the case.
             */
            const segment = "_" + path.shift() + "-";   // This here consumes the path, step by step.

            const entities = constants.relations    // Find entities by selecting all links, starting
                .filter(relation => relation[from] === entity.id &&     // from me, that point to the
                                    relation[to].includes(segment))     // type of entities we want.
                .map(   relation => constants.inventory[relation[to]])  // Get those entities, except
                .filter(entity   => !!entity);                          // those missing in inventory.

            if (path.length === 0)              // The path is empty now, so we got what we need.
                return entities;
            else if (entities.length === 0)     // No entities here. Nothing to traverse further.
                return [];
            else         // We're not there yet! Find everything behind each intermediate entity.
                return entities.map(entity => traverseRelations(entity, path, from, to)).flat();
        }

        const entityType   = entity.entityId.split("-")[0];
        const placeInChain = entityChain.indexOf(entityType);
        const foundInChain = entityChain.indexOf(relatedType);

        let entities = null;
        if (foundInChain > placeInChain) {
            // Get the path from 'me' to the type of entity we want.
            const path = entityChain.slice(placeInChain + 1, foundInChain + 1);

            // Now we traverse the chain to get the entities we need.
            entities = traverseRelations(entity, path, "target", "source");
        }
        else if (foundInChain < placeInChain) {
            // Get the path from 'me' to the type of entity we want.
            const path = entityChain.slice(foundInChain, placeInChain).reverse();

            // Now we traverse the chain to get the entities we need.
            entities = traverseRelations(entity, path, "source", "target");
        }
        else {
            // We are a <TYPE> and we want all related <TYPE>s. Get all of them, except 'me'.
            const segment = "_" + relatedType + "-";

            entities = Object.keys(constants.inventory)
                .filter(id => id !== entity.id && id.includes(segment))
                .map(id => constants.inventory[id]);
        }

        return entities;
    }

    function resolveString(value, entity, constants, keepArray, previousID) {
        if (typeof value !== "string")
            return value;

        /*  We 'while' because replacements may (theoretically) contain macros, and more
            realistically because values may contain multiple macros. Additionally, OR-ed
            macros are supported, and we may need to try multiple options (the first macro
            in such lists that leads to a value will be used, and the others are removed).
            When a macro resolves to an object, we exit the loop because there is no good
            way to combine objects into a string. Everything else will become stringified.
         */
        let oldValue = null;            // Track if we are still changing.
        while (oldValue !== value) {
            oldValue = value;

            let switchStart = value.indexOf("<");
            let switchEnd   = value.indexOf(">");
            let macroStart  = value.indexOf("[");
            let macroEnd    = value.indexOf("]");

            if (switchStart !== -1 && switchEnd !== -1 &&
                (macroStart === -1 || switchStart < macroStart)) {
                /*  There is a context switch before the first upcoming macro (or no macro).
                    We need to heed that first. Note: we are going to produce a collection
                    of replacements now. That may need to become an array, or an object.
                    Since we may need the IDs, we have no other options but to create an
                    object with keys that correspond to the found entity IDs.
                    Since we may do this recursively, we will have to build up the ID from
                    a previous one (if we were called recursively already). If so, every ID
                    we find is prefixed with that previous ID.
                 */
                const beforeSwitch = value.substring(0, switchStart);
                const afterSwitch  = value.substring(switchEnd + 1);
                const relatedType  = value.substring(switchStart + 1, switchEnd);

                value = beforeSwitch + afterSwitch;    // Remove the entity type from the value.

                if (entityTypes.includes(relatedType)) {
                    const idPrefix = previousID ? (previousID + ":") : "";
                    const entities = getRelatedEntities(entity, relatedType, constants);

                    const replacement = entities.reduce((collection, entity) => {
                        // Note: 'entity' is now a related entity. This is where the context switch happens.
                        const result = resolveString(value, entity, constants, keepArray, entity.id);

                        if (typeof result === "object") {
                            for (let p in result) {
                                if (!result.hasOwnProperty(p)) continue;

                                collection[idPrefix + p] = result[p];
                            }
                        }
                        else {
                            collection[idPrefix + entity.id] = result;
                        }
                        return collection;
                    }, {});

                    return keepArray ? Object.values(replacement) : replacement;
                }
                else {
                    // All we could do is remove this <SWITCH> from the value and let nature take its course...
                }
            }

            if (macroStart !== -1 && macroEnd !== -1 && macroStart < macroEnd) {
                const macro       = value.substring(macroStart+1, macroEnd);
                let macroParts    = macro.split("=>");       // Only supported for metrics!
                const mappedField = constants.metadata && macroParts[1];
                macroParts        = macroParts[0].split(".");
                const category    = macroParts.shift();

                let replacement   = null;

                switch (category) {
                    case "job":
                        // Only certain properties from the job definition are exposed.
                        replacement = getGenericProp(constants, macroParts);
                        break;

                    case "prop":
                        // Field may start with an entity type and macro may contain a property chain.
                        if (entityTypes.includes(macroParts[0])) {
                            const type = macroParts.shift();
                            const entities = getRelatedEntities(entity, type, constants);

                            replacement = keepArray
                                        ? entities.map(entity =>
                                                getEntityProp(entity, constants, macroParts)
                                            )
                                        : entities.reduce((acc, entity) => {
                                                acc[entity.id] = getEntityProp(entity, constants, macroParts);
                                                return acc;
                                            }, {});
                        }
                        else {
                            replacement = getEntityProp(entity, constants, macroParts);
                        }
                        break;

                    case "tag":
                        // Field may start with an entity type. Tag references are 2-property chains.
                        if (entityTypes.includes(macroParts[0])) {
                            const type = macroParts.shift();
                            const entities = getRelatedEntities(entity, type, constants);

                            replacement = keepArray
                                        ? entities.map(entity =>
                                                getEntityTag(entity, constants, macroParts)
                                            )
                                        : entities.reduce((acc, entity) => {
                                                acc[entity.id] = getEntityTag(entity, constants, macroParts);
                                                return acc;
                                            }, {});
                        }
                        else {
                            replacement = getEntityTag(entity, constants, macroParts);
                        }
                        break;

                    case "metric":
                        //  Metric names may contain '.' - those are IDs, not property chains.
                        // If we have a mapping and metadata, we may get a list of values.
                        replacement = getEntityMetric(constants.metrics, macroParts.join("."), mappedField, constants.metadata);
                        break;

                    default:
                        // Maybe it's in the defaults defined in the destination. TODO: We need to add the property chain!!
                        replacement = getGenericProp({ locals: {}, globals: constants.globals }, [ category ]);
                        break;
                }

                if (replacement) {
                    const endOfMacro  = value.charAt(macroEnd+1) === '|'
                                      ? value.search(lastBracket)  // Up to first "]*" where '*' is not a '|'.
                                      : macroEnd;                  // Or where this macro ends if no OR options.
                    const beforeMacro = value.substring(0, macroStart);
                    const afterMacro  = endOfMacro === -1 ? "" : value.substring(endOfMacro+1);

                    // It may already have been an array, or may have just become an array. If the value we need
                    // to replace into has more stuff in it, we have no option but to stringify this replacement.
                    if (Array.isArray(replacement) && afterMacro.length > 0) {
                        replacement = join(replacement);    // Stringify. Objects will be JSON-ized.
                    }

                    if (typeof replacement != "string" &&
                        typeof replacement != "number" &&
                        typeof replacement != "boolean") {
                        // Replacement is not a string, so don't replace the macro but just return this value.
                        return replacement;
                    }
                    else {
                        // We have a string-compatible replacement for the macro.
                        value = beforeMacro + replacement + afterMacro;
                    }
                }
                else {
                    // No replacement found. If we have an OR option, remove this macro and try the next one in
                    // the '|'-separated list.
                    value = value.charAt(macroEnd+1) === '|'
                          ? value.replace("[" + macro + "]|", "")    // That part did not work. Try next option.
                          : value.replace("[" + macro + "]",  "");   // Else we don't have a replacement at all.
                }
            }
            else {
                // It's a literal value - no need to replace anything in the value.
            }
        }

        return value;
    }

    function resolveObject(resolvableObject, entity, constants, ledger, keepArray) {

        const analyzeValue = function (prop, value, metadata) {
            // Establish what type of value we're dealing with.
            const type = Array.isArray(value) ? "array" : typeof value;

            // Rename props if there is a mapping in the template (e.g. "*=>planum-metric-name")
            // The mapping may also refer to a property on an entity if the prop is an entity ID.
            const mappedField = metadata && prop.split("=>")[1];
            const prefix = prop.substring(0, prop.indexOf("*"));
            const values = {};
            const ledger = {};

            if (prop.indexOf("*") !== -1) {
                /*  We want to allow entries like "*": "[prop.*]" to be used in multiple places in
                    the same (immediate) parent object. To allow that, you can use "*1", "*2", etc.
                    to prevent the entries from collapsing into one entry. You can also use prefixes,
                    like this: "K8S_*": "[tag.KUBERNETES.*]". The part before the "*" is used as
                    prefix for the property names (eg. K8S_Cloud = "TheCloud", K8S_DC = "TheDC").
                 */
                if (type === "object") {
                    for (let p in value) {
                        if (!value.hasOwnProperty(p)) continue;

                        // There may be a mapping of a prop name to something else.
                        // Note: p may be an entity ID, but also a ':'-separated list of IDs...
                        const meta = mappedField
                                   ? metadata[p] || metadata[p.split("_")[0]] || constants.inventory[p] || null
                                   : null;
                        const prop = meta
                                   ? meta[mappedField] || p
                                   : p.includes(":")        // is it a list if entity IDs?
                                   ? p.split(":").map(id => constants.inventory[id] && constants.inventory[id][mappedField] || id).join(":")
                                   : p;

                        // Update the ledger using the appropriate prefix.
                        ledger[prefix + prop] = true;

                        // Add the property and value to the resolved object.
                        values[prefix + prop] = value[p];
                    }
                }
                else if (type === "array") {
                    for (let i=0; i < value.length; i++) {
                        // For consistency with the above...
                        const prop = i;

                        // Update the ledger using the appropriate prefix.
                        ledger[prefix + prop] = true;

                        // Add the property and value to the resolved object.
                        values[prefix + prop] = value[i];
                    }
                }
                else {
                    // Update the ledger. We keep the prop name as is.
                    ledger[prop] = true;

                    // Just assign the value as an object.
                    values[prop] = type !== "array" || keepArray ? value : join(value);
                }
            }
            else {
                if (prop.includes("=>join")) {
                    const mapp = prop.split("=>")[1];
                    const regx = /'(.*)'/;
                    const sepa = regx.exec(mapp)[1] || ",";
                    const list = type === "object"
                               ? Object.values(value)
                               : type === "array"
                               ? value
                               : [value];

                    const bareProp  = prop.split("=>")[0];

                    ledger[bareProp] = true;
                    values[bareProp] = list.join(sepa);
                }
                else {
                    // No wildcard or mapping in prop name. Update the ledger.
                    ledger[prop] = true;

                    // Convert value to native type if necessary, and add to resolved object.
                    values[prop] = type === "array" && !keepArray ? join(value)
                                 : type !== "string"      ? value
                                 : value === ""           ? value
                                 : value === "true"       ? true
                                 : value === "false"      ? false
                                 : Number(value) == value ? Number(value)
                                 : value;
                }
            }

            return { properties: ledger, values: values };
        };

        if (!resolvableObject)
            return null;

        if (typeof resolvableObject === "string" ||
            typeof resolvableObject === "number" ||
            typeof resolvableObject === "boolean")
            return resolveString(resolvableObject, entity, constants, keepArray);

        /*  This here captures situations where the template has an array directly inside
            an array. We inherently flatten that into one array - I don't see a clean way
            to keep an array inside of an array while preserving other expected behavior.
            So, we will not support that. The code below merely serves to avoid treating
            an array as an object when we shouldn't.
         */
        if (Array.isArray(resolvableObject)) {
            const values = Object.values(resolvableObject);

            return values.map((val, i) =>
                resolveObject(val, entity, constants, {}, true)
            ).flat();
        }

        // OK. The resolvable object is legitimately an object.
        let resolvedObject = {};

        const props = Object.keys(resolvableObject);
        for (let prop of props) {
            const value = resolvableObject[prop];
            const type  = Array.isArray(value) ? "array" : typeof value;

            if (type === "number" ||
                type === "boolean") {
                // Update the ledger.
                ledger[prop] = true;

                resolvedObject[prop] = value;   // Nothing to resolve here.
            }
            else if (type === "array") {
                const bareProp = prop.split("=>")[0];

                if (prop.includes("=>join")) {
                    // Update the ledger.
                    ledger[bareProp] = {};

                    const mapp = prop.split("=>")[1];
                    const regx = /'(.*)'/;
                    const sepa = regx.exec(mapp)[1] || ",";
                    const list = value.map((val, i) =>
                        resolveObject(val, entity, constants, ledger[bareProp][i] = {}, true)
                    ).flat();

                    resolvedObject[bareProp] = list.join(sepa);
                }
                else {
                    // Update the ledger.
                    ledger[bareProp] = {};

                    resolvedObject[bareProp] = value.map((val, i) =>
                        resolveObject(val, entity, constants, ledger[bareProp][i] = {}, true)
                    ).flat();
                }
            }
            else  {
                /*  It's a string or a nested object. Nested objects are resolved recursively. Both
                    strings and nested objects can resolve to objects.

                    For tabular objects the ledger just contains the top-level property names, and so this
                    code path is not hit. For complex objects, the ledger should be a complex object too.
                    I don 't know how we would use it, but we should at least keep the ledger consistent
                    with the structure of the generated results. How we do this may require improvement.
                 */
                let resolvedValue = type === "string"
                                  ? resolveString(value, entity, constants)
                                  : resolveObject(value, entity, constants, ledger);

                // Turn the resolved value into an object that either just has 'the' property,
                // or that has a number of properties if the target property name contains a "*".
                let { properties, values } = analyzeValue(prop, resolvedValue, constants.metadata);

                Object.assign(ledger, properties);
                Object.assign(resolvedObject, values);
            }
        }
        return resolvedObject;
    }

    function usesRelatedEntities(template, job, answers) {
        const type = job.type;

        if (!answers.hasOwnProperty(type)) {
            const stringified  = JSON.stringify(template);
            const noReferences = entityTypes.every(type =>
                !stringified.includes("." + type + ".") &&
                !stringified.includes("<" + type + ">")
            );

            answers[type] = !noReferences && !!job.linkages;
        }
        return answers[type];
    }

    return function (format, defaults) {
        defaults = defaults  || {};

        if (!format)
            return {
                ledger: {}, // Initialize a ledger to track all fields placed in messages.
                format: "default",
                isTypeSupported: function () {
                    return true;
                },
                createMessage: function (job, entity, metrics) {
                    const locals = (({ name, timestamp, source, destination, type, TYPE, granularity, mzName }) =>
                                    ({ name, timestamp, source, destination, type, TYPE, granularity, mzName }))(job);

                    locals.TYPE = job.type.toUpperCase().replace(/ /g, "_");
                    locals.type = job.type.toLowerCase().replace(/_/g, " ");
                    locals.timestamp  = job.endTime;
                    locals.localtime  = (new Date(job.endTime)).toLocaleString();
                    locals.timestring = (new Date(job.endTime)).toISOString();

                    const globals = Object.assign({}, defaults, {
                        format:     "unknown",
                        tenant:     job.source,
                        timestamp:  job.timestamp,
                        timestring: job.timestring
                        // Other things can be added as we evolve.
                    });

                    const template = {
                        tags: {
                            id:   "[prop.entityId]",
                            name: "[prop.displayName]",
                            type: "[job.type]"
                        },
                        fields: {
                            "*": "[metric.*]"
                        },
                        timestamp: "[job.timestring]"
                    };

                    return resolveObject(template, entity, {
                        locals:    locals,
                        globals:   globals,
                        metadata:  null,
                        metrics:   metrics,
                        relations: [],
                        related:   [],
                        inventory: {}
                    },  {});
                },
                getFields: function () {
                    return [];
                }
            };
        else if (!templates.hasOwnProperty(format))
            return {
                format: format,
                isTypeSupported: function () {
                    return false;
                },
                createMessage: function () {
                    return {};
                },
                getFields: function () {
                    return [];
                }
            };
        else
            return {
                ledger: {},     // Initialize a ledger to track all fields placed in messages.
                format: format,
                template: null,
                metadata: null,
                usesRelated: {},
                isTypeSupported: function (type) {
                    this.template = templates[format][type] || templates[format]["default"];

                    // This is a good time to load any metadata.
                    const metadata  = templates[format]._metadata_
                                    ? YAML.parse(fs.readFileSync("../config/" + templates[format]._metadata_, 'utf8'))
                                    : null;

                    this.metadata   = metadata && Array.isArray(metadata.metrics)
                                    ? metadata.metrics.reduce((meta, metric) => {
                                            const xName = metric["time-series-id"].replace(dots, "-").replace(hashes, "_");
                                            meta[xName] = metric;
                                            return meta;
                                        }, {})
                                    : null;

                    return !!this.template;
                },
                createMessage: function (job, entity, metrics) {
                    // Copy select things from the job. Not all of its properties are/should be exposed.
                    const locals = (({ name, timestamp, source, destination, type, granularity, mzName }) =>
                                    ({ name, timestamp, source, destination, type, granularity, mzName }))(job);

                    locals.TYPE = job.type.toUpperCase().replace(/ /g, "_");
                    locals.type = job.type.toLowerCase().replace(/_/g, " ");
                    locals.timestamp  = job.endTime;
                    locals.localtime  = (new Date(job.endTime)).toLocaleString();
                    locals.timestring = (new Date(job.endTime)).toISOString();

                    const globals = Object.assign({}, defaults, {
                        format:     format,
                        tenant:     locals.source,
                        timestamp:  locals.timestamp,
                        timestring: locals.timestring
                        // Other things can be added as we evolve.
                    });

                    const {
                        related,
                        inventory,
                        relations
                    } = usesRelatedEntities(this.template, job, this.usesRelated)
                      ? getRelatedBranches(entity, job.entities, job.linkages)
                      : {};

                    // We will add the entity itself to the inventory also. Some scenarios need it.
                    const self = {};
                    self[entity.id] = entity;

                    return resolveObject(this.template, entity, {
                        locals:    locals,
                        globals:   globals,
                        metadata:  this.metadata,
                        metrics:   metrics,
                        relations: relations || [],
                        related:   related   || [],
                        inventory: Object.assign(inventory || {}, self)
                    },  this.ledger);
                },
                getFields: function (prefixSet) {
                    function getFieldNames(obj) {
                        const fields = [];
                        for (let field in obj) {
                            if (typeof obj[field] == "object")
                                fields.push(getFieldNames(obj[field]));
                            else
                                fields.push(field);
                        }
                        return fields;
                    }

                    /*  Collect the fields to be used for 'stringify()' requests.
                        The 'columns' object is supposed to look like this:
                            ["colA", "colB", "SomePrefix.*", "*"]
                        The use of "*" is optional. It refers to the remaining fields and
                        can be placed anywhere in the array to indicate where variable column
                        names will go. Note that if "SomePrefix.*" is used, those have to come
                        before the "*", as the "*" will gobble up all remaining fields.
                     */
                    prefixSet = prefixSet ? prefixSet.map(prefix => prefix + ".") : [""];

                    const fields  = [];
                    const columns = templates[format]._columns_;

                    // If there are no columns, then we return an array of all encountered field names - hierarchical if the ledger is hierarchical.
                    // If there are columns, we're specifically talking about a tabular output. In that case, if the ledger is hierarchical, we flatten it.
                    if (!columns) {
                        return getFieldNames(this.ledger);
                    }

                    for (let i=0; i < columns.length; i++) {
                        if (!columns[i].includes("*")) {
                            // No wildcard. Just an actual column name.
                            fields.push(columns[i]);
                        }
                        else {
                            // There's a wilcard in the column name. Insert matching fields from the message.
                            let prefix  = columns[i].includes(".*")
                                        ? columns[i].replace(".*", ":")
                                        : null;

                            for (let prop in this.ledger) {
                                if (this.ledger.hasOwnProperty(prop) &&
                                    !fields.includes(prop) &&                   // Don't insert a field name twice.
                                    (!prefix || prop.startsWith(prefix))) {     // If we have a prefix, require it.
                                    fields.push(prop);
                                }
                            }
                        }
                    }
                    return {
                        columns: fields.flat(),
                        headers: prefixSet.map(p => fields.map(f => p + f)).flat()
                    };
                }
            };
    };
};

module.exports = FormatterFactory;
