'use strict';

const IndexedSet = function(iterator, key, name) {
    let set = new Set(iterator);
    let map = new Map();

    let i       = -1;
    let arr     = [];
    let proxy = null;
    let isDirty = false;

    set.forEach(value => map.set(value[key], value));
    arr = [ ...map.keys() ];

    const overrides = {
        name: function () {
            return name || "unknown";
        },
        add: function (value) {
            if (map.size !== set.size) {
                console.error("ERROR: Map and set size not the same! Repairing...");
                set = new Set(map.entries());
            }
            // Do we have a previous version?
            if (map.has(value[key])) {
                // Remove the original object.
                set.delete(map.get(value[key]));
                map.delete(value[key]);
            }
            // Add the new object.
            map.set(value[key], value);
            set.add(value);

            isDirty = true;
            return proxy;
        },
        clear: function () {
            map.clear();
            set.clear();
            arr = [];
            isDirty = false;
        },
        delete: function (value) {
            map.delete(value[key]);
            set.delete(value);
            isDirty = true;

            if (map.size !== set.size) {
                // Sometimes we need to bring the set and the map into alignment
                // again because the set did not hold the value to be deleted.
                set = new Set(map.entries());
            }
        },
        has: function (value) {
            return map.has(value[key])
        },
        get: function (key) {
            return map.get(key);
        },
        get ids() {
            return [ ...map.keys() ];
        },
        get length() {
            if (isDirty) {
                arr = [ ...map.keys() ];
                isDirty = false;
            }
            return arr.length;
        },
        at: function (i) {
            if (isDirty) {
                arr = [...map.keys()];
                isDirty = false;
            }
            return map.get(arr[i]);
        },
        getIterator: function () {
            return {
                i: 0,
                get size() {
                    if (isDirty) {
                        arr = [ ...map.keys() ];
                        isDirty = false;
                    }
                    return arr.length;
                },
                next: function () {
                    if (isDirty) {
                        arr = [ ...map.keys() ];
                        isDirty = false;
                    }
                    if (arr.length === 0)
                        return null;

                    if (this.i < arr.length-1)
                        this.i++;
                    else
                        this.i = 0;

                    return arr[this.i];
                }
            };
        },
        next: function() {
            if (isDirty) {
                arr = [ ...map.keys() ];
                isDirty = false;
            }
            if (arr.length === 0)
                return null;

            if (i < arr.length-1)
                i++;
            else
                i = 0;

            return arr[i];
        },
        map: function(callback) {
            const result = [];
            set.forEach(value => result.push(callback(value)));
            return result;
        },
        top: function (count) {
            const result = [];
            for (let [key, value] of map) {
                if (result.length >= count)
                    break;
                result.push(key);
            }
            return result;
        }
    };

    const interceptor = {
        get (target, key) {
            if (typeof key === 'symbol')
                return target[key].bind(target);

            if (overrides.hasOwnProperty(key))
                return overrides[key];

            if (Number(key) == key)
                return overrides.at(key);

            if (typeof target[key] === "function")
                return target[key].bind(target);

            return target[key];
        }
        /*  For reference: these are all the traps that an interceptor can, uh, intercept.
            - getPrototypeOf()      A trap for Object.getPrototypeOf.
            - setPrototypeOf()      A trap for Object.setPrototypeOf.
            - isExtensible()        A trap for Object.isExtensible.
            - preventExtensions()   A trap for Object.preventExtensions.
            - getOwnPropertyDescriptor()    A trap for Object.getOwnPropertyDescriptor.
            - defineProperty()      A trap for Object.defineProperty.
            - has()                 A trap for the 'in' operator.
            - get()                 A trap for getting property values. Since methods are properties,
                                    this is where we check if we want one of our overrides to be
                                    called instead of the native method on the proxied object.
            - set()                 A trap for setting property values.
            - deleteProperty()      A trap for the 'delete' operator.
            - ownKeys()             A trap for Object.getOwnPropertyNames and Object.getOwnPropertySymbols.
            - apply()               A trap for a function call (directly on the proxy - not for a method).
            - construct()           A trap for the 'new' operator.
         */
    };

    proxy = new Proxy(set, interceptor);
    return proxy;
}

module.exports = IndexedSet;