'use strict';
require("./Extensions.js");
const fs = require('fs');
const async = require('async');
const IndexedSet = require('./IndexedSet.js');

const executeSerially = function (work, resolve, reject, results) {
    results = results || { errors: [], data: [] };

    const task = work.shift();
    if (task) {
        task(
            data => {
                if (data) results.data.push(data);
                setTimeout(executeSerially, 1, work, resolve, reject, results);
            },
            error => {
                if (error) results.errors.push(error);
                setTimeout(executeSerially, 1, work, resolve, reject, results);
            }
        );
    }
    else {
        if (results.errors.length > 0)
            reject(results);
        else
            resolve(results);
    }
}

function add(first, second) {
    // This inherently removes duplicates. Modifies the original!
    for (let item of second)
        first.add(item);
}

function union(...iterables) {
    // This inherently removes duplicates. Does not mutate sets.
    const result = new Set();

    for (let iterable of iterables)
        for (let item of iterable)
            result.add(item);

    return result;
}

function subtract(first, second) {
    // The deletion may safely fail. Modifies the original!
    for (let item of second)
        first.delete(item);
}

function without(first, second) {
    // Selectively add to result. Does not mutate sets.
    const result = new Set();

    for (let item of first)
        if (!second.has(item))
            result.add(item);

    return result;
}

function intersect(sets) {
    if (sets.length === 0)
        return null;
    if (sets.length === 1)
        return Array.from(sets[0]);

    const result = [];
    const first  = sets.pop();
    const second = sets.pop();
    for (const elem of first) {
        if (second.has(elem)) result.push(elem);
    }
    if (sets.length > 0) {
        sets.push(new Set(result));
        return intersect(sets);
    }
    return result;
}

class Indexer {
    #hasInitialized = false;
    #indexes = {};
    // These are the properties we will certainly NOT index:
    #unindexables = [
        "entityId", 
        "displayName", "discoveredName", "customizedName",
        "firstSeenTimestamp", "lastSeenTimestamp",
        "fromRelationships", "toRelationships" 
    ];

    #validateEntity = function (tenant, type, entityProxy, entity,log) {
        const self = this;
        const indexEntries = this.indexer.getIndexes(tenant, type);

        function checkRelationships(entityProxy, thisEntity, entityType, relationshipType, direction) {
            // The problem is that we constantly get updates, and new entities may be referenced
            // that we haven't sync-ed yet. And that leads to errors here. So, we skip this check.

            let relationshipAttr  = direction === "IN" ? "toRelationships" : "fromRelationships";
            let oppositeDirection = direction === "IN" ? "OUT" : "IN";

            let relation = self.relationships[entityType][relationshipType][direction];
            if (!relation) return true; // That's OK.

            let idList      = thisEntity[relationshipAttr][relation.prop] || [];
            let entityLists = relation.accept.map(acceptedType => self.curator.getEntities(tenant, acceptedType));

            // Replace all IDs with entity references - insofar we can at this point.
            let referrers = [];
            for (var i=0; i < idList.length; i++) {
                let thatEntity = entityLists.reduce((entity, entityList) => {
                    return entity || entityList.get(idList[i]);
                }, null);
                if (!thatEntity) {   // The referenced entity does not exist, but it should have!
                    log.messages.push("Could not find entity with ID " + idList[i]);
                    return false;
                }
                referrers.push(thatEntity);
            };
            if (referrers.length === entityProxy[relationshipType][direction].length)
                return true;

            // Better than the above: create ID lists, sort, and then compare to check for equality.
            // It is possible one reference may have been replaced with another one, keeping the length the same.

            // Hmm. There;s a difference. We need to report what the difference is and return false.
        }

        let referrers = new Set();
        let entry     = null;

        // To be sure...
        entity.tags = entity.tags || [];

        let goodTags = entity.tags.filter(tag => {
            // Create a new entry for this tag if it does not exist yet.
            tag.context = tag.context || "CONTEXTLESS";

            if (!indexEntries.hasOwnProperty(tag.context)) {
                log.messages.push(tag.context + " does not exist in indexes");
                return false;
            }

            if (!indexEntries[tag.context].hasOwnProperty(tag.key)) {
                log.messages.push(tag.key + " does not exist in index for " + tag.context);
                return false;
            }

            entry = indexEntries[tag.context][tag.key];

            if (!entry._exists.has(entityProxy)) {
                log.messages.push("Index entry for " + tag.context + ":" + tag.key + " does not reference " + entity.entityId + " (" + entity.displayName + ")");
                return false;
            }
            // Take note that this reference should exist.
            referrers.add(entry._exists);

            tag.value = tag.value || "_empty";
            let goodValues = tag.value.split(",").filter(value => {
                if (!entry.hasOwnProperty(value)) {
                    log.messages.push("Index for " + tag.context + ":" + tag.key + " does not have an entry for " + value);
                    return false;
                }

                if (!entry[value].has(entityProxy)) {
                    log.messages.push("Index for " + tag.context + ":" + tag.key + ":" + value + " does not reference " + entity.entityId + " (" + entity.displayName + ")");
                    return false;
                }

                // Take note that this reference should exist.
                referrers.add(entry[value]);

                return true;
            });
            if (goodValues.length !== tag.value.split(",").length)
                return false;

            return true;
        });
        if (goodTags.length !== entity.tags.length)
            return false;

        // Next, property values!
        let context = "PROPERTY";
        if (!indexEntries.hasOwnProperty(context)) {
            log.messages.push(context + " does not exist in indexes");
            return false;
        }

        if (!entity.hasOwnProperty("hostGroup")) {
            referrers.add(indexEntries["HOST_GROUP"]._empty);
        }

        for (var prop in entity) {
            if (prop === "toRelationships") {
                checkRelationships(entityProxy, entity, type, "flows", "IN");
                checkRelationships(entityProxy, entity, type, "stack", "IN");
                checkRelationships(entityProxy, entity, type, "group", "IN");
            }
            else if (prop === "fromRelationships") {
                checkRelationships(entityProxy, entity, type, "flows", "OUT");
                checkRelationships(entityProxy, entity, type, "stack", "OUT");
                checkRelationships(entityProxy, entity, type, "group", "OUT");
            }
            else if (prop === "hostGroup") {
                if (!indexEntries["HOST_GROUP"].hasOwnProperty(entity.hostGroup.name)) {
                    log.messages.push("Host group " + entity.hostGroup.name + " does not exist");
                    return false;
                }
                // Take note that these references should exist.
                referrers.add(indexEntries["HOST_GROUP"]._exists);
                referrers.add(indexEntries["HOST_GROUP"][entity.hostGroup.name].members);
            }
            else if (typeof entity[prop] === "string" && !this.#unindexables.includes(prop)) {

                if (!indexEntries[context].hasOwnProperty(prop)) {
                    log.messages.push(prop + " does not exist in index for " + context);
                    return false;
                }

                entry = indexEntries[context][prop];

                // Take note that this reference should exist.
                referrers.add(entry._exists);

                entity[prop] = entity[prop] || "_empty";

                // Create entry for this value.
                if (!entry.hasOwnProperty(entity[prop])) {
                    log.messages.push("Index for " + context + ":" + prop + ":" + entity[prop] + " does not reference " + entity.entityId + " (" + entity.displayName + ")");
                    return false;
                }

                // Take note that this reference should exist.
                referrers.add(entry[entity[prop]]);
            }
        }

        // Finally the management zones.
        if (!indexEntries.hasOwnProperty("ZONE")) {
            log.messages.push("Index entry for ZONE not found");
            return false;
        }

        if (!entity.managementZones || entity.managementZones.length === 0) {

            if (!indexEntries["ZONE"]._nonmembers.has(entityProxy)) {
                log.messages.push("No-zone entity " + entity.entityId + " (" + entity.displayName + ")" + " not referenced as outsider");
                return false;
            }
            // Take note that this reference should exist.
            referrers.add(indexEntries["ZONE"]._nonmembers);

            if (indexEntries["ZONE"]._members.has(entityProxy)) {
                log.messages.push("No-zone entity " + entity.entityId + " (" + entity.displayName + ")" + " incorrectly referenced as insider");
                return false;
            }
        }
        else {
            if (!indexEntries["ZONE"]._members.has(entityProxy)) {
                log.messages.push("Entity " + entity.entityId + " (" + entity.displayName + ")" + " not referenced as insider");
                return false;
            }
            // Take note that this reference should exist.
            referrers.add(indexEntries["ZONE"]._members);

            if (indexEntries["ZONE"]._nonmembers.has(entityProxy)) {
                log.messages.push("Entity " + entity.entityId + " (" + entity.displayName + ")" + " incorrectly referenced as outsider");
                return false;
            }

        let goodZones = entity.managementZones.filter(zone => {
            /*  Create a new entry for this zone if it does not exist yet.
                Note: Zones have an ID and a name. In order to allow for filtering by name
                as well as by ID, we create two entries in the zoneEntry: one by ID and one
                by name, both pointing to the very same object (and thus the same set).
            */
            if (!indexEntries["ZONE"].hasOwnProperty(zone.id)) {
                    log.messages.push("Index entry for zone ID " + zone.id + " not found for " + entity.entityId + " (" + entity.displayName + ")");
                return false;
            }
            if (!indexEntries["ZONE"][zone.id].members.has(entityProxy)) {
                    log.messages.push("Index entry for zone ID " + zone.id + " does not contain " + entity.entityId + " (" + entity.displayName + ")");
                return false;
            }
            // Take note that this reference should exist.
            referrers.add(indexEntries["ZONE"][zone.id].members);
            referrers.add(indexEntries["ZONE"][zone.name].members);  // This validates that both are in fact the same set.

            /*  IMPORTANT NOTE: Sometimes, entities have a zone with different names and ID combinations
                than originally reported by Dynatrace (when retrieving the MZs). We seem not to be able
                to rely on names being unique. Consequently, we are not going to validate membership by
                zone NAME. In the creation of indexes we always create a primary one by ID, and a copy
                one by NAME. But ID is leading. So, entity zone memberships will be evaluated by ID only.
                Therefore, the below is commented out to avoid false positives for those few scenarios
                where entities have bad zone ID/name combinations.

                if (!indexEntries["ZONE"].hasOwnProperty(zone.name)) {
                    log.messages.push("Index entry for zone name " + zone.name + " not found for " + entity.entityId);
                    return false;
                }
                if (!indexEntries["ZONE"][zone.name].members.has(entityProxy)) {
                    log.messages.push("Index entry for zone name " + zone.name + " does not contain " + entity.entityId);
                    return false;
                }
                referrers.add(indexEntries["ZONE"][zone.name].members); // Is a set - no dup will occur.
            */
            return true;
        });
        if (goodZones.length !== entity.managementZones.length)
            return false;
        }

        // Now let's check that the actual and expected 'referrer' lists are identical.
        if (referrers.size > entityProxy.referrers.size) {
            let expectedReferrers = [...referrers].map(referrer => referrer.name());
            let actualReferrers   = [...entityProxy.referrers].map(referrer => referrer.name());
            let missingReferrers  = expectedReferrers.without(actualReferrers);

            log.messages.push("Entity " + entity.entityId + " (" + entity.displayName + ") should also be referenced by " + missingReferrers.join(", "));
            return false;
        }
        if (referrers.size < entityProxy.referrers.size) {
            let expectedReferrers = [...referrers].map(referrer => referrer.name());
            let actualReferrers   = [...entityProxy.referrers].map(referrer => referrer.name());
            let missingReferrers  = actualReferrers.without(expectedReferrers);

            log.messages.push("Entity " + entity.entityId + " (" + entity.displayName + ") should not be referenced by " + missingReferrers.join(", "));
            return false;
        }
        // The above were easy, quick checks.  Now let's see if there is an actual one-for-one match.
        let goodReferrers = Array.from(referrers).filter(referrer => {
            if (!entityProxy.referrers.has(referrer)) {
                log.messages.push("Referrer " + referrer.name() + " is missing from the actual referrer list for entity " + entity.entityId + " (" + entity.displayName + ")");
                return false;
            }
            return true;
        });
        if (goodReferrers.length !== referrers.size)
            return false;

        return true;
    }    

    #validateEntities = async function (tenant, type) {
        const self = this;

        return new Promise(async function (resolve, reject) {
            const log = {
                tenant:   tenant,
                type:     type,
                start:    new Date(),
                finish:   null,
                failed:   false,
                messages: []
            };

            function getAllSets(obj) {
                let sets = [];
                if (obj) for (let prop in obj) {
                    if (obj[prop] instanceof Set)
                        sets.push(obj[prop]);
                    else if (typeof obj[prop] === "object")
                        sets = sets.concat(getAllSets(obj[prop]));
                }
                return sets;
            }

            try {
                const allEntities = this.curator.getEntities(tenant, type);
                if (allEntities.size === 0) {
                    log.messages.push("No " + this.friendly(type) + "s in " + tenant + " were found");
                    log.failed = true;
                    return resolve(log);
                }

                if (!this.#indexes[tenant] || !this.#indexes[tenant][type]) {
                    log.messages.push("No indexes were found for " + this.friendly(type) + "s in " + tenant);
                    log.failed = true;
                    return resolve(log);
                }

                const allIndexes  = this.#indexes[tenant][type];
                const allSets     = getAllSets(allIndexes);
                const foundCount  = await this.dataAccess.countEntities(tenant, type);

                // Let's see if the database agrees.
                if (foundCount != allEntities.size) {
                    log.messages.push("There are " + foundCount + " " + 
                    this.friendly(type) + "s in the database, " + allEntities.size + " in memory");
                    log.failed = true;
                }

                let setsWithBadReferences = allSets.filter(set =>
                    Array.from(set)
                    .filter(item => !allEntities.has(item))
                    .length !== 0
                );
                if (setsWithBadReferences.length !== 0) {
                    let badProxies = setsWithBadReferences.map(set =>
                        Array.from(set).map(entity =>
                            entity.displayName + " (" + entity.entityId + ")"
                        )
                    );
                    badProxies = [].concat(...badProxies).unique();

                    log.messages.push("Some indexes point to the following unknown or purged " +
                        this.friendly(type) + "s in " + tenant + ": " + badProxies.join(", "));
                    log.failed = true;
                };

                let allValidations = allEntities.map(entityProxy => {
                    // Not a promise! Just a callback.
                    return (resolve, reject) => {
                        if (entityProxy.entity) {
                            let isCorrect = this.#validateEntity(tenant, type, entityProxy, entityProxy.entity, log);
                            log.failed = log.failed || !isCorrect;
                            resolve();
                        }
                        else {
                            this.dataAccess.loadEntity(entityProxy.entityId, tenant, type)
                            .then(entity => {
                                let isCorrect = this.#validateEntity(tenant, type, entityProxy, entity, log);
                                log.failed = log.failed || !isCorrect;
                                resolve();
                            })
                            .catch(err => {
                                log.messages.push("Could not load record for " + 
                                    entityProxy.displayName + " in " + tenant + ": " + err);

                                this.entityAccess.removeEntity(tenant, type, entityProxy);      // Cleanup
                                resolve();  // No reject!
                            });
                        }
                    };
                });

                executeSerially(allValidations,
                    () => {
                        log.finish  = new Date();
                        log.elapsed = log.finish.getTime() - log.start.getTime();

                        log.messages.push("Validated all " + allEntities.size + " " + 
                            self.friendly(type) + "s in " + tenant + ", indexed along " + 
                            allSets.length + " dimensions, in " + log.elapsed + "ms");
                        resolve(log);
                    },
                    ({errors, data}) => {
                        reject(errors);
                    }
                );
            }
            catch (ex) {
                log.messages.push(ex.message);
                log.failed = true;
                resolve(log);
            }
        });
    }
   
    get curator() {
        return this.globals.curator;
    }
    get dataAccess() {
        return this.globals.dataAccess;
    }    
    get entityAccess() {
        return this.globals.entityAccess;
    }       

    constructor(globals) {
        this.globals       = globals;
        this.friendly      = globals.friendly;
        this.logger        = globals.logger;
        this.tenants       = globals.tenants;
        this.types         = globals.types;
        this.relationships = globals.relationships;

        this.tenants.forEach(tenant => {
            this.#indexes[tenant] = {};
     
            this.types.forEach(type => {
                this.#indexes[tenant][type] = {
                    ZONE: {
                        _members:  new Set(),
                        _nonmembers: new Set()
                    }
                };
            });
        });        
    }
    
    indexZones(tenant, zones) {
        this.types.forEach(type => {
            const indexes = this.#indexes[tenant][type];

            zones.forEach(zone => {
                if (!indexes["ZONE"].hasOwnProperty(zone.id)) {
                    indexes["ZONE"][zone.id] = indexes["ZONE"][zone.name] = {
                        id:      zone.id,
                        name:    zone.name,
                        members: new Set()
                    }
                }
            });
        });
    }

    getZones(tenant, type, startsWithValue) {
        const zones = [];

        const indexes = this.#indexes[tenant][type];
        for (let zone in indexes["ZONE"]) {
            // Skip the ID entries - only use name entries.
            if (indexes["ZONE"][zone].name === zone)
                zones.push(zone);
        }
        return startsWithValue ? zones.filter(zone => zone.name.startsWith(startsWithValue)) : zones;
    }

    getZoneId(tenant, type, name) {
        if (name === undefined) return null;

        const indexes = this.#indexes[tenant][type];
        if (indexes.hasOwnProperty("ZONE") &&
            indexes["ZONE"].hasOwnProperty(name))
            return indexes["ZONE"][name].id;

        return null;
    }

    addRelationship(firstProxy, secondProxy, relationship, direction) {
        let opposite = direction === "IN" ? "OUT" : "IN";

        // Make sure we have the arrays we need.
        firstProxy[relationship] = firstProxy[relationship] || {};
        firstProxy[relationship][direction] = firstProxy[relationship][direction] || [];
        firstProxy[relationship][opposite]  = firstProxy[relationship][opposite]  || [];

        if (firstProxy[relationship][direction].includes(secondProxy.entityId)) {
            firstProxy[relationship][direction].replace(secondProxy.entityId, secondProxy);
        }
        if (!firstProxy[relationship][direction].includes(secondProxy)) {
            firstProxy[relationship][direction].push(secondProxy);
        }
    }

    delRelationships(firstProxy, relationship, direction) {
        let opposite = direction === "IN" ? "OUT" : "IN";

        // Make sure we have the arrays we need.
        firstProxy[relationship] = firstProxy[relationship] || {};
        firstProxy[relationship][direction] = firstProxy[relationship][direction] || [];

        firstProxy[relationship][direction].forEach(secondProxy => {
            firstProxy[relationship][direction].remove(secondProxy);

            if (typeof secondProxy === "string") return;

            secondProxy[relationship] = secondProxy[relationship] || {};
            secondProxy[relationship][opposite] = secondProxy[relationship][opposite] || [];

            secondProxy[relationship][opposite].remove(firstProxy);
        });
    }

    forgetEntity(entityProxy) {
        // Remove proxy from all sets that referred to it.
        this.removeReferences(entityProxy);

        // Also relationships! Who points to us and who did we point to?!
        this.delRelationships(entityProxy, "flows", "IN");
        this.delRelationships(entityProxy, "stack", "IN");
        this.delRelationships(entityProxy, "group", "IN");

        this.delRelationships(entityProxy, "flows", "OUT");
        this.delRelationships(entityProxy, "stack", "OUT");
        this.delRelationships(entityProxy, "group", "OUT");
    }
        
    removeReferences(entityProxy) {
        // Remove proxy from all sets that referred to it.
        if (entityProxy.referrers)
            entityProxy.referrers.forEach(referrer => referrer && referrer.delete && referrer.delete(entityProxy));

        delete entityProxy.referrers;
    }    

    analyze(tenant, type, entityProxy, entity) {
        const self = this;

        function parseRelationships(entityProxy, thisEntity, entityType, relationshipType, direction) {
            const findNode = function (lists, id) {
                let node = null;
                for (let i=0; !node && i < lists.length; i++) {
                    node = lists[i] ? lists[i].get(id) : null;
                }
                return node;
            }

            let opposite = direction === "IN" ? "OUT" : "IN";
            let relation = self.relationships[entityType][relationshipType][direction];
            if (!relation) return;

            /*  For services, the 'accept' value may be an array, because both applications
                of various types as well as services can call services. Thus, we need to
                consult several entity lists to find the right one. For other types there
                will be only one value. The below handles this generically.

                The 'idList' is what the DynatraceAPI tells us, in terms of related entities.
                We'll map that list to actual entities, if we can find them. And where not,
                we'll keep the IDs.

                NOTE: relation.prop consists of two parts: fromRelationships or toRelationships
                and the property underneath that (e.g. 'runsOn'), reparated with a '.'.
                NOTE: Sometimes it consists if just a single property (e.g. 'hostGroup'). We're
                handling that as well. The assumption is that we're getting an array.
            */
            let parentProp   = relation.prop.split(".")[0];
            let relationProp = relation.prop.split(".")[1];

            try {
                let idList = (relationProp ? thisEntity[parentProp][relationProp] : thisEntity[parentProp]) || [];
                if (!Array.isArray(idList)) idList = [idList];

                let entityLists = relation.accept.map(type => self.curator.getEntities(tenant, type));

                /*  Replace all IDs with entity references - insofar we can at this point.
                    Also, in the same pass we will add related entities to the 'discovered' list.
                */
                entityProxy[relationshipType][direction] = idList.map(id => {
                    let referencedProxy = findNode(entityLists, id);
                    if (referencedProxy) {
                        // The referenced entity already exists. Make sure it points back to us.
                        self.addRelationship(referencedProxy, entityProxy, relationshipType, opposite);

                        // Add that opposite reference list to our list of referrers.
                        // Not needed - it's not of the right type.
                        // entityProxy.referrers.add(referencedProxy[relationshipType][opposite]);

                        return referencedProxy; // We have the referenced proxy, and it is pointing back at us.
                    }
                    else {
                        /*  The referenced entity does not exist yet. We'll reference it upon its
                            birth and leave the ID in place for now. We do need to put it on the
                            'discovered' list so that it will be retrieved later.

                            For IN-flows of services can't we unambigously determine the type of the
                            referenced entitiy, but we don't do 'discovery' of services anyway. They
                            are just retrieved. For the rest, we can simpy take the first member of
                            the 'accept' list and use that as the type.
                        */
                        self.curator.addDiscovered(tenant, [relation.accept[0]], id);

                        return id; // Sadly, we couldn't find the proxy yet, so we keep the ID.
                    }
                });
            }
            catch (ex) {
                this.logger.warn("Error while parsing relationships: " + ex.message);
            }
        }

        function addReference(entityProxy, list) {
            list.add(entityProxy);
            entityProxy.referrers.add(list);
        }

        const indexes = this.#indexes[tenant][type];

        // Add the new version to all tag sets that it matches.
        (entity.tags || []).forEach(tag => {
            // Create a new entry for this tag if it does not exist yet.
            tag.context = tag.context || "CONTEXTLESS";

            if (!indexes.hasOwnProperty(tag.context))
                indexes[tag.context] = {};

            if (!indexes[tag.context].hasOwnProperty(tag.key))
                indexes[tag.context][tag.key] = {
                    _exists: new IndexedSet([], "entityId", tenant+"_"+type+"_"+tag.context+"_"+tag.key+"_exists"),
                    _empty:  new IndexedSet([], "entityId", tenant+"_"+type+"_"+tag.context+"_"+tag.key+"_empty")
                };

            const entry = indexes[tag.context][tag.key];

            // Add this entity to it, meaning: Yep, this entity has this tag at all.
            addReference(entityProxy, entry._exists);

            tag.value = tag.value || "_empty";
            tag.value.split(",").forEach(value => {
                // Create a new entry for this tag value if it does not exist yet.
                if (!entry.hasOwnProperty(value))
                    entry[value] = new IndexedSet([], "entityId", tenant+"_"+type+"_"+tag.context+"_"+tag.key+"_"+value);

                // Add this entity to it.
                addReference(entityProxy, entry[value]);
            });
        });

        // Collect property values (except un-index-able ones).
        if (!indexes.hasOwnProperty("PROPERTY"))
            indexes["PROPERTY"] = {};
        if (!indexes.hasOwnProperty("HOST_GROUP"))
            indexes["HOST_GROUP"] = {
                _exists: new IndexedSet([], "entityId", tenant+"_"+type+"_HOST_GROUP_exists"),
                _empty:  new IndexedSet([], "entityId", tenant+"_"+type+"_HOST_GROUP_empty")
            };

        if (!entity.hasOwnProperty("hostGroup")) {
            addReference(entityProxy, indexes["HOST_GROUP"]._empty);
        }

        let entry = null;

        for (var prop in entity) {
            if (prop === "toRelationships") {
                parseRelationships(entityProxy, entity, type, "flows", "IN");
                parseRelationships(entityProxy, entity, type, "stack", "IN");
                parseRelationships(entityProxy, entity, type, "group", "IN");
            } 
            else if (prop === "fromRelationships") {
                parseRelationships(entityProxy, entity, type, "flows", "OUT");
                parseRelationships(entityProxy, entity, type, "stack", "OUT");
                parseRelationships(entityProxy, entity, type, "group", "OUT");
            }
            else if (prop === "hostGroup") {
                addReference(entityProxy, indexes["HOST_GROUP"]._exists);

                if (!indexes["HOST_GROUP"].hasOwnProperty(entity.hostGroup.name))
                    indexes["HOST_GROUP"][entity.hostGroup.name] = {
                        id:      entity.hostGroup.meId,
                        name:    entity.hostGroup.name,
                        members: new Set()
                    };
                    // TODO: We have a host group (_exists) and we are a member of _this_ one.

                entry = indexes["HOST_GROUP"][entity.hostGroup.name];

                // Add this entity to it, meaning: Yep, this entity has this property at all.
                addReference(entityProxy, entry.members);
            }
            else if (typeof entity[prop] === "string" && !this.#unindexables.includes(prop)) {
                // TODO: Some properties are arays of strings. For example: ipAddresses.
                // It would be good to capture that, too, and create distinct property
                // value indexes that point to this entity.

                if (!indexes["PROPERTY"].hasOwnProperty(prop))
                    indexes["PROPERTY"][prop] = {
                        _exists: new IndexedSet([], "entityId", tenant+"_"+type+"_PROPERTY_"+prop+"_exists"),
                        _empty:  new IndexedSet([], "entityId", tenant+"_"+type+"_PROPERTY_"+prop+"_empty")
                    };

                entry = indexes["PROPERTY"][prop];

                // Add this entity to it, meaning: Yep, this entity has this property at all.
                addReference(entityProxy, entry._exists);

                entity[prop] = entity[prop] || "_empty";

                // Create entry for this value.
                if (!entry.hasOwnProperty(entity[prop]))
                    entry[entity[prop]] = new IndexedSet([], "entityId", tenant+"_"+type+"_PROPERTY"+"_"+prop+"_"+entity[prop]);

                // Add this entity to it.
                addReference(entityProxy, entry[entity[prop]]);
            }
        }

        // Finally the management zones.
        if (!indexes.hasOwnProperty("ZONE"))
            indexes["ZONE"] = {
                _members:  new Set(),
                _nonmembers: new Set()
            };

        // First add the entity to the 'insiders' or 'outsiders' sets. This is for easy filtering later.
        if (!entity.managementZones || entity.managementZones.length === 0) {
            addReference(entityProxy, indexes["ZONE"]._nonmembers);
            return;
        }

        addReference(entityProxy, indexes["ZONE"]._members);

        (entity.managementZones || []).forEach(zone => {
            /*  Create a new entry for this zone if it does not exist yet.
                Note: Zones have an ID and a name. In order to allow for filtering by name
                as well as by ID, we create two entries in the zoneEntry: one by ID and one
                by name, both pointing to the very same object (and thus the same set).
            */
            if (!indexes["ZONE"].hasOwnProperty(zone.id))  {
                // It is possible that another zone with the same name but a different
                // ID exists. In that case, should be clear that out? Or check about
                // the other ID it seems to have? The other thing has references that
                // may not be valid anymore... Maybe just log an error so that we know
                // that this scenario happens.
                if (indexes["ZONE"][zone.name])
                    this.logger.warn("Zone " + zone.name + " already exists with ID '" + indexes["ZONE"][zone.name].id + "' - adopting new ID");

                indexes["ZONE"][zone.id] = indexes["ZONE"][zone.name] = {
                    id:      zone.id,
                    name:    zone.name,
                    members: new Set()   // This needs to be reconsidered. Should we not adopt the known members?
                };
            }
            // Add this entity to it.
            addReference(entityProxy, indexes["ZONE"][zone.id].members);
        });
    }

    validate() {
        let start = new Date();

        this.logger.info("Starting a validation of all indexes and proxies against the database");
        this.logger.info("    Validation will run asynchronously and report when finished");

        const affectedPipelines = this.curator.getPipelines(this.tenants) || [];

        this.logger.info("    Pausing " + affectedPipelines.length + " affected pipelines");
        affectedPipelines.forEach(pipeline => pipeline.pause());

        return new Promise((resolve, reject) => {
            const promises = [];

            this.tenants.forEach(tenant => {
                this.types.forEach(type => {
                    promises.push(this.#validateEntities(tenant, type));
                });
            });

            Promise
            .all(promises)
            .then(logs => {
                let finish = new Date();
                let messages = logs.map(log => log.messages);
                let failedTenants = logs.filter(log => log.failed).map(log => log.tenant).unique();

                this.logger.info("    Validation report:");
                this.logger.info("        Known tenants: " + this.tenants.join(", "));

                for (let i=0; i < messages.length; i++) {
                    for (let j=0; j < messages[i].length; j++) {
                        this.logger.warn("        " + messages[i][j]);
                    }
                }

                this.logger.info("        Problematic tenants: " + failedTenants.length > 0 ? failedTenants.join(", ") : "none");
                this.logger.info("    Validating all entities in " + this.tenants.length + " tenants took " + (finish.getTime() - start.getTime()) + "ms");
                this.logger.info("    Resuming " + affectedPipelines.length + " affected pipelines");

                affectedPipelines.forEach(pipeline => pipeline.resume());

                resolve(failedTenants);
            })
            .catch(error => {
                this.logger.error("   Validation of indexes and proxies failed: " + error);

                this.logger.info("    Resuming " + affectedPipelines.length + " affected pipelines");
                affectedPipelines.forEach(pipeline => pipeline.resume());

                reject(error);
            });
        });
    }

    clear(tenant, type) {
        // Better than 'entities[tenant][type].clear()' because this cleans the
        // indexes as well. To be sure, we clear them afterwards. When we save,
        // we skip those with empty collections.
        const proxies = this.curator.getEntities(tenant, type);

        for (let i=0; i < proxies.length; i++) {
            this.entityAccess.removeEntity(tenant, type, proxies[i]);
        }

        // Now clear the indexes. They should all be emptied already because 
        // we removed all entities, and that clears their indexes as well.
        const indexes = this.#indexes[tenant][type];

        for (let ctx in indexes) {
            if (!indexes.hasOwnProperty(ctx)) continue;

            const context = indexes[ctx];

            for (let prop in context) {
                if (!context.hasOwnProperty(prop)) continue;

                if (context[prop] instanceof Set) {
                    context[prop].clear();
                }
                else if (typeof context[prop] === "object") {
                    for (let subprop in context[prop]) {
                        if (!context[prop].hasOwnProperty(subprop)) continue;

                        const entry = context[prop][subprop];   // CONTEXTLESS.MyTag

                        if (entry instanceof Set) {             // _empty, _exists and (some value)
                            entry.clear();
                        }
                        else if (typeof entry === "object") {   // SomeValue = (not a list)
                            // Does not happen
                        }
                        else {
                            // No action needed
                        }
                    }
                }
                else {
                    // No action needed
                }
            }
        }
    }

    cleanup(tenant, type) {
        const indexes = this.#indexes[tenant][type];

        for (let ctx in indexes) {
            if (!indexes.hasOwnProperty(ctx)) continue;

            const context = indexes[ctx];

            for (let prop in context) {
                if (!context.hasOwnProperty(prop)) continue;

                // If prop is a zone name/id and it has no members, remove the zone.
                if (ctx === "ZONE") {
                    if (prop !== "_members" &&
                        prop !== "_nonmembers" &&
                        context[prop].members &&
                        context[prop].members.size === 0) {
                        const zone = context[prop];
                        delete context[zone.id];
                        delete context[zone.name];
                        continue;
                    }
                }
                else if (ctx === "HOST_GROUP") {
                    // TODO
                    let a = 0;
                }
                else {
                    if (context[prop]._exists && context[prop]._exists.size === 0) {
                        delete context[prop];
                        continue;
                    }

                    const container = context[prop];
                    for (let subprop in container) {
                        if (!container.hasOwnProperty(subprop) || subprop === "_exists" || subprop === "_empty") continue;

                        if (container[subprop] instanceof Set && container[subprop].size === 0) {
                            delete container[subprop];
                            continue;
                        }
                    }
                }
            }
        }
    }
    
    load() {
        /*  Per tenant:
            Step 1: Load the entities for all types as proxies.
                1.a Replace the IN and OUT IDs with references (after everything in a tenant has been loaded).
                1.b Update the referrers array per referenced proxy.
            Step 2: Per type:
                2.a Load the indexes per type.
                2.b Replace the IDs with references to proxies.
                2.c Update the referrers array per referenced proxy.
            Step 3: Run validation
                3.a Do this per type, and do it asynchronously, because,
                    technically, the library is ready to handle requests.
                3.b The validation process will load from the database and check
                    if everything is up-to-date with what is found there.
                3.c If inconsistencies are found, just update the administration.
         */
        let failedTenants = [];

        this.logger.info("    Loading all proxies and indexes from file");

        try {
            this.tenants.forEach(tenant => {
                const allProxies = new IndexedSet([], "entityId", tenant+"_all");

                this.types.forEach(type => {
                    try {
                        let data = fs.readFileSync(
                            "../data/entities_" + tenant + "_" + type + ".json",
                            { encoding: "utf8" }
                        );
                        const list = JSON.parse(data);
                        this.logger.info("    Found " + list.length + " " + this.friendly(type) + "s in " + tenant);

                        this.curator.setEntities(tenant, type, list);

                        for (let i=0; i < list.length; i++) {
                            allProxies.add(list[i]);
                        }
                    }
                    catch (ex) {
                        failedTenants.push(tenant);
                        this.logger.error("   Could not load proxies for " + this.friendly(type) + "s in " + tenant + ": " + ex.message);
                    }
                });

                // All entities for this tenant have been loaded. Now we fix the IN and OUT references.
                // Once we have cross-tenant links, we need to do this _after_ all tenants have been loaded.
                // Note: The 'lastSeen' timestamp is reset because we may start to delete too much stuff that
                // appears old but really isn't. We do preserve the 'lastUpdated' timestamp.
                this.types.forEach(type => {
                    const proxies = this.curator.getEntities(tenant, type);

                    // This ought to be the fastest way, faster than using an index. But that's fast, too.
                    for (const proxy of proxies) {
                        proxy.type        = type;
                        proxy.tenant      = tenant;
                        // Don't do this, because lastUpdated is to tell about deltas: 
                        // proxy.lastUpdated = new Date().getTime();
                        proxy.lastSeen    = proxy.lastSeen || new Date().getTime();
                        proxy.lastUsed    = 0;
                        proxy.referrers   = new Set();
                        proxy.entity      = null;
                        proxy.flows.IN    = proxy.flows.IN .map(entityId => allProxies.get(entityId) || entityId);
                        proxy.flows.OUT   = proxy.flows.OUT.map(entityId => allProxies.get(entityId) || entityId);
                        proxy.stack.IN    = proxy.stack.IN .map(entityId => allProxies.get(entityId) || entityId);
                        proxy.stack.OUT   = proxy.stack.OUT.map(entityId => allProxies.get(entityId) || entityId);
                        proxy.group.IN    = proxy.group.IN .map(entityId => allProxies.get(entityId) || entityId);
                        proxy.group.OUT   = proxy.group.OUT.map(entityId => allProxies.get(entityId) || entityId);
                    };

                    try {
                        let data = fs.readFileSync(
                            "../data/indexes_" + tenant + "_" + type + ".json",
                            { encoding: "utf8" }
                        );
                        this.#indexes[tenant][type] = JSON.parse(data);

                        // Now we need to resolve all IDs to references again.
                        const indexes = Object.entries(this.#indexes[tenant][type]) || [];
                        indexes.forEach(entry => {
                            const context = entry[0];
                            const keys    = entry[1];

                            const keyValues = Object.entries(keys) || [];
                            keyValues.forEach(value => {
                                const key  = value[0];
                                const coll = value[1];

                                if (Array.isArray(coll)) {
                                    // We need to resolve IDs to entities and add back-references.
                                    this.#indexes[tenant][type][context][key] = new IndexedSet(coll.map(
                                        entityId => allProxies.get(entityId) || entityId
                                    ), "entityId", tenant+"_"+type+"_"+context+"_"+key);

                                    this.#indexes[tenant][type][context][key].forEach(
                                        proxy => proxy && proxy.referrers
                                               ? proxy.referrers.add(this.#indexes[tenant][type][context][key])
                                               : null
                                    );
                                }
                                else if (typeof coll === "object") {
                                    const collEntries = Object.entries(coll) || [];
                                    collEntries.forEach(entry => {
                                        const subkey = entry[0];
                                        const subval = entry[1];

                                        if (Array.isArray(this.#indexes[tenant][type][context][key][subkey])) {
                                            let list = this.#indexes[tenant][type][context][key][subkey];
                                            // We need to resolve IDs to entities and add back-references.
                                            this.#indexes[tenant][type][context][key][subkey] = new IndexedSet(list.map(
                                                entityId => allProxies.get(entityId) || entityId
                                            ), "entityId", tenant+"_"+type+"_"+context+"_"+key+"_"+subkey);
                                            this.#indexes[tenant][type][context][key][subkey].forEach(
                                                proxy => proxy && proxy.referrers
                                                       ? proxy.referrers.add(this.#indexes[tenant][type][context][key][subkey])
                                                       : null
                                            );
                                        }
                                    });
                                }
                                else {
                                    // Is a simple value. Those we can leave alone.
                                }

                                // NOTE: if context is "ZONE", keys are zone IDs or names (or _members or _nonmembers).
                                // We need to make sure the two entries point to the very same object.
                                if (context === "ZONE" && key !== "_members" && key !== "_nonmembers") {
                                    let zoneName   = this.#indexes[tenant][type][context][key].name;
                                    this.#indexes[tenant][type][context][zoneName] = this.#indexes[tenant][type][context][key];
                                }
                            });
                        });
                    }
                    catch (ex) {
                        failedTenants.push(tenant);
                        this.logger.error("   Could not load indexes for " + this.friendly(type) + "s in " + tenant + ": " + ex.message);
                    }
                });
            });

            if (failedTenants.length > 0) {
                return failedTenants.unique();
            }

            this.logger.info("    Indexes and proxies for all tenants have been loaded");
            return null;
        }
        catch (ex) {
            this.logger.error("    Not all indexes and proxies could be loaded: " + ex.message);
            return this.tenants;
        }
        finally {
            this.#hasInitialized = true;
        }
    }
    
    save(tenant, onDone = function () {}) {
        const self = this;
        const isAlpha = function() {
            const hasAlpha  = /[A-Za-z]/;
            return function(value) {
                return hasAlpha.test(value);
            }
        }();

        if (!this.#hasInitialized) return onDone();

        const tasks = [];
        const tenants = tenant ? [ tenant ] : this.tenants || [];

        try {
            this.logger.info("Saving entity administration");

            tenants.forEach(tenant => {
                this.types.forEach(type => {
                    tasks.push(function (callback) {
                        self.logger.info("    Saving " + self.friendly(type) + " proxies for " + tenant);

                        const serializableEntities = [];
                        const proxies = self.curator.getEntities(tenant, type);

                        let proxy = null;
                        for (let i=0; i < proxies.length; i++) {
                            proxy = proxies[i];

                            serializableEntities.push({
                                entityId:    proxy.entityId,
                                displayName: proxy.displayName,
                                lastSeen:    proxy.lastSeen,
                                lastUpdated: proxy.lastUpdated,
                                flows:    {  IN:  proxy.flows.IN .map(proxy => proxy.entityId || proxy),
                                             OUT: proxy.flows.OUT.map(proxy => proxy.entityId || proxy)
                                          },
                                stack:    {  IN:  proxy.stack.IN .map(proxy => proxy.entityId || proxy),
                                             OUT: proxy.stack.OUT.map(proxy => proxy.entityId || proxy)
                                          },
                                group:    {  IN:  proxy.group.IN .map(proxy => proxy.entityId || proxy),
                                             OUT: proxy.group.OUT.map(proxy => proxy.entityId || proxy)
                                          }
                            });
                        };

                        fs.writeFile(
                            "../data/entities_" + tenant + "_" + type + ".json",
                            JSON.stringify(serializableEntities, null, 4),
                            'utf8',
                            callback
                        );
                    });

                    tasks.push(function (callback) {
                        self.logger.info("    Saving " + self.friendly(type) + " indexes for " + tenant);

                        try {
                            // First, remove empty stuff.
                            self.cleanup(tenant, type);

                            // Now, serialize what remains.
                            const serializableIndexes = {};

                            const indexes = Object.entries(self.#indexes[tenant][type]) || [];

                            let entry = null;
                            for (let i=0; i < indexes.length; i++) {
                                entry = indexes[i];

                                const context = entry[0];
                                const keys    = entry[1];

                                serializableIndexes[context] = {};

                                const keyValues = Object.entries(keys) || [];
                                keyValues.forEach(value => {
                                    const key        = value[0];
                                    const collection = value[1];

                                    // ZONE entries that are names are not going to be saved. We recreate those on load.
                                    if (context === "ZONE" &&
                                        key !== "_members" &&
                                        key !== "_nonmembers" &&
                                        isAlpha(key))
                                        return;

                                    serializableIndexes[context][key] = {};

                                    if (collection instanceof Set) {
                                        serializableIndexes[context][key] = [ ...collection.values() ]
                                        .map(entityProxy => entityProxy ? entityProxy.entityId : null)
                                        .filter(id => !!id);    // Remove null values
                                    }
                                    else {
                                        const indexCollections = Object.entries(collection) || [];
                                        indexCollections.forEach(collection => {
                                            const name = collection[0];
                                            const set  = collection[1];

                                            if (set instanceof Set) {
                                                serializableIndexes[context][key][name] = [ ...set.values() ]
                                                    .map(entityProxy => entityProxy ? entityProxy.entityId : null)
                                                    .filter(id => !!id);  // Remove null values
                                            }
                                            else if (typeof set === "object") {
                                                serializableIndexes[context][key][name] = {};
                                                for (let subkey in set) {
                                                    if (set[subkey] instanceof Set)
                                                        serializableIndexes[context][key][name][subkey] = [ ...set[subkey].values() ]
                                                        .map(entityProxy => entityProxy ? entityProxy.entityId : null)
                                                        .filter(id => !!id);  // Remove null values
                                                    else
                                                        serializableIndexes[context][key][name][subkey] = set[subkey];
                                                }
                                            }
                                            else {
                                                serializableIndexes[context][key][name] = set; // Or do nothing
                                            }
                                        });
                                    }
                                });
                            };

                            fs.writeFile(
                                "../data/indexes_" + tenant + "_" + type + ".json",
                                JSON.stringify(serializableIndexes, null, 4),
                                'utf8',
                                callback
                            );
                        }
                        catch (ex) {
                            callback(ex.message);
                        }
                    });
                });
            });

            async.series(tasks, err => {
                if (err)
                    self.logger.error("The entity administration could not be saved: " + err);
                else
                    self.logger.info("The entity administration was successfully saved");

                onDone();
            });
        }
        catch (ex) {
            this.logger.error("The entity administration could not be saved: " + ex.message);

            onDone();
        }
    }

    // QUERY INTERFACES
    filterIndexes(tenant, name) {
        const self = this;

        return new Promise((resolve, reject) => {
            if (!self.#indexes[tenant]) return reject("Unknown tenant " + tenant);

            const pattern = name ? new RegExp(name) : null;
            const isName  = /[A-Za-z]/;

            // Initialize the matches with the types we know of.
            // Is a convenience for new API users.
            let matches = self.types.reduce((acc, type) => {
                acc[type] = {
                    displayName: type
                };
                return acc;
            },{});

            // Now let's find tag contexts, properties and zones.
            for (let i=0; i < self.types.length; i++) {
                let typeIndex = self.#indexes[tenant][self.types[i]];

                for (let idx in typeIndex) {
                    let keys   = null;
                    let values = null;

                    if (typeIndex.hasOwnProperty(idx) &&
                        (!pattern || pattern.test(idx))) {

                        if (idx === "ZONE") {
                            keys   = Object.keys(typeIndex[idx]).filter(key => !isName.test(key));
                            values = keys.map(key => {
                                return { entityId: key, displayName: typeIndex[idx][key].name }
                            });

                            if (!matches.hasOwnProperty(idx))
                                matches[idx] = {
                                    displayName: idx,
                                    values: []
                                };

                            // Because we are iterating over entity types, we need to maintain uniqueness of zones.
                            matches[idx].values = matches[idx].values
                                .concat(values)
                                .filter((v, i, a) => a.map(a => a.entityId).indexOf(v.entityId) === i);

                            matches[idx].values.sort((a, b) => { return b.entityId - a.entityId });
                        }
                        else {
                            keys = Object.keys(typeIndex[idx]);

                            if (!matches.hasOwnProperty(idx))
                                matches[idx] = {
                                    displayName: idx,
                                    values: []
                                };

                            matches[idx].values = matches[idx].values
                                .concat(keys)
                                .filter((v, i, a) => a.indexOf(v) === i);

                            matches[idx].values.sort();
                        }
                    }
                }
            };
            resolve(Object.values(matches));
        });
    }

    filterZones(tenant, filters, name) {
        const self = this;

        const isAlpha = function() {
            const hasAlpha  = /[A-Za-z]/;
            return function(value) {
                // "_nonmembers" is considered an internal, non-alpha key.
                return value !== "_nonmembers" || hasAlpha.test(value);
            }
        }();

        const nameMatch = function (name) {
            if (!name) {
                return function(value) {
                    return true;
                };
            }
            else {
                const pattern = new RegExp(name);
                return function(value) {
                    return !value || pattern.test(value);
                };
            }
        }(name);

        const idMatch  = function (filter) {
            // Returns a function that will return true or false given an ID.
            if (!filter) {
                return function(value) {
                    return true;
                };
            }
            else {
                const negate = filter.operator === "!=";
                const idList = filter.values;
                return function(value) {
                    return negate ? !idList.includes(value) : idList.includes(value);
                };
            }
        }((filters || []).filter(f => f.context === "VALUE" && f.name === "entityId")[0]);

        return new Promise((resolve, reject) => {
            if (!self.#indexes[tenant]) return reject("Unknown tenant " + tenant);

            let matches = {};
            // TODO: What if there are no zones? Then everything is by definition an outsider.

            for (let i=0; i < self.types.length; i++) {
                let zoneIndex = self.#indexes[tenant][types[i]].ZONE || {};

                for (let zone in zoneIndex) {
                    if (!zoneIndex.hasOwnProperty(zone) ||
                        zone === "_members"             ||  // "_members" is just everybody that is in a zone.
                        !isAlpha(zone)                  ||  // Textual entries are ignored, except "_nonmembers".
                        !idMatch(zone)                  ||  // If specific zones requested, skip everybody else.
                        !nameMatch(zoneIndex[zone].name))   // If name match requested, skip everbody else.
                        continue;

                    // TODO: We are going to have to treat members and non-members just like everything else. Maybe.
                    if (zone === "_nonmembers") {
                        matches[zone] = {
                            entityId: zone,
                            displayName: "No Management Zone"
                        };

                        // Add the entities that are not a member of any zone to a new property (ie. hosts: []).
                        // UPDATE: NO, we are no longer doing that. If you need non-members of a managemant
                        // zone, use the ZONE:id=(empty) query parameter. If you need everything, use /all.
                        // DISABLED: matches[zone][types[i]] = Array.from(zoneIndex[zone]).map(member => member.entityId);
                    }
                    else {
                        if (!matches.hasOwnProperty(zone))
                            matches[zone] = {
                                entityId: zone,
                                displayName: zoneIndex[zone].name
                            };

                        // Add the entities of current type to a new property (ie. hosts: []).
                        // UPDATE: NO, we are no longer doing that. If you need members of a managemant
                        // zone, use the ZONE:id=... query parameter. If you need everything, use /all.
                        // DISABLED: matches[zone][types[i]] = Array.from(zoneIndex[zone].members).map(member => member.entityId);
                    }
                }
            }
            resolve(Object.values(matches));
        });
    }

    filterTags(tenant, context, name) {
        const self = this;
        const pattern = name ? new RegExp(name) : null;

        function remove(arr, val) {
            let i = arr.indexOf(val);
            if (i > -1) arr.splice(i,1);
        }

        return new Promise((resolve, reject) => {
            if (!self.#indexes[tenant]) return reject("Unknown tenant " + tenant);

            let matches = {};

            for (let i=0; i <self. types.length; i++) {
                let typeIndex = self.#indexes[tenant][self.types[i]][context];

                for (let idx in typeIndex) {
                    if (typeIndex.hasOwnProperty(idx) &&
                        (!pattern || pattern.test(idx))) {

                        let keys = Object.keys(typeIndex[idx]);
                        remove(keys, "_exists");
                        remove(keys, "_empty");

                        if (!matches.hasOwnProperty(idx))
                            matches[idx] = {
                                displayName: idx,
                                values: []
                            };

                        // Add the values found for the current entity type to the list (no duplicates).
                        matches[idx].values = matches[idx].values
                            .concat(keys)
                            .filter((v, i, a) => a.indexOf(v) === i);

                        // Add the list of values for current type that have this tag.
                        matches[idx][self.types[i]] = keys;

                        matches[idx].values.sort();
                    }
                }
            }
            resolve(Object.values(matches));
        });
    }    

    collectEntities(tenant, type, filters, name, requestedFields, since, getProxies) {
        const self = this;

        function getNameMatches(candidates, field, value) {
            if (!value)
                return candidates;

            // RegExp.test() is faster than String.includes() or String.indexOf(), they say.
            const result = [];
            const pattern = new RegExp(value);
            for (let i = 0; i < candidates.length; i++) {
                if (pattern.test(candidates[i][field])) result.push(candidates[i]);
            }
            return result;
        }

        function getSinceMatches(candidates, value) {
            if (!value)
                return candidates;

            return candidates.filter(proxy => proxy.lastUpdated > value);
        }

        function getZoneMembers(allEntities, zonesEntry, operator, values) {
            /*  Note: values is an array. The following conditions are supported:
                - Zone=         Entity must not be a member of any zone.
                - Zone=*        Entity must be a member of at least one zone.
                - Zone!=        Entity must be a member of at least one zone.
                - Zone!=*       Entity must not be a member of any zone.
                - Zone=a,b,c    Entity must be in zone 'a', 'b' or 'c'.
                - Zone!=a,b,c   Entity must not be in zone 'a', 'b' or 'c'.
             */
            let candidates = new Set();

            for (let v = 0; v < values.length; v++) {
                let matches = null;
                let value = values[v] === "empty" ? "" : values[v];

                switch (operator) {
                    case "=":
                        if (value === "")
                            // Entity must not be a member of any zone (zero).
                            matches = zonesEntry._nonmembers;
                        else if (value === ".")
                            // Entity must be a member of zero or more zones.
                            matches = allEntities;
                        else if (value === "*")
                            // Entity must be a member of at least one zone.
                            matches = zonesEntry._members;
                        else
                            // Entity must be a member of the given zone.
                            matches = zonesEntry[value] ? zonesEntry[value].members : [];
                        break;
                    case "!=":
                        if (value === "")
                            // Entity must be a member of at least one zone.
                            matches = zonesEntry._members;
                        else if (value === ".")
                            // Entity must not be a member of zero or more zones. That's nobody.
                            matches = new Set();
                        else if (value === "*")
                            // Entity must not be a member of any zone.
                            matches = zonesEntry._nonmembers;
                        else
                            // Entity must not be a member of the given zone.
                            matches = without(allEntities, zonesEntry[value] ? zonesEntry[value].members : []);
                        break;
                    default:
                        logger.warn("Operator '" + operator + "' is not supported yet");
                        break;
                }
                add(candidates, matches);
            }

            return candidates;
        }

        function getFilterMatches(allEntities, hashEntry, operator, values) {
            /*  Note: values is an array. The following conditions are supported:
                - MyTag=        MyTag must exist and be empty.
                - MyTag=*       MyTag must exist and have a value.
                - MyTag=.       MyTag must exist at all (empty or not).
                - MyTag!=       MyTag must exist and have a value (not empty).
                - MyTag!=*      MyTag must exist and have no value (be empty).
                - MyTag!=.      MyTag must not exist at all.
                - MyTag=a,b,c   MyTag must equal 'a', 'b' or 'c'.
                - MyTag!=a,b,c  MyTag must exist and not equal 'a', 'b' or 'c'.

                The following can only be used in POSTed requests because they are incompatible
                with the query parameter syntax.
                - MyTag>"       MyTag must have a value larger than 'a'.
                - MyTag!>a      MyTag must have a value smaller or equal to 'a'.
                - MyTag<a       MyTag must have a value smaller than 'a'.
                - MyTag!<a      MyTag must have a value larger or equal to 'a'.
                - MyTag*=a      MyTag must have a value that starts with 'a'.

                The notion of 'between' or 'outside' can be expressed by using two separate filters
                that specify the same tag name: one for '>' and one for '<'.

                Of course, this is not only for tags. Properties and zones can be specified as well.
             */
            let candidates = new Set();

            // It is possible that this tag value does not exist at all!
            hashEntry = hashEntry || {};
            hashEntry._exists = hashEntry._exists || new IndexedSet([], "entityId");
            hashEntry._empty  = hashEntry._empty  || new IndexedSet([], "entityId");

            if (values.length === 0) // No values at all means one, empty value.
                values = [""];

            for (let v = 0; v < values.length; v++) {
                let matches = null;
                let value = values[v];

                switch (operator) {
                    case "=":
                        if (value === "")
                            // MyTag must exist and be empty.
                            matches = hashEntry._empty;
                        else if (value === ".")
                            // MyTag must exist at all (empty or not).
                            matches = hashEntry._exists;
                        else if (value === "*")
                            // MyTag must exist and have a value.
                            matches = without(hashEntry._exists, hashEntry._empty);
                        else
                            // MyTag must have the specified value.
                            matches = hashEntry[value] || new Set();
                        break;
                    case "!=":
                        if (value === "")
                            // MyTag must exist and have a value (not empty).
                            matches = without(hashEntry._exists, hashEntry._empty);
                        else if (value === ".")
                            // MyTag must not exist at all.
                            matches = without(allEntities, hashEntry._exists);
                        else if (value === "*")
                            // MyTag must exist and have no value (be empty).
                            matches = hashEntry._empty;
                        else
                            // MyTag must NOT have the specified value.
                            matches = without(hashEntry._exists, hashEntry[value] || new Set())
                        break;
                    case ">":
                    case "!>":
                    case "<":
                    case "!<":
                    default:
                        logger.warn("Operator '" + operator + "' is not supported yet");
                        matches = new Set();
                        break;
                }
                add(candidates, matches);
            }

            return candidates;
        }

        function getValueMatches(allEntities, field, operator, values) {
            let candidates = new Set();
            let negated = operator === "!=";

            allEntities = Array.from(allEntities);
            for (let i = 0; i < allEntities.length; i++) {
                if ((negated && !values.includes(allEntities[i][field])) ||
                    (!negated && values.includes(allEntities[i][field])))
                    candidates.add(allEntities[i]);
            }
            return candidates;
        }

        return new Promise((resolve, reject) => {
            if (!self.#indexes[tenant]) return reject("Unknown tenant " + tenant);

            if (type === "ALL") {
                // The user wants to query across entities. We call ourselves for each type.
                // We return an object that has a property per type, each containing matches.
                let response = {};
                let promises = self.types.map(t => {
                    return new Promise((resolve, reject) => {
                        collectEntities(tenant, t, filters, name, requestedFields, since, getProxies)
                            .then(result => {
                                response[t] = result;
                                resolve(response);
                            })
                            .catch(error => {
                                response[t] = error;
                                resolve(response);
                            });
                    })
                });
                Promise
                    .all(promises)
                    .then(() => resolve(response))
                    .catch(error => reject(error));

                // That's it. Let the recursion do its work.
                return;
            }

            if (!self.#indexes[tenant][type]) return reject("Unknown type " + type);

            if (filters.length > 0 && Array.isArray(filters[0])) {
                /*  Typically, there is one filter with AND-ed criteria, but multiple (OR-ed) groups
                    of filters are supported as well. We treat having one or multiple groups in the
                    same way: we call ourselves for each group and join the results.
                 */
                const seen = {};          // Tracks uniqueness.
                const matches = []; // Will contain the joined set of unique results.

                const promises = filters.map(filterGroup => {
                    return new Promise((resolve, reject) => {
                        self.collectEntities(tenant, type, filterGroup, name, requestedFields, since, getProxies)
                            .then(result => {
                                for (let i = 0; i < result.length; i++) {
                                    if (!seen[result[i].entityId]) {
                                        seen[result[i].entityId] = true;
                                        matches.push(result[i]);
                                    }
                                }
                                resolve();  // The overall set of matches is kept outside the promises.
                            })
                            .catch(error => reject(error));
                    })
                });
                // Now wait for all the OR-ed groups to produce their results, and aggregate them.
                Promise
                    .all(promises)
                    .then(() => resolve(matches))
                    .catch(error => reject(error));

                // That's it. Let the recursion do its work.
                return;
            }

            let allEntities = self.curator.getEntities(tenant, type);
            let candidates = [];

            /*  An optimization: If the filters contain "VALUE:entityId" then that's
                the first set of candidates right there. But also,  'allEntities' can
                now immediately be constrained to that sub-collection now. That filter
                can then be removed from the filters, and things can continue as usual.
             */
            for (let j = 0; j < filters.length; j++) {
                if (filters[j].context === "VALUE" &&
                    filters[j].key === "entityId" &&
                    filters[j].operator === "=") {

                    let subSet = new Set();
                    let proxy = null;
                    let values = filters[j].values;
                    for (let v = 0; v < values.length; v++) {
                        proxy = allEntities.get(values[v]);
                        if (proxy)
                            subSet.add(proxy);
                    }
                    // The specified entities are a set of candidates in their own right.
                    candidates.push(subSet);
                    // But they also consititute a limit on what we have to consider for
                    // 'negative' searches in getFilterMatches() and getZoneMembers().
                    allEntities = new IndexedSet(subSet, "entityId");

                    // We've handled this case, so remove this filter. We're done.
                    filters.splice(j, 1);
                    break;
                }
            }

            // Collect sets of candidates per filter. At the end we will calculate the intersection.
            for (let i = 0; i < filters.length; i++) {
                /*  The default context for a name is "CONTEXTLESS", which makes it refer to the default
                    namespace for tags in Dynatrace. If we have never heard of the specified context, then
                    there are no matches at all and we're done immediately.
                    The exception is "VALUE": that's our internal 'imaginary' context for criteria
                    that require that we iterate over all entities and compare a field against a list of
                    values. The primary use case is filtering by 'entityId', but that case is actually
                    already handled above for positive comparisons. We leave this here because it may
                    come in handy for other cases.
                 */
                let context = filters[i].context || "CONTEXTLESS";
                if (!self.#indexes[tenant][type][context] && context !== "VALUE")
                    return resolve([]);

                if (context === "ZONE")
                    candidates.push(getZoneMembers(
                        allEntities, // Used for negative filters.
                        self.#indexes[tenant][type][context], // Points to all zones.
                        filters[i].operator, // Can be "=" or "!=".
                        filters[i].values // Zone names (OR).
                    ));
                else if (context === "VALUE")
                    candidates.push(getValueMatches(
                        allEntities, // The list to walk through.
                        filters[i].key, // The field to look in.
                        filters[i].operator, // Can be "=" or "!=".
                        filters[i].values // Values it may (not) have.
                    ));
                else
                    candidates.push(getFilterMatches(
                        allEntities, // Used for negative filters.
                        self.#indexes[tenant][type][context][filters[i].key],
                        filters[i].operator, // ^ Points to specific tag.
                        filters[i].values // Values it may (not) have.
                    ));
            }

            let matches = candidates.length > 0                 // Did we have any filters at all?
                        ? intersect(candidates)                 // Then intersect the result sets.
                        : Array.from(entities[tenant][type]);   // Else everybody is a possible match.

            matches = getNameMatches(matches, "displayName", name);
            matches = getSinceMatches(matches, since);

            /*  At this point we have all matches (i.e. matching proxies). We now need to replace
                those proxies with actual entities. We'll acquire them from the cache (which will
                consult the database for those not yet cached). After that's done, we return the
                full entity objects.
                NOTE: We may not actually have to obtain the entities themselves - see below.
             */
            if (getProxies) {
                /*  Explicit request to NOT give us the entities, but just the proxies.
                 */
                resolve(matches);
            }
            else if (requestedFields &&
                     requestedFields.length > 0 &&
                     requestedFields.filter(field => !proxyFields.includes(field)).length === 0) {
                /*  No fields have been requested that are not already available in the proxies.
                    Thus, no need to go to the cache (and thus maybe the database). However, if
                    the 'entity' field has been requested specifically, we will need to get them
                    from the database after all (where missing).
                 */
                if (requestedFields.includes("entity")) {
                    self.dataAccess.cacheEntities(tenant, type, matches)
                        .then(() => resolve(matches))
                        .catch(error => reject(error));
                }
                else {
                    resolve(matches);
                }
            }
            else if (type === "PROCESS_GROUP_INSTANCES" && config.dataAccess === "sql") {
                /*  For entity types where we typically have way too many entities, we stream
                    straight from the database to the output, using an event emitter.
                    TODO: We may be able to make this the norm, at some point.
                 */
                self.dataAccess.streamEntities(tenant, type, matches)
                    .then(emitter => resolve(emitter))
                    .catch(error => reject(error));
            }
            else {
                /*  The is the normal scenario: get the entities from the cache and return them.
                 */
                self.dataAccess.cacheEntities(tenant, type, matches)
                    .then(() => resolve(matches.map(match => match.entity)))
                    .catch(error => reject(error));
            }
        });
    }

    relatedEntities(tenant, type, entityId, relationshipType, requestedFields) {
        const self = this;

        return new Promise((resolve, reject) => {
            if (!self.#indexes[tenant]) return reject("Unknown tenant " + tenant);

            if (!self.#indexes[tenant][type]) return reject("Unknown type " + type);

            const visited = {};

            const findNode = function (lists, id) {
                let node = null;
                for (let i=0; !node && i < lists.length; i++) {
                    node = lists[i].get(id);
                }
                return node;
            }

            function getRelated(entity, relationshipType, direction) {
                /*  It could be that some references are still unresolved. Then the traversal has reached
                    a dead end at this branch. We'll attempt to load this node later, and traverse again.
                 */
                if (entity.isMissing)
                    return { edges: [], nodes: [] };

                const opposite = direction === "IN" ? "OUT" : "IN";

                // Make sure we have the arrays we need.
                entity[relationshipType] = entity[relationshipType] || {};
                entity[relationshipType][direction] = entity[relationshipType][direction] || [];
                entity[relationshipType][opposite]  = entity[relationshipType][opposite]  || [];

                let nodes = entity[relationshipType][direction].map(node => {
                    if (typeof node !== "string")
                        return node;    // The related node is a proxy, which is the expected scenario.

                    /*  The related node is a string (i.e. an ID). So we need to retrieve it. To do that
                        efficiently, we look up the types of entities we can be related to in the given
                        direction. We then only need to consider those lists in locating the node.

                        Here are the scenarios we have to deal with at this point:
                        1) We have to replace the ID of this node with a pointer to it.
                        2) The referenced node does not seem to point back to this node.
                        2) We don't have that referenced node yet.

                        Whatever the reasons, we're fixing these things here as we walk through the links.
                     */
                    const id = node;
                    const relatedTypes = self.relationships[entity.type][relationshipType][direction];

                    node = relatedTypes
                         ? findNode(relatedTypes.accept.map(type => entities[tenant][type]), id)
                         : null;

                    if (node) {
                        // Scenario (1) and (2): The referenced entity exists but may or may not know of us.
                        self.addRelationship(node, entity, relationshipType, opposite);
                    }
                    else {
                        // Scanario (3): Create a placeholder for later retrieval.
                        node = {
                            entityId:  id,
                            tenant:    tenant,
                            type:      id.split("-")[0],  // We don't know the type, but this works too.
                            isMissing: true               // Means: this should be retrieved later.
                        };
                    }

                    return node;
                });

                let edges = nodes.map(node => direction === "IN"
                    ? { source: node,   target: entity }
                    : { source: entity, target: node   }
                );

                /*  Track and filter out what we visited already, for efficiency and also to
                    protect against (theoretically possible) circular references.
                 */
                nodes = nodes.filter(node => {
                    if (visited[node.entityId]) return false;
                    return visited[node.entityId] = true;
                });

                /*  Also add whatever is 'behind' each node we found. The 'nodes' array will grow
                    as we walk through it. Thus, we capture the initial length before we start.
                 */
                for (let i=0, l=nodes.length; i < l; i++) {
                    const related = getRelated(nodes[i], relationshipType, direction);
                    edges = edges.concat(related.edges);
                    nodes = nodes.concat(related.nodes);
                }

                return { edges: edges, nodes: nodes };
            }

            const entity = findNode((type ? [type] : types).map(type => self.curator.getEntities(tenant, type)), entityId);
            if (!entity)
                return reject("A " + type + " with ID " + entityId + " was not found in " + tenant);

            let nodes = [ entity ];   // We start from here.
            let edges = [];

            visited[entity.entityId] = true;

            let incoming = getRelated(entity, relationshipType, "IN");
            edges = edges.concat(incoming.edges);
            nodes = nodes.concat(incoming.nodes);

            let outgoing = getRelated(entity, relationshipType, "OUT");
            edges = edges.concat(outgoing.edges);
            nodes = nodes.concat(outgoing.nodes);

            /*  In principle we now have the complete graph. However, it is possible that there are
                entities in it that have not yet been 'seen' (retrieved) by us. At those points the
                graph will have been truncated. We'll need to load the missing entities explicitly,
                via the relevant Dynatrace APIs, out-of-bounds. After that, we will re-run this query
                again, and again, until we the graph is truly complete and all entities in it have
                been loaded into our database. Because that moment is inevitable... right?!

                ACTUALLY: We have observed during testing that the Dynatrace API occasionally states
                that it does not know of an entity that IS IN FACT REFERENCED by another entity...
                That really, really should NEVER happen, but it does, and that presents a conundrum.
                We will remediate this unfortunate situation by marking a proxy as "unknown" when the
                Dynatrace API does not recognize its ID. When asked to load an unknown entity, the
                'cacheEntities()' method will instead return a dummy version of it. ¯\_(ツ)_/¯

                So, the missing nodes are the ones we couldn't find in our admin AND that have not
                been officially declared 'unknown' (because never mind trying to retrieve those -
                the Dynatrace API already told us they don't seem to exist).
            */
            let missingNodes = nodes.filter(node => node.isMissing && !node.isUnknown);
            if (missingNodes.length > 0) {
                // Create groups of distinct tenant/type nodes from the collection of missing nodes.
                const uniqueTenantsAndTypes = {};
                for (let i=0; i < missingNodes.length; i++) {
                    let node = missingNodes[i];
                    let tenantType  = node.tenant + "_" + node.type;

                    uniqueTenantsAndTypes[tenantType] = uniqueTenantsAndTypes[tenantType] || {
                        tenant: node.tenant,
                        type:   node.type,
                        nodes:  []
                    };

                    uniqueTenantsAndTypes[tenantType].nodes.push(node);
                }

                // Collect the promises to get the missing nodes from the DynatraceAPI.
                const promises = Object
                    .values(uniqueTenantsAndTypes)
                    .map(tenantType => self.dataAccess.getEntities(
                        tenantType.tenant, tenantType.type,
                        tenantType.nodes.map(node => node.entityId),
                        true   // Udate database and internal admin.
                    ));

                // Wait for all the promises to resolve.
                Promise
                .all(promises)
                .then(() => {
                    /*  At this point, all missing entities have been loaded (hopefully), and we can
                        re-run the query. Note that this works recursively, because the relationship
                        chain may now be extended. But that will stop at some point...
                    */
                    self.relatedEntities(tenant, type, entityId, relationshipType, requestedFields)
                        .then(resolve)
                        .catch(reject);
                })
                .catch(error => {
                    // Some entities were not found. If that's so, throw in the towel. Tell the user.
                    reject("Could not retrieve all related entities: " + error);
                });

                /*  We've addressed the 'unknown nodes' scenario. We don't need to do anything else
                    because the recursive call will do it (ie. the code below) - and then we resolve.
                 */
                return;
            }

            /*  At this point we have the full graph. We now need to replace the proxies with actual
                entities. We'll acquire them from the cache (which will query the database for those
                not yet cached). After that's done, we return the full entity objects.

                NOTE: We don't have to obtain the entities if all requested fields are in the proxies.
             */
            if (requestedFields &&
                requestedFields.length > 0 &&
                requestedFields.filter(field => !proxyFields.includes(field)).length === 0)
            {   // No fields have been requested that are not already available in the proxies.
                let nodeset = nodes.map(node => {
                    // Add the graph ID that the edges use, plus a label.
                    node.id = node.tenant + "_" + node.entityId;
                    node.label = node.displayName;
                    return node;
                });
                let edgeset = edges.map(edge => {
                    // Replace the source and target entity references with graph IDs.
                    return {
                        id:     edge.source.tenant + "_" + edge.source.entityId + "_"
                              + edge.target.tenant + "_" + edge.target.entityId,
                        source: edge.source.tenant + "_" + edge.source.entityId,
                        target: edge.target.tenant + "_" + edge.target.entityId
                    }
                });
                resolve({
                    graph: {
                        nodes: nodeset,
                        edges: edgeset
                    }
                });
            }
            else {
                self.dataAccess.cacheEntities(null, null, nodes)
                    .then(() => {
                        let nodeset = nodes.map(node => {
                            node.entity = node.entity || self.entityAccess.createEntity(node);

                            // Add the graph ID that the edges use, plus a label.
                            node.entity.id    = node.tenant + "_" + node.entityId;
                            node.entity.label = node.displayName;
                            return node.entity;
                        });
                        let edgeset = edges.map(edge => {
                            // Replace the source and target entity references with graph IDs.
                            return {
                                id:     edge.source.tenant + "_" + edge.source.entityId + "_"
                                      + edge.target.tenant + "_" + edge.target.entityId,
                                source: edge.source.tenant + "_" + edge.source.entityId,
                                target: edge.target.tenant + "_" + edge.target.entityId
                            }
                        });
                        resolve({
                            graph: {
                                nodes: nodeset,
                                edges: edgeset
                            }
                        });
                    })
                    .catch(error => {
                        reject(error);
                    });
            }
        });
    }

    filterEntities(tenant, type, filters, name, requestedFields, since, follow) {
        const self = this;

        function constructGraph(matchingEntities) {
            return new Promise((resolve, reject) => {
                /*  The matches are just a list so far. If we have to follow relationships, we're
                    going to use each match as individual starting point to get all sub-graphs. The
                    function 'relatedEntities()' does that. After that we'll unify them into one.

                    Each call to 'relatedEntities()' may find missing nodes. The more starting points
                    are used, the higher the chances that each finds (some of) the same missing nodes.

                    Because 'relatedEntities()' will retrieve any missing nodes in the graph itself,
                    and because that happens ASYNCHRONOUSLY (using the DynatraceAPI), parallel calls
                    to 'relatedEntities()' will lead to parallel (and thus superfluous) attempts to
                    retrieve some of the very same entities. We must therefore make sure that each call
                    to 'relatedEntities()' happens AFTER the previous execution of 'relatedEntities()'
                    has finished retrieving whatever it found missing.

                    Promise.all() would attempt to resolve promises in parallel, and that won't do here.
                    Instead, we use "matches.reduce()" to create a chained series of '.then()' calls,
                    effectively serializing the attempts to resolve the promises.
                 */
                const allGraphs = [];

                matchingEntities                 // Each entity becomes a promise to get
                    .reduce(async (prevPromise, entity) => {         // related entities.
                        try {
                            const graph = await prevPromise;
                            if (graph) allGraphs.push(graph);
                        }
                        catch (ex) {}           // Ignore failure. Let later calls retry.
                        finally {               // We ask for 'entityId', to get proxies.
                            return self.relatedEntities(tenant, type, entity.entityId, follow, ["entityId"]);
                        }
                    }, Promise.resolve())       // Use an empty promise as starting point.
                    .then(lastGraph => {        // This is when the last promise resolved.
                        if (lastGraph) allGraphs.push(lastGraph);

                        const unifiedGraph = {  // Now we de-duplicate the graphs into one.
                            nodes: allGraphs.map(g => g.graph.nodes).flatUnique("entityId"),
                            edges: allGraphs.map(g => g.graph.edges).flatUnique("id")
                        };

                        self.dataAccess.cacheEntities(tenant, type, unifiedGraph.nodes)
                            .then(() => {
                                /*  Now everything we need is in the cache. Map the proxies to
                                    the actual entities. Note that the graph-IDs may already be
                                    set for things that were cached earlier. But for new ones,
                                    we still need to set it. So, we'll do that while mapping.
                                 */
                                unifiedGraph.nodes = unifiedGraph.nodes.map(node => {
                                    node.entity.id = node.id; // The edges use these IDs.
                                    return node.entity;
                                });

                                // Aaand... now we're totally done. Return the resulting graph.
                                resolve({ graph: unifiedGraph });
                            })
                            .catch(error => reject(error));
                    })
                    .catch(error => reject(error));
            });
        }

        return new Promise(function (resolve, reject) {
            if (follow === "flows" || follow === "stack" || follow === "group") {
                self.collectEntities(tenant, type, filters, name, requestedFields, since, true)     // Get proxies only.
                    .then(matches => {
                        constructGraph(matches)                                  // This will get actual entities.
                            .then(graph => resolve(graph))
                            .catch(error => reject(error));
                    })
                    .catch(error => reject(error));
            }
            else {
                self.collectEntities(tenant, type, filters, name, requestedFields, since, false) // Get actual entities.
                    .then(matches => resolve(matches))
                    .catch(error => reject(error));
            }
        });
    }

}

module.exports = Indexer;