'use strict';

const Ring = require('./Ring.js');

class Janitor {
    #cleaners   = [];
    #cleanTimer = null;
    #purgers    = [];
    #purgeTimer = null;
    #savers     = [];
    #saveTimer  = null;
    #purgeInfo  = null;
    #lastPurge  = null;

    get curator() {
        return this.globals.curator;
    }
    get indexer() {
        return this.globals.indexer;
    }
    get dataAccess() {
        return this.globals.dataAccess;
    }    

    constructor(globals, curator) {
        const self = this;

        this.globals   = globals;
        this.config    = globals.config;
        this.tenants   = globals.tenants;
        this.types     = globals.types;
        this.scheduler = globals.scheduler;

        const savers   = [];
        const cleaners = [];
        const purgers  = [];

        this.tenants.forEach(tenant => {
            savers.push(function(tenant) { 
                return () => self.indexer.save(tenant) 
            }(tenant));

            this.types.forEach(type => {
                cleaners.push(function(tenant, type) {
                    return cutoffDate => {
                        const proxies = self.curator.getEntities(tenant, type);

                        let proxy = null;
                        for (let i=0; i < proxies.length; i++) {
                            proxy = proxies[i];
                            if (proxy.lastUsed < cutoffDate) {
                                delete proxy.entity;
                            }
                        }
                    }
                }(tenant, type));

                purgers.push(function(tenant, type) {
                    return cutoffDate => {
                        self.#lastPurge = new Date();
                        self.dataAccess.retireEntities(tenant, type, cutoffDate);
                    }
                }(tenant, type));
            });
        });

        this.#cleaners = new Ring(cleaners);
        this.#purgers  = new Ring(purgers);
        this.#savers   = new Ring(savers);
    }

    runNextCleaner(cacheCycle) {
        let cleaner = this.#cleaners.next();
        if (cleaner)
            cleaner((new Date()).getTime() - cacheCycle);
    }

    runNextPurger(purgeCycle) {
        let purger = this.#purgers.next();
        if (purger)
            purger((new Date()).getTime() - purgeCycle);
    }

    runAllPurgers(purgeCycle) {
        const cutOffTime = (new Date()).getTime() - purgeCycle;
        for (let i=0; i < this.#purgers.length; i++)
            this.#purgers.next()(cutOffTime);
    }

    runNextSaver(saveCycle) {
        let saver = this.#savers.next();
        if (saver)
            saver();
    }

    getPurgerHealth() {
        return this.#purgeInfo + ", last run=" + (this.#lastPurge ? this.#lastPurge.toLocaleTimeString() : "(N/A)");
    }

    start(refreshRate) {
        refreshRate = refreshRate || this.config.library.refreshRate;

        if (this.#purgers.length === 0) return;

        this.stop();

        // Make sure we're not purging more frequently than 3 times the refreshRate.
        let timeToPurge = this.config.library.timeToPurge;
        let purgeSpan   = this.scheduler.getGranularityTime(timeToPurge);
        let refreshSpan = this.scheduler.getGranularityTime(refreshRate);
        while (refreshSpan * 3 > purgeSpan) {
            timeToPurge = this.scheduler.getNextGranularity(timeToPurge);
            purgeSpan   = this.scheduler.getGranularityTime(timeToPurge);
        }

        let purgeTime    = Math.round(purgeSpan / this.#purgers.length);
        this.#purgeTimer = setInterval(this.runNextPurger.bind(this), purgeTime, purgeSpan);

        // Do it now for all (well, in 1 second)...
        setTimeout(this.runAllPurgers.bind(this), 1000, purgeSpan);

        this.#purgeInfo = "Entity purging: max age=" + Math.round(purgeSpan/60000) + " mins, purgers=" + this.#purgers.length +", interval=" + Math.round(purgeTime/60000) + " mins";

        let cacheSpan    = this.scheduler.getGranularityTime(this.config.library.cachePeriod);
        let cacheTime    = Math.round(cacheSpan / this.#cleaners.length);
        this.#cleanTimer = setInterval(this.runNextCleaner.bind(this), cacheTime, cacheSpan);

        // Save the indexes and entities periodically for quick startup later.
        let saveSpan     = this.scheduler.getGranularityTime(this.config.library.savePeriod);
        let saveTime     = Math.round(saveSpan / this.#savers.length);
        this.#saveTimer  = setInterval(this.runNextSaver.bind(this), saveTime, saveSpan);
    }
    
    stop() {
        if (this.#cleanTimer) clearInterval(this.#cleanTimer);
        this.#cleanTimer = null;

        if (this.#purgeTimer) clearInterval(this.#purgeTimer);
        this.#purgeTimer = null;

        if (this.#saveTimer) clearInterval(this.#saveTimer);
        this.#saveTimer = null;
    }

    clear() {
        this.stop();
        this.#cleaners = [];
        this.#purgers  = [];
        this.#savers   = [];
    }

    get isRunning() {
        return !!this.#cleanTimer;
    }
};

module.exports = Janitor;