'use strict';

require("./Extensions.js");

const EntityAccess    = require('./EntityAccess.js');
const DynatraceAPI    = require('./DynatraceAPI.js');
const DataAccess      = require('./DataAccess.js');
const DatabaseManager = require('./DatabaseManager.js');
const Curator         = require('./Curator.js');
const Indexer         = require('./Indexer.js');

const proxyFields = ["entityId", "displayName", "checksum", "lastSeen", "lastUpdated", "entity"];

const apiEndPoints = {
    APPLICATION:   'entity/applications',
    SERVICE:       'entity/services',
    PROCESS_GROUP: 'entity/infrastructure/process-groups',
    PROCESS_GROUP_INSTANCE: 'entity/infrastructure/processes',
    HOST:          'entity/infrastructure/hosts',
    CUSTOM_DEVICE: null,
    CUSTOM_APPLICATION: 'entity/applications',
    MOBILE_APPLICATION: 'entity/applications',
    APPLICATION_METHOD: null,
    DYNAMO_DB_TABLE: null,
    ELASTIC_LOAD_BALANCER: null,
    AWS_LAMBDA_FUNCTION: null,
    RELATIONAL_DATABASE_SERVICE: null,
    DCRUM_APPLICATION: null,
    DCRUM_SERVICE: null,
    DOCKER_CONTAINER_GROUP_INSTANCE: null,
    SERVICE_METHOD: null,
    SYNTHETIC_TEST: null,
    SYNTHETIC_TEST_STEP: null,
    EBS_VOLUME: null,
    EC2_INSTANCE: null,
    OPENSTACK_PROJECT: null
};

const relationships = {
    // NOTE:
    // The 'IN' props are presumed to be under "toRelationships".
    // The 'OUT' props are presumed to be under "fromRelationships".
    CUSTOM_APPLICATION: {
        apiEndpoint: 'entity/applications',
        retrieval:   [ "all" ],
        flows: { IN:  null,
                 OUT: { prop: "fromRelationships.calls", accept: ["SERVICE"] }
        },
        stack: { IN:  null,
                 OUT: null
        },
        group: { IN:  null,
                 OUT: null
        }
    },
    MOBILE_APPLICATION: {
        apiEndpoint: 'entity/applications',
        retrieval:   [ "all" ],
        flows: { IN:  null,
                 OUT: { prop: "fromRelationships.calls", accept: ["SERVICE"] }
        },
        stack: { IN:  null,
                 OUT: null
        },
        group: { IN:  null,
                 OUT: null
        }
    },
    APPLICATION: {
        apiEndpoint: 'entity/applications',
        retrieval:   [ "all" ],
        flows: { IN:  null,
                 OUT: { prop: "fromRelationships.calls", accept: ["SERVICE"] }
        },
        stack: { IN:  null,
                 OUT: null
        },
        group: { IN:  null,
                 OUT: null
        }
    },
    SERVICE: {
        apiEndpoint: 'entity/services',
        retrieval:   [ "all", "chunked", "discover" ],
        flows: { IN:  { prop: "toRelationships.calls", accept: ["SERVICE", "APPLICATION", "MOBILE_APPLICATION", "CUSTOM_APPLICATION"] },
                 OUT: { prop: "fromRelationships.calls", accept: ["SERVICE"] }
        },
        stack: { IN:  null,
                 OUT: { prop: "fromRelationships.runsOn", accept: ["PROCESS_GROUP"] }
        },
        // NOTE: This here is so that we have a second path to connect SERVICES to HOSTS.
        infra: { IN:  null,
                 OUT: { prop: "fromRelationships.runsOnProcessGroupInstance", accept: ["PROCESS_GROUP_INSTANCE"] }
        },
        group: { IN:  null,
                 OUT: null
        }
    },
    PROCESS_GROUP: {
        apiEndpoint: 'entity/infrastructure/process-groups',
        retrieval:   [ "all", "chunked", "entity", "discover" ],
        flows: { IN:  { prop: "toRelationships.isNetworkClientOfProcessGroup", accept: ["PROCESS_GROUP"] },
                 OUT: { prop: "fromRelationships.isNetworkClientOfProcessGroup", accept: ["PROCESS_GROUP"] }
        },
        stack: { IN:  { prop: "toRelationships.runsOn", accept: ["SERVICE"] },
                 OUT: { prop: "toRelationships.isInstanceOf", accept: ["PROCESS_GROUP_INSTANCE"] }
        },
        // NOTE: This here is so that we have a second path to connect PROCESS_GROUPS to HOSTS.
        infra: { IN:  { prop: "fromRelationships.runsOn", accept: ["HOST"] },
                 OUT: null
        },
        group: { IN:  null,
                 OUT: null
        }
    },
    PROCESS_GROUP_INSTANCE: {
        apiEndpoint: 'entity/infrastructure/processes',
        retrieval:   [ "all", "chunked", "entity", "discover" ],
        flows: { IN:  { prop: "toRelationships.isNetworkClientOf",   accept: ["PROCESS_GROUP_INSTANCE"] },
                 OUT: { prop: "fromRelationships.isNetworkClientOf", accept: ["PROCESS_GROUP_INSTANCE"] }
        },
        stack: { IN:  { prop: "toRelationships.isInstanceOf",   accept: ["PROCESS_GROUP"] },
                 OUT: { prop: "fromRelationships.isProcessOf",  accept: ["HOST"], }
        },
        // NOTE: This here is so that we have a second path to connect SERVICES to HOSTS.
        infra: { IN:  { prop: "toRelationships.runsOnProcessGroupInstance", accept: ["SERVICE"] },
                 OUT: { prop: "fromRelationships.isProcessOf",              accept: ["HOST"] }
        },
        group: { IN:  { prop: "fromRelationships.isInstanceOf", accept: ["PROCESS_GROUP"] },
                 OUT: null
        }
    },
    HOST: {
        apiEndpoint: 'entity/infrastructure/hosts',
        retrieval:   [ "all", "chunked", "discover" ],
        flows: { IN:  { prop: "toRelationships.isNetworkClientOfHost",   accept: ["HOST"] },
                 OUT: { prop: "fromRelationships.isNetworkClientOfHOst", accept: ["HOST"] }
        },
        stack: { IN:  { prop: "toRelationships.isProcessOf", accept: ["PROCESS_GROUP_INSTANCE"] },
                 OUT: { prop: "fromRelationships.runsOn",    accept: ["HOST"] }
        },
        // NOTE: This here is so that we have a second path to connect PROCESS_GROUPS to HOSTS.
        infra: { IN:  { prop: "toRelationships.runsOn", accept: ["PROCESS_GROUP"] },
                 OUT: null
        },
        // NOTE: We don't actually have 'hostGroup' in the 'fromRelationships' property.
        // We list it here to show what we capture the relation - but not per the below.
        group: { IN:  null,
                 OUT: { prop: "hostGroup", accept: ["HOST_GROUP"] }
        }
        // MISSING, STILL: what do we do with "isSiteOf"?
    }
};

const friendly = function (name, length = 1) {
    if (!name) return "entitie";
    return name.toLowerCase().replace(/_/g, " ").padEnd(length, ' ');
}

const Library = function (config, scheduler, socket, logger, isFailover) {
    logger.info("    Instantiating library" + (config.library.name ? " '" + config.library.name + "'" : ""));

    const self = this;
    self.events = {};

    const emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const event = self.events[eventName] || [];
        event.forEach(fn => { fn.call(null, data); });
    };

    const subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);
    };
    
    // If we run in failover mode, we start out in standby mode.
    let isOnStandby = isFailover;
    let isRunning = false;

    const globals = {
        socket        : socket,
        friendly      : friendly,
        tenants       : null,
        types         : Object.keys(apiEndPoints).filter(type => !!apiEndPoints[type]),
        proxyFields   : proxyFields,
        relationships : relationships,
        apiEndPoints  : apiEndPoints,
        databasePool  : null,
        dynatraceAPI  : null,
        libraryAPI    : null,
        dataAccess    : null,
        entityAccess  : null,
        logger        : logger,
        scheduler     : scheduler,
        indexer       : null,
        curator       : null
    }

    async function initialize() {
        // This here will ensure that we don't touch the original config object.
        // If we do, we confuse the Controller, which checks every minute if the
        // config has changed on disk relative to in memory. So, don't touch it.
        globals.config = JSON.parse(JSON.stringify(config));

        globals.name = globals.config.library.name;
        globals.config.library.idListLimit  = globals.config.library.idListLimit  || 15;  // Serves as a max.
        globals.config.library.refreshRate  = globals.config.library.refreshRate  || "_hour";
        globals.config.library.timeToPurge  = globals.config.library.timeToPurge  || "_6hours";
        globals.config.library.cachePeriod  = globals.config.library.cachePeriod  || "_15mins";
        globals.config.library.savePeriod   = globals.config.library.savePeriod   || "_6hours";
        globals.config.library.healthCheck  = globals.config.library.healthCheck  || 60;
        globals.config.database = Object.assign({
            connectionLimit: 100,
            connectTimeout:  10000,
            pipelining:      true
        }, globals.config.database);

        // Extract the known tenants from the configuration and create admins for each.
        // Note that if 'tenants' is an array, only those tenants will be handled.
        // If 'tenants' is null or undefined, all known tenants will be handled.
        globals.tenants = Object.keys(globals.config.Tenants).filter(tenant =>
            globals.config.library.tenants ? globals.config.library.tenants.includes(tenant) : true
        );
        logger.info("    Library '" + globals.name + "' handles " + globals.tenants.join(", "));

        try {
            const cacheCycle = globals.scheduler.getGranularityTime(globals.config.library.cachePeriod);

            // Create a database pool and create the database if necessary.
            const adaptorClass = globals.config.database.class || "SQLiteAdaptor";
            logger.info("Preparing database connection pool");

            try {
                const DatabaseAdaptor = require("./databases/" + adaptorClass + ".js");

                globals.databasePool = new DatabaseManager(DatabaseAdaptor, globals.config.database, logger);

                // Create DataAccess object to perform DB queries.
                switch (globals.config.database.class) {
                    // TODO: NoSQL implementation 
                    case "SQLiteAdaptor":
                        globals.dataAccess = new DataAccess.SQLite(globals, cacheCycle);
                        break;
                    case "MySQLAdaptor":
                    case "MariaDBAdaptor":
                    default:
                        globals.dataAccess = new DataAccess.InnoDB(globals, cacheCycle);
                        break;
                }

                logger.info("    Connecting to the database");

                await globals.dataAccess.createTable();

                logger.info("    Data access layer started successfully");
            }
            catch (ex) {
                emit("fatal", "Database initialization failed: " + (ex.message || ex));
                return;
            }

            // This thing here encapsulates CRUD actions to entities.
            globals.entityAccess = new EntityAccess(globals);

            if (!globals.config.library.offline) {
                logger.info("Preparing Dynatrace API connection pool");

                // Initialize a DynatraceAPI object that knows how to connect to each tenant.
                // Also initialize a dedicated version of it to sync with a primary instance.
                globals.dynatraceAPI = new DynatraceAPI(globals.config.Tenants, globals.config);
                globals.dynatraceAPI.getHealth(5000, (healthReport) => socket.emitStats("DynatraceAPI", healthReport));

                globals.libraryAPI = isFailover
                    ? new DynatraceAPI({
                        Primary: {     // Always called "Primary". There is only one for us.
                            host: globals.config.library.primaryIP,  // Because we may be behind a DNS and therefore we need to know the IP to specifically connect to that host.
                            port: globals.config.library.port,
                            token: "notused",
                            noSSL: true
                        }
                      }, {
                        reqRateLimit: 10,   // Max requests made per minute.
                        requestLimit: 10,   // Max no. of parallel requests.
                        maxQueueTime: 20,   // Max minutes waiting in queue.
                        retryLimit:   3,    // Max retries on HTTP failure.
                        maxQueueSize: 10    // Max requests in queue total.
                      })
                    : null;

                logger.info("    Dynatrace API pool created successfully");
            }
            else {
                logger.info("The Library will run in offline mode");
            }

            logger.info("Creating the curator");
            globals.curator = new Curator(globals);

            logger.info("Creating the indexer");
            globals.indexer = new Indexer(globals);

            let failedTenants = globals.indexer.load();

            if (failedTenants && failedTenants.length !== 0) {
                logger.warn("    Indexes for the following tenants could not be loaded: " + failedTenants.join(", "));
                logger.warn("    Library will load entities for those tenants from the database and rebuild their indexes");

                try {
                    await globals.dataAccess.loadEntities(failedTenants);

                    logger.info("    All missing tenants' entities have been loaded from the database and re-indexed");
                }
                catch (ex) {
                    logger.error("   Loading and indexing of tenants " + failedTenants.join(", ") + " from the database failed: " + ex.message);
                    // But we can stay alive.
                }
            }

            try {
                await globals.curator.loadManagementZones();
            }
            catch (ex) {
                // NOP - just go on. We'll recreate them from the data.
            }         

            // Perform validation.
            if (globals.config.library.validateOnStartup) {
                try {
                    let failedTenants = await globals.indexer.validate();

                    if (failedTenants && failedTenants.length !== 0) {
                        logger.warn("The following tenants had problems: " + failedTenants.join(", "));
                        logger.warn("Library will reload entities for those tenants from the database and rebuild their indexes");

                        try {
                            await globals.dataAccess.loadEntities(failedTenants);

                            logger.info("    All problematic tenants have been reindexed from the database");
                        }
                        catch (ex) {
                            logger.error("   Reload and reindex of all tenants from the database failed: " + ex.message);
                        }

                        try {
                            failedTenants = await globals.indexer.validate();

                            if (failedTenants && failedTenants.length !== 0) {
                                logger.warn("    The following tenants still have problems after reindexing: " + failedTenants.join(" "));
                            }
                            else {
                                logger.info("    All problems were corrected");
                            }
                        }
                        catch (ex) {
                            logger.error("    Re-validation failed: " + err);
                        }
                    }
                }
                catch (ex) {
                    logger.error("Validation failed: " + ex.message);
                }
            }

            // Only start the curator if we are not in offline mode.
            if (!globals.config.library.offline)
                globals.curator.start(isOnStandby);

            // At this point we're properly ready.
            isRunning = true;
            emit("ready");
        }
        catch (ex) {
            emit("fatal", ex.message || ex);
        }
    }

/*******
 *  Querying our administration.
 *  - filterIndexes:     Return all the things we know about: management zones, properties, different types of tags.
 *  - filterZones:       Return all zones we know about, and what is in each.
 *  - FilterProperties:  Return all properties we know about, and what values they have.
 *  - filterTags:        Return all tags we know about (for a given context), and what values they have.
 *  - filterEntities:    Returns all entities (of a given type) that match specified criteria.
 *  - relatedEntities:   Returns all entities related to a given entity along a specified dimension.
 */
    function filterIndexes(tenant, name) {
        return globals.indexer.filterIndexes(tenant, name);
    }

    function filterZones(tenant, filters, name) {
        return globals.indexer.filterZones(tenant, filters, name);
    }

    function filterProperties(tenant, name) {
        // TODO: This is not yet implemented.
        let matches = {};

        return new Promise(function(resolve, reject) {
            resolve(Object.values(matches));
        })
    }

    function filterTags(tenant, context, name) {
        return globals.indexer.filterTags(tenant, context, name);
    }

    function filterEntities(tenant, type, filters, name, requestedFields, since, follow) {
        return globals.indexer.filterEntities(tenant, type, filters, name, requestedFields, since, follow);
    }

/*******
 *  Interactions with the Dynatrace API directly.
 */
    function loadMetricDefinitions(tenant) {
        return new Promise((resolve, reject) => {
            /*  Retrieve all known management zones in this tenant. Our chunking
                options include chunking by a list of MZs, but we need to know about
                their names and IDs first. This is a one-tinme thing.
            */
            globals.dynatraceAPI.request({
                    tenant: tenant,
                    endPoint: "/api/v1/timeseries",
                    queryData: ''
                },
                function (err, metrics) {
                    if (err)
                        reject(err.message);
                    else
                        resolve(metrics);
                }
            );
        });
    }

    function loadUsqlRules(tenant) {
        return new Promise((resolve, reject) => {
           const spec = require("../client/spec2.json");
           resolve(spec);
        });
    }

/*******
 *  On instantiation, load everything we've got from files or the database.
 *  The 'ready' event will be emitted once this is complete.
 */
    initialize(config);

    // Return our interface.
    return {
        on:     subscribe,
        start:  () => {
            if (globals.curator)
                globals.curator.start();
        },
        stop:   (onStopped) => {
            if (globals.curator)
                globals.curator.stop();                 // Runs synchronously.
                
            globals.databasePool.close(() => {          // Runs asynchronously.
                if (globals.indexer)
                    globals.indexer.save(null, onStopped);  // Runs asynchronously.
                else 
                    onStopped();
            });
        },
        get isReady () {
            return isRunning;
        },
        get inStandby () {
            return isOnStandby;
        },
        set inStandby (value) {
            // Make sure value is a boolean.
            if (isOnStandby === !value) {
                isOnStandby = !!value;
                // Recreate the pipelines - it uses the current 'standby' mode.
                globals.curator.start(isOnStandby);
            }
        },
        set config (value) {
            if (isRunning && !isOnStandby) {
                logger.info("Please restart the Library for configuration changes to take effect");
            }
            else {
                config = value;
                initialize();
            }
        },
        get config () {
            return config;
        },
        filter: (tenant, type, filters, name, requestedFields, since, follow) => {
            if (globals.types.includes(type) || type === "ALL") {
                return filterEntities(tenant, type, filters, name, requestedFields, since, follow);
            }
            else if (type === "ZONE") {
                return filterZones(tenant, filters, name);
            }
            else if (type === "PROPERTY") {
                // return filterProperties(tenant, name);
                return filterTags(tenant, type, name);
            }
            else if (type === "CONTEXT") {
                return filterIndexes(tenant, name);
            }
            else { // It's some kind of tag context.
                return filterTags(tenant, type, name);
            }
        },
        getDeletions: (tenant, type, since) => {
            return globals.curator.getDeletedEntities(tenant, type, since);
        },
        flush: (tenantList) => {
            tenantList = tenantList || globals.tenants;

            // Remove entities from cache.
            return new Promise(function (resolve, reject) {
                tenantList.forEach(tenant => {
                    globals.types.forEach(type => {
                        const proxies = globals.curator.getEntities(tenant, type);

                        let proxy = null;
                        for (let i=0; i < proxies.length; i++) {
                            proxy = proxies[i];
                            delete proxy.entity;
                        }
                    });
                });
                resolve("OK");
            });
        },
        reset: function (tenantList) {
            tenantList = tenantList || globals.tenants;

            // Remove all entities altogether - also from database.
            // Forget about things we discovered but didn't load yet.
            return new Promise(function (resolve, reject) {
                tenantList.forEach(tenant => {
                    globals.curator.controlPipeline(tenant, null, "stop");

                    globals.types.forEach(type => {
                        globals.dataAccess.retireEntities(tenant, type);
                    });

                    globals.curator.controlPipeline(tenant, null, "start");
                });
                resolve("OK");
            });
        },
        validate: function (tenantList) {
            return globals.indexer.validate(tenantList || globals.tenants);
        },
        getHealthReport: function () {
            if (!isRunning) return null;

            return {
                identity:  globals.name,
                database:  globals.databasePool ? globals.databasePool.getHealth() : "Database not available",
                pipelines: globals.curator.getPipelineHealth(),
                dynatrace: globals.dynatraceAPI.getHealth()
            };
        },
        getHealthMetrics: function () {
            if (!isRunning) return [];

            const now = new Date();
            now.setSeconds(0, 0);

            const healthMetrics = { 
                timestamp: now,
                events:  [],
                metrics: []
            };

            const databaseHealth = globals.databasePool.getHealthMetrics();

            healthMetrics.metrics.push({
                metricId: "activeDbConnections", 
                dimensions: [ globals.name ],
                value: databaseHealth.active
            });
            healthMetrics.metrics.push({
                metricId: "queuedDbConnections", 
                dimensions: [ globals.name ],
                value: databaseHealth.queued
            });            

            healthMetrics.metrics = healthMetrics.metrics.concat(globals.curator.getHealthMetrics());

            return healthMetrics;
        },
        getPipelineStats: function (tenant, type) {
            return globals.curator.getPipelineStats(tenant, type);
        },
        getPipelines: async function () {
            return globals.curator.getPipelinesInfo();
        },
        controlPipeline: async function (tenant, type, action) {
            return globals.curator.controlPipeline(tenant, type ? type.toUpperCase() : null, action);
        },
        getMetricDefinitions: loadMetricDefinitions,
        getUsqlRules: loadUsqlRules
    }
};

module.exports = Library;
