const cleanup = require('node-cleanup');
const Diagnostics = require("./Diagnostics.js");
const ConfigReader = require('./ConfigReader.js');

// Fun Javascript extensions.
require("./Extensions.js");

// TODO: The below allows response.send(any object). 
// Also request.body is already parsed
//   webapi.use(express.json());  

const LibraryService = function (libraryName = "LibraryService", ownProcess = null, args = {}) {
    const express   = require('express');
    const webapi    = express();
    const cors      = require('cors');
    const http      = require('http');
    const server    = http.Server(webapi);
    const socket    = require('socket.io')(server);
    const winston   = require('winston');
    const Scheduler = require('./Scheduler.js');
    const Library   = require('./Library.js');

    const requests    = [];
    const contentType = 'application/json';

    const isFailover = args.mode === "failover";
    const isUnderETL = args.mode === "auto" && ownProcess && ownProcess.send;

    // This scheduler won't execute jobs. It just has some useful methods for us.
    let scheduler = new Scheduler();
    let listener  = null;
    let library   = null;

    // Setup infra for logging.
    const loggerTransports = {
        console: new winston.transports.Console({ level: 'silly' }),
        file:    new winston.transports.File({ filename: '../log/' + libraryName + '.log', maxsize: 1048576, maxFiles: 10 }),
        silent:  new winston.transports.Console({ level: 'error', silent: true })
    };

    const libraryLogger = winston.createLogger({
        format: winston.format.combine(
            winston.format.label({ label: 'Library' }),
            winston.format.timestamp({ format: 'YYYY-MM-DD HH:mm:ss.SSS' }),
            winston.format.printf(({ level, message, label, timestamp }) => {
                return `${timestamp} [${label}] ${level}: ${message}`;
            })
        ),
        transports: [ loggerTransports.silent ]  // We may add transports later, as the config specifies.
    });

    function initLogging(transports) {
        transports = transports || [];

        if (libraryLogger.transports.length === 1) {
            // It only has the 'silent' logger. Let's see what
            // we need to log to, per the config.json.
            transports.forEach(transport => {
                if (loggerTransports[transport])
                    libraryLogger.add(loggerTransports[transport]);
            });
        }
    }

    // Setup infra for loading the configuration.
    let config = null;

    const configReader = new ConfigReader(args.config);
    configReader.on("loaded", (section, config) => {
        // Find the entry in the "Libraries" section that is ours.
        config.library = libraryName ? config.Libraries[libraryName] : config.Libraries;
        if (!config.library)
            throw new Error("There is no library configuration" + (libraryName ? " named '" + libraryName + "'" : ""));

        loadConfig(config);
    });
    configReader.on("error", (msg, err) => {
        console.error("Configuration could not be loaded: " + msg + ": " + err);
        setTimeout(process.exit(125), 1000); // Trigger PM2 to restart us.
    });
    configReader.on("changed", (section, data) => {
        console.warn("Configuration change detected. This service will restart so that the change can take effect");
        setTimeout(process.exit(0), 1); // Trigger PM2 to restart us.
    });

    // Called by a ConfigReader event, or by a parent process 'configure' message.
    function loadConfig(newConfig) {
        if (library && JSON.stringify(library.config) === JSON.stringify(newConfig)) 
            return;
        
        config = newConfig;

        if (!library) {
            // Set up where we will be logging to.
            initLogging(config.logging);

            libraryLogger.info("LibraryService is starting...");
        }
        else {
            library.config = config;
        }

        if (!listener) {          
            const port = isFailover ? config.library.failover.port : config.library.port;

            listener = server.listen(port, () => {
                library = new Library(config, scheduler, socketWrapper, libraryLogger, isFailover);
                library.on('error', error => {
                    libraryLogger.error(error);
                    if (isUnderETL)
                        process.send({ status: "error", error: error });
                });
                library.on('ready', () => {
                    libraryLogger.info("LibraryService " + (libraryName ? "'" + libraryName + "' " : "") + "is ready and listening on port " + port);

                    if (isUnderETL)
                        process.send({ status: "ready" }); 
                    else if (process.send)
                        process.send("ready"); 
                });
                library.on('fatal', error => {
                    libraryLogger.error("The library cannot initialize: " + error);

                    if (isUnderETL)
                        process.send({ status: "error", error: "Library cannot initialize: " + error });
                    else
                        process.exit(126);
                });

                if (ownProcess) {
                    // Start the diagnostics to track memory.
                    diagnostics = new Diagnostics(config, library, libraryLogger);
                    diagnostics.start();
                }
            });

            listener.on("error", err => {
                const error = "LibraryService cannot listen on port " + config.library.port + ": " + err.message;

                libraryLogger.error(error);
                listener.close();

                if (isUnderETL)
                    process.send({ status: "error", error: error });
                else
                    process.exit(127);
            });
        }

        if (isFailover && !failoverTimer) {
            // Failover mode: periodically check the primary instance's health.
            failoverTimer = setInterval(checkHealth, 10000, config.library.failover);

            libraryLogger.info("This failover instance will synchronize its entities with the primary instance at " + config.failover.primary);
        }
    }

    // Make sure we have a process object we can work with.
    let asyncIF = null;   // Declare now, assign when we return.

    const process = ownProcess || {
        events: {},
        on: function (eventName, fn) {
            this.events[eventName] = this.events[eventName] || [];
            this.events[eventName].push(fn);
        },
        emit: function (eventName, data) {
            const eventHandlers = this.events[eventName] || [];
            eventHandlers.forEach(fn => { fn.call(null, data); });
        },
        send: function (message) {
            asyncIF.emit("message", message);
        },
        disconnect: function () {
            asyncIF.emit("disconnect");
        },
        exit: function () {
            stop();   // Will call 'kill'.
        },
        kill: function () {
            asyncIF.emit("exit");
        }
    };

    // Send things to the browser for live updates.
    const socketWrapper = {
        emit: function (message) {
            message = Object.assign({
                status: "progress",
                processID: libraryName
            }, message);

            // The ServiceProxy will send this to the browser when it sees 'status = "progress"'.
            if (isUnderETL)
                process.send(message);

            // In standalone mode, we'll have to do it ourselves.
            socket.emit("progress", message);
        },
        emitStats: function (emitter, stats) {
            message = {
                status: "stats",
                processID: libraryName,
                component: emitter,
                stats: stats
            };

            // The ServiceProxy will send this to the browser when it sees 'status = "html"'.
            if (isUnderETL)
                process.send(message);

            // In standalone mode, we'll have to do it ourselves.
            socket.emit("stats", message);
        }
    };

    /*  FAILOVER HANDLING
        If we run in failover mode, then we need to stay in 'offline' mode until we
        take over from the primary (and go back to offline mode when we return to
        standby mode again). This is to avoid duplicating the work the primary
        instance's library is doing.

        In order to hit the ground running when we take over, we will need to receive
        continuous updates from the primary instance's embedded library (which is a
        copy of ourselves) with respect to all the changes that are happening in it.

        The approach is that is that we ask for a list of known entities of a type,
        with checksums to efficiently check if entities (by entityId) that we know of
        have changed. We can then ask for fresh, full copies of the things that have
        changed or that appear new, and delete entities the primary library did not
        report. Then on an ongoing basis, we can ask for updates since some timestamp.

        Note that we remain responsible for maintaining our own indexes.

        So, the exchange goes like this:
         1) Ask for a list of all host IDs and their checksums that the primary instance
            knows of. Keep the timestamp of the request.
         2) Go through own admin, delete what wasn't confirmed, and ask for full objects
            for things that are new or have a different checksum than we have. We may use
            batching to retrieve the full entity defintions. We don't update the database.
         3) Periodically, ask for a list of all host IDs and their checksums that have
            changed or have been added or deleted since the previous request.
            Track the timestamp of the request for use in "since=..." in the next request.

        If we are the primary instance (which we also may temporarily become when we take
        over), we need to expose this interface and have a way to answer the requests:

        Request 1) Return all entities of a given type (only include selected fields -
                   in this scenario we'd be asked for IDs and checksums).
        Request 2) Return full entities for a list of entity IDs (already in place).
        Request 3) Return all entities of a given type that have been deleted, changed
                   or added since the provided timestamp ("since=...").
                   Note that request (3) is essentially a special case of request (1)
                   with a "since=..." query parameter provided, which will cause the
                   response to contains two resultsets: 'updates' and 'deletions'
                   since the specified timestamp. The 'deletions' resultset can only
                   contain information available in the proxies, as the actual entities
                   have been deleted.

        To handle request (3) we record a timestamp for every entity that indicates when
        it was added or updated. We also maintain a separate list of deleted entities (ID
        and timestamp of deletion). We should keep that list limited to things deleted
        during the last 24 hours (dependent on config).

        Checksums and timestamps are in the proxies and are also stored in the database.
    */
    const primaryDownMax = 5;
    let failoverTimer    = null;
    let primaryDownCount = 0;
    let configSentToFailover = null;

    // Called periodically to sync with the primary instance and 
    // to determine if we need to take over from the primary.
    function checkHealth(failoverConfig) {
        const address = "http://" + failoverConfig.primaryIP + ":" + failoverConfig.primaryPort;

        function handleResponse(error, answer)  {
            if (error) {
                // Let's count successive failures. We take over when beyond primaryDownMax...
                primaryDownCount++;

                // Only if we have not already have taken over, do our continuing failures need to be handled.
                if (library.inStandby) {
                    libraryLogger.warn("LibraryService failover instance could not connect to " + address +
                        " " + primaryDownCount + " times (" + (primaryDownMax - primaryDownCount) +
                        " attempts left)"
                    );
                    if (primaryDownCount >= primaryDownMax) {
                        libraryLogger.warn("LibraryService failover instance is taking over from  " + address);
                        library.inStandby = false;  // Makes the failover instance connect to Dynatrace directly.
                    }
                }
            }
            else {
                // The primary instance is working (again) - reset the failure counter.
                primaryDownCount = 0;

                // If we had taken over, then it's time to go back to standby mode.
                if (!library.inStandby) {
                    libraryLogger.warn("LibraryService failover instance is returning to standby mode");
                    library.inStandby = true;       // Reverts the failover instance to 'sync with primary' mode.
                }

                // Important note: The library has pipelines that check for additions, updates and deletes.
                // We don't have to do that here. What we do need to do is adopt the configuration of the
                // primary instance.
                if (answer && answer.config) {
                    library.config = answer.config;
                }
            }
        }

        try {
            // We ask for the primary instance's status to test for availability.
            const request = http.get(address + "/status?secret=" + failoverConfig.secret, response => {
                let data = [];
                response.setEncoding('utf8');

                response.on('data', (chunk) => {
                    data.push(chunk);
                });
                response.on('end', () => {
                    response.body = JSON.parse(data.join(""));
                    handleResponse(response.statusCode !== 200, response.body);
                });
            });
            request.setTimeout(9000, () => {
                request.destroy();
            });
            request.on("abort", () => {
                handleResponse(true);
            });
            request.on("error", () => {
                handleResponse(true);
            });
        }
        catch (ex) {
            handleResponse(true);
        }
    }

    // Call by the failover instance to sync and see if we're still up.
    function healthStatus(request, response) {
        let answer = {
            status: !library
                ? "preparing"
                : library.isReady
                ? "alive"
                : library.inStandby
                ? "standby"
                : "preparing"
        };

        try {
            const requestingHostIP = request.ip ? request.ip.split(":").pop() : null;
            if (config.library.failover &&
                requestingHostIP     === config.library.failover.failoverIP &&
                request.query.secret === config.library.failover.secret) {

                // If the request comes from the host that runs the failover instance,
                // and it contains the secret, include our config if we haven't sent it yet.
                const stringifiedConfig = JSON.stringify(config);

                if (answer.config === "alive" &&
                    configSentToFailover !== stringifiedConfig) {
                    configSentToFailover   = stringifiedConfig;
                    answer.config = config;
                }

                response.setHeader('Content-Type', "application/json");
                response.write(JSON.stringify(answer));
            }
            else {
                response.setHeader('Content-Type', "text/plain");
                response.write(answer.status);
            }

            response.statusCode = 200;
        }
        catch (ex) {
            response.statusCode = 500;
        }
        finally {
            response.end();
        }
    }

    // Called every minute by the StreamerService or by the Dynatrace Solution Server.
    function healthMetrics(request, response) {
        const healthMetrics = library.getHealthMetrics();

        if (!response) return healthMetrics;

        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(healthMetrics, undefined, 4));
        response.end();
    }

    // Called by the UI. I think. TODO To be checked.
    function healthReport(request, response) {
        try {
            const healthReport = library.getHealthReport();
            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(healthReport, undefined, 4));
            response.end();
        }
        catch (ex) {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({ error: ex.message }));
            response.end();
        }
    }

    // Functions that implement the REST services we expose.
    function executeLibraryRequest(request, response, filters, name, inMemory) {
        const stringify    = inMemory ? obj => obj : JSON.stringify;
        const responseType = inMemory ? "" : contentType;

        const libAPI = {
            filter:  library.filter,
            check:   library.validate,
            stats:   library.statistics,
            flush:   library.flush,
            reset:   library.reset,
            deletes: library.getDeletions
        };

        // What to select from result set and render into the response.
        // TODO: includeDeletes (isn't that a 'since' thing?)
        let since   = request.query.hasOwnProperty("since") && request.query.since !== ""
                    ? Number(request.query.since)
                    : null;
        let follow  = request.query.hasOwnProperty("follow") && request.query.follow !== "" && request.query.follow !== "none"
                    ? request.query.follow
                    : null;
        let select  = request.query.hasOwnProperty("select") && request.query.select !== ""
                    ? request.query.select.split(",")
                    : null;
        let pretty  = request.query.hasOwnProperty("pretty")
                    ? 4
                    : undefined;
        let extract = function (key, value) {
            if (key === "checksum") {
                if (!select || !select.includes("checksum"))
                    return undefined;
                else
                    return value;
            }

            // If the entity itself, as a property on the proxy, is to be extracted,
            // stringify it right straight up here and now so that we don't extract
            // the entity's properties. Check on 'checksum' to ensure this is the proxy.
            if (key === "entity" && this.hasOwnProperty("checksum")) {
                if (!select || !select.includes("entity"))
                    return undefined;
                else
                    return JSON.stringify(value);
            }

            // Never return the relationships. They are bulky, and the library has
            // already processed them and made them part of the query options.
            if (["fromRelationships", "toRelationships"].includes(key))
                return undefined;

            // We use "_empty" internally to index empty strings. But the external value is still "".
            if (value === "_empty") value = "";

            /*  If we have no 'select' value, include everything. If the 'select' value is "stringables",
                only include string values. Some things can be stringified - we'll implement that later.
                Otherwise, only stringify the specified keys on entities (wich have "entityId" and/or
                "displayName" properties).
            */
            if (select && select[0] === "stringables") {
                if (this.hasOwnProperty("entityId")  || this.hasOwnProperty("displayName"))
                    return typeof value === "string" || typeof value === "number" ? value : undefined;
                else
                    return value;
            }

            if (!select || select.includes(key)
                || Array.isArray(this)
                || (!this.hasOwnProperty("entityId") && !this.hasOwnProperty("displayName"))
                || key === "")
                return value;

            return undefined;
        };

        try {
            if (request.params.tenantKey === "admin") {
                // Get the 'tenant' query parameter, if any.
                let tenants = request.query.hasOwnProperty("tenant")
                            ? request.query.tenant.split(",")
                            : null;

                // Call the appropriate library interface and return the response.
                libAPI[request.params.entityType](tenants)
                .then(result => {
                    response.statusCode = 200;
                    response.setHeader('Content-Type', responseType);
                    response.write(stringify(result, undefined, 4));
                    response.end();
                })
                .catch(error => {
                    response.statusCode = 500;
                    response.setHeader('Content-Type', responseType);
                    response.write(stringify({ error: error }));
                    response.end();
                });
            }
            else {
                libAPI["filter"](
                    request.params.tenantKey,
                    request.params.entityType.toUpperCase(),
                    filters, name, select, since, follow
                )
                .then(result => {
                    // Note: 'result' can be a graph or a list!
                    function extractField(result, select) {
                        if (select && select.length === 1 && select[0] !== "stringables") {
                            if (result.graph) {
                                result.graph.nodes = result.graph.nodes.map(elem => elem[select[0]]);
                            }
                            else {
                                result = result.map(elem => elem[select[0]]);
                            }
                        }
                        return result;
                    }

                    response.statusCode = 200;
                    response.setHeader('Content-Type', responseType);

                    let updates = null;
                    let deletes = since
                                ? libAPI.deletes(
                                    request.params.tenantKey,
                                    request.params.entityType.toUpperCase(),
                                    since
                                  )
                                : null;

                    if (typeof result.on === "function") {
                        // This is when we get the result as a database result stream.
                        let rowReceived = null;

                        result
                        .on("fields", meta => {
                            // Meta data - we don't need to do anything with this.
                            // logger.info("Resultset stream reports fields: " + meta);
                        })
                        .on("data", row => {
                            if (inMemory) {
                                updates = (updates || []).concat(JSON.parse(row.entity));
                            }
                            else {
                                response.write(rowReceived ? ',' : since ? '{ "updates": [' : '[');

                                if (select && select.length === 1 && select[0] !== "stringables") {
                                    rowReceived = JSON.parse(row.entity);
                                    response.write(JSON.stringify(rowReceived[select[0]]));
                                }
                                else if (select) {
                                    rowReceived = JSON.parse(row.entity);
                                    response.write(JSON.stringify(rowReceived, extract, pretty));
                                }
                                else {
                                    // Return just the raw JSON data from database.
                                    rowReceived = true;
                                    response.write(row.entity);
                                }
                            }
                        })
                        .on("error", err => {
                            response.write(err.message);
                            response.end();
                        })
                        .on("end", () => {
                            if (inMemory) {
                                response.write(since ? {
                                    updates: updates,
                                    deletes: deletes
                                } : updates, true);
                            }
                            else {
                                response.write(rowReceived ? "]" : "");

                                if (since) {
                                    response.write(', "deletes": ');

                                    deletes = extractField(deletes, select);
                                    response.write(JSON.stringify(deletes, extract, pretty));

                                    response.write('}');
                                }
                            }

                            response.end();
                        });
                    }
                    else {
                        // This is when we get the result as a complete list or graph.
                        if (inMemory) {
                            updates = extractField(result, select);

                            response.write(since ? {
                                updates: updates,
                                deletes: deletes
                            } : updates, true);
                        }
                        else {
                            if (since)
                                response.write('{ "updates": ');

                            updates = extractField(result, select);
                            response.write(JSON.stringify(updates, extract, pretty));

                            if (since) {
                                response.write(', "deletes": ');

                                deletes = extractField(deletes, select);
                                response.write(JSON.stringify(deletes, extract, pretty));

                                response.write('}');
                            }
                        }

                        response.end();
                    }
                })
                .catch(error => {
                    response.statusCode = 500;
                    response.setHeader('Content-Type', responseType);
                    response.write(stringify({ error: error.message || error }), true);
                    response.end();
                });
            }
        }
        catch (ex) {
            response.statusCode = 500;
            response.setHeader('Content-Type', responseType);
            response.write(stringify({ error: ex.message }));
            response.end();
        }
    }

    function handleLibraryRequest(request, response, inMemory) {
        // NodeJS adds the 'next' handler to the end of the argument list.
        // We must not interpret that as 'truthy'...
        inMemory = typeof inMemory === "boolean" ? inMemory : false;

        const stringify = inMemory
                        ? obj => obj
                        : JSON.stringify;

        if (!library || !library.isReady) {
            response.statusCode = 503; // 503 Service Unavailable
            response.setHeader('Content-Type', contentType);
            response.write(stringify({ error: "The library is not ready yet" }));
            response.end();

            return;
        }

        let filtersInQuery = [];
        let filtersInBody  = [];
        let name = null;

        for (var param in request.query) {
            if (request.query.hasOwnProperty(param)) {
                if (param === "pretty") {
                    // Ignore - Will be used later to format the response.
                }
                else if (param === "since") {
                    // Ignore - Will be used later to only report entity changes since a timestamp.
                }
                else if (param === "select") {
                    // Ignore - Will be used later to determine what fields to include in the response.
                }
                else if (param === "follow") {
                    // Ignore - Will be used later to add related entities.
                }
                else if (param === "name") {
                    // This is treated separately for RegExp-based name-matching.
                    name = request.query[param];
                }
                else if (param === "id") {
                    // Treated separately here because IDs are not indexed.
                    filtersInQuery.push({
                        context:  "VALUE",
                        key:      "entityId",
                        values:   Array.isArray(request.query[param]) ? request.query[param] : request.query[param].split(","),
                        operator: "="
                    });
                }
                else if (param === "ZONE") {
                    // This is treated separately because the 'id' or 'name' part may be missing.
                    filtersInQuery.push({
                        context:  "ZONE",
                        key:      "id",
                        values:   Array.isArray(request.query[param]) ? request.query[param] : request.query[param].split(","),
                        operator: "="
                    });
                }
                else {
                    // Everything else: "context:key=value" and "context:key!=value".
                    let operator = "=";
                    let paramParts = param.split(":");
                    if (paramParts.length === 1) {
                        paramParts.unshift("CONTEXTLESS");
                    }
                    if (paramParts[1].endsWith('!')) {
                        paramParts[1] = paramParts[1].slice(0, -1);
                        operator = "!=";
                    }
                    filtersInQuery.push({
                        context:  paramParts[0],
                        key:      paramParts[1],
                        values:   request.query[param].split(","),
                        operator: operator
                    });
                }
            }
        }

        /*  POST requests will have filters in the body. If we have multiple groups of filters,
            whatever is in the query parameters applies to all groups. So we add them to each.
        */
        if (request.method === "POST") {
            if (request.body) {
                filtersInBody = JSON.parse(request.body) || [];

                let filters = filtersInBody.length > 0 && Array.isArray(filtersInBody[0])
                            ? filtersInBody.map(filterGroup => filtersInQuery.concat(filterGroup))
                            : filtersInQuery.concat(filtersInBody);

                executeLibraryRequest(request, response, filters, name, inMemory);
            }
            else {
                let data = [];
                request.on('data', (chunk) => {
                    data.push(chunk);
                });
                request.on('end', () => {
                    filtersInBody = JSON.parse(data.join("")) || [];

                    let filters = filtersInBody[0] && Array.isArray(filtersInBody[0])
                                ? filtersInBody.map(filterGroup => filtersInQuery.concat(filterGroup))
                                : filtersInQuery.concat(filtersInBody);

                    executeLibraryRequest(request, response, filters, name, inMemory);
                });
            }
        }
        else {
            executeLibraryRequest(request, response, filtersInQuery, name, inMemory);
        }
    }

    function handleTopologyRequest(request, response) {
        if (!library || !library.isReady) {
            response.statusCode = 503; // 503 Service Unavailable
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({ error: "The library is not ready yet" }));
            response.end();

            return;
        }

        let select = request.query.hasOwnProperty("select") && request.query.select !== ""
                    ? request.query.select.split(",")
                    : null;
        let pretty  = request.query.hasOwnProperty("pretty")
                    ? 4
                    : undefined;

        if (!["flows", "stack", "group"].includes(request.query.follow)) {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({ error: "Invalid 'follow' attribute. Use 'flows' or 'stack'." }));
            response.end();
        }

        library.follow(
            request.params.tenantKey,
            (request.params.entityType || "").toUpperCase(),
            request.query.id,
            request.query.follow)
                .then(result => {
                    response.statusCode = 200;
                    response.setHeader('Content-Type', contentType);
                    response.write(JSON.stringify(result, function (key, value) {
                        // If we have no 'select' value, include everything.
                        // Else only stringify the specified keys on entities.
                        if (["fromRelationships", "toRelationships"].includes(key))
                            return undefined;

                        if (!select ||
                            Array.isArray(this) ||
                            !this.hasOwnProperty("entityId") ||
                            key === "" ||
                            select.includes(key))
                            return value;
                        else
                            return undefined;
                    }, pretty));
                    response.end();
                })
                .catch(error => {
                    response.statusCode = 500;
                    response.setHeader('Content-Type', contentType);
                    response.write(JSON.stringify({ error: error.message || error }));
                    response.end();
                });
    }

    function handlePipelineRequest(request, response) {
        if (request.query.action === "stats")
            library
            .getPipelineStats(
                request.params.tenantKey,
                request.params.entityType
            )
            .then(data => {
                response.statusCode = 200;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify(data));
                response.end();
            })
            .catch(error => {
                response.statusCode = 503; // 503 Service Unavailable
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify({ error: error }));
                response.end();
            });
        else
            library
            .controlPipeline(
                request.params.tenantKey,
                request.params.entityType,
                request.query.action
            )
            .then(count => {
                response.statusCode = 200;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify({ affected: count }));
                response.end();
            })
            .catch(error => {
                response.statusCode = 503; // 503 Service Unavailable
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify({ error: error }));
                response.end();
            });
    }

    function getPipelines(request, response) {
        library
        .getPipelines()
        .then(pipelines => {
            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({
                processID: libraryName,
                pipelines: pipelines
            }));
            response.end();
        })
        .catch(error => {
            response.statusCode = 503; // 503 Service Unavailable
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({ error: error }));
            response.end();
        });
    }

    // Called by the UI to ask for all metrics it can allow the user to select.
    function getMetrics(request, response) {
        if (!library || !library.isReady) {
            response.statusCode = 503; // 503 Service Unavailable
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({
                error: "The library is not ready yet"
            }));
            response.end();
            return;
        }

        library
        .getMetricDefinitions(request.params.tenantKey)
        .then(metrics => {
            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify(metrics));
            response.end();
        })
        .catch(error => {
            response.statusCode = 503; // 503 Service Unavailable
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({ error: error }));
            response.end();
        });
    }

    // Called by the UI to ask for all USQL rules it can allow the user to use.
    function getUsqlRules(request, response) {
        if (!library || !library.isReady) {
            response.statusCode = 503; // 503 Service Unavailable
            response.setHeader('Content-Type', contentType);
            response.write(JSON.stringify({
                error: "The library is not ready yet"
            }));
            response.end();
            return;
        }

        library
            .getUsqlRules(request.params.tenantKey)
            .then(metrics => {
                response.statusCode = 200;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify(metrics));
                response.end();
            })
            .catch(error => {
                response.statusCode = 503; // 503 Service Unavailable
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify({
                    error: error
                }));
                response.end();
            });
    }

    // Register the HTTP routes we expose.
    webapi.use( cors());
    webapi.get( '/status',                          cors(), healthStatus);
    webapi.get( '/api/health',                      cors(), healthMetrics);
    webapi.get( '/library/health',                  cors(), healthReport);
    webapi.get( '/library/pipelines',               cors(), getPipelines);
    webapi.get( '/library/pipeline/:tenantKey/:entityType', cors(), handlePipelineRequest);
    webapi.get( '/library/:tenantKey/metrics',      cors(), getMetrics);
    webapi.get( '/library/:tenantKey/usql',         cors(), getUsqlRules);
    webapi.get( '/library/:tenantKey/topo',         cors(), handleTopologyRequest);
    webapi.get( '/library/:tenantKey/:entityType',  cors(), handleLibraryRequest);
    webapi.post('/library/:tenantKey/:entityType',  cors(), handleLibraryRequest);
    webapi.use( '/',                                express.static("../client"));

    // Listen for messages from our parent or instantiator.
    process.on('message', message => {
        switch (message.command) {
            case "configure":
                loadConfig(message.config);
                break;

            case "health":
                process.send({
                    status:   "health",
                    code:     200,
                    id:       message.id,
                    headers:  {},
                    request:  message.request,
                    response: healthMetrics()
                });
                break;

            case "request":
                // The message must have a 'query' property with the request parameters.
                // The message should also have an 'ID' property that the caller can use to
                // correlate our response.
                // Also, if there may be a 'body' property with JSON-stringified data.
                requests.push(message.id);
                message.request.method = message.request.body ? "POST" : "GET";

                handleLibraryRequest(
                    message.request,
                    {   // 'response' stub.
                        headers:   {},
                        response:  ownProcess ? "" : [],
                        setHeader: function (header, value) {
                            this.headers[header] = value;
                        },
                        write: function (response, singularWrite) {
                            if (singularWrite)
                                this.response = response;
                            else if (ownProcess)
                                this.response += response;
                            else
                                this.response = this.response.concat(response);
                        },
                        end: function () {
                            process.send({
                                status:   "response",
                                code:     this.statusCode,
                                id:       message.id,
                                headers:  this.headers,
                                request:  message.request,
                                response: this.response
                            });
                            requests.remove(message.id);
                            setTimeout(retire, 1);   // If we were asked to retire, do so now.
                        }
                    },
                    !ownProcess    // If in-memory request (No ownProcess), don't stringify reponse!
                );
                break;

            case "retire":
                // Terminate as soon as all pending requests have been handled.
                retire(true);
                break;
        }
    });
    process.on('disconnect', function() {
        libraryLogger.warn('StreamerService exited - terminating LibraryService');
        process.exit(0);    
    });

    let waitingToRetire = false;
    function retire(prepareToRetire) {
        waitingToRetire = waitingToRetire || !!prepareToRetire;

        if (waitingToRetire && requests.length <= 0) {
            process.disconnect();
            process.exit(0);    
        }
    }

    function stop(signal) {
        libraryLogger.info("LibraryService is exiting...");

        if (listener)
            listener.close();     // Don't leave HTTP port open.

        if (library)
            library.stop(function () {
                libraryLogger.info("The Library has been stopped");

                if (ownProcess) {
                    libraryLogger.info("Terminating self...  Goodbye");
                    libraryLogger.info("----------- :-( ------------");
                }
                process.kill(process.pid, signal);
            });
        else 
            process.kill(process.pid, signal);
    }

    // This is what gets returned to parties instantiating us inside their process.
    asyncIF = {
        events: {},
        on: function (eventName, fn) {
            this.events[eventName] = this.events[eventName] || [];
            this.events[eventName].push(fn);
        },
        emit: function (eventName, data) {
            const eventHandlers = this.events[eventName] || [];
            eventHandlers.forEach(fn => { fn.call(null, data); });
        },
        send: function (eventName, data) {
            if (process.emit) process.emit(eventName, data);
        },
        start: function () {
            configReader.load();
        },
        stop: stop,
        getLibrary: function () {
            return library || {};
        }
    };

    /*  Everything is in place. If we have a parent process, tell it to send us
        the config now. If we don't, we have to load the configuration ourselves.
     */
    if (isUnderETL) {
        setTimeout(() => process.send({ status: "preparing" }), 1);
    }
    else {
        configReader.load();
    }
    
    return asyncIF;
};
// End of LibraryService class.


/*  Now some startup stuff.
    Command line arguments:
    - name=<libraryName>: The library name. Used to select the proper config for this instance.
    - mode=auto:          This library was started by the StreamerService.
    - mode=failover:      Runs the library in 'offline' mode (no sync with Dynatrace) until
                          the primary instance (as specified in the first argument) goes down.
                          In the meantime, it synchronizes its admin with the primary instance.
    - config=<fileName>:  Load the config from somewhere else than '../config/config.json'.
 */
const processArgs = process.argv.slice(1).reduce((args, arg) => {
    if (!arg.includes("=")) {
        // The only relevant thing here is to see if
        // we have been started as a LibraryService.
        if (arg.includes("LibraryService.js")) {
            args.embedded = false;
        }
    }
    else {
        const argParts = arg.split("=");
        args[argParts[0]] = argParts[1];
    }
    return args;
},{
    embedded: true
});

if (processArgs.embedded) {
    // We are embedded in another process. Enable it to instantiate us
    // as follows: new LibraryService(name);
    module.exports = LibraryService;
}
else {
    // We were launched as a stand-alone process. Do what we need to do to start.
    const libraryService = new LibraryService(
        processArgs.name,
        process,
        processArgs
    );

    cleanup(function (exitCode, signal) {
        libraryService.stop();
        cleanup.uninstall();  // Don't call cleanup handler again.
        return false;
    });
}
