'use strict';

const { Writable } = require('stream');

class Persister extends Writable {

    constructor(options) {
        super({
            objectMode: true,
            highWaterMark: 1
        });

        this.curator  = options.curator;
        this.persist  = options.persist;
        this.tracker  = options.tracker || { progress: () => {} };  // For live diagnostic messages.
        this.pipeline = null;
    }

    _write(runData, unused, onDone) {
        /*  Because we only call 'onDone()' once we're done, the Persister
            causes backpressure to happen when things take too long. And
            that will slow down calls to '_write()'.
         */
        const cycle = runData.cycle;
        const data  = runData.data;

        if (!data || data.length === 0) {
            this.tracker.progress('persister', {
                stage:     "error",
                cycle:     cycle,
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                error:     "No data received"
            });

            onDone();
            return;
        }

        this.tracker.progress('persister', {
            stage:     "begin",
            cycle:     cycle,
            timestamp: (new Date()).getTime(),
            pipeline:  this.pipeline,
            data:      data.length
        });

        this.persist(data)
        .then(() => {
            this.tracker.progress('persister', {
                stage:     "end",
                cycle:     cycle,
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                data:      data.length,
            });

            onDone();
        })
        .catch(err => {
            this.tracker.progress('persister', {
                stage:     "error",
                cycle:     cycle,
                timestamp: (new Date()).getTime(),
                pipeline:  this.pipeline,
                error:     err
            });

            onDone();
        });
    }

    start(pipeline) {
        this.pipeline = pipeline;
        // There's nothng else we need to do because we are a Writable.
    }
    stop() {
        this.end();             // No more data will be written.
        this.persist(false);    // Release whatever we're using to persist.
    }
}

module.exports = Persister;