'use strict';

const Persister       = require('./Persister.js');
const Collector       = require('./Collector.js');
const Ticker          = require('./Ticker.js');

// Internal class for the PipelineManager.
const PipelineTracker = function (tenant, type, socket, logger) {
    const self = this;
    this.reports = {};

    const friendly = function (name) {
        if (!name) return "entitie";
        return name.toLowerCase().replace(/_/g, " ");
    }

    return {
        progress: function (component, update) {
            update.cycle = update.cycle || new Date().getTime();

            try {
                let report = self.reports[update.cycle] = self.reports[update.cycle] || {
                    id:     "Pipeline_" + tenant + "_" + type,
                    tenant: tenant,
                    type:   type,
                    name:   "Synchronize " + friendly(type) + "s in " + tenant,
                    cycle:  update.cycle,
                    log: {
                        results: {},
                        errors:  [],
                        size:    0
                    }
                };

                if (update.error) {
                    report.log.errors.push(update.error);

                    setImmediate(() => logger.error(update.error));
                }
                else if (update.warning) {
                    report.log.errors.push(update.warning);

                    setImmediate(() => logger.warn(update.warning));
                }
                else if (update.info) {
                    setImmediate(() => logger.info(update.info));
                }

                switch (component) {
                    case "ticker":
                        switch (update.stage) {
                            case "start":
                                // This when the Trigger starts the pipeline.
                                report.progress = "start";
                                report.log.started = update.timestamp;
                                break;
                            case "stop":
                                // This when the Trigger stops the pipeline.
                                report.progress = "stop";
                                report.log.stopped = update.timestamp;
                                break;
                            case "pause":
                                // This when the Trigger pauses the pipeline.
                                report.progress = "pause";
                                report.log.paused = update.cycle;
                                break;
                            case "resume":
                                // This when the Trigger resumes the pipeline.
                                report.progress = "resume";
                                report.log.resumed = update.resumed;
                                break;
                            case "plan":
                                report.progress = "plan";
                                report.log.planned = update.timestamp;
                                break;
                            case "fire":
                                report.progress = "fire";
                                report.log.fired = update.timestamp;
                                report.log.waiting = report.log.fired - report.log.planned;
                                break;
                        }
                        break;

                    case "collector":
                        switch (update.stage) {
                            case "init":
                                // This when the Collector initializes.
                                break;

                            case "begin":
                                report.progress = "initiated";
                                report.log.initiated = update.timestamp;
                                report.log.results.requested = update.data;
                                break;

                            case "start":
                                report.progress = "requested";
                                report.log.requested = update.timestamp;
                                break;

                            case "end":
                                report.progress = "collected";
                                report.log.collected  = update.timestamp;
                                report.log.collecting = report.log.collected - report.log.requested;
                                report.log.results.collected = update.data;
                                break;

                            case "warning":
                                report.progress = "done";
                                report.log.duration = new Date().getTime() - (report.log.requested || report.log.initiated || report.log.fired);
                                report.log.results.collected = update.data;
                                break;

                            case "error":
                                report.progress = "failed";
                                report.log.duration = new Date().getTime() - (report.log.requested || report.log.initiated || report.log.fired);
                                break;
                        }
                        break;

                    case "persister":
                        switch (update.stage) {
                            case "begin":
                                report.progress = "persisting";
                                report.log.submitted = update.timestamp;
                                report.log.results.persisted = update.data;
                                break;

                            case "end":
                                report.progress = "done";
                                report.log.persisted  = update.timestamp;
                                report.log.persisting = report.log.persisted - report.log.submitted;
                                report.log.duration   = report.log.persisted - report.log.requested;
                                break;

                            case "error":
                                report.progress = "failed";
                                report.log.duration = new Date().getTime() - report.log.requested;
                                break;
                        }
                        break;
                }

                socket.emit(report);

                if (report.progress === "done" || report.progress === "failed") {
                    delete self.reports[update.cycle];
                }
            }
            catch (ex) {
                logger.error("Could not track progress in pipeline for " + friendly(type) + "s in " + tenant + ": " + ex.message);
            }
        }
    };
}

class PipelineManager {
    #healthCheckInterval = null;
    #pipelineCreator = null;
    #pipelines  = [];

    #monitor = async function() {
        /*  Pipelines may theoretically stop and not schedule a next run. If that happens,
            we need to intervene. What we will do is check each pipeline every minute to
            see if it is still working. If not, we stop and restart that pipeline.
            */
        try {
            for (let i=0; i < this.#pipelines.length; i++) {
                let pipeline = this.#pipelines[i];
                const info = pipeline.getDefinition();
                if (!pipeline.isAlive()) {
                    try {
                        this.logger.warn("Restarting stalled pipeline for " + info.type + " in " + info.tenant + " - should run every " + info.interval + "ms");
                        pipeline.stop();

                        // Pipelines are never restarted. We create a whole new pipeline and enforce reinitialization.
                        pipeline = this.#pipelineCreator(info.tenant, info.type);
                        this.#pipelines.splice(i, 1, pipeline);
                        await pipeline.start(100 + 10 * i);  // Lets stagger it and give some breathing room.
                    }
                    catch (ex) {
                        this.logger.error("    Could not restart pipeline for " + info.type + " in " + info.tenant + ": " + ex.message + " - Please restart the library");
                    }
                }
            };
        }
        catch (ex) {
            this.logger.error("Periodic pipeline health check failed: " + ex.message);
        }
    }

    #createDynatracePipeline = function(tenant, type) {
        const self = this;

        this.logger.info("    Creating pipeline for " + this.friendly(type) + "s in " + tenant);

        if (!this.relationships[type] || !this.relationships[type].retrieval) {
            this.logger.error("    Entity type " + type + " is unknown or does not have any retrieval strategies");
            return;
        }
        // The tracker will collect the progress of each pipeline.
        const tracker = new PipelineTracker(tenant, type, this.socket, this.logger);

        /*  The persister is the last step in the pipeline, but we create it first.
            It writes entities to the database and updates the internal administration.
            */
        const persister = new Persister({
            curator:  this.curator,
            tracker:  tracker,
            persist: (function (tenant, type) {
                // Return a curried version of 'refreshEntities()', with tenant and type already bound.
                return (data) => self.dataAccess.refreshEntities(tenant, type, data);
            })(tenant, type)
        });

        let chunkingOptions = this.config.Tenants[tenant].chunking || {};

        /*  The collector is the second step in the pipeline. A collector will retrieve
            (all at once or in chunks) entities of a specific type in a specific tenant.
            */
        const collectorOptions = {
            logger:       this.logger,
            pageSize:     this.config.Tenants[tenant].pageSize,
            curator:      this.curator,
            scheduler:    this.scheduler,
            tracker:      tracker,
            tenant:       tenant,
            type:         type,
            friendlyType: this.friendly(type),
            dynatraceAPI: this.dynatraceAPI,
            config:       this.config,
            endPoint:     this.apiEndPoints[type],
            chunking:     chunkingOptions[type],
            consumer:     persister
        };
        const collector = chunkingOptions[type] == "NONE"
                        ? new Collector.NoCollector(collectorOptions)
                        : this.relationships[type].retrieval.includes("entity") &&
                            chunkingOptions[type] == "HOST"   // Only 'host=' is supported (PGs and PGIs only).
                        ? new Collector.PerEntityCollector(collectorOptions)
                        : this.relationships[type].retrieval.includes("chunked") &&
                            chunkingOptions[type] &&
                            isNaN(chunkingOptions[type]) &&   // A number means: as discovered, X at a time.
                            chunkingOptions[type] != "ALL" &&
                            chunkingOptions[type] != "LIST"
                        ? new Collector.ChunkedCollector(collectorOptions)
                        : this.relationships[type].retrieval.includes("discover") &&
                            chunkingOptions[type] &&
                            !isNaN(chunkingOptions[type])     // A number means: as discovered, X at a time.
                        ? new Collector.DiscoverCollector(collectorOptions, self.config.library.idListLimit)
                        : this.relationships[type].retrieval.includes("all")
                        ? new Collector.SimpleCollector(collectorOptions)
                        : null;

        if (collector) {
            this.logger.info(
                "    Pipeline for " + this.friendly(type) + "s in " + tenant + " uses the " +
                collector.constructor.name + " (" + collector.chunking + ")"
            );
        }
        else {
            this.logger.error("   Entity type " + type + " has an invalid retrieval strategy for " + tenant);
            return;
        }

        /*  The ticker is the first step in the pipeline. It drives the pace at which
            collectors are triggered, but it will slow down as backpressure happens.
            */
        return new Ticker.ReadableTicker({
            curator:  this.curator,
            tracker:  tracker,
            logger:   this.logger,
            consumer: collector
        });
    }

    #createLibraryPipeline = function (tenant, type) {
        this.logger.info("    Creating pipeline for " + this.friendly(type) + "s in " + tenant);

        const self = this;

        // The tracker will collect the progress of each pipeline.
        const tracker = new PipelineTracker(tenant, type, this.socket, this.logger);

        /*  The persister is the last step in the pipeline, but we create it first.
            It it only updates the internal administration - no database writes.
            */
        const persister = new Persister({
            curator:  this.curator,
            tracker:  tracker,
            persist: (function(tenant, type) {
                return (data) => {
                    return new Promise(function (resolve, reject) {
                        if (data && Array.isArray(data.updates)) {
                            let proxy = null;
                            for (let i=0; i < data.updates.length; i++) {
                                proxy = data.updates[i];

                                self.entityAccess.updateEntity(tenant, type, proxy.entity, null, proxy.checksum, proxy.lastUpdated);
                            };
                        }

                        if (data && Array.isArray(data.deletes)) {
                            let proxy = null;
                            for (let i=0; i < data.deletes.length; i++) {
                                proxy = self.curator.getEntity(tenant, type, data.deletes[i]);

                                self.entityAccess.removeEntity(tenant, type, proxy);
                            };
                        }

                        resolve();
                    });
                };
            })(tenant, type)
        });

        /*  The collector is the second step in the pipeline. A collector will retrieve
            (all at once or in chunks) entities of a specific type in a specific tenant.

            The LibraryCollector will basically just asks another Library for updates.
            */
        const collectorOptions = {
            pageSize:     0,
            curator:      this.curator,
            scheduler:    this.scheduler,
            tracker:      tracker,
            tenant:       tenant,
            type:         type,
            friendlyType: this.friendly(type),
            dynatraceAPI: this.libraryAPI,
            config:       this.config,
            endPoint:     "/library/" + tenant + "/" + type,
            chunking:     null,
            consumer:     persister,
            primary:      "Primary",    // Name of the Primary Library instance.
            pollTime:     5             // Poll every 5 minutes
        };
        const collector = new Collector.LibraryCollector(collectorOptions);

        this.logger.info(
            "    Pipeline for " + this.friendly(type) + "s in " + tenant + " uses the " +
            collector.constructor.name + " (synchronization with primary library)"
        );

        /*  The ticker is the first step in the pipeline. It drives the pace at which
            collectors are triggered, but it will slow down as backpressure happens.
            */
        return new Ticker.ReadableTicker({
            curator:  this.curator,
            tracker:  tracker,
            logger:   this.logger,
            consumer: collector
        });
    }

    get curator() {
        return this.globals.curator;
    }  
    get dataAccess() {
        return this.globals.dataAccess;
    }    
    get entityAccess() {
        return this.globals.entityAccess;
    }     

    constructor(globals) {
        this.globals       = globals;
        this.config        = globals.config;
        this.friendly      = globals.friendly;
        this.scheduler     = globals.scheduler;
        this.logger        = globals.logger;
        this.socket        = globals.socket;
        this.relationships = globals.relationships;
        this.tenants       = globals.tenants;
        this.types         = globals.types;
        this.dynatraceAPI  = globals.dynatraceAPI;
        this.libraryAPI    = globals.libraryAPI;
        this.apiEndPoints  = globals.apiEndPoints;
    }

    start(isOnStandby) {
        this.logger.info("Creating synchronization pipelines");

        this.#pipelineCreator = isOnStandby
            ? this.#createLibraryPipeline
            : this.#createDynatracePipeline;

        this.#pipelines = [];
        this.tenants.forEach(tenant => {
            this.types.forEach(type => {
                this.#pipelines.push(this.#pipelineCreator(tenant, type));
            });
        });
        this.logger.info("    Starting " + this.#pipelines.length + " pipelines");

        this.#pipelines.forEach((pipeline, i) => {
            // We start the pipelines in a staggered manner to spread their execution.
            pipeline.start(i * 1000);
        });

        this.#healthCheckInterval = setInterval(() => this.#monitor(), this.config.library.healthCheck * 1000);
    }

    stop(quiet) {
        if (!quiet && this.#pipelines.length > 0)
            this.logger.info("Stopping " + this.#pipelines.length + " pipelines");

        clearInterval(this.#healthCheckInterval);

        this.#pipelines.forEach((pipeline) => pipeline.stop());

        this.logger.info(this.#pipelines.length + " pipelines were stopped");
        this.#pipelines = [];
    }

    getPipelines(tenants, types) {
        if (typeof tenants === "string")
            tenants = [ tenants ];
        if (typeof types   === "string")
            types   = [ types   ];

        return this.#pipelines
            .filter(pipeline => !tenants || tenants.includes(pipeline.consumer.tenant))
            .filter(pipeline => !types   || types  .includes(pipeline.consumer.type  ));
    }

    async getPipelinesInfo() {
        const self = this;

        return this.#pipelines.map(pipeline => {
            const definition   = pipeline.getDefinition();
            definition.id      = "Pipeline_" + definition.tenant + "_" + definition.type;
            definition.name    = "Synchronize " + self.friendly(definition.type) + "s in " + definition.tenant;
            definition.library = self.config.library.name;
            return definition;
        });
    }

    async getPipelineStats(tenant, type) {
        const self = this;

        for (let i=0; i < this.#pipelines.length; i++) {
            const pipeline   = this.#pipelines[i];
            const definition = pipeline.getDefinition();

            if (definition.tenant === tenant && definition.type === type) {
                const statistics = pipeline.getStatistics();
                statistics.id      = "Pipeline_" + tenant + "_" + type;
                statistics.name    = "Synchronize " + self.friendly(type) + "s in " + tenant;
                statistics.library = self.config.library.name;
                return statistics;
            }
        }
        return "Pipeline for " + type + " in " + tenant + " not found";
    }

    getHealthMetrics() {
        const healthMetrics = [];

        this.#pipelines.forEach(pipeline => {
            const metrics = pipeline.getStatistics();
            const tenant  = pipeline.consumer.tenant;
            const type    = pipeline.consumer.type;

            healthMetrics.push({
                metricId: "fullCycleTime", 
                dimensions: [ tenant, type ],
                value: metrics.fullCycle
            });
            healthMetrics.push({
                metricId: "avgChunkTime", 
                dimensions: [ tenant, type ],
                value: metrics.avgChunkTime
            });
            healthMetrics.push({
                metricId: "avgChunkSize", 
                dimensions: [ tenant, type ],
                value: metrics.avgChunkSize
            });
            healthMetrics.push({
                metricId: "failedChunkRequests", 
                dimensions: [ tenant, type ],
                value: metrics.failedChunks
            });
            healthMetrics.push({
                metricId: "totalChunkRequests", 
                dimensions: [ tenant, type ],
                value: metrics.totalChunks
            });
            healthMetrics.push({
                metricId: "isAlive", 
                dimensions: [ tenant, type ],
                value: pipeline.isAlive()
            });    
        });

        return healthMetrics;
    }

    getPipelineHealth() {
        const self = this;

        const health = [ 
            "active=" + this.#pipelines.filter(pipeline => pipeline.isAlive()).length + 
            ", inactive=" + this.#pipelines.filter(pipeline => !pipeline.isAlive()).length 
        ];
        this.#pipelines.forEach(pipeline => {
            const stats = pipeline.getStatistics();
            health.push(stats.tenant + " has " + (""+stats.count).padStart(5) + " " + self.friendly(stats.type + "s", 23) + " - pipe is " + stats.state + "/" + (stats.isBusy ? "busy" : "idle") + ": chunks=" + stats.chunkCount + ", full cycle=" + Math.round(stats.fullCycle/60) + " mins, last run=" + stats.lastRun + ", last result=" + stats.lastResult + ", next run in " + stats.nextRun + " secs");
        });

        return health;
    }

    async controlPipeline(tenant, type, action) {
        /*  The 'pause' and 'resume' actions are easy. However, since a stopped
            pipeline cannot be restarted, we have to create a new one.
            */
        let count = 0;
        for (let i=0; i < this.#pipelines.length; i++) {
            const info = this.#pipelines[i].getDefinition();
            if (info.tenant === tenant &&
                (!type || info.type === type) &&
                typeof this.#pipelines[i][action] === "function") {

                try {
                    this.logger.info("User-requested " + action + " of pipeline for " + this.friendly(info.type) + "s in " + info.tenant);

                    if (action === "start") {
                        let pipeline = this.#pipelineCreator(info.tenant, info.type);
                        this.#pipelines.splice(i, 1, pipeline);

                        await pipeline.start();
                    }
                    else {
                        this.#pipelines[i][action]();
                    }
                    count++;
                }
                catch (ex) {
                    throw new Error("Could not " + action + " pipeline for " + this.friendly(info.type) + "s in " + info.tenant + ": " + ex.message);
                }
            }
        }
        return count;
    }
}

module.exports = PipelineManager;