'use strict';

class Resolver {

    constructor() {
        this.dots = /\./g;
        this.hashes = /#/g;
        this.lastBracket = /\][^|]/;
    }

    resolveString(resolvable, values) {
        if (!resolvable || typeof resolvable !== "string")
            return resolvable;

        const followPath = function (path, value) {
            let field = null;
            while (value && path.length > 0) {
                field = path.shift();
                value = !value || typeof value !== "object"
                      ? value
                      : value[field];
            }
            return value;
        };

        /*  We 'while' because values may contain multiple macros. Additionally, OR-ed
            macros are supported, and we may need to try multiple options (the first macro
            in such lists that leads to a value will be used, and the others are removed).
         */
        let oldValue = null; // Track if we are still changing.
        while (oldValue !== resolvable) {
            oldValue = resolvable;

            let macroStart = resolvable.indexOf("[");
            let macroEnd   = resolvable.indexOf("]");

            if (macroStart !== -1 && macroEnd !== -1 && macroStart < macroEnd) {
                const macro = resolvable.substring(macroStart + 1, macroEnd);
                const path  = macro.split(".");
                const repl  = followPath(path, values);

                if (repl) {
                    if (typeof repl === "object")       // No further resolution! Specifically because arrays will
                        return JSON.stringify(repl);    // be treated as macros, which would abviously screw it up.

                    const endOfMacro  = resolvable.charAt(macroEnd + 1) === '|'
                                      ? resolvable.search(this.lastBracket) // Up to first "]*" where '*' is not a '|'.
                                      : macroEnd;                           // Or where this macro ends if no OR options.
                    const beforeMacro = resolvable.substring(0, macroStart);
                    const afterMacro  = endOfMacro === -1 ? "" : resolvable.substring(endOfMacro + 1);

                    resolvable = beforeMacro + repl + afterMacro;
                }
                else {
                    // No replacement found. If we have an OR option, remove this macro and try the next one in
                    // the '|'-separated list.
                    resolvable = resolvable.charAt(macroEnd + 1) === '|'
                               ? resolvable.replace("[" + macro + "]|", "") // That part did not work. Try next option.
                               : resolvable.replace("[" + macro + "]", ""); // Else we don't have a replacement at all.
                }
            }
            else {
                // It's a literal value - no need to replace anything in the value.
            }
        }
        return resolvable;
    }

    resolveObject(resolvable, values) {
        if (!resolvable)
            return undefined;

        if (typeof resolvable !== "object")
            return this.resolveString(resolvable, values);

        const resolved = {};
        const properties = Object.getOwnPropertyNames(resolvable);

        for (const prop of properties) {
            resolved[prop] = this.resolveString(resolvable[prop], values);
        }
        return resolved;
    }
}

module.exports = Resolver;