// eslint.enable

require("./Extensions.js");
const DatabaseTracker = require('./DatabaseTracker.js');
const metricAggregations = require("../config/metricAggregations.json");

const maxDisplayNameSize = 512;
const maxStringValueSize = 256;
const purgeOldData = "DELETE FROM ServiceFlows WHERE endTime < CUTOFFTIME";
const createAppMethodsTableSQL = "CREATE TABLE `AppMethods` ( "
    + " `id` varchar(50) NOT NULL, "
    + " `name` nvarchar(" + maxDisplayNameSize + ") NOT NULL, "
    + " `tenant` nvarchar(50) NOT NULL, "
    + " `timestamp` bigint(20) unsigned NOT NULL, "
    + " `methods` longtext NOT NULL, "
    + " PRIMARY KEY (`id`) "
    + ") ENGINE=InnoDB DEFAULT CHARSET=utf8";
const createServiceFlowsTableSQL = "CREATE TABLE `ServiceFlows` ( "
    + " `index` bigint NOT NULL AUTO_INCREMENT, "
    + " `tenant` nvarchar(50) NOT NULL, "
    + " `startTime` bigint(20) unsigned NOT NULL, "
    + " `endTime` bigint(20) unsigned NOT NULL, "
    + " `appMethodId` nvarchar(50) NOT NULL, "
    + " `appMethodName` nvarchar(" + maxDisplayNameSize + ") NOT NULL, "
    + " `applicationId` nvarchar(50) NOT NULL, "
    + " `applicationName` nvarchar(" + maxDisplayNameSize + ") NOT NULL, "
    + " `id` nvarchar(50) NOT NULL, "
    + " `name` nvarchar(" + maxDisplayNameSize + ") NOT NULL, "
    + " `type` nvarchar(50) NOT NULL, "
    + " `icon` nvarchar(50) NOT NULL, "
    + " `async` bool, "
    + " `serviceType` nvarchar(50) NOT NULL, "
    + " `servicePath` nvarchar(10922) NOT NULL, "
    + " `pathLength` int NOT NULL, "
    + " `callCount` int NOT NULL, "
    + " `metrics` longtext NOT NULL, "
    + " PRIMARY KEY (`index`) "
    + ") ENGINE=InnoDB DEFAULT CHARSET=utf8";

class SFQueryServer {
    constructor(config, logger) {
        this.logger = logger;
        this.events = {};
        this.emit   = function (eventName, data) {
            // Call the subscribed event handlers.
            const eventHandlers = this.events[eventName] || [];
            eventHandlers.forEach(fn => { fn.call(null, data); });
        };

        // Initialize the database: create a pool and create the tables if necessary.
        this.logger.info("Establishing connection to the database");

        this.database = new DatabaseTracker(Object.assign({
            host: "localhost",
            port: 3306,
            user: "root",
            password: "",
            database: "StreamingAPI",
            connectionLimit: 100,
            connectTimeout: 10000,
            pipelining: true
        }, config.database), undefined, this.logger);

                      this.createTable("ServiceFlows", createServiceFlowsTableSQL)
        .then( res => this.createTable("AppMethods",   createAppMethodsTableSQL)
        .then( res => this.emit("ready")))
        .catch(err => this.emit("error", err));

        // Cleanup every hour.
        setInterval(this.cleanupTable.bind(this), 60*60*1000, "ServiceFlows", purgeOldData, config.daysToRetain || 14);
    }

    createTable(name, sql) {
        return new Promise((resolve, reject) => {
            this.logger.info("    Looking for '" + name + "' table");

            this.database.connect("Create table " + name)
                .then(connection => {
                    connection.query(sql)
                        .then(result => {
                            connection.release(() => {
                                this.logger.info("    Created table '" + name + "' in database");
                                resolve();
                            });
                        })
                        .catch(error => {
                            connection.release(() => {
                                if (error.code === "ER_TABLE_EXISTS_ERROR") {
                                    this.logger.info("    Found table '" + name + "' in database");
                                    resolve();
                                }
                                else {
                                    reject(error);
                                }
                            });
                        });
                })
                .catch(error => {
                    this.logger.error("Could not connect to database: " + (error.message || error));
                    reject(error);
                });
        });
    }

    cleanupTable(name, sql, days) {
        return new Promise((resolve, reject) => {
            this.logger.info("Purging " + name + " older than '" + days + "' days");
            const cutoffTime = (new Date()).getTime() - days * 24 * 60 * 60 * 1000;

            this.database.connect("Cleanup table " + name)
                .then(connection => {
                    connection.query(sql.replace("CUTOFFTIME", cutoffTime))
                        .then(result => {
                            connection.release(() => {
                                this.logger.info("    Purging of " + name + " succeeded");
                                resolve();
                            });
                        })
                        .catch(error => {
                            connection.release(() => {
                                this.logger.warn("    Purging of " + name + " failed: " + error.message);
                                resolve();
                            });
                        });
                })
                .catch(error => {
                    this.logger.error("Could not connect to database: " + (error.message || error));
                    reject(error);
                });
        });
    }

    on(eventName, fn) {
        // Add the subscription (event handler) for this event.
        this.events[eventName] = this.events[eventName] || [];
        this.events[eventName].push(fn);
    };

    getServiceFlows(tenant, appMethodId, startTime, endTime, serviceId, async) {
        return new Promise((resolve, reject) => {
            let sql = "SELECT * FROM ServiceFlows WHERE tenant='" + tenant + "' "
                + "AND appMethodId='" + appMethodId + "' "
                + "AND startTime >= " + startTime + " "
                + "AND endTime   <= " + endTime + " "
                + (serviceId !== undefined
                    ? "AND (id= '" + serviceId + "' OR servicePath LIKE '%" + serviceId + "%') "
                    : "")
                + (async !== undefined
                    ? "AND async=" + async + " "
                    : "")
                + "ORDER BY pathLength, startTime ASC";

            this.database.connect("ServiceFlowQuery for " + appMethodId)
                .then(connection => {
                    const times = [];           // Holds the timeframes actually found in the data.
                    const nodes = new Map();    // Holds the nodes in the tree, indexed by servicePath.
                    const tree = {             // The object that we will return after building it out.
                        timeframe: {
                            startTime: startTime,
                            endTime: endTime,
                            duration: endTime - startTime
                        }
                    };

                    connection.stream(sql)
                        .on("error", error => {
                            // If an error occurs while streaming, close the connection and fail.
                            connection.release(() => reject(error));
                        })
                        .on("fields", columns => {
                            // Here we receive an accounting of the columns. We don't need it.
                        })
                        .on("data", row => {
                            // A next row is being streamed to us from the database. Add it to the tree.
                            const nodeIndex = row.servicePath
                                            ? row.servicePath + "." + row.id
                                            : row.id;
                            let node = nodes.get(nodeIndex);
                            if (node) {
                                // The node exists. Add this row's callCount to get the sum across the timeframe.
                                node.callCount += row.callCount;
                            }
                            else {
                                // Not yet in tree. Create new node using only a subset of the row's properties.
                                node = (({ id, name, type, icon, async, serviceType, servicePath, callCount }) =>
                                    ({ id, name, type, icon, async, serviceType, servicePath, callCount }))(row);

                                // Prepare to store its children plus its metrics for different time periods.
                                node.timeframes = new Map();
                                node.children = [];

                                // Find the parent under which we need to add this new node.
                                const parent = nodes.get(node.servicePath);
                                if (node.id === appMethodId) {
                                    // This is the root node.
                                    tree.transactionId = appMethodId;
                                    tree.transactionName = row.appMethodName;
                                    tree.applicationId = row.applicationId;
                                    tree.applicationName = row.applicationName;

                                    tree.root = node;
                                }
                                else if (parent) {
                                    parent.children.push(node);
                                }
                                else {
                                    // This should never happen because the rows have been sorted in order
                                    // of increasing path length, and we should have all the nodes as rows.
                                    throw new Error("Cannot find parent for " + node.id + " under " + node.servicePath);
                                }

                                // Add node to the map so that we can look it up quickly, later.
                                nodes.set(nodeIndex, node);
                            }

                            // Add the metrics to the (now) existing node. We add the callCount to the 
                            // metrics object because that's easier for later.
                            const metrics =  JSON.parse(row.metrics);
                            metrics.callCount = row.callCount;
                            node.timeframes.set(row.startTime, metrics);

                            // Store this timeframe we found in the data.
                            if (!times.includes(row.startTime))
                                times.push(row.startTime);
                        })
                        .on("end", () => {
                            // Now we've consumed all the rows. Release the database connection.
                            connection.release(() => {
                                // We're almost ready to return the tree. We only need to merge
                                // the metrics across the timeframes and then we're done.
                                const add = (x, y) => x + y;
                                let node = null;
                                let series = null;

                                // We don't traverse the tree, but instead we iterate over the map
                                // of nodes - it's the same collection, but this is faster.
                                for (let nodeEntry of nodes) {
                                    // The key (servicePath) is at nodeEntry[0], but we don't need it.
                                    node = nodeEntry[1];
                                    node.metrics = {};

                                    // We don't iterate over the metrics in the data, but over the
                                    // metrics found in the metric aggregation strategy list.
                                    for (let metricName in metricAggregations) {
                                        const aggregation  = metricAggregations[metricName].aggregation;
                                        const weightFactor = metricAggregations[metricName].weight;

                                        /*  The metric series consists of an entry per found timeframe.
                                            Note that if we don't have a weight factor that depends on
                                            some other metric value, the weight factor defaults to 1.
                                            If we don't find a value for a specific timeframe, the
                                            weight factor will be 0 (the value will carry no weight).
                                         */
                                        series = times.map(time => ({
                                            value  : node.timeframes.has(time) &&
                                                     node.timeframes.get(time).hasOwnProperty(metricName)
                                                   ? node.timeframes.get(time)[metricName]
                                                   : -1,
                                            weight : node.timeframes.has(time)
                                                   ? (weightFactor ? node.timeframes.get(time)[weightFactor] || 0 : 1)
                                                : 0,
                                        })).filter(entry => entry.value !== -1);

                                        // Don't set the metric if we don't have any valid values for it.
                                        if (series.length === 0) continue;

                                        // Aggregate the series per the strategy specified.
                                        switch (aggregation) {
                                            case "NONE":
                                                node.metrics[metricName] = series.map(entry => entry.value * entry.weight);
                                                break;
                                            case "SUM":
                                                node.metrics[metricName] = series.map(entry => entry.value * entry.weight).reduce(add);
                                                break;
                                            case "AVG":
                                                node.metrics[metricName] = series.map(entry => entry.value * entry.weight).reduce(add)
                                                                         / (series.map(entry => entry.weight).reduce(add) || 1);
                                                break;
                                            case "MAX":
                                                node.metrics[metricName] = Math.max(...series.map(entry => entry.value * entry.weight));
                                                break;
                                            case "MIN":
                                                node.metrics[metricName] = Math.min(...series.map(entry => entry.value * entry.weight));
                                                break;
                                            default:
                                            // Unknown aggregation strategy. Don't set the metric in the result.
                                        }
                                    }
                                    node.timeframes = node.timeframes.size; // Report how many timeframes were found.
                                    // delete node.timeframes; // No longer needed and should not be in the output.
                                }

                                // Return the tree. It consists of the very same nodes we've just been updating.
                                resolve(tree);
                            });
                        });
                })
                .catch(err => {
                    // No connection to the database could be created. Fail.
                    reject(err);
                });
        });
    }

    getApplications(tenant, id) {
        return new Promise((resolve, reject) => {
            let sql = "SELECT * FROM AppMethods WHERE tenant='" + tenant + "'"
                + (id ? " AND id='" + id + "'" : "");

            this.database.connect("AppMethodQuery for " + id)
                .then(connection => {
                    connection.query(sql)
                        .then(rows => {
                            connection.release(() => {
                                if (rows.length > 0) {
                                    resolve(rows.map(row => ({
                                        id:   row.id,
                                        name: row.name,
                                        methods: JSON.parse(row.methods).map(method => ({
                                            id:   method.transactionId,
                                            name: method.transactionName
                                        }))
                                    })));
                                }
                                else {
                                    reject(404);
                                }
                            });
                        })
                        .catch(err => {
                            connection.release(() => {
                                reject(err);
                            });
                        });
                })
                .catch(err => {
                    // No connection to the database could be created. Fail.
                    reject(err);
                });
        });
    }
}
module.exports = SFQueryServer;