'use strict';

const { promises } = require('dns');
const fs = require('fs');
const axios = require('axios').default;

if (!Array.prototype.replace)
    Array.prototype.replace = function (oldElem, newElem) {
        let idx = this.indexOf(oldElem);
        if (idx >= 0) this.splice(idx, 1, newElem);
        return idx;
    };

if (!Array.prototype.remove)
    Array.prototype.remove = function (value) {
        for (var i=0; i < this.length; i++)
            if (this[i] === value)
                return this.splice(i, 1)[0];
        return null;
    };

if (!Array.prototype.indexOfByField)
    Array.prototype.indexOfByField = function (value, field) {
        for (var i=0; i < this.length; i++)
            if (this[i][field] === value)
                return i;
        return -1;
    };

if (!Array.prototype.unique)
    Array.prototype.unique = function () {
        const seen = {};
        return this.filter(item => seen.hasOwnProperty(item) ? false : (seen[item] = true));
    };

const Scheduler = function (config, library, logger, jobs, onTuningUpdate) {
    config = config || {};

    logger = logger || {
        info:  function () {},
        warn:  function () {},
        error: function () {}
    };

    const tuningInfoFile = library ? "../config/tuning.json" : null;
    const sortTags  = job => {
        if (!job.tagsAreSorted) {
            job.tags = (job.tags || []).sort((a, b) => (a.context + a.key).localeCompare(b.context + b.key));
            job.tagsAreSorted = true;
        }
    };
    const jobSorter = (a, b) => {
        sortTags(a); sortTags(b);
        return a.entityId < b.entityId ? -1 : a.entityId > b.entityId ? 1 : 0;
    };

    function loadTuningInfo() {
        if (!tuningInfoFile)
            return  {
                maxJobSize: Number.POSITIVE_INFINITY,
                minJobSize: 100,
                maxThreads: 16
            };

        let tuningInfo = null;
        try {
            // Startup, so we have do do this synchronously.
            tuningInfo = JSON.parse(fs.readFileSync(tuningInfoFile, "utf8"));
        }
        catch (ex) {
            logger.warn("Scheduler could not find tuning information file " + tuningInfoFile + " - starting with defaults");
            tuningInfo = {
                maxJobSize: Number.POSITIVE_INFINITY,
                minJobSize: 100,
                maxThreads: 16
            };
        }
        return tuningInfo;
    }

    function updateTuningInfo() {
        if (onTuningUpdate) onTuningUpdate(tuningInfo);

        // To be sure that we don't write before we finished last time.
        if (!tuningInfoFile || updateTuningInfo.writing) return;

        updateTuningInfo.writing = JSON.stringify(tuningInfo, null, 4);
        fs.writeFile(tuningInfoFile, updateTuningInfo.writing, "utf8", (err) => {
            if (err)
               logger.warn("Scheduler could not save tuning information: " + err);

            logger.info("Tuning info adjusted and written to " + tuningInfoFile + ": ");
            logger.info("    maxJobSize: "  + tuningInfo.maxJobSize);
            logger.info("    minJobSize: "  + tuningInfo.minJobSize);
            logger.info("    maxThreads: "  + tuningInfo.maxThreads);

            const capacities = [];
            for (let granularity in tuningInfo.capacities) {
                capacities.push(tuningInfo.capacities[granularity] + " for " + getGranularityName(granularity));
            }
            logger.info("    Workers per schedule: " + capacities.join(", "));

            updateTuningInfo.writing = null;
        });
    }

    let tuningInfo    = loadTuningInfo();
    let activeCycles  = {};      // Indexed by tick (timestamp). Old cycles get purged.
    let metricLatency = Number(config.jobLatency || 1) * 60000;
    let currentSchedules  = {};
    let originalSchedules = {};
   
    const granularities = [
        { label: "1 minute",   value: "_min",     bucket: 1,       dtSupported: true  },
        { label: "5 minutes",  value: "_5mins",   bucket: 5,       dtSupported: true  },
        { label: "10 minutes", value: "_10mins",  bucket: 10,      dtSupported: true  },
        { label: "15 minutes", value: "_15mins",  bucket: 15,      dtSupported: true  },
        { label: "30 minutes", value: "_30mins",  bucket: 30,      dtSupported: true  },
        { label: "1 hour",     value: "_hour",    bucket: 60,      dtSupported: true  },
        { label: "2 hours",    value: "_2hours",  bucket: 60*2,    dtSupported: true  },
        { label: "3 hours",    value: "_3hours",  bucket: 60*3,    dtSupported: false },
        { label: "6 hours",    value: "_6hours",  bucket: 60*6,    dtSupported: true  },
        { label: "12 hours",   value: "_12hours", bucket: 60*12,   dtSupported: false },
        { label: "24 hours",   value: "_day",     bucket: 60*24,   dtSupported: true  },
        { label: "48 hours",   value: "_2days",   bucket: 60*48,   dtSupported: false },
        { label: "72 hours",   value: "_3days",   bucket: 60*72,   dtSupported: false },
        { label: "1 year",     value: "_year",    bucket: 60*8760, dtSupported: false },
        { label: "disabled",   value: "_disabled",bucket: 0,       dtSupported: true }
    ];

    const timespans = granularities.reduce(function(result, granularity) {
        result[granularity.value] = granularity.bucket * 60 * 1000;
        return result;
    }, {});

    function isEntityJob(job) {
        return !["usersession", "useraction", "PROBLEM"].includes(job.type);
    }

    function getGranularityName(frequency) {
        for (var i = 0; i < granularities.length; i++) {
            if (granularities[i].value === frequency)
                return granularities[i].label;
        }
        return "unknown duration";
    }

    function getGranularityTime(frequency) {
        let bucket = null;

        for (var i = 0; !bucket && i < granularities.length; i++) {
            if (granularities[i].value === frequency)
                bucket = granularities[i].bucket;
        }
        bucket = bucket || 60*24;   // If not found, make it 1 day.

        return bucket*60*1000;
    }

    function getGranularityValue(timespan, dtSupported) {
        // The frequncy can be in minutes or millisconds.
        if (timespan > 2000)
            timespan = Math.ceil(timespan / 60000);   // Now it is minutes.

        const count = dtSupported ? granularities.length-1 : granularities.length;

        for (var i = 0; i < count; i++) {
            if (granularities[i].bucket === timespan)
                return dtSupported && !granularities[i].dtSupported ? granularities[i+1].value : granularities[i].value;
        }
        // No exact match. Then we find the nearest match (ceiled).
        for (var i = 0; i < count; i++) {
            if (granularities[i].bucket > timespan)
                return dtSupported && !granularities[i].dtSupported ? granularities[i+1].value : granularities[i].value;
        }

        return "_day";  // Fine. Let's do 'day'.
    }

    function getNextGranularity(frequency, dtSupported) {
        let granularity = null;

        for (var i = 0; !granularity && i < granularities.length; i++) {
            if (granularities[i].value === frequency)
                granularity = dtSupported && !granularities[i+1].dtSupported ? granularities[i+2] : granularities[i+1];
        }

        return granularity ? granularity.value : dtSupported ? "_day" : "_year";
    }

    function createSchedules(jobList) {
        // Create a schedule per supported granularity.
        const schedules = granularities.reduce((map, granularity) => {
            if (granularity.dtSupported)    // Skip timeframes that Dynatrace does not support.
                map[granularity.value] = [];
            return map;
        }, {});

        jobList.forEach(job => {
            if (!schedules.hasOwnProperty(job.granularity))
                job.granularity = "_min";   // Just to be sure.

            schedules[job.granularity].push(prepareJob(job));
        });

        return schedules;
    }

    function init(jobList) {
        logger.info("    Creating schedules using a latency of " + Math.round(metricLatency / 60000)+ " minutes");

        // The maxJobSize is initially taken form the config, but after that becomes a tuning parameter.
        tuningInfo.maxJobSize = tuningInfo.maxJobSize || config.maxJobSize || Number.POSITIVE_INFINITY;

        // The minJobSize is also taken from the config, but we don't change that afterwards.
        tuningInfo.minJobSize = config.minJobSize || 100;

        // The maxThreads is also taken from the config, but after that becomes a tuning parameter.
        tuningInfo.maxThreads = tuningInfo.maxThreads || config.maxThreads || 16;

        // Initialize the recommended worker counts per schedule if we don't have it yet.
        tuningInfo.capacities = tuningInfo.capacities || granularities.reduce((map, granularity) => {
            map[granularity.value] = 1;
            return map;
        }, {});

        // Write it out to file so that we can skip this step next time we start.
        updateTuningInfo();

        // Now create the job schedules.
        jobs = jobList || [];

        // Create a schedule per supported granularity.
        logger.info("    Allocating jobs to schedules");
        currentSchedules  = createSchedules(jobs);
        originalSchedules = createSchedules(jobs);

        for (let granularity in currentSchedules) {
            const jobNames = currentSchedules[granularity].map(job => job.name);
            if (jobNames.length > 0)
                logger.info("    Every " + getGranularityName(granularity) + ":\t " + jobNames.join(", "));
        }
    };

    function prepareJob(job) {
        // We don't have to 'prepare' jobs that are for user sessions/actions or for
        // problems. We also don't have to prepare jobs that already have been prepared.
        if (job.query || job.problems || job.requestOptions) return job;

        /*  We want to capture here what the mappings are that will be used by the
            destination's formatter, because when we go ask the Library for the
            entity data (ID based as well as criteria-based), we want to actually
            ask for as few fields as possible. At a minumum we are going to request
            that non-indexable fields (everything that is not a string) are excluded.
            */
        job.requestOptions = {
            name:   job.name,
            params: {
                tenantKey:  job.source,
                entityType: job.type,
            },
            query: {
                ZONE:   job.mz
            }
        };

        // Ask the API to return a graph with related entities, if so specified.
        if (job.related === "stack" || job.related == "flows") {
            job.requestOptions.query.follow = job.related;
        }

        if (job.mode === "explicit") {
            job.entities = job.entities || [];
            // FIX: We can't put it all in a query parameter - potentially too much!
            //     job.requestOptions.query.id = job.entities.map(entity => entity.entityId || entity);
            // So we add it to the body of the request instead, as a proper filter.
            job.requestOptions.body = JSON.stringify([{
                context:  "VALUE",
                key:      "entityId",
                values:   job.entities.map(entity => entity.entityId || entity),
                operator: "="
            }]);

            // Important note: From this point on, 'job.entities' will point to an array of full entities.
            // The original ID list will only be kept now in 'job.requestOptions.query.id'. Nowhere else!
        }
        else if (job.mode === "implicit") {
            job.criteria = job.criteria || [];
            job.requestOptions.body = JSON.stringify(job.criteria);
        }
        else {
            logger.error("Streamer encountered a job with mode '" + job.mode + "', "
                       + "which is not supported. Only 'implicit' and 'explicit' are allowed");
        }

        return job;
    }

    function getLastCycle(previousTick) {
        return activeCycles[previousTick];
    }

    function createNewCycle(tick) {
        /*  Depending on the time of the hour, different sets of jobs get included.
            Note that we look at the previous minute, not this one. The reason is
            that our Workers always stay some time in arrear with respect to the
            cycle time in the absolute times they use to collect their metrics.
            So, we need to check if that previous minute was on a relevant schedule
            boundary.
         */
        let endTime  = new Date(tick - metricLatency);
        let minutes  = endTime.getHours() * 60 + endTime.getMinutes();
        let schedule = null;

        for (var i = granularities.length; !schedule && i > 0; i--) {
            if (minutes % granularities[i-1].bucket === 0)
                schedule = granularities[i-1];
        }

        activeCycles[tick] = {
            progress:     "planned",
            cycle:        tick,
            id:           "Cycle_" + tick,
            name:         "Cycle at " + new Date(tick).toLocaleTimeString(),
            endTime:      endTime.getTime(),
            minutes:      minutes,
            schedule:     schedule ? schedule.value : "unknown",
            scheduleName: schedule ? schedule.label : "Unscheduled",
            capacity:     tuningInfo.capacities[schedule ? schedule.value : "unknown"] || 1,
            elapsedTime:  null,
            errors:       0,
            metrics:      {},
            assignments:  null,
            workers:      [],
            log: {
                planned:     (new Date()).getTime(),
                schedule:    schedule ? schedule.label : "Unscheduled",
                capacity:    tuningInfo.capacities[schedule ? schedule.value : "unknown"] || 1,
                assignments: null,
                workers:     null,
                results:     [],
                errors:      [],
                durations:   []
            },
            worker: "Scheduler",
            player: "Controller"
        };
        logger.info(
            "This cycle runs the " + activeCycles[tick].scheduleName + " schedule ending at " +
            endTime.toLocaleTimeString() + ", and should require " + activeCycles[tick].log.capacity + " workers"
        );

        return activeCycles[tick];
    }

    function reduceCycle(cycle) {
        let granularity = cycle.schedule;

        if (currentSchedules[granularity] && currentSchedules[granularity].length > 0) {
            // Note: We try to demote a job only once, but it's not always possible.
            let biggest =
                // What is the biggest (entity-based) job here that wasn't already demoted?
                currentSchedules[granularity].reduce((biggest, jobInSchedule) => {
                    if (!isEntityJob(jobInSchedule) || jobInSchedule.isDemoted || !jobInSchedule.entities)
                        return biggest;

                    if (!biggest || !biggest.entities || biggest.entities.length < jobInSchedule.entities.length)
                        return jobInSchedule;

                    return biggest;
                }, null) ||
                // No 'entity' jobs to demote? Well, then get the first never-demoted USQL job.
                currentSchedules[granularity].reduce((biggest, jobInSchedule) => {
                    if (!isEntityJob(jobInSchedule) && !jobInSchedule.isDemoted)
                        return jobInSchedule;
                    else
                        return biggest;
                }, null) ||
                // There is not a single job left that hasn't already been demoted.
                // Then take the first one and demote it again. Sorry dude, we have no choice.
                currentSchedules[granularity][0];

            if (!biggest) {
                return;    // How can this cycle be too big and yet nothing in it?! Impossible!
            }

            currentSchedules[granularity].remove(biggest);
            granularity = getNextGranularity(granularity, true);
            biggest.granularity = granularity;
            biggest.isDemoted = true;
            currentSchedules[granularity].push(biggest);

            logger.warn("Job '" + biggest.name + "' has been selected for downgrade to the " + getGranularityName(granularity) + " schedule");

            for (let granularity in currentSchedules) {
                const jobNames = currentSchedules[granularity].map(job => job.name);
                if (jobNames.length > 0)
                    logger.info("    Every " + getGranularityName(granularity) + ":\t " + jobNames.join(", "));
            }
        }
    }

    function evaluateCycle(previousTick) {
        // First, cleanup old cycles. Should be quick.
        var distantPast = Number(previousTick) - 60*10*1000;  // 10 minutes ago.

        for (let cycle in activeCycles) {
            if (Number(cycle) < distantPast)
                delete activeCycles[cycle];
        }

        // Next, examine how the previous cycle fared.
        let previousCycle = activeCycles[previousTick];
        if (!previousCycle) {
            return {
                retirees: [],
                fallen:   []
            };
        }

        /*  Things to consider when changing the recommended number of workers:
            - Non-fatal errors are reported as part of each job report. That stuff gets logged.
            - When a worker reports a serious error, it gets killed right away by the WorkerProxy.
            In that case, it would not have been able to report one or more jobs as completed.
            As a result we also would not have removed them from the cycle's worker list. Thus,
            for (remaining) each worker in the previous cycle we need to check if it is still
            alive and in the rosteredWorker list.
            - If so, it's simply overworked and we need to scale up.
            - If not, it died. We're going to have to replace it.
            - If we completed in less than half the allotted time, then scale back.
            - Else we have juuust the right amount of workers.
        */
       let fallen   = [];
       let retirees = [];

        try {
            if (previousCycle.workers && previousCycle.workers.length > 0) {
                // Some workers have not finished yet. Retire them (but let them finish).
                previousCycle.workers.forEach(worker => {
                    if (worker.isAlive)
                        retirees.push(worker);  // This one is overworked.
                    else
                        fallen.push(worker);    // This one has died.
                });

                logger.info("Retiring " + retirees.length + " overworked workers, mourning "
                                        + fallen.length   + " fallen workers");

                if (retirees.length > 0) {
                    /*  If there are overworked workers, we increase the number of workers.
                        If we need more workers than we are allowed to have, we 'throttle' some
                        jobs: we move a job to a less-frequent bucket and temporarily adjust its
                        granularity.
                        That begs the question: when will we move it back to its original bucket?
                        We are going to do that (implicitly) whenever the jobs.json file changes,
                        and also when the user moves to jobs to different buckets in the health UI.
                    */
                    logger.warn("Previous cycle did not complete in the allocated time");

                    if (previousCycle.capacity + 1 > config.maxWorkers) {
                        logger.warn("Maximum capacity has been reached for the "
                                + previousCycle.scheduleName + " schedule.");

                        reduceCycle(previousCycle);
                    }
                    else {
                        // Increase the capacity, and make sure all courser granularities (which have at least
                        // as much work to do as this granularity/schedule) are not smaller than the new value.
                        tuningInfo.capacities[previousCycle.schedule] = Math.min(previousCycle.capacity + 1, config.maxWorkers);

                        let idx = granularities.indexOfByField(previousCycle.schedule, "value");
                        for (let i = idx+1; i > 0 && i < granularities.length; i++) {
                            tuningInfo.capacities[granularities[i].value] = Math.max(
                                tuningInfo.capacities[previousCycle.schedule],
                                tuningInfo.capacities[granularities[i].value]
                            );
                        }
                        updateTuningInfo();     // Write to file also, for after restarts.

                        logger.info("Increasing recommended capacity for the "
                                + previousCycle.scheduleName + " schedule (and courser) to "
                                + tuningInfo.capacities[previousCycle.schedule] + " workers");
                    }
                }
            }
            else if (previousCycle.errors === 0 && previousCycle.elapsedTime < 30000 && previousCycle.capacity > 1) {
                logger.info("Previous cycle was successful, but over-resourced");

                tuningInfo.capacities[previousCycle.schedule] = previousCycle.capacity - 1;
                updateTuningInfo();     // Write to file also, for after restarts.

                logger.info("Reducing recommended capacity for the "
                        + previousCycle.scheduleName + " schedule to "
                        + tuningInfo.capacities[previousCycle.schedule] + " workers");
            }
            else if (previousCycle.errors > 0) {
                tuningInfo.capacities[previousCycle.schedule] = previousCycle.capacity;

                logger.info("Previous cycle had " + previousCycle.errors + " errors");
                logger.info("Recommended capacity for the "
                        + previousCycle.scheduleName + " schedule remains at "
                        + tuningInfo.capacities[previousCycle.schedule] + " workers");
            }
            else {
                tuningInfo.capacities[previousCycle.schedule] = previousCycle.capacity;

                logger.info("Previous cycle was successful");
                logger.info("Capacity for the " + previousCycle.scheduleName
                        + " schedule is in goldilocks zone with "
                        + tuningInfo.capacities[previousCycle.schedule] + " workers");
            }
        }
        catch (ex) {
            logger.error("Evaluation for the " + previousCycle.scheduleName + " schedule failed: " + ex.message);
        }
        finally {
            return {
                retirees: retirees,
                fallen:   fallen
            };
        }
    }

    function refreshJob(job) {
        return new Promise((resolve, reject) => {
            prepareJob(job);

            // Only entity-based jobs need to be refreshed. USQL jobs don't.
            if (!job.requestOptions)
                return resolve(job);

            logger.info("Refreshing entity data for job " + job.name);

            library
                .request(job.requestOptions)
                .then(listOrGraph => {
                    job.entities = !listOrGraph
                                 ? []
                                 : listOrGraph.graph
                                 ? (listOrGraph.graph.nodes || []).sort(jobSorter)
                                 : listOrGraph.sort(jobSorter);
                    job.linkages = !listOrGraph
                                 ? null
                                 : listOrGraph.graph
                                 ? listOrGraph.graph.edges
                                 : null;
                    resolve(job);
                })
                .catch(err => {
                    job.entities = job.entities || [];  // Make sure that we have at least an actual array.

                    reject("Unable to refresh entity data for job " + job.name + ": " + err);
                });
        });
    }

    function refreshJobs(jobs, doParallel) {
        return new Promise((resolve, reject) => {
            if (!jobs || jobs.length === 0) resolve();

            let errors = [];
            let i;

            try {
                if (doParallel) {
                    // Only entity-based jobs need to be refreshed. USQL jobs don't.
                    jobs = jobs.filter(job => !!job.requestOptions);

                    i = jobs.length;
                    const onJobRefreshed = function (error) {
                        if (error) {
                            logger.error(error);
                            errors.push(error);
                        }

                        i--;
                        if (i === 0) {
                            if (errors.length > 0)
                                reject(errors)
                            else
                                resolve();
                        }
                    };

                    // Kick off all job refresh requests at the same time.
                    jobs.forEach(job => {
                        library
                        .request(job.requestOptions)
                        .then(listOrGraph => {
                            // Adopt the refreshed list of entities as provided by the library.
                            job.entities = !listOrGraph
                                         ? []
                                         : listOrGraph.graph
                                         ? (listOrGraph.graph.nodes || []).sort(jobSorter)
                                         : listOrGraph.sort(jobSorter);
                            job.linkages = !listOrGraph
                                         ? null
                                         : listOrGraph.graph
                                         ? listOrGraph.graph.edges
                                         : null;
                            onJobRefreshed();
                        })
                        .catch(err => {
                            // Fallback: reuse previous list of entities (ensure it's an array).
                            job.entities = job.entities || [];

                            onJobRefreshed(
                                "Scheduler was unable to refresh entity data for job " + job.name + ": " +
                                err + " - reusing previous contents (" + job.entities.length + " entities)"
                            );
                        });
                    });
                }
                else {
                    // We serialize the requests to the library so that it doesn't get overwhelmed.
                    const refreshNextJob = function (error) {
                        if (error) {
                            logger.error(error);
                            errors.push(error);
                        }

                        i++;
                        if (i >= jobs.length) {
                            if (errors.length > 0)
                                reject(errors)
                            else
                                resolve();
                        }
                        else if (!jobs[i].requestOptions) {
                            // Only entity-based jobs need to be refreshed. USQL jobs are skipped.
                            refreshNextJob();
                        }
                        else {
                            library
                            .request(jobs[i].requestOptions)
                            .then(listOrGraph => {
                                // Adopt the refreshed list of entities as provided by the library.
                                jobs[i].entities = !listOrGraph
                                                 ? []
                                                 : listOrGraph.graph
                                                 ? (listOrGraph.graph.nodes || []).sort(jobSorter)
                                                 : listOrGraph.sort(jobSorter);
                                jobs[i].linkages = !listOrGraph
                                                 ? null
                                                 : listOrGraph.graph
                                                 ? listOrGraph.graph.edges
                                                 : null;
                                refreshNextJob();
                            })
                            .catch(err => {
                                // Fallback: reuse previous list of entities (ensure it's an array).
                                jobs[i].entities = jobs[i].entities || [];

                                refreshNextJob(
                                    "Scheduler was unable to refresh entity data for job " + jobs[i].name + ": " +
                                    err + " - reusing previous contents (" + jobs[i].entities.length + " entities)"
                                );
                            });
                        }
                    }

                    // Kick off a refresh request for the first job in the list.
                    i = -1;
                    refreshNextJob();
                }
            }
            catch (ex) {
                logger.error(ex.message);
                errors.push(ex.message);
            }
        });
    }

    function divideJob(jobs, job, count) {
        if (!job || !job.entities) return tuningInfo.maxJobSize;

        logger.info("Splitting job " + job.name + " up into " + count + " parts");

        let i = 0;
        let chunk = Math.ceil(job.entities.length / count);

        while (job.entities.length) {
            let subJob = Object.assign({}, job);    // This is to make sure we don't miss anything.
            subJob.id       = job.id   + "." + i;
            subJob.name     = job.name + "." + i;
            subJob.entities = job.entities.splice(0, chunk);
            subJob.requests = null;

            jobs.push(subJob);
            i++;
        }
        return chunk;
    }

    function resizeJobs(jobs, otherJobsCount, capacity) {
        if (jobs.length === 0) return;

        /*  First, we make the jobs small enough to not exceed the maximum job size.
            We do this by splitting up the largest too-large job into even-sized
            sub-jobs that are each smaller than the batch size limit.

            Note: Jobs that have linkages must not be split - the worker will have to do that.
            TODO: There may be an explicit marker that says: don't split this job. The worker
                  will handle it. This maybe useful when a job's output HAS to be produced in
                  one step (into a single file, for instance).
         */
        const jobsWithLinkages = [];
        for (let i=0; i < jobs.length; /* no inc */) {
            if (!!jobs[i].linkages || jobs[i].keepasone)
                jobsWithLinkages.push(jobs.splice(i, 1)[0]);
            else
                i++;
        }
        jobs.sort((jobA, jobB) => jobB.entities.length - jobA.entities.length);

        while (jobs.length > 0 && jobs[0].entities.length > tuningInfo.maxJobSize) {
            let job = jobs.shift();     // We are sorted, so this removes the biggest one.

            divideJob(jobs, job, Math.ceil(job.entities.length / tuningInfo.maxJobSize)); // How many chunks.

            jobs.sort((jobA, jobB) => jobB.entities.length - jobA.entities.length);
        }

        /*  Next, we create at least as many jobs as the number of workers we think we need.
            At some point we need to stop splitting because the jobs are surely small enough.
            We thus stop at a minimum size of tuningInfo.minJobSize.
         */
        while (jobs.length > 0 &&
               otherJobsCount + jobs.length < capacity &&
               jobs[0].entities.length > tuningInfo.minJobSize) {
            let job = jobs.shift();     // Take the biggest one.

            divideJob(jobs, job, 2);    // Split in half.

            jobs.sort((jobA, jobB) => jobB.entities.length - jobA.entities.length);
        }

        // Add the jobs with linkages back in.
        jobs.splice(0, 0, ...jobsWithLinkages);
    }

    function resolveJobTimeframes(jobs, cycleInfo) {
        return new Promise((resolve, reject) => {
            // We cannot use 'map' because any HTTP requests would be async.
            const jobTimeframeRequestors = jobs.map(job => {
                job.cycle   = cycleInfo.cycle;

                return new Promise((resolve, reject) => {
                    if (job.timeEndpoint) {
                        axios.get(job.timeEndpoint + "?tenant=" + job.source + "&job=" + job.name)
                        .then(response => {
                            const timeframes = response.data;

                            // Create as many job copies as there are timeframes, and resolve with them.
                            resolve(timeframes.map(timeframe => {
                                const newJob = JSON.parse(JSON.stringify(job));
                                newJob.startTime = timeframe.startTime || timeframe.endTime - timespans[job.granularity];
                                newJob.endTime   = timeframe.endTime;
                                return newJob;
                            }));
                        })
                        .catch(err => {
                            logger.warn("Could not call " + job.timeEndpoint + " for job " + job.id + ": " + err.message || err);
                            resolve([]);    // If we can't call that endpoint, we don't execute the job.
                        });
                    }
                    else {
                        job.startTime = cycleInfo.endTime - timespans[job.granularity]; 
                        job.endTime   = cycleInfo.endTime;
                        resolve([job]);     // No need to call an endpoint to get the times.
                    }
                });
            });

            // Now go get the jobs their timeframes. Most requestors will resolve immediately, of course.
            Promise
            .all(jobTimeframeRequestors)
            .then(jobLists => resolve((jobLists || []).flat()))
            .catch(err => reject(err));
        });
    }

    async function collectJobs(cycleInfo) {
        const result = {
            jobs:     null,
            capacity: null,
            errors:   null,
            sizes:    null
        };

        let capacity = tuningInfo.capacities[cycleInfo.schedule] || 1;

        let scheduledJobs = granularities
            // First we get the granularities whose 'buckets' match this minute.
            .filter(granularity => {
                return cycleInfo.minutes % granularity.bucket === 0;
            })
            // Next we reduce that list to a single array that contains the jobs from
            // the corresponding schedules.
            .reduce((jobs, granularity) => {
                return jobs.concat(currentSchedules[granularity.value]);
            }, [])
            // And finally we leave out everything that is disabled.
            .filter(job => job && job.destination != "None");

        let entityJobs = scheduledJobs.filter(job =>  isEntityJob(job));
        let otherJobs  = scheduledJobs.filter(job => !isEntityJob(job));

        logger.info("This cycle runs the following jobs: " + otherJobs.concat(entityJobs).map(job => job.name).join(", "));

        // Get the latest entity versions from the library.
        try {
            await refreshJobs(entityJobs);
        }
        catch (errors) {
            // Take note of any errors, but continue.
            result.errors = errors;
            logger.error("Refreshing the " + entityJobs.length + " entity jobs produced " + errors.length + " errors (details logged below)");
        }

        // We log this primarily to obtain a timestamp to see how long it took to refresh all jobs.
        // logger.info("The entity jobs contain " + entityJobs.map(job => job.entities.length).reduce((a, b) => a + b, 0) + " entity references");

        // Make sure the jobs are properly sized to efficiently run across the workers.
        resizeJobs(entityJobs, otherJobs.length, capacity);
         
        /*  Each job in the collection of jobs for this schedule gets an end time so that
            there is no confusion when a worker executes it: The granularity determines how
            far back in time the job reaches in its collection of metric values, but the
            end time determines when that timespan ends.
            NEW: If the job contains a reference to a REST API at which timespans for data
            collection should be obtained, call that job and create as many subjobs as needed.
            Note that each job is complete in its own right. WHat means taht each job has all
            the entities in it. That's quite a bit of duplication, but otherwise we cannot
            effectively distribute the jobs across the available workers. So, we're going to
            have to accept the dduplications.
         */
        scheduledJobs = otherJobs.concat(entityJobs);
        scheduledJobs.forEach(job => {
            job.cycle   = cycleInfo.cycle;
            job.endTime = cycleInfo.endTime;
        });

        try {
            const jobs = await resolveJobTimeframes(scheduledJobs, cycleInfo);
            result.capacity = Math.min(jobs.length, capacity);
            result.jobs     = jobs;
            result.sizes    = entityJobs.filter(job => !!job.entities).map(job => job.entities.length);

            logger.info("After adjusting sizes and accounting for timeframes, this amounts to " + jobs.length + " jobs");
        }
        catch (ex) {
            logger.error("Could not resolve the job timeframes: " + ex.message);
        }

        return result;    
    }

    function splitJob(jobId) {
        let job = null;
        let schedule = null;

        for (var i=0; !job && i < granularities.length; i++) {
            schedule = currentSchedules[granularities[i].value] || [];
            for (var j=0; !job && j < schedule.length; j++) {
                if (schedule[j].id == jobId) {
                    job = schedule.splice(j, 1)[0];
                    break;
                }
            }
        }

        const newJobSize = divideJob(schedule, job, 2);    // Split in half.

        if (tuningInfo.maxJobSize > newJobSize) {
            tuningInfo.maxJobSize = newJobSize;
            updateTuningInfo();         // Save this new insight...
        }
    }

    function dingJob(jobId, cycleInfo) {
        // TODO: Keep track of how often a job fails. We may have to do something, eventually.
        // job.errors++;  // Track failure count. Exceeding a threshold should trigger some remedial action.
        cycleInfo.errors++;
    }

    // If we got jobs on instantiation, initialize them now. We'll rinse and repeat every
    // time we receive and updated jobs list.
    if (jobs)
        init(jobs);

    return {
        get jobs() {
            return jobs;
        },
        set jobs (jobList) {
            init(jobList);
        },
        get libraries() {
            return library.instances;
        },
        get schedules() {
            function getSummary(schedules) {
                const summarizedSchedule = {};

                for (let granularity in schedules) {
                    summarizedSchedule[granularity] = schedules[granularity].map(scheduleEntry => {
                        return {
                            name: scheduleEntry.name,
                            type: scheduleEntry.type,
                            source: scheduleEntry.source,
                            destination: scheduleEntry.destination
                        };
                    })
                };
                return summarizedSchedule;
            }
            return {
                latency: metricLatency,
                labels: granularities.reduce((map, granularity) => {
                    if (granularity.dtSupported)    // Skip timeframes that Dynatrace does not support.
                        map[granularity.value] = granularity.label;
                    return map;
                }, {}),
                configured: getSummary(originalSchedules),
                current:    getSummary(currentSchedules)
            };
        },
        set schedules(newSchedules) {
            logger.info("Updating schedules as instructed ");
            for (let granularity in currentSchedules) {
                currentSchedules[granularity] = [];

                for (let i=0; i < jobs.length; i++) {
                    if (newSchedules[granularity].includes(jobs[i].name))
                        currentSchedules[granularity].push(jobs[i]);
                }
                const jobNames = currentSchedules[granularity].map(job => job.name);
                if (jobNames.length > 0)
                    logger.info("    Every " + getGranularityName(granularity) + ":\t " + jobNames.join(", "));
            }
        },
        set config (newConfig) {
            config = newConfig;
            init(jobs);
        },
        get tuning () {
            return tuningInfo;
        },
        set tuning (info) {
            tuningInfo = info;
            updateTuningInfo();
        },
        getCycleInfo: function (tick) {
            return activeCycles[tick];
        },
        getDummyCycle: function (tick, progress, error) {
            const cycleInfo = {
                progress:  progress,
                timestamp: (new Date()).getTime(),
                cycle:     tick,
                id:        "Cycle_" + tick,
                name:      "Cycle at " + new Date(tick).toLocaleTimeString(),
                log:     {
                    planned: tick,
                    errors: error? [ error ] : null
                },
                worker: "Scheduler",
                player: "Controller"
            };
            cycleInfo.log[progress] = (new Date()).getTime();

            return cycleInfo;
        },
        getGranularityName:  getGranularityName,
        getGranularityTime:  getGranularityTime,
        getGranularityValue: getGranularityValue,
        getNextGranularity:  getNextGranularity,
        refreshJob:      refreshJob,
        collectJobs:     collectJobs,
        splitJob:        splitJob,
        dingJob:         dingJob,
        createNewCycle:  createNewCycle,
        evaluateCycle:   evaluateCycle,
        getLastCycle:    getLastCycle
    };
}

module.exports = Scheduler;