const ServiceProxy = require('./ServiceProxy.js');

const ServiceBroker = function (name, config, socket, logger, autoRestart) {
    const self             = this;
    const identity         = name;
    const type             = "Library";
    const serviceInstances = [];
    const servicePerTenant = {};

    // Event emitting.
    self.events  = {};

    const subscribe = function (eventName, func, once) {
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push({ func: func, once: once });
    };

    const emit = function (eventName, data) {
        const eventHandlers = self.events[eventName] || [];
        eventHandlers.forEach(handler => {
            if (handler.once && handler.called) return;
            handler.func.call(null, data);
            handler.called = true;
        });
    };

    // Create individual config objects - one for each library.
    const serviceConfigs = Object
        .values(config.Libraries)
        .map(library => {
            const clonedConfig = JSON.parse(JSON.stringify(config));
            delete clonedConfig.Libraries;
            clonedConfig.library = library;
            return clonedConfig;
        });

    serviceConfigs.forEach(config => {
        logger.info("    Library '" + config.library.name + "' handles " + config.library.tenants.join(", "));

        const serviceProxy = new ServiceProxy(config.library.name, type, config, socket, logger, autoRestart);

        serviceProxy.once("ready", function (serviceInstance) {
            serviceInstances.push(serviceInstance);         // To get a full list.

            config.library.tenants.forEach(tenant => {
                servicePerTenant[tenant] = serviceInstance; // For quick access.
            });

            if (serviceInstances.length === serviceConfigs.length)
                emit("ready", asyncIF);
        });
        serviceProxy.on("died", instance => {
            const i = serviceInstances.indexOf(instance);
            if (i > -1) serviceInstances.splice(i, 1);
        });
    });

    const asyncIF = {
        get id() {
            return identity;
        },
        get instances() {
            return serviceInstances;
        },
        set config(value) {
            // We basically need to re-establish the relationships between tenants and instances.
            logger.error("Setting or changing the configuration for the Library without restarting is currently not supported");
            return;
        },
        request: function (value) {
            if (servicePerTenant[value.params.tenantKey] && servicePerTenant[value.params.tenantKey].isAlive)
                return servicePerTenant[value.params.tenantKey].request(value);
            else
                return new Promise((resolve, reject) => {
                    reject("No library instance is available for " + value.params.tenantKey);
            });
        },
        retire: function (onRetired) {
            if (serviceInstances.length === 0)
                setTimeout(onRetired, 1);
            else
                serviceInstances.forEach(instance => {
                    instance.retire(() => {
                        logger.info("Library " + instance.id + " has been retired");
                        
                        const i = serviceInstances.indexOf(instance);
                        if (i > -1) serviceInstances.splice(i, 1);
                        if (serviceInstances.length === 0) onRetired();
                    });
                });
        },
        get isAlive() {
            return serviceInstances.some(instance => instance.isAlive);
        },
        getHealthReport: function () {
            return serviceInstances.map(instance => instance.getHealthReport());
        },
        getHealthMetrics: function () {
            return serviceInstances.map(instance => instance.getHealthMetrics());
        }
    };

    return {
        on: function (evt, handler) {
            subscribe(evt, handler);
            return this;    // For chaining.
        },
        once: function (evt, handler) {
            subscribe(evt, handler, true);
            return this;    // For chaining.
        },
        retire: asyncIF.retire
    };
}

module.exports = ServiceBroker;
