const { fork } = require('child_process');
const http     = require('http');

const ServiceProxy = function (name, type, config, socket, logger, autoRestart) {
    const self           = this;
    const identity       = name;    // + "_" + ServiceProxy.idGenerator++;
    const servicePath    = process.cwd() + "/LibraryService.js";
    const restartDelay   = 1000;    // Wait a second before restarting.

    let serviceConfig    = config;
    let serviceProcess   = null;
    let requestIdGenerator = 0;

    // Both 'launch' and 'mode' are supported. Documentation uses 'mode'.
    const mode = serviceConfig.library.mode || serviceConfig.library.launch;  
       
    // We only support auto-restart if we're launching the service ourselves.
    autoRestart = mode !== "spawn" && mode !== "auto" ? false : !!autoRestart;

    // If we don't have a socket, use a dummy.
    socket = socket || { emit: function () {} };

    self.events  = {};

    const subscribe = function (eventName, func, once) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push({ func: func, once: once });
    };

    const emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const eventHandlers = self.events[eventName] || [];
        eventHandlers.forEach(handler => {
            if (handler.once && handler.called) return;

            handler.func.call(null, data);
            handler.called = true;
        });
    };

    let promisesMade = {};
    let requestTimeout = (serviceConfig.library.timeout || 30) * 1000; // 30 seconds...
    let isReady      = false;
    let isDead       = false;
    let isDying      = false;
    let isRetiring   = false;
    let isPreparing  = true;
    let killTimer    = null;
    let onRetired    = null;

    function scheduleForcedKill() {
        killTimer = setTimeout(function () {
            logger.info("Terminating service " + identity);
            serviceProcess.kill();
        }, 3 * 60000);
    }
    function cancelForcedKill() {
        clearTimeout(killTimer);
        killTimer = null;
    }
    function onServiceDied(code, signal) {
        // If onRetired has a value, we should call that and stay dead.
        if (autoRestart && !onRetired) {
            // Dead, but to be ressurected...!
            setTimeout(() => {
                serviceProcess = startService();
                },  restartDelay);

            return;
        }
        else if (isDead) {
            // We've been declared dead already - no need to do anything.
        }
        else {
            isDead = true;

            logger.info("Service " + identity + " died");
            emit("died", identity);

            // Also tell any browsers that are listening we're dead now.
            socket.emit('died', identity);

            cancelForcedKill();
        }
        if (onRetired)
            onRetired();
    }
    function cancelRequest(requestId) {
        let promise = promisesMade[requestId];
        if (promise) {
            delete promisesMade[requestId];

            socket.emit('request', {
                status:    "failed",
                id:        requestId,
                code:      500,
                error:     "Request timed out",
                requests:  Object.keys(promisesMade).length,
                requested: promise.timestamp,
                timestamp: new Date().getTime(),
                worker:  {
                    id:   identity,
                    name: identity,
                    type: "Library"
                },
                player:    "Controller"
            });

            promise.reject("Request timed out");
        }
    }

    function startService() {
        const service = mode === "embed" || mode === "embedded"
                      ? embedService()
                      : mode === "spawn" || mode === "auto"
                      ? spawnService()
                      : findService();   // mode === "find"

        service.on("message", message => {
            let promise = null;

            switch (message.status) {
                case "preparing":
                    if (serviceConfig)
                        service.send({
                            command: "configure",
                            config: serviceConfig
                        });

                    isPreparing = false;

                    // Also tell any browsers that are listening.
                    socket.emit('born', identity);
                    break;

                case "unavailable":
                    isReady = false;
                    logger.warn("Connection with " + identity + " not available: " + message.error);
                    socket.emit('down', identity);
                    break;

                case "ready":
                    isReady = true;
                    emit("ready", asyncIF);
                    socket.emit('ready', identity);
                    break;

                case "health":
                    promise = promisesMade[message.id];
                    if (promise) {
                        delete promisesMade[message.id];

                        clearTimeout(promise.timer);

                        if (message.headers['Content-Type'] == "application/json")
                            message.response = JSON.parse(message.response);

                        if (message.code !== 200 || message.error || message.response.error)
                            promise.reject(message.code + ": " + (message.error || message.response.error));
                        else
                            promise.resolve(message.response);
                    }
                    break;

                case "response":
                    promise = promisesMade[message.id];
                    if (promise) {
                        delete promisesMade[message.id];

                        clearTimeout(promise.timer);

                        if (message.headers['Content-Type'] == "application/json")
                            message.response = JSON.parse(message.response);

                        if (message.code !== 200 || message.error || message.response.error)
                            promise.reject(message.code + ": " + (message.error || message.response.error));
                        else
                            promise.resolve(message.response);
                    }

                    socket.emit('request', {
                        status:    message.response && message.response.error ? "failed": "done",
                        id:        "Request_" + message.id,
                        name:      message.request.name,
                        code:      message.code,
                        error:     message.response && message.response.error ? message.response.error : message.error,
                        headers:   JSON.stringify(message.headers),
                        request:   JSON.stringify(message.request),
                        result:    message.response ? message.response.length : 0,
                        requested: promise ? promise.timestamp : new Date().getTime(),
                        completed: new Date().getTime(),
                        worker: {
                            id:   identity,
                            name: identity,
                            type: "Library"
                        },
                        player: "Controller"
                    });
                    break;

                case "error":
                    logger.warn("Killing " + identity + " on error: " + message.error);
                    service.kill();
                    break;

                case "progress":
                    // Tell the diagnostics UI.
                    message.processID = identity;
                    socket.emit("progress", message);
                    break;

                case "html":
                    // Tell the diagnostics UI.
                    message.processID = identity;
                    socket.emit("html", message);
                    break;
            }
        });
        return service;
    }

    function findService() {
        /*  Failover algorithm:
            0)  isFailoverAvailable = isFailover defined

            1)  Ping primary periodically
                Response:
                    - Fail:
                        If isFailoverAvailable
                            connectedService = failoverService
                            // Wait for the next ping to there.
                        else
                            if connectedService === failoverService
                                isFailoverAvailable = false
                                connectedService = primaryService

                            emit "UNAVAILABLE"
                    - Back to primary:
                        connectedService = primaryService
                        // Wait for the next ping to there.
                    - Success:
                        Do the usual. BTW: Don't emit 'ready' or 'preparing twice.

         */
        const primaryService = {
            host: serviceConfig.library.host,
            port: serviceConfig.library.port
        };
        const failoverService = serviceConfig.library.failover
            ? {
                host: serviceConfig.library.failover.hostIP,
                port: serviceConfig.library.failover.port
            }
            : null;

        let inFailoverMode      = false;
        let isFailoverAvailable = !!failoverService;

        let connectedService = primaryService;
        let pinger = null;

        function pingService(self, service) {
            service = service || connectedService;

            function tryFailover(error) {
                if (!inFailoverMode && isFailoverAvailable) {
                    connectedService = failoverService
                    inFailoverMode = true;
                    // Wait for the next ping to there.
                }
                else if (connectedService === failoverService) {
                    isFailoverAvailable = false;
                    connectedService = primaryService;
                    inFailoverMode = false;
                    // Wait for the next ping to there.
                }
                else {
                    self.emit("message", { status: "unavailable", error: error });
                }
            }

            const ping = http.request({
                    hostname: service.host,
                    port:     service.port,
                    path:     '/status',
                    method:   'GET'
                },  pong => {
                pong.setEncoding('utf8');

                var data = [];

                pong.on('data', chunk => {
                    data.push(chunk);
                });
                pong.on('end', () => {
                    const response = data.join("");
                    switch (response) {
                        case "alive":
                            // Primary is back up. Stop assuming the failover service
                            // is unavailable - in case that was previously detected.
                            isFailoverAvailable = !!failoverService;
                            self.emit("message", { status: "ready" });
                            break;
                        case "preparing":
                            self.emit("message", { status: "preparing" });
                            break;
                        case "standby":
                            // Only the failover instance tells us this:
                            // "I've gone back to standby mode because the primary instance is back up".
                            connectedService = primaryService;
                            inFailoverMode = false;
                            break;
                        default:
                            // Unavailable.
                            tryFailover(response);
                            break;
                    }
                });
            });
            ping.on('error', err => {
                tryFailover(err.message || err);
            });
            ping.on('abort', err => {
                tryFailover(err.message || err);
            });

            ping.end();  // Required to actually send the request.
        }

        const service = {
            events: {},
            on:   function(eventName, fn) {
                this.events[eventName] = self.events[eventName] || [];
                this.events[eventName].push(fn);
            },
            emit: function (eventName, data) {
                const eventHandlers = this.events[eventName] || [];
                eventHandlers.forEach(fn => { fn.call(null, data); });
            },
            kill: function() {},        // We'll never kill the standalone service.
            send: function(message) {
                const self = this;

                switch (message.command) {
                    case "configure":
                        // Ignore. Standalone process is responsible for its own config.
                        break;

                    case "health":
                    case "request":
                        let method = "GET";
                        let query  = [];
                        let body   = null;
                        let name   = message.id;

                        if (message.request) {
                            name   = message.request.name;
                            body   = message.request.body;
                            method = message.request.body ? "POST" : "GET";

                            for (let param in message.request.query) {
                                const value = Array.isArray(message.request.query[param])
                                            ? message.request.query[param].join(",")
                                            : message.request.query[param];
                                query.push(param + "=" + value);
                            }
                        }

                        const status = message.command == "health"
                                    ? "health"
                                    : "response";
                        const path  = message.command == "health"
                                    ? "/api/health?" + query.join("&")
                                    : "/library/" + message.request.params.tenantKey +
                                        "/" + message.request.params.entityType +
                                        "?" + query.join("&");

                        const respond = (message, result, data, error) => {
                            self.emit("message", {
                                status:   status,
                                id:       message.id,
                                name:     name,
                                code:     !result || error ? 500 : result.statusCode,
                                error:    error,
                                headers:  {},
                                request:  message.request,
                                response: data
                            });
                        };

                        const request = http.request({
                            hostname: connectedService.host,
                            port:     connectedService.port,
                            path:     path,
                            method:   method,
                            timeout:  requestTimeout
                        }, 
                        result => {
                            result.setEncoding('utf8');

                            var data = [];

                            result.on('data', chunk => {
                                data.push(chunk);
                            });
                            result.on('end', () => {
                                try {
                                    var parsedData = JSON.parse(data.join(""));
                                    if (parsedData.error)
                                        throw new Error(parsedData.error);

                                    respond(message, result, parsedData, null);
                                }
                                catch (ex) {
                                    respond(message, result, parsedData, identity + " says: " + ex.message + ". Response: " + data.join("").substring(0, 100));
                                }
                            });
                        });
                        request.on('error', err => {
                            respond(message, null, null, "Request to " + identity + " has error: " + err);
                        });
                        request.on('abort', err => {
                            respond(message, null, null, "Request to " + identity + " was aborted: " + err);
                        });

                        if (body) {
                            request.write(body);
                        }

                        request.end();  // Required to actually send the request.                        
                        break;

                    case "retire":
                        // Ignore. We can't just stop the standalone process.
                        break;
                }
            }
        };

        // Start (re-)finding the library and emit 'preparing' and 'ready' messages.
        setTimeout(function (self) {
            self.emit("message", { status: "preparing" });

            pingService(self);
            pinger = setInterval(pingService, 10*1000, self);
        }, 10, service);  // 10ms

        return service;
    }

    function spawnService() {
        const debugEnabled = serviceConfig.library.debug && (typeof v8debug === 'object' || /--debug|--inspect/.test(process.execArgv.join(' ')));
        const service      = fork(servicePath, [ "name="+name, "mode=auto" ], { silent: true, stdio: 'pipe', execArgv: debugEnabled ? [ "--inspect" ] : [] });

        /*  If the process cannot be created, we call 'onServiceDied()' asynchronously, so that
            callers have time to subscribe to the 'died' event.
         */
        if (!service) {
            logger.info("Could not start " + identity);
            setTimeout(onServiceDied, 1);
            return null;
        }

        service.on("exit",  onServiceDied);

        /*  We use the process' stdout to capture its log output. But in exceptional cases stdout
            is not available. That signals a very bad situation, so in that case, we'll kill this
            process immediately yet asynchronously, so that callers have time to subscribe to the
            'died' event.
         */
        if (!service.stdout) {
            logger.info("No stdout - Killing " + identity + " on creation");
            setTimeout(service.kill, 1);
            return null;
        }

        /*  The child process should have its own logging. But if it hasn't, it's
            output can be intercepted here.
         */
        service.stdout.on('data', data => {
        //  logger.info( "    Output from " + identity + ": " + ("" + data).trim());
        });
        service.stderr.on('data', data => {
            logger.error("    Error from "  + identity + ": " + ("" + data).trim());
        });

        service.on("disconnect", () => {
            isDying = true;

            if (autoRestart)
                logger.warn("Restarting " + identity + " on disconnect");
            else
                logger.warn("Killing " + identity + " on disconnect");

            service.kill();
        });
        service.on("error", err => {
            if (autoRestart)
                logger.warn("Restarting " + identity + " on error: " + err);
            else
                logger.warn("Killing " + identity + " on error: " + err);

            socket.emit('retiring', identity);

            service.kill();
        });

        return service;
    }

    function embedService() {
        const LibraryService = require('./LibraryService.js');
        const libraryService = new LibraryService();

        const service = {
            on:   libraryService.on.bind(libraryService),
            kill: libraryService.stop.bind(libraryService),
            send: function (data) {
                libraryService.send("message", data);
            },
            resourceUsage: function () {
                const library = libraryService.getLibrary();
                return library ? library.resourceUsage() : "(N/A)";
            }
        };
        service.on("exit", onServiceDied);

        return service;
    }

    serviceProcess = startService();

    const asyncIF = {
        get id() {
            return identity;
        },
        get identity() {
            return {
                id:   identity,
                name: identity,
                type: type
            }
        },
        set config(value) {
            serviceConfig = value;
            if (!isPreparing)
                serviceProcess.send({
                    command: "configure",
                    config: serviceConfig
                });
        },
        request: function (value) {
            return new Promise((resolve, reject) => {
                if (!isReady) {
                    reject("Service is not ready to handle request");
                    return;
                }

                const requestId = requestIdGenerator++;
                promisesMade[requestId] = {
                    resolve: resolve,
                    reject:  reject,
                    timestamp: new Date().getTime(),
                    timer:     setTimeout(cancelRequest, requestTimeout+1000, requestId)
                };

                // The value must be an object that contains the query parameters.
                serviceProcess.send({
                    command: "request",
                    id:      requestId,
                    request: value
                });

                // Also tell any browsers that are listening.
                socket.emit('request', {
                    status:    "started",
                    id:        "Request_" + requestId,
                    name:      value.name,
                    request:   JSON.stringify(value),
                    requested: promisesMade[requestId].timestamp,
                    worker: {
                        id:   identity,
                        name: identity,
                        type: "Library"
                    },
                    player:    "Controller"
                });
            });
        },
        retire: function (onRetiredCallback) {
            onRetired = onRetiredCallback;

            if (isDead) {
                if (onRetired) setTimeout(onRetired, 1);
                return;
            }
            logger.warn("Asking " + identity + " to retire");

            serviceProcess.send({
                command: "retire"
            });
            isRetiring = true;

            // Also tell any browsers that are listening.
            socket.emit('stopping', identity);

            scheduleForcedKill();  // Backup plan...
        },
        get isAlive() {
            return !isRetiring && !isDying && !isDead;
        },
        getHealthReport: function () {
            // TODO: Fix this.
            return serviceProcess.getHealthReport ? serviceProcess.getHealthReport() : "(N/A)";
        },
        getHealthMetrics: function () {
            return new Promise((resolve, reject) => {
                if (!isReady) {
                    reject("Service is not ready to handle request");
                    return;
                }

                if (serviceProcess.getHealthMetrics) {
                    resolve(serviceProcess.getHealthMetrics());
                    return;
                }

                const requestId = requestIdGenerator++;
                promisesMade[requestId] = {
                    resolve: resolve,
                    reject:  reject,
                    timestamp: new Date().getTime(),
                    timer:     setTimeout(cancelRequest, requestTimeout+1000, requestId)
                };

                serviceProcess.send({
                    command: "health",
                    id:      requestId,
                    request: null
                });
            });
        }
    };

    /*  This chainable sync interface will be returned to the caller when a new instance is created.
        The 'ready' event will provide the async interface that contains all regular APIs.
     */
    const syncIF = {
        on: function (evt, handler) {
            subscribe(evt, handler);
            return this;    // For chaining.
        },
        once: function (evt, handler) {
            subscribe(evt, handler, true);
            return this;    // For chaining.
        },
        retire: asyncIF.retire
    };

    return syncIF;
};
ServiceProxy.idGenerator = 0;

module.exports = ServiceProxy;
