// eslint.enable

const fs      = require('fs');
const crypto  = require('crypto');
const selfsigned  = require('selfsigned');
const cleanup = require('node-cleanup');
const express = require('express');
const session = require('express-session');
const webui   = express();
const https   = require('https');

// HTTPS provisions.
let sslKey  = null;
let sslCert = null;

if (fs.existsSync(__dirname + '/../data/generated.crt')) {
    sslKey  = fs.readFileSync(__dirname + '/../data/generated.key', { encoding: 'utf8' });
    sslCert = fs.readFileSync(__dirname + '/../data/generated.crt', { encoding: 'utf8' });
}
else {
    const signedCert = selfsigned.generate(
        [{ name: 'commonName', value: 'services.dynatrace.corp' }], 
        { days: 365*2 }
    );

    sslKey  = signedCert.private;
    sslCert = signedCert.cert;

    fs.writeFileSync(__dirname + '/../data/generated.key', sslKey);
    fs.writeFileSync(__dirname + '/../data/generated.crt', sslCert);
}

// Create an HTTPS server that can fallback to HTTP.
const server  = https.createServer({
    key:  sslKey,
    cert: sslCert
}, webui);

const socket  = require('socket.io')(server);
const cookies = require('cookie-parser');
const winston = require('winston');
const { authentication, authorize } = require("@dt-esa/authorizer");

const WorkerProxy   = require('./WorkerProxy.js');
const Controller    = require('./Controller.js');
const DynatraceUI   = require('./DynatraceUI.js');
const OpenAPIHelper = require('./OpenAPIHelper.js');
const ConfigReader  = require('./ConfigReader.js');



const parseQueryString = data =>
    data.split("&")
        .reduce((params, param) => {
            const paramParts = param.split("=");
            params[paramParts.shift()] = decodeURIComponent(paramParts.join("="));
            return params;
        }, {});

const loggerTransports = {
    console: new winston.transports.Console({ level: 'silly' }),
    file:    new winston.transports.File({ filename: '../log/StreamerService.log', maxsize: 1048576, maxFiles: 10 }),
    silent:  new winston.transports.Console({ level: 'error', silent: true })
};

const controllerLogger = winston.createLogger({
    format: winston.format.combine(
        winston.format.label({ label: 'Controller' }),
        winston.format.timestamp({ format: 'YYYY-MM-DD HH:mm:ss.SSS' }),
        winston.format.printf(({ level, message, label, timestamp }) => {
            return `${timestamp} [${label}] ${level}: ${message}`;
        })
    ),
    transports: [ loggerTransports.silent ]  // We may add transports later, as the config specifies.
});

const contentType  = 'application/json';
const debugEnabled = typeof v8debug === 'object' || /--debug|--inspect/.test(process.execArgv.join(' '));
let jobsFileName   = "jobs.json";
let configReader   = null;
let openAPIHelper  = null;
let controller     = null;
let listener       = null;
let config         = null;

// For failover.
const excludedJobProperties = [ "cycle", "endTime", "assigned", "firstSeenTimestamp", "lastSeenTimestamp" ];

const primaryDownMax     = 5;
let primaryDownCount     = 0;
let jobsFromPrimary      = null;
let schedulesFromPrimary = null;
let tuningFromPrimary    = null;
let configSentToFailover = null;
let jobsSentToFailover   = null;
let schedSentToFailover  = null;
let tuningSentToFailover = null;

// This is called on a timer if we are the failover instance.
// It serves to sync with the primary instance and to detect
// when we should take over from the primary.
function checkHealth(failoverConfig) {
    const address = "http://" + failoverConfig.primaryIP + ":" + failoverConfig.primaryPort;

    function handleResponse(isFailure, answer)  {
        if (isFailure) {
            // Let's count successive failures. We take over when beyond primaryDownMax...
            primaryDownCount++;

            // If we already have taken over, our continuing failures don't need to be handled.
            if (controller && !controller.isRunning) {
                controllerLogger.warn("StreamerService failover instance could not connect to " + address +
                    " " + primaryDownCount + " times (" + (primaryDownMax - primaryDownCount) +
                    " attempts left)"
                );
                if (primaryDownCount >= primaryDownMax) {
                    controllerLogger.warn("StreamerService failover instance is taking over from  " + address);
                    controller.start();   // We need to start working!
                }
            }
            else {
                if (!controller && primaryDownCount >= primaryDownMax) {
                    controllerLogger.error("StreamerService failover instance wants to take over from  " + address + " but hasn't received its config yet!");
                }
            }
        }
        else {
            // The primary instance is working (again) - reset the failure counter.
            primaryDownCount = 0;

            // If we had taken over, then it's time to go back to standby mode.
            if (controller && controller.isRunning) {
                controllerLogger.warn("StreamerService failover instance is returning to standby mode");
                controller.stop();       // Stop creating new cycles.
            }

            // Finally adopt any new config, jobs and schedules we just received.
            if (answer) {
                // We are taking the jobs and schedules from the answer, or, if not there,
                // we're keeping any unused values until we've used and cleared them.
                tuningFromPrimary    = answer.tuning    || tuningFromPrimary;
                jobsFromPrimary      = answer.jobs      || jobsFromPrimary;
                schedulesFromPrimary = answer.schedules || schedulesFromPrimary;

                if (answer.config && !controller) {
                    controllerLogger.info("Initial config received - creating Controller...");
                    controller = new Controller(answer.config, socket, controllerLogger, -1);
                    controller.on("ready", message => {
                        controllerLogger.info("StreamerService is ready. " + message);

                        if (jobsFromPrimary) {
                            controller.jobs = jobsFromPrimary;
                            jobsFromPrimary = null;
                        }
                        if (schedulesFromPrimary) {
                            controller.scheduler.schedules = schedulesFromPrimary;
                            schedulesFromPrimary = null;
                        }
                        if (tuningFromPrimary) {
                            controller.tuningInfo = tuningFromPrimary;
                            tuningFromPrimary = null;
                        }
                    });
                }
                else if (controller) {
                    if (answer.config) {
                        controller.config = answer.config;
                    }
                    if (jobsFromPrimary) {
                        controller.jobs   = jobsFromPrimary;
                        jobsFromPrimary   = null;
                    }
                    if (schedulesFromPrimary && controller.scheduler) {
                        controller.scheduler.schedules = schedulesFromPrimary;
                        schedulesFromPrimary = null;
                    }
                    if (tuningFromPrimary) {
                        controller.tuningInfo = tuningFromPrimary;
                        tuningFromPrimary = null;
                    }
                }
                else {
                    // No controller yet. If we have schedules or jobs in this answer,
                    // we'll keep them around until we get a config so that we can
                    // instantiate the controller. And then we'll use them.
                }
            }
        }
    }

    try {
        // We ask for the primary instance's status to test for availability.
        // Periodically, we should ask for everything. Or maybe only on the first request.
        const request = http.get(address + "/status?secret=" + failoverConfig.secret, response => {
            let data = [];
            response.setEncoding('utf8');

            response.on('data', (chunk) => {
                data.push(chunk);
            });
            response.on('end', () => {
                response.body = JSON.parse(data.join(""));
                handleResponse(response.statusCode !== 200, response.body);
            });
        });
        request.setTimeout(9000, () => {
            request.destroy();
        });
        request.on("abort", () => {
            handleResponse(true);
        });
        request.on("error", () => {
            handleResponse(true);
        });
    }
    catch (ex) {
        handleResponse(true);
    }
}

// Requested by the failover instance (i.e. by the function above).
// Strictly speaking we don't need this if we are running under PM2.
// However, the advantage of doing it ourselves is that the failover 
// instance is always in sync with the primary instance in terms of 
// jobs, schedules and worker capacities.
function healthStatus(request, response) {
    let answer = {};

    try {
        if (controller.isRunning) {
            response.statusCode = 200;
            answer.status = "running";

            const requestingHostIP = request.ip ? request.ip.split(":").pop() : null;
            if (config.failover &&
                requestingHostIP     === config.failover.failoverIP &&
                request.query.secret === config.failover.secret) {

                const schedules = Object.keys(controller.scheduler.schedules.current).reduce((schedules, granularity) => {
                    schedules[granularity] = controller.scheduler.schedules.current[granularity].map(job => job.name);
                    return schedules;
                }, {});

                const stringifiedSched  = JSON.stringify(schedules);
                const stringifiedTuning = JSON.stringify(controller.tuningInfo);
                const stringifiedConfig = JSON.stringify(controller.config);
                const stringifiedJobs   = JSON.stringify(controller.jobs, (key, value) => {
                    if (excludedJobProperties.includes(key))
                        return undefined;
                    if (value === "_empty")
                        return "";
                    return value;
                });

                if (schedSentToFailover  !== stringifiedSched) {
                    schedSentToFailover    = stringifiedSched;
                    answer.schedules = schedules;
                }

                if (tuningSentToFailover !== stringifiedTuning) {
                    tuningSentToFailover   = stringifiedTuning;
                    answer.tuning = controller.tuningInfo;
                }

                if (configSentToFailover !== stringifiedConfig) {
                    configSentToFailover   = stringifiedConfig;
                    answer.config = controller.config;
                }

                if (jobsSentToFailover !== stringifiedJobs) {
                    jobsSentToFailover   = stringifiedJobs;
                    answer.jobs   = controller.jobs;
                }
            }
            else {
                controllerLogger.warn("Health check from " + requestingHostIP + " does not originate from the failover instance");
            }
        }
        else {
            response.statusCode = 404;
            answer.status = "stopped";
        }
    }
    catch (ex) {
        response.statusCode = 500;
        answer.status = "error";
        answer.message = ex.message;
    }
    finally {
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(answer));
        response.end();
    }
}

// Requested by the Dynatrace Solution Server.
function healthMetrics(request, response) {
    controller.getHealthMetrics()
    .then(healthMetrics => {
        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(healthMetrics, undefined, 4));
        response.end();
    })
    .catch(err => {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify({ status: "error", message: err }, undefined, 4));
        response.end();
    });
}

let workerProxy = null;

function executeJob(jobData, capture, response) {
    controllerLogger.info("Starting worker for " + jobData.name + (jobData.entities ? (" containing " + jobData.entities.length + " entities") : ""));

    // Options for 'capture':
    // - report: Run the job as usual and return the execution report (default).
    // - result: Redirect the data to the client instead of to the usual destination.
    // - all:    Redirect the data to the client and include the execution report.
    const whatToCapture = capture;   // Required for closure.
    if (whatToCapture === "result" || whatToCapture === "all")
        jobData.destination = "Console";

    // This is so that we can keep the 'interactive' worker alive for debugging.
    if (!workerProxy || !workerProxy.worker.isAlive || !workerProxy.worker.isReady) {
        workerProxy = new WorkerProxy(config, controller.scheduler.tuning, null, controllerLogger, debugEnabled);
        workerProxy
        .on("ready", worker => {
                worker.assignment = {
                    jobs: [ jobData ],
                    cycle: jobData.cycle
                };
            })
        .on("done", jobStatus => {
            controllerLogger.info(jobStatus.name + " has completed and produced " + (jobStatus.output ? (jobStatus.output.length + " rows") : "no output"));

            if (jobStatus.log.summary.errors && jobStatus.log.summary.errors.length > 0) {
                response.status(500).send({ status: "error", error: jobStatus.log.summary.errors.join("\n") });
            }
            else {
                const output = jobStatus.output;
                delete jobStatus.output;

                let captured = null;
                switch (whatToCapture) {
                    case "all": 
                        captured = {
                            report: jobStatus,
                            result: output
                        }; 
                        break;
                    case "result": 
                        captured = output; 
                        break;
                    case "report": 
                    default:
                        captured = jobStatus;
                        break;
                }

                response.statusCode = 200;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify(captured, undefined, 4));
                response.end();                
            }

            // If we want to keep the 'interactive' worker alive, disable the below two lines.
            workerProxy.retire();
            workerProxy = null;
        })
        .on("died", () => {
            // controllerLogger.info("Worker has died peacefully.");
        });
    }
    else {
        // For debugging convenience: if the worker is still alive, and ready, just assign it the work.
        workerProxy.worker.assignment = {
            jobs: [ jobData ],
            cycle: jobData.cycle
        };
    }
}

try {
    /*  Command line arguments:
        - config=<filename>: Load the config from somewhere else than '../config/config.json'.
        - once: Runs each job once, but only when you click 'Start'.
        - failover: Does not start the library or execute jobs at all until the
            primary instance (as specified in the 'failover' option) goes down.
     */
    const cmdArgs = process.argv.slice(2).reduce((args, arg) => {
        const argParts = arg.split("=");
        args[argParts[0]] = argParts[1];
        return args;
    },{});

    // Everything that follows requires a config, so we do this synchronously.
    configReader = new ConfigReader();
    configReader.on("error", (msg, err) => {
        console.error("Configuration could not be loaded: " + msg + ": " + err);
        process.exit(125);              // PM2 should not try to restart us.
    });
    configReader.on("changed", (section, data) => {
        if (section === jobsFileName) {
            controllerLogger.info("Job list has been updated. Reloading...");
            controller.jobs = data;
        }
        else {
            console.warn("Configuration change detected. This service will restart so that the change can take effect");
            setTimeout(process.exit(0), 1); // Trigger PM2 to restart us.
        }
    });

    config = configReader.load();

    openAPIHelper = new OpenAPIHelper(config);

    // Set up where we will be logging to.
    (config.logging || []).forEach(transport => {
        if (loggerTransports[transport])
            controllerLogger.add(loggerTransports[transport]);
    });

    controllerLogger.info("StreamerService is starting...");

    const startOption = cmdArgs.hasOwnProperty("failover")
                      ? -1
                      : cmdArgs.hasOwnProperty("once")
                      ? 1
                      : 0;

    if (startOption === -1) {
        // Failover mode: periodically check the primary instance's health.
        setInterval(checkHealth, 10000, config.failover);

        controllerLogger.info("This failover instance will receive its config and jobs from the primary instance at " + config.failover.primary);

        // Note: The failover instance will not listen for web requests.
    }
    else {
        // Primary mode.
        controller  = new Controller(config, socket, controllerLogger, startOption);
        controller.on("ready", message => {
            // The Controller is ready to run.
            controller.start();

            try {
                // Read the jobs file and give it to the Controller (exclude extraction jobs).
                // TODO: We need to better handle extraction jobs in the UI.
                controller.jobs = configReader
                    .load(jobsFileName)
                    .filter(job => job.subtype !== "Extraction");

                // And finally, start listening for HTTP requests.
                listener = server.listen(config.streamerPort, () => {
                    controllerLogger.info("StreamerService is listening on port " + config.streamerPort);
                });
                listener.on("error", err => {
                    controllerLogger.error("StreamerService cannot listen on port " + config.streamerPort + ": " + err.message);
                    listener.close();
                });
                controllerLogger.info("    " + message);
                controllerLogger.info("StreamerService is ready.");
            }
            catch (ex) {
                controllerLogger.error("Error reading jobs: " + ex.message);
            }
        });
    }
}
catch (ex) {
    controllerLogger.info("StreamerService could not start: " + ex.message);
    process.exit(126);
}

// User session store.
let sessionSecret;
const sessionFile = __dirname + `/../data/session.secret`;
if (fs.existsSync(sessionFile))
    sessionSecret = fs.readFileSync(sessionFile, { encoding: "utf-8" });
else {
    sessionSecret = crypto.randomBytes(32).toString();
    fs.writeFileSync(sessionFile, sessionSecret);
} 

webui.use(express.json());
webui.use(cookies());
webui.use(session({ secret: sessionSecret, resave: true, saveUninitialized: false }));
webui.use(express.urlencoded({ extended: true }));
webui.use(authentication(config.Authorizations));

const unprotectedRoutes = express.Router();
const protectedRoutes   = express.Router();

// The below entry points are accessible without authentication.
unprotectedRoutes.get("/status",     healthStatus);    // Our 'checkHealth()' function calls this periodically.
unprotectedRoutes.get("/api/health", healthMetrics);   // Called every minute by the Dynatrace Solution Server. 
unprotectedRoutes.use('/api/libraryapi', (req, res) => openAPIHelper.getLibraryAPI(req, res));
unprotectedRoutes.use('/api/configapi',  (req, res) => openAPIHelper.getConfigAPI(req, res));
unprotectedRoutes.use('/doc', express.static("../doc"));    // The documentation files.

// Request for a service provided by a Library. Redirect to appropriate Library instance.
unprotectedRoutes.use("/library/:tenantKey",  function (request, response) {
    const name = Object
        .keys(config.Libraries)
        .find(name => config.Libraries[name].tenants.includes(request.params.tenantKey));
    const libInfo = config.Libraries[name];

    if (!libInfo) {
        response.statusCode = 404;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify({
            error: "No library instance available for " + request.params.tenantKey
        }));
        response.end();
    }
    else {
        // Default HTTP redirect status is 302, but we'll use 307 to tell the browser to use the same HTTP method.
        response.redirect(307,
            request.protocol + "://" +
            (libInfo.host !== "localhost" ? libInfo.host : request.hostname) + 
            ":" + libInfo.port + "/library/" + request.params.tenantKey + request.url
        );
    }
});

// Static files for the UI.
protectedRoutes.use("/", express.static("../client")); 

protectedRoutes.get("/controller/kill/:workerID",  function (request, response) {
    try {
        const killed = controller.killWorker(request.params.workerID);

        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(killed));
        response.end();
    }
    catch (ex) {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(ex));
        response.end();
    }
});

protectedRoutes.get("/controller/:participantType",  function (request, response) {
    try {
        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(controller[request.params.participantType]));
        response.end();
    }
    catch (ex) {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(ex));
        response.end();
    }
});

protectedRoutes.post("/controller/schedules", authorize(["detl-schedules-write"]), function (request, response) {
    try {
        let data = [];
        request.on('data', chunk => {
            data.push(chunk);
        });
        request.on('end', () => {
            try {
                const schedules = JSON.parse(data.join(""));
                controller.scheduler.schedules = schedules;

                response.statusCode = 200;
                response.write(JSON.stringify("OK"));
                response.end();
            }
            catch (ex) {
                response.statusCode = 500;
                response.setHeader('Content-Type', contentType);
                response.write(JSON.stringify(ex));
                response.end();
            }
        });
    }
    catch (ex) {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(ex));
        response.end();
    }
});

// Valid timespans for job executions.
const spans = [ "min", "5mins", "10mins", "15mins", "30mins", "hour", "2hours", "6hours", "day" ];

// Run a job directly and return the results.
protectedRoutes.get('/controller/run/:tenantKey/:jobName', function (request, response) {
    if (!controller.scheduler) {
        response.status(503).send(JSON.stringify({ error: "Please wait until the service is ready" }));
        return;
    }

    const tenant  = request.params.tenantKey;
    const name    = request.params.jobName;
    const span    = request.query.span;
    const end     = new Date();
    const capture = request.query.capture; 

    switch (request.query.end) {
        case "min":
            end.setSeconds(0, 0);   // Start of this minute.
            break;
        case "hour":
            end.setMinutes(0, 0, 0);
            break;
        case "day":
            end.setHours(0, 0, 0, 0);
            break;
        default:
            end.setMinutes(0, 0, 0);
            if (end.getHours() < Number(request.query.end))
                end.setTime(end.getTime() - 24*60*60*1000);
            end.setHours(request.query.end);
            break;
    }

    if (!spans.includes(span)) 
        throw new Error("Invalid granularity");
        
    const jobs = configReader.load(jobsFileName);
    const jobData = jobs.find(job => job.source === tenant && job.name === name);

    jobData.cycle = end.getTime();
    jobData.granularity = "_" + span;            
    jobData.endTime = jobData.cycle - Number(config.jobLatency || 1) * 60000;  
    
    controller.scheduler
        .refreshJob(jobData)
        .then(jobData => {
            executeJob(jobData, capture, response);
        })
        .catch(error => {
            response.status(500).send({ status: "error", error: error });
        });
});
protectedRoutes.post('/controller/run', function (request, response) {
    let data = [];
    request.on('data', chunk => {
        data.push(chunk);
    });
    request.on('end', () => {
        if (!controller.scheduler) {
            response.status(503).send(JSON.stringify({ error: "Please wait until the service is ready" }));
            return;
        }

        const capture = request.query.capture; 
        const jobData = JSON.parse(data.join(""));
        jobData.endTime = jobData.cycle - Number(config.jobLatency || 1) * 60000;

        controller.scheduler
            .refreshJob(jobData)
            .then(jobData => {
                executeJob(jobData, capture, response);
            })
            .catch(error => {
                response.status(500).send({ status: "error", error: error });
            });
    });
});

protectedRoutes.get('/config.json', function (request, response) {
    const configCopy = JSON.parse(JSON.stringify(config));
    delete configCopy.failover; // Let's not send the failover secret.

    response.statusCode = 200;
    response.setHeader('Content-Type', contentType);
    response.write(JSON.stringify(configCopy));
    response.end();
});

protectedRoutes.get('/jobs.json', authorize(["detl-jobs-read"]), function (request, response) {
    try {
        const jobs = configReader.load(jobsFileName);
        response.statusCode = 200;
        response.setHeader('Content-Type', contentType);
        response.write(JSON.stringify(jobs));
        response.end();
    }
    catch (ex) {
        response.statusCode = 404;
        response.setHeader('Content-Type', contentType);
        response.write('{ "error": "' + ex.message + '" }');
        response.end();
    }
});
protectedRoutes.post('/jobs.json', function (request, response) {
    let data = [];

    request
    .on('error', error => {
        response.statusCode = 500;
        response.setHeader('Content-Type', contentType);
        response.write('{ "error": "' + error + '" }');
        response.end();
    })
    .on('data', chunk => {
        data.push(chunk);
    })
    .on('end', () => {
        data = Buffer.concat(data).toString();

        try {
            // Always make a backup copy...
            fs.writeFileSync("../config/backup/jobs_" + (new Date()).toUTCString().replace(/ /g, "_").replace(/,/g, "").replace(/:/g, "") + ".json", data);
            fs.writeFileSync("../config/" + jobsFileName, data);

            response.statusCode = 200;
            response.setHeader('Content-Type', contentType);
            response.write('{ "status": "OK" }');
            response.end();
        }
        catch (ex) {
            response.statusCode = 500;
            response.setHeader('Content-Type', contentType);
            response.write('{ "status": "FAILED", message: "' + ex.message + '" }');
            response.end();
        }
    });
});

webui.use(unprotectedRoutes);

webui.use(authorize(), protectedRoutes);

// Catch-all error handler.
webui.use((req, res, next) => {
    next({ status: 404, message: "Not Found"});
});
webui.use((err, req, res, next) => {
    const error = err.stack || err.message || err.toString();

    // Don't bother logging 404 errors in the general log.
    if (err.status != 404)
        controllerLogger.error(error);

    // Render the error page.
    res.status(err.status || 500);

    res.send(error);
});

cleanup(function (exitCode, signal) {
    controllerLogger.info("StreamerService is exiting...");

    if (listener)
        listener.close();     // Don't leave HTTP port open.

    if (controller)
        controller.end(function () {
            controllerLogger.info("The Controller has been ended");
            controllerLogger.info("Terminating self...   Goodbye");
            controllerLogger.info("----------- :-( -------------");

            process.kill(process.pid, signal);
        });
    else {
        controllerLogger.info("The Controller has been ended");
        controllerLogger.info("Terminating self...   Goodbye");
        controllerLogger.info("----------- :-( -------------");

        process.kill(process.pid, signal);
    }

    cleanup.uninstall();    // Don't call cleanup handler again.
    return false;
});
