'use strict';

require("./Extensions.js");

const async = require('async');
const DynatraceAPI = require('./DynatraceAPI.js');
const BlackHoleAPI = require('./BlackHoleAPI.js');
const FormatterFactory = require('./FormatterFactory.js');

const friendly = function (name) {
    return name.toLowerCase().replace(/_/g, " ");
}

const logger = {
    log: function (job, message, severity) {
        if (!job)
            return console.log(severity + ": " + message);

        job.log = job.log || {};
        job.log.messages = job.log.messages || [];
        job.log.summary = job.log.summary || {}
        job.log.summary.errors = job.log.summary.errors || [];

        job.log.messages.push((severity || "INFO") + ": " + message);
    },
    info: function (job, message) {
        this.log(job, message, "INFO");
    },
    warn: function (job, message) {
        this.log(job, message, "WARNING");
    },
    error: function (job, message) {
        this.log(job, message, "ERROR");
    },
    errors: function (job, max) {
        let errorCount = 0;
        let errors = [];

        job.log.messages = job.log.messages.filter(message => {
            if (message.indexOf("ERROR") === 0) {
                if (errors.length < (max || 1))
                    errors.push(message);

                errorCount++;
                return false;
            }
            else {
                return true;
            }
        });
        if (errors.length < errorCount)
            errors.push("(" + (errorCount - errors.length) + " more)");

        return errors;
    }
};

const destinationFactory = {
    instances: {},
    createDestinations: function (tenants, destinations, templates, onReady) {
        const outputs = [ "entities", "linkages", "graphs" ];
        let pending = 0;

        this.getFormatter = new FormatterFactory(templates);

        Object.keys(destinations).forEach(destinationName => {
            const destination = destinations[destinationName];

            // For backwards compatibility and sanity.
            if (!outputs.includes(destination.produce)) destination.produce = "entities";

            try {
                // Load dynamically what we need.
                let classModule = require("./destinations/" + destination.class + ".js");
                this.instances[destinationName] = new classModule(destination, tenants);
            }
            catch (ex) {
                logger.error(null, "DestinationAPI class " + destination.class + " could not be loaded: " + ex.message);

                // Use a fallback to ensure we have a destination...
                this.instances[destinationName] = new BlackHoleAPI(destination);
            }

            pending++;

            this.instances[destinationName].on('error', function (err) {
                pending = -1;
                onReady({ status: "error", error: destinationName + ": " + err });
            });
            this.instances[destinationName].on('ready', function () {
                pending--;
                if (pending === 0)
                    onReady({ status: "ready" });
            });
        });

        if (pending === 0)
            setTimeout(onReady, 1, { status: "error", error: "No destinations found in configuration" });
    },
    getDestination: function (job) {
        let destinationName = job.destination || job.target || "unknown";

        if (this.instances[destinationName]) {
            logger.info(job, "Destination for data from " + job.name + " is " + destinationName);

            return this.instances[destinationName];
        }
        else if (destinationName == "Console") {
            logger.info(job, "Data from " + job.name + " will be returned to caller");

            return {
                produce: (job.template === "test" || job.related !== "flows") ? "entities" : "graphs",
                open: function (job, fieldGetter, onConnectionOpen) {
                    job.output = [];

                    setTimeout(onConnectionOpen, 1, null, {
                        send: function (messages, onComplete) {
                            job.output = messages || [];
                            setTimeout(onComplete, 1);
                        },
                        close: function () {
                            return true;
                        }
                    });
                }
            };
        }
        else {
            logger.warn(job, "There is no known destination for data from " + job.name);

            return {
                produce: "entities",
                open: function (job, fieldGetter, onConnectionOpen) {
                    logger.warn(job, "Opening the void to flush data from job " + job.name);

                    setTimeout(onConnectionOpen, 1, null, {
                        send: function (messages, onComplete) {
                            setTimeout(onComplete, 1);
                        },
                        close: function () {
                            return true;
                        }
                    });
                }
            };
        }
    }
};

// This list only contains granularities that the Dynatrace API accepts.
const granularities = [
       { label: "minute",     value: "_min",    bucket: 1
    }, { label: "5 minutes",  value: "_5mins",  bucket: 5
    }, { label: "10 minutes", value: "_10mins", bucket: 10
    }, { label: "15 minutes", value: "_15mins", bucket: 15
    }, { label: "30 minutes", value: "_30mins", bucket: 30
    }, { label: "hourly",     value: "_hour",   bucket: 60
    }, { label: "2 hours",    value: "_2hours", bucket: 60*2
    }, { label: "6 hours",    value: "_6hours", bucket: 60*6
    }, { label: "daily",      value: "_day",    bucket: 60*24
    }, { label: "disabled",      value: "_disabled",    bucket: 0
}
];
const timespans = granularities.reduce(function (result, granularity) {
    result[granularity.value] = granularity.bucket * 60 * 1000;
    return result;
}, {});

let streamerConfig = {};
let tuningInfo = { maxThreads: 8, maxJobSize: 10000 };
let dynatraceAPI = null;
let jobCount = 0;
let jobQueue = [];

function divideJob(job) {
    const count = Math.ceil(job.relevant.length / tuningInfo.maxJobSize); // How many chunks.
    if (count < 2)
        return [ job ];

    logger.info(job, "Splitting job " + job.name + " up into " + count + " parts");

    const list = job.relevant.slice();             // Keep original entity list in tact.
    const chunk = Math.ceil(list.length / count);   // How big should each subjob be?
    const jobs = [];

    let i = 0;
    while (list.length) {
        let subJob = Object.assign({}, job);        // Make sure we don't miss anything.
        subJob.id = job.id + "." + i;
        subJob.name = job.name + "." + i;
        subJob.relevant = list.splice(0, chunk);
        subJob.requests = null;

        jobs.push(subJob);
        i++;
    }
    return jobs;
}

function triggerNextJob(threadID) {
    // Trigger next job, but create a bit of breathing room in the timing, like 2 seconds
    if (jobQueue.length > 0) {
        let job = jobQueue.pop();
        setTimeout(performJob, 2000, job, threadID);
    }
}

function performJob(job, threadID) {
    if (!job) return;  // Should never happen, but to be sure.
    if (timespans[job.granularity] === 0) return;

    // Create a new log for this run.
    job.thread = threadID;
    job.log = {
        summary: {
            assigned: job.assigned,
            started: (new Date()).getTime(),
            errors: []
        },
        messages: []
    };
    jobCount++;
    try {
        if (job.problems) {
            // Problem feed.
            job.size = job.limit || 1000;
            job.dims = 1;

            logger.info(job, job.name + " specifies problems");

            const filters = job.criteria.length === 0
                ? [[]]
                : Array.isArray(job.criteria[0])
                    ? job.criteria
                    : [job.criteria];

            // We need one 'curried' collector per filter group.
            // They will all run in parallel, and then come together again.
            job.collectors = filters.map(function (filterGroup) {
                return function (onComplete) {
                    let queryParams = ["expandDetails=true"];

                    for (let i=0; i < filterGroup.length; i++) {
                        if (filterGroup[i].context === "PROBLEM") {
                            queryParams.push(filterGroup[i].key + "=" + filterGroup[i].values[0]);
                        }
                        else if (filterGroup[i].context === "CONTEXTLESS") {
                            filterGroup[i].values.forEach(value => {
                                queryParams.push("tag=" + filterGroup[i].key + ":" + value);
                            });
                        }
                        else {
                            filterGroup[i].values.forEach(value => {
                                queryParams.push("tag=[" + filterGroup[i].context + "]" + filterGroup[i].key + ":" + value);
                            });
                        }
                    }
                    collectProblemsByTags(job, {
                        startTimestamp: job.endTime - timespans[job.granularity],
                        endTimestamp: job.endTime,
                        queryString: queryParams.join("&")
                    }, onComplete);
                };
            });
        }
        else if (job.query) {
            // For USQL-based jobs there is no need to do anything to the metrics.
            job.size = job.limit || 1000;
            job.dims = job.metrics.length;

            logger.info(job, job.name + " specifies " + job.dims + " metrics for up to " + job.size + " " + friendly(job.type) + "s");

            // When we have a USQL query, we only need one 'curried' collector.
            job.collectors = [ onComplete => collectMetricsForSessions(job, {
                    startTimestamp: job.endTime - timespans[job.granularity],
                    endTimestamp: job.endTime,
                    queryString: encodeURIComponent(job.query)
            }, onComplete) ];
            
        }
        else if (job.entities) {
            // For entity-based jobs we split out multiple aggregations on a single metric into multiple metrics.
            job.metrics = job.metrics.reduce((acc, metric) => {
                let metricParts = metric.split("#");

                if (metricParts[1]) {
                    metricParts[1].split(",").forEach(aggregator => {
                        acc.push(metricParts[0] + "#" + aggregator);
                    });
                }
                else {
                    acc.push(metricParts[0]);
                }
                return acc;
            }, []);

            job.relevant = job.entities.filter(entity => entity.entityId.startsWith(job.type + "-"));

            job.size = job.relevant.length;
            job.dims = job.metrics.length;
            job.sent = 0;

            logger.info(job, job.name + " specifies " + job.dims + " metrics for " + job.size + " " + friendly(job.type) + "s");

            /*  We extract the entity IDs to be used in the requests first, because all collectors
                need to request their respective metrics' values across the same set of IDs.
                Note that we filter the entities based on their entityId prefix because jobs where
                'job.related' has the value "stack" also contain entities of other types, and we
                don't need to collect metrics for those. That stems from a fundamental design
                decision: each job has a type that constrains which metrics can be collected.

                Note that we use the expression "id.split('_').slice(0, -1).join('_') === job.type"
                rather than "id.startsWith(job.type)" to avoid matching "PROCESS_GROUP_INSTANCE"
                entities against the "PROCESS_GROUP" type. Because that expression is expensive,
                we only check for that if 'job.related' is "stack" - that's when we need to worry.

                Note that this doesn't remove entities from the job itself - just from the request
                to the timeseries API. As a result, excluded entities simply don't get metrics.
             */

            // Here we create a 'curried' collector per metric, which we can fire off in parallel.
            job.collectors = job.metrics.map(function (metric) {
                return function (onComplete) {
                    let metricParts = metric.split("#");
                    let aggregation = metricParts[1] ? metricParts[1].split("=")[0] : null; // Remove alias from aggregator.
                    let aggregator = aggregation ? aggregation.split("|")[0] : null; // In case there's a percentile.
                    let percentile = aggregation ? aggregation.split("|")[1] : null; // Get the percentile, if avail.

                    collectMetricForEntities(job, metric, {
                        timeseriesId   : metricParts[0],
                        aggregationType: aggregator === "NUM" ? null : aggregator,          // Aggregator "NUM" is for the UI.
                        percentile     : percentile,
                        includeData    : true,
                        queryMode      : "TOTAL",
                        startTimestamp : job.endTime - timespans[job.granularity],
                        endTimestamp   : job.endTime
                    }, onComplete);
                };
            });
        }
        else {
            // TODO: Unknown job type
        }

        // Tell the parent process we started working on this job.
        process.send({
            status: "started",
            id: job.id,
            name: job.name,
            thread: job.thread,
            size: job.size,
            sent: job.sent,
            dims: job.dims,
            cycle: job.cycle,
            log: job.log
        });

        if (job.size === 0) throw new Error("Job " + job.name + " does not contain entities");
       
        if (job.dims === 0 && !job.query ) throw new Error("Job " + job.name + " does not contain metrics");

        // Fire the collectors in parallel, and wait until all metrics have been collected.
        // TODO: For problems, it's technically not metric data.
        async.parallel(job.collectors, function (err, allMetricData) {
            if (err)
                logger.error(job, err);

            // Even if there are errors, we're going to send out the metrics we did collect.
            onMetricsCollected(job, allMetricData);

            // Once all metrics have been sent out, 'onJobProcessed()' will be called.
        });
    }
    catch (ex) {
        /*  An exception occurred! Then sign off asynchronously because
            we don't want to inadvertently trigger premature retirement...
         */
        logger.error(job, ex.message);

        triggerNextJob(threadID);   // Don't let this 'thread' go to waste.
        setTimeout(onJobProcessed, 10, job);
    }
}

function collectProblemsByTags(job, requestData, addProblemData) {
    job.log.query = {
        started: (new Date()).getTime()
    };

    /*
        The filters can be OR-ed (groups). In that case we have to run multiple times
        and then aggregate. Also note that some filters have a context named "PROBLEM".
        Those become actual query parameters.
     */

    dynatraceAPI.request({
        timeout: (streamerConfig.metricTimeout || 10) * 1000,
        tenant: job.source,
        endPoint: '/api/v1/problem/feed',
        queryData: 'startTimestamp=' + requestData.startTimestamp
            + '&endTimestamp=' + requestData.endTimestamp
            + '&' + requestData.queryString
    },
        function (err, problemData) {
            job.log.query.collected = (new Date()).getTime();

            if (err)
                err = "Unable to read problem feed. Reason: " + err.message.replace("!", ".");

            addProblemData(err, problemData);
        }
    );
}

function collectMetricsForSessions(job, requestData, addMetricData) {
    job.log.query = { started: (new Date()).getTime() };

    dynatraceAPI.request({
        timeout: (streamerConfig.metricTimeout || 10) * 1000,
        tenant: job.source,
        endPoint: '/api/v1/userSessionQueryLanguage/table',
        queryData: 'startTimestamp='  + requestData.startTimestamp 
            + '&endTimestamp=' + requestData.endTimestamp
            + '&query=' + requestData.queryString
    },
        function (err, metricData) {
            job.log.query.collected = (new Date()).getTime();

            if (err)
                err = "Unable to execute " + job.query + ". Reason: " + err.message.replace("!", ".");

            addMetricData(err, metricData);
        }
    );
}

function collectMetricForEntities(job, metric, requestData, addMetricData) {
    job.log[metric] = { started: (new Date()).getTime() };

    if (job.relevant.length > tuningInfo.maxJobSize) {
        /*  It may happen that jobs are too large because the Controller cannot split all jobs.
            Specifically, jobs that have linkages and jobs otherwise marked as 'non-splittable'
            are not split. Whenever we encounter a job that is too big, we split it into chunks
            that individually are smaller than whatever the current maxJobSize is. We then call
            ourselves recursively for each, consolidate the results, and call 'addMetricData()'
            once we have everything.
         */
        const subjobs = divideJob(job);
        const accumulator = {
            count: subjobs.length,
            errors: [],
            metricData: null
        }

        subjobs.forEach(subjob => {
            // Each subjob gets its own (shallow) copy of the requestData.
            collectMetricForEntities(subjob, metric, { ...requestData }, (err, metricData) => {
                /*  This here is our intermediate 'addMetricData()' implementation. Note that
                    it will call the original 'addMetricData()' once we have everything.
                */
                if (err) {
                    // Add errors if any.
                    accumulator.errors.push(err);
                }
                else if (!accumulator.metricData) {
                    //  Adopt first metricData as-is.
                    accumulator.metricData = metricData;
                }
                else {
                    //  For subsequent metricDatas we need to merge.
                    const amd = accumulator.metricData.result;
                    Object.assign(amd.entities, metricData.result.entities);
                    Object.assign(amd.dataPoints, metricData.result.dataPoints);
                }

                accumulator.count--;
                if (accumulator.count == 0) {
                    if (accumulator.errors.length > 0)
                        addMetricData(accumulator.errors.join(", "), null);
                    else
                        addMetricData(null, accumulator.metricData);
                }
            });
        });

        return;     // Let the recursion handle it.
    }

    // Convert the entity list to IDs for the Dynatrace API.
    requestData.entities = job.relevant.map(entity => entity.entityId);

    dynatraceAPI.request({
        timeout: (streamerConfig.metricTimeout || 10) * 1000,
        tenant: job.source,
        endPoint: '/api/v1/timeseries/' + metric,
        queryData: null,
        postData: JSON.stringify(requestData)
    },
        (err, metricData) => {
            job.log[metric].collected = (new Date()).getTime();

            /*  Async.parallel uses the calls to 'addMetricData()' to collect all results into
                one array. It will eventually call 'onMetricsCollected()' when all invocations
                of 'collectMetricForEntities()' have called 'addMetricData()'.
             */
            if (err)
                addMetricData("Retrieving values for " + metric + " yields: " + err.message, null);
            else
                addMetricData(null, metricData);
        }
    );
}

function onMetricsCollected(job, allMetricData) {
    job.log.summary.collected = (new Date()).getTime();

    try {
        let destination = destinationFactory.getDestination(job);
        let formatter = destinationFactory.getFormatter(job.template, destination.defaults);

        // Make sure that the entity type is supported by the message formatter.
        if (!formatter.isTypeSupported(job.type)) {
            throw new Error("Template '" + formatter.format + "' does not support " + job.type + " jobs");
        }

        let allEntityData = null;

        if (job.problems) {
            /*
                We must merge all problem sets into one collection. After that, we may have to
                transform (rotate) the array from problems to entities, if the job says so.
                For now, we just merge all problems into one array of problems.
             */
            const problemArrays = allMetricData.map(data => data.result.problems);
            allEntityData = []
                .concat(...problemArrays)
                .map(problem => formatter.createMessage(job, problem, {}) );
        }
        else if (job.query) {
            /**** Below is for USQL-based query results ****/

            function findMatchBeforeSep(list, name, sep) {
                name = name.toLowerCase();
                for (var i=0; i < list.length; i++) {
                    if (list[i].split(sep)[0].toLowerCase() == name) {
                        return {
                            index: i,
                            name: list[i]
                        };
                    }
                }
                return null;
            }

            // For UQSL queries: Find the fields in the returned list of column names.
            // Also, the fields together constitute the "id" of this message.
            // Because async.parallel() puts the USQL reponse object in an array...
            allMetricData = allMetricData[0] || { columnNames: [], values: [] };

            let fieldIndices = job.fields.reduce((indices, field) => {
                if (field.values && field.values.length > 0) {
                    field.values.forEach(value => {
                        let valParts = value.split("=");
                        let name = valParts[1] || (valParts[0].split("|")[0] + "(" + (field.alias || field.value));
                        let match = findMatchBeforeSep(allMetricData.columnNames, name, ",");
                        if (match) indices[match.name] = match.index;
                    });
                }
                else {
                    let name = field.alias || field.value;
                    let match = findMatchBeforeSep(allMetricData.columnNames, name, ",");
                    if (match) indices[match.name] = match.index;
                }
                return indices;
            }, {});

            allEntityData = allMetricData.values.map(row => {
                const fields = {};
                const idParts = [];
                const metrics = {};

                for (let colName in fieldIndices) {
                    if (!fieldIndices.hasOwnProperty(colName)) continue;

                    let fieldValue = Array.isArray(row[fieldIndices[colName]])
                        ? row[fieldIndices[colName]].join(", ")
                        : row[fieldIndices[colName]];
                    fields[colName] = fieldValue;
                    idParts.push(fieldValue);
                }

                fields.id = idParts.join("_");
                fields.entityId = idParts.join("_");

                for (let i=0; i < allMetricData.columnNames.length; i++) {
                    if (fieldIndices.hasOwnProperty(allMetricData.columnNames[i])) continue;

                    metrics[allMetricData.columnNames[i]] = row[i];
                }

                return formatter.createMessage(job,fields, metrics);

            });
        }
        else if (job.entities) {
            /**** Below is for entity-based query results (rotate the matrix) ****/

            // Let's first extract the dimensions for each metric so that we can
            // efficiently iterate over them as we create a message per entity.
            allMetricData = allMetricData.map(metricData => {
                if (!metricData)
                    return null;

                let metricInfo = metricData.result
                    ? metricData.result
                    : metricData[0] && metricData[0].result
                        ? metricData[0].result
                        : null;
                if (!metricInfo)
                    return null;

                const dimensions = [ null ];    // Empty entry to cover metrics without dimensions.

                for (let entityOrDimension in metricInfo.entities) {
                    if (!entityOrDimension.startsWith(job.type))
                        dimensions.push(entityOrDimension);
                }
                // Fallback mechanism for when dimesnions are not explicitly provided.
                for (let dataPointFor in metricInfo.dataPoints) {
                    const dimCandidate = dataPointFor.split(", ")[1];
                    if (dimCandidate && !dimensions.includes(dimCandidate)) {
                        dimensions.push(dimCandidate);
                    }
                }

                return {
                    timeseriesId: metricInfo.timeseriesId,
                    aggregation: metricInfo.aggregationType,
                    dataPoints: metricInfo.dataPoints,
                    entities: metricInfo.entities,
                    dimensions: dimensions
                };
            }, []);

            // Create a message per entity.
            const entities = job.related === "stack"
                ? job.entities.filter(entity => entity.entityId.startsWith(job.type + "-"))
                : job.entities;

            allEntityData = entities.map(function (entity) {
                entity.metrics = [];

                // Collect the metrics across the various requests.
                const metrics = {};
                for (let i=0; i < job.metrics.length; i++) {
                    let metricInfo = allMetricData[i];
                    let metric = job.metrics[i];
                    let metricId = metric.split("=")[0].split("|")[0];
                    let metricAlias = metric.split("=")[1] || metricId;

                    // NUM is an internally used aggregator. The DynatraceAPI does not use it.
                    if (metricId.split("#")[1] === "NUM")
                        metricId = metricId.split("#")[0];

                    let fullID = !metricInfo
                        ? null
                        : metricInfo.aggregation
                            ? metricInfo.timeseriesId + "#" + metricInfo.aggregation
                            : metricInfo.timeseriesId;
                    if (fullID !== metricId) {
                        logger.error(job, "Expected " + metricId + " - found " + fullID + " at index " + i);
                        break;  // We just move on to the next metric.
                    }

                    /*  This works with and without dimensions. If no dimensions were found for this metric,
                        it will have a single 'null' entry. That allows us to just look for the metric value
                        by the entityId (rather than "<entityId>, <dimension>").
                     */
                    for (let j = 0; j < metricInfo.dimensions.length; j++) {
                        let dimension = metricInfo.dimensions[j];        // Will be 'null' when j==0.
                        let dimensionLabel = dimension ? metricInfo.entities[dimension] || dimension : null;
                        let metricName = dimensionLabel ? metricAlias + "#" + dimensionLabel : metricAlias;
                        let metricEntryId = entity.entityId + (dimension ? ", " + dimension : "");
                        let dataPoint = metricInfo.dataPoints[metricEntryId];

                        if (!job.optimize || !dimensionLabel || (dataPoint && dataPoint.length > 0)) {
                            metrics[metricName] = !dataPoint || dataPoint.length === 0
                                ? 0                     // Not found? Set it to 0.
                                : dataPoint.length > 1  // Pair of timestamp-value.
                                    ? dataPoint[1]          // We don't need the timestamp - take the value.
                                    : dataPoint[0][1];      // It commonly also sits in a nested array...

                            entity.metrics.push({
                                name: metricInfo.timeseriesId,
                                aggr: metricInfo.aggregation || null,
                                dim: dimensionLabel,
                                value: metrics[metricName]
                            });
                        }
                    }
                };

                // Ask the formatter to create a message for us.
                return formatter.createMessage(job, entity, metrics);
            });
        }
        else {
            // Unknown job type
        }

        /*  Now we have a message per entity. Next, we have to consider several scenarios:
            1) If we have a list (no relationships in the job), then it's business as usual:
                Send out a message per entity (we'll modify this rule below, but bear with me).
            2) If we have a graph (we have linkages in the job), we let the destination control
                how we produce rows. A destination may have a "produce" property that can have
                three options:
                - "entities": This is the default. We output a message for each entity individually
                                and ignore the edges.
                - "linkages": We iterate over the edges and produce a message per edge that contains
                                both entities: the 'source' entities and the 'target' entities.
                - "graph":    We send all nodes (i.e. messages) and edges as-is. Note that destination
                                adaptors that cannot handle graphs should revert to "linkages".

            Generalized and harmonized rules for the "produce" property:
            1) If the data is a list:
                entities) Produce the entities as                 [ nodes ]
                linkages) Produce the entities as                 [{ source: nodeFrom, target: null }, ...]
                graphs)   Produce the graph as                    { nodes: [ nodes ], edges: null }

            2) If the data is a graph:
                entities) Produce the list (not the linkages) as  [ nodes ]
                linkages) Produce a list with linkages as         [{ source: nodeFrom, target: nodeTo }, ...]
                graphs)   Produce the graph as                    { nodes: [ nodes ], edges: [ edges ] }

            This means that destination adaptors may receive any of the following types of data:
            1) entities: Array of nodes.
            2) linkages: Array of objects with 'source' and 'target' properties, each pointing to a node.
            3) graphs:   An object with 'nodes' and 'edges' properties.

            A destination adaptor can tell us whether it can accept a 'produce' value or not.
            If it doesn't, it may choose a fallback option, or throw an error. For instance,
            the CVSAdaptor cannot accept "graph" as output, so it will fall back to "linkages",
            as that is the appropriate output there.
         */
        let dataToSend = null;

        // TODO: Base this on a negotiation between what the destination can support
        // and what the job prefers. This also eliminates the need to make this
        // dynamic somehow inside the destination API.
        const produce = job.problems ? "entities" : destination.produce;
        switch (produce) {
            case "entities": // This is also for problems.
                dataToSend = allEntityData;
                job.sent = dataToSend.length;
                break;

            case "linkages":
                dataToSend = job.linkages
                    ? job.linkages.map(linkage => ({
                        source: allEntityData[job.entities.getIndexByField("id", linkage.source)],
                        target: allEntityData[job.entities.getIndexByField("id", linkage.target)]
                    }))
                    : allEntityData.map(entityData => ({
                        source: entityData,
                        target: null
                    }));
                job.sent = dataToSend.length;
                break;

            case "graphs":
                dataToSend = {
                    nodes: allEntityData,
                    edges: job.linkages
                };
                job.sent = dataToSend.nodes.length;
                break;
        }

        job.log.summary.processed = (new Date()).getTime();

        // Send to destination.
        destination.open(job, formatter.getFields.bind(formatter), (err, connection) => {
            if (err) {
                logger.error(job, "Destination could not be opened to send data for " + job.name + ": " + err);
                onJobProcessed(job);
                return;
            }

            try {
            connection.send(dataToSend, (err, data) => {
                logger.info(job, "Job produced " + job.sent + " results for " + job.size + " entities/rows");

                job.log.summary.published = (new Date()).getTime();

                if (err) {
                    let destinationName = job.destination || job.target || "unknown";
                        logger.error(job, destinationName + " failed while sending data for " + job.name  + ": " + err);
                }

                connection.close();
                onJobProcessed(job);
            });
            }
            catch (ex) {
                job.log.summary.published = (new Date()).getTime();

                let destinationName = job.destination || job.target || "unknown";
                logger.error(job, destinationName + " failed while sending data for " + job.name  + ": " + ex.message);

                connection.close();
                onJobProcessed(job);
            }
        });
    }
    catch (ex) {
        /*  An exception occurred! Then sign off asynchronously because
            we don't want to inadvertently trigger premature retirement...
         */
        logger.error(job, ex.message);

        setTimeout(onJobProcessed, 10, job);
    }
    finally {
        // Trigger next job asynchronously: it can start while this one is still sending its data.
        triggerNextJob(job.thread);
    }
};

function onJobProcessed(job) {
    try {
        for (var metric in job.log) {
            // These things are not metrics. We'll treat them separately.
            if (metric == "summary" || metric == "timestamp" || metric == "messages") continue;

            job.log[metric].collecting = job.log[metric].collected
                - job.log[metric].started;
        }

        job.log.summary.collecting = job.log.summary.collected
            - job.log.summary.started;
        job.log.summary.processing = job.log.summary.processed
            - job.log.summary.collected;
        job.log.summary.publishing = job.log.summary.published
            - job.log.summary.processed;
        job.log.summary.duration = job.log.summary.published
            - job.log.summary.started;
        job.log.summary.errors = logger.errors(job, 3);

        if (job.log.summary.errors &&
            job.log.summary.errors.filter(error => error.includes("Entity must not be larger than")).length > 0)
            job.log.feedback = "split";
    }
    catch (ex) {
        logger.error(job, "Error while compiling results for '" + job.name + "': " + ex);

        job.log.summary.collecting = job.log.summary.collecting || 0;
        job.log.summary.processing = job.log.summary.processing || 0;
        job.log.summary.publishing = job.log.summary.publishing || 0;
        job.log.summary.duration = (new Date()).getTime()
            - job.log.summary.started;
        job.log.summary.errors = logger.errors(job, 3);
    }
    finally {
        process.send({
            status: job.log.summary.errors && job.log.summary.errors.length > 0 ? "failed" : "done",
            id: job.id,
            name: job.name,
            thread: job.thread,
            assigned: job.assigned,
            size: job.size,
            sent: job.sent,
            dims: job.dims,
            cycle: job.cycle,
            log: job.log,
            output: job.output
        });
        jobCount--;
        retire();
    }
}

let waitingToRetire = false;

function retire(prepareToRetire) {
    waitingToRetire = waitingToRetire || !!prepareToRetire;

    /*  We test for both the jobCount and the jobQueue because even when the jobQueue is
        empty, not all jobs may be done. Reversely, jobCount may have returned to 0, but
        there may still be jobs waiting to be taken from the queue for processing.
     */
    if (waitingToRetire && jobCount <= 0 && jobQueue.length === 0) {
        process.disconnect();
        process.exit();
    }
}
/*  Each job can have a different destination. The configuration contains the
    list of destinations that will be used, so we will instantiate all of them.
    Each job will ask for the destination object it needs when it gets executed.
 */
process.on('message', async message => {
    switch (message.command) {
        case "configure":
            console.log("initializing");

            // DEBUG aid to give debugger time to attach: 
            // if (message.config.debugWorker)
            //     await new Promise(resolve => setTimeout(resolve, 3000));

            streamerConfig = message.config;

            dynatraceAPI = new DynatraceAPI(streamerConfig.Tenants, streamerConfig);

            destinationFactory.createDestinations(
                streamerConfig.Tenants,
                streamerConfig.Destinations,
                streamerConfig.Templates,
                function (status) {
                    console.log("reporting for duty");
                    process.send(status);
                }
            );
            break;

        case "tuning":
            // Update the tuning information.
            tuningInfo = Object.assign(tuningInfo, message.tuning);
            break;

        case "perform":
            jobQueue = message.jobs;
            console.log("has " + jobQueue.length + " jobs for cycle " + message.cycle);

            jobQueue.splice(0, tuningInfo.maxThreads)     // Extract first batch of jobs.
                .forEach(performJob);                   // Start them.
            // The rest of the jobs will be done, staggered, as individual jobs finish.
            break;

        case "retire":
            // Terminate as soon as all jobs are complete.
            retire(true);
            break;
    }
});
process.on('disconnect', function () {
    console.log('StreamingAPI exited - stopping Worker');
    process.exit();
});

process.send({ status: "preparing" });



/*  TODO:
    If we are not connected to a parent process (ie. the StreamerService),
    then we should read the config.json and jobs.json ourselves, and connect
    to the relevant library to obtain job entities.
    The command line will then tell us which jobs to execute. After we've
    executed the relevant (or all) jobs, we exit.
    This is useful for debugging, particularly when customers add their own
    DestinationAPI implementations. Or just to run some jobs offline.
 */

    // Setup logging.
    // Create a dummy 'process' object at which the work can be received and reported to.
    // Load the config and connect to the relevant libraries via HTTP.
    // Load the jobs.json file.
    // Read the command line arguments for a comma-separated list of job names.
    // For each job to be executed:
    //      Ask the library for its contents.
    //      Execute the job (process.emit("message", jobs)).
    // Terminate(process.emit("message", 'retire'')).
