const { fork } = require('child_process');

const WorkerProxy = function (streamerConfig, tuningInfo, socket, logger, debugWorker) {
    const self          = this;
    const config        = streamerConfig;
    const identity      = WorkerProxy.idGenerator++;
    const workerPath    = process.cwd() + '/' + (config.workerName || 'Worker.js');
    const debugEnabled  = (debugWorker || config.debugWorker) && (typeof v8debug === 'object' || /--debug|--inspect/.test(process.execArgv.join(' ')));
    const workerProcess = fork(workerPath, [], { silent: true, stdio: 'pipe', execArgv: debugEnabled ? [ "--inspect" ] : [] });

    // If we don't have a socket, use a dummy.
    socket = socket || { emit: function () {} };

    // Just to be sure...
    if (!workerProcess) return null;

    this.events  = {};

    const subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);
    };

    const emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const eventHandlers = self.events[eventName] || [];
        eventHandlers.forEach(fn => { fn.call(null, data); });
    };

    let jobsInProgress = [];
    let isReady     = false;
    let isDead      = false;
    let isDying     = false;
    let isRetiring  = false;
    let isPreparing = true;
    let killTimer   = null;

    function scheduleForcedKill() {
        killTimer = setTimeout(function () {
            logger.warn("Terminating uncooperative worker " + identity);
            workerProcess.kill();
        }, 3 * 60000);
    }
    function cancelForcedKill() {
        clearTimeout(killTimer);
        killTimer = null;
    }
    function onWorkerDied(code, signal) {
        if (isDead) return;
        isDead = true;

        logger.info("Worker " + identity + " died");
        emit("died", identity);

        // Also tell any browsers that are listening.
        socket.emit('died', {
            player: "Controller",
            worker: {
                id:   identity,
                name: "Worker " + identity,
                type: "Worker"
            }
        });

        cancelForcedKill();
    }

    workerProcess.on("exit",  onWorkerDied);
    workerProcess.on("close", onWorkerDied);

    /*  We use the worker's stdout to capture its log output. But in exceptional cases stdout
        is not available. That signals a very bad situation, so in that case, we'll kill this
        process immediately yet asynchronously, so that callers have time to subscribe to the
        'died' event.
     */
    if (!workerProcess.stdout) {
        logger.warn("Killing worker " + identity + " on creation (no stdout found)");
        setTimeout(workerProcess.kill, 1);
        return syncIF;
    }

    workerProcess.stdout.on('data', data => {
        logger.info(  "Worker " + identity + " " + ("" + data).trim());
    });
    workerProcess.stderr.on('data', data => {
        logger.error("Worker " + identity + " " + ("" + data).trim());
    });
    workerProcess.on("disconnect", msg => {
        isDying = true;
        logger.warn("Killing worker " + identity + " on disconnect");
        workerProcess.kill();
    });
    workerProcess.on("error", err => {
        logger.warn("Killing worker " + identity + " on error: " + err);
        workerProcess.kill();
    });
    workerProcess.on("message", message => {
        switch (message.status) {
            case "preparing":
                if (config)
                    workerProcess.send({
                        command: "configure",
                        config: config
                    });

                isPreparing = false;

                // Also tell any browsers that are listening.
                socket.emit('born', identity);
                break;

            case "ready":
                // TODO: we also may get a 'ready' if the Worker just processed
                // a new configuration. In that case, no need to emit a 'ready'
                // again. Rather, we may have to hold a new assignment until
                // we received this signal here.
                workerProcess.send({
                    command: "tuning",
                    tuning: tuningInfo
                });
                isReady = true;
                emit("ready", asyncIF);
                break;

            case "started":
                // This means that an assigned job was just kicked off.
                message.player = "Controller";
                message.worker = {
                    id:   identity,
                    name: "Worker " + identity,
                    type: "Worker"
                };

                // Tell any browsers that are listening.
                message.id = "Job_" + message.id;
                socket.emit('job', message);
                break;

            case "failed":
                // This means that an assigned job just failed.
            case "done":
                // This means that an assigned job just completed.
                var i = jobsInProgress.indexOf(message.id);
                if (i >= 0)
                    jobsInProgress.splice(i, 1);

                message.player = "Controller";
                message.worker = {
                    id:   identity,
                    name: "Worker " + identity,
                    type: "Worker"
                };

                // Tell the Controller.
                emit("done", message);

                // Tell any browsers that are listening.
                message.id = "Job_" + message.id;
                socket.emit('job', message);
                break;

            case "error":
                logger.error("Killing worker " + identity + " on error: " + message.error);

                if (!isReady)
                    emit("ready", null);    // Let stakeholders know we couldn't create the Worker.

                workerProcess.kill();
                break;
        }
    });

    const asyncIF = {
        get id() {
            return identity;
        },
        get identity() {
            return {
                id:   identity,
                name: "Worker " + identity,
                type: "Worker"
            }
        },
        set config(value) {
            config = value;
            if (!isPreparing)
                workerProcess.send({
                    command: "configure",
                    config: config
                });
        },
        set tuning(value) {
            tuningInfo = value;
            workerProcess.send({
                command: "tuning",
                tuning: tuningInfo
            });
        },
        set assignment(value) {
            // TODO: We should wait until we're ready (again), and then send the assignment.
            // Otherwise, just hold the assignment until we can send it.
            if (!isReady)
                logger.warn("Worker is assigned work while not yet ready");

            jobsInProgress = value.jobs.map(job => job.id);

            workerProcess.send({
                command: "perform",
                cycle:   value.cycle,
                jobs:    value.jobs
            });

            // Also tell any browsers that are listening.
            value.jobs.forEach(job => {
                socket.emit('job', {
                    status:    "assigned",
                    assigned:  job.assigned,
                    id:        "Job_" + job.id,
                    name:      job.name,
                    dims:      job.metrics  ? job.metrics.length  : 0,
                    size:      job.entities ? job.entities.length : job.limit || 0,
                    cycle:     value.cycle,
                    log:       null,
                    worker: {
                        id:   identity,
                        name: "Worker " + identity,
                        type: "Worker"
                    },
                    player:   "Controller"
                });
            });
        },

        retire: function () {
            if (isDead) return;

            logger.info("Asking Worker " + identity + " to retire...");

            workerProcess.send({
                command: "retire"
            });
            isRetiring = true;

            // Also tell any browsers that are listening.
            socket.emit('retiring',  {
                id:   identity,
                name: "Worker " + identity,
                type: "Worker"
            });

            scheduleForcedKill();  // Backup plan...
        },
        get isAlive() {
            return !isRetiring && !isDying && !isDead;
        },
        get isReady() {
            return isReady;
        }
    };

    /*  This chainable sync interface will be returned to the caller when a new instance is created.
        The 'ready' event will provide the async interface that contains all regular APIs.
     */
    const syncIF = {
        on: function (evt, handler) {
            subscribe(evt, handler);
            return this;    // For chaining.
        },
        get worker() {
            return asyncIF;
        },
        retire: asyncIF.retire
    };

    return syncIF;
};
WorkerProxy.idGenerator = 0;

module.exports = WorkerProxy;
