/* MariaDB Adaptor */
'use strict';
const mariadb = require('mariadb');

 /**
  * A MariaDB-specific implementation of the DatabaseAPI.
  * @extends DatabaseAPI
  * @private
  */
class MariaDBAdaptor {
    constructor(config, logger, tableName) {
        logger.info("    Initializing MariaDB adaptor");
        this.logger = logger;
        this.pool   = mariadb.createPool(config);
        this.table  = tableName;
        logger.info("    MariaDB adaptor initialized");
    }

    isRetryable(error) {
        /*  There are errors returned by data operations where a retry may make sense.
            These are the errors we are looking for:
            NOTE: THIS ULTIMATELY WILL NEED TO GO TO THE ADAPTORS.

            errno       code                                message
            -------------------------------------------------------------------------------------------------------------------------
            1213        ER_LOCK_DEADLOCK                    Deadlock found when trying to get lock; try restarting transaction.
            1614	    ER_XA_RBDEADLOCK                    XA_RBDEADLOCK: Transaction branch was rolled back: deadlock was detected.
            1205        ER_LOCK_WAIT_TIMEOUT                Lock wait timeout exceeded; try restarting transaction.
            1206        ER_LOCK_TABLE_FULL                  The total number of locks exceeds the lock table size.
            1613        ER_XA_RBTIMEOUT                     XA_RBTIMEOUT: Transaction branch was rolled back: took too long.
            1875        ER_STOP_SLAVE_SQL_THREAD_TIMEOUT    STOP SLAVE command execution is incomplete: Slave SQL thread got the
                                                            stop signal, thread is busy, SQL thread will stop once the current task
                                                            is complete.
            1876        ER_STOP_SLAVE_IO_THREAD_TIMEOUT     STOP SLAVE command execution is incomplete: Slave IO thread got the stop
                                                            signal, thread is busy, IO thread will stop once current task is complete.
            1969        ER_STATEMENT_TIMEOUT                Query execution was interrupted (max_statement_time exceeded).
            1040        ER_CON_COUNT_ERROR                  Too many connections.
            1203        ER_TOO_MANY_USER_CONNECTIONS        User %s already has more than 'max_user_connections' active connections.
            1927        ER_CONNECTION_KILLED 70100          Connection was killed.
         */
        const retriables = [1040, 1203, 1927, 1213, 1614, 1205, 1206, 1613, 1875, 1876, 1969];
        return retriables.includes(error.errno);
    }

    createConnection(tenant, taskName, resolve, reject, cleanup) {
        const self = this;

        return {
            name:    taskName,
            tenant:  tenant,
            resolve: resolve,
            reject:  reject,
            cleanup: cleanup,
            state:   "Queued",
            initialize: function (conn) {
                this.state  = "Connected";
                this.conn   = conn;
                this.query  = function (...args) {
                    args[0] = args[0].replace(self.table, "Entities_" + this.tenant);

                    return conn.query.apply(conn, args);
				};
                this.stream  = function (...args) {
                    args[0] = args[0].replace(self.table, "Entities_" + this.tenant);
                    
                    return conn.queryStream.apply(conn, args);
				};                
                this.begin  = conn.beginTransaction;
            },
            commit: function (onDone, onError) {
                if (this.conn)
                    this.conn.commit()
                    .then(() => {
                        this.release(() => onDone && onDone());
                    })
                    .catch(error => {
                        this.release(() => onError && onError(error));
                    });
                else
                    onDone && onDone();
            },
            rollback: function (onDone, onError) {
                if (this.conn)
                    this.conn.rollback()
                    .then(() => {
                        this.release(() => onDone && onDone());
                    })
                    .catch(error => {
                        this.release(() => onError && onError(error));
                    });
                else
                    onDone && onDone();
            },
            release: function (onReleased, noCleanup) {
                const self = this;
                const finish = function () {
                    if (!noCleanup) self.cleanup(self);
                    if (onReleased) setTimeout(onReleased, 1);
                    onReleased = null;
                    self.conn = null;
                };

                if (this.conn) {
                    this.conn.end().then(finish).catch(finish);
                }
                else {
                    finish();   // Yes, always.
                }
            }
        };
    }

    acquire(connection, onAcquired) {
        connection.state = "Connecting";

        this.pool.getConnection()
        .then(conn => {
            connection.initialize(conn);
            connection.resolve(connection); 
            if (onAcquired) onAcquired();
        })
        .catch(err => {
            this.logger.error("Could not obtain a database connection for " + connection.name + ": " + err);
            connection.cleanup(connection);
            connection.reject(err);
        });
    }

    get activeConnections() {
        return this.pool.activeConnections();
    }

    get totalConnections() {
        return this.pool.totalConnections();
    }

    close(resolve) {
        this.pool.end()
        .then(()  => { resolve && resolve(); })
        .catch(() => { resolve && resolve(); });
    }
}

module.exports = MariaDBAdaptor;