/* MySQL Adaptor */
'use strict';
const mysql = require('mysql');

 /**
  * A MySQL-specific implementation of the DatabaseAPI.
  * @extends DatabaseAPI
  * @private
  */
class MySQLAdaptor {
    constructor(config, logger, tableName) {
        logger.info("    Initializing MySQL adaptor");
        this.logger = logger;
        this.pool   = mysql.createPool(config);
        this.table  = tableName;        
        logger.info("    MySQL adaptor initialized");
    }

    isRetryable(error) {
        /*  There are errors returned by data operations where a retry may make sense.
            These are the errors we are looking for:
            NOTE: THIS ULTIMATELY WILL NEED TO GO TO THE ADAPTORS.

            errno       code                                message
            -------------------------------------------------------------------------------------------------------------------------
            1213        ER_LOCK_DEADLOCK                    Deadlock found when trying to get lock; try restarting transaction.
            1614	    ER_XA_RBDEADLOCK                    XA_RBDEADLOCK: Transaction branch was rolled back: deadlock was detected.
            1205        ER_LOCK_WAIT_TIMEOUT                Lock wait timeout exceeded; try restarting transaction.
            1206        ER_LOCK_TABLE_FULL                  The total number of locks exceeds the lock table size.
            1613        ER_XA_RBTIMEOUT                     XA_RBTIMEOUT: Transaction branch was rolled back: took too long.
            1875        ER_STOP_SLAVE_SQL_THREAD_TIMEOUT    STOP SLAVE command execution is incomplete: Slave SQL thread got the
                                                            stop signal, thread is busy, SQL thread will stop once the current task
                                                            is complete.
            1876        ER_STOP_SLAVE_IO_THREAD_TIMEOUT     STOP SLAVE command execution is incomplete: Slave IO thread got the stop
                                                            signal, thread is busy, IO thread will stop once current task is complete.
            1969        ER_STATEMENT_TIMEOUT                Query execution was interrupted (max_statement_time exceeded).
            1040        ER_CON_COUNT_ERROR                  Too many connections.
            1203        ER_TOO_MANY_USER_CONNECTIONS        User %s already has more than 'max_user_connections' active connections.
            1927        ER_CONNECTION_KILLED 70100          Connection was killed.
         */
        const retriables = [1040, 1203, 1927, 1213, 1614, 1205, 1206, 1613, 1875, 1876, 1969];
        return retriables.includes(error.errno);
    }

    createConnection(tenant, taskName, resolve, reject, cleanup) {
        const self = this;

        return {
            name:    taskName,
            tenant:  tenant,
            resolve: resolve,
            reject:  reject,
            cleanup: cleanup,
            state:   "Queued",
            initialize:    function(conn) {
                this.state  = "Connected";
                this.conn   = conn;
                this.query  = function (...args) {
                    args[0] = args[0].replace(self.table, "Entities_" + this.tenant);

					return new Promise((resolve, reject) => {
						// Add the callback that will resolve or reject.
						args.push((err, results, fields) => {
							if (err)
								reject(err);
							else {
                                results.meta = fields;
                                results.warningStatus = results.warningCount;
								resolve(results, fields);
							}
						});
						conn.query.apply(conn, args);
					});
				};
                this.stream = function (...args) {
                    args[0] = args[0].replace(self.table, "Entities_" + this.tenant);

					// Create an interceptor so that we can translate the events.
					const interceptor = {
						events: {},
						on: function (eventName, handler) {
							// Add the subscription (event handler) for this event.
							this.events[eventName] = this.events[eventName] || [];
							this.events[eventName].push(handler);
							return this;	// For chaining.
						},
						emit: function (eventName, data) {
							// Call the subscribed event handlers.
							const event = this.events[eventName] || [];
							event.forEach(handler => { handler.call(null, data); });
						}
					}

					// We subscribe to all relevant events and simply re-emit them,
					// except that we replace the 'result' event with a 'data' event.
					conn.query.apply(conn, args)
					   .on("fields", meta => interceptor.emit("fields", meta))
                       .on("result", row  => interceptor.emit("data",   row))
                       .on("error",  err  => interceptor.emit("error",  err))
                       .on("end",    ()   => interceptor.emit("end"));

					return interceptor;
				};
                this.begin = function (...args) {
					return new Promise((resolve, reject) => {
						// Add the callback that will resolve or reject.
						args.push(err => {
							if (err)
								reject(err);
							else
								resolve();
						});
						conn.beginTransaction.apply(conn, args);
					});
				}
            },
            commit: function (onDone, onError) {
                if (this.conn)
                    this.conn.commit(err => {
						if (err)
							this.release(() => onError && onError(err));
						else
							this.release(() => onDone && onDone());
					});
                else
                    onDone && onDone();
            },
            rollback: function (onDone, onError) {
                if (this.conn)
                    this.conn.rollback(err => {
						if (err)
							this.release(() => onError && onError(err));
						else
							this.release(() => onDone && onDone());
					});
                else
                    onDone && onDone();
            },
            release: function (onReleased, noCleanup) {
                const self = this;
                const finish = function () {
                    if (!noCleanup) self.cleanup(self);
                    if (onReleased) setTimeout(onReleased, 1);
                    onReleased = null;
                    self.conn = null;
                };

                if (this.conn) {
                    this.conn.release();
                }
                finish(); // Yes, always.
            }
        };
    }

    acquire(connection, onAcquired) {
        connection.state = "Connecting";

        this.pool.getConnection((err, conn) => {
            if (err) {
                this.logger.error("Could not obtain a database connection for " + connection.name + ": " + err);
                connection.cleanup(connection);
                connection.reject(err);
            } else {
                connection.initialize(conn);
                connection.resolve(connection);
                if (onAcquired) onAcquired();
            }
        });
    }

    get activeConnections() {
        return this.pool._allConnections.length - this.pool._freeConnections.length;
    }

    get totalConnections() {
        return this.pool.config.connectionLimit;
    }

    close(resolve) {
        this.pool.end(err => {
            resolve && resolve();
        });
    }
}

module.exports = MySQLAdaptor;