/* sqlite Adaptor */
'use strict';
const sqlite = require('better-sqlite3');
const { Pool } = require("better-sqlite-pool");

 /**
  * A sqlite3-specific implementation of the DatabaseAPI.
  * @extends DatabaseAPI
  * @private
  */
class SQLiteAdaptor {
    constructor(config, logger, tableName) {
        logger.info("    Initializing SQLite adaptor");
        this.config = config;
        this.logger = logger;
        this.table  = tableName;              
        this.pools  = {};   
        logger.info("    SQLite adaptor initialized");
    }

    isRetryable(error) {
        /*  There are errors returned by data operations where a retry may make sense.
            These are the errors we are looking for:
            NOTE: THIS ULTIMATELY WILL NEED TO GO TO THE ADAPTORS.

            errno       code                                message
            -------------------------------------------------------------------------------------------------------------------------
            0000        SQLITE_BUSY                    
            0000	    SQLITE_LOCKED                    
            0000        SQLITE_BUSY_RECOVERY                
            0000        SQLITE_ERROR_RETRY                  
            0000        SQLITE_INTERRUPT                    
         */
        const retriables = ['SQLITE_BUSY', 'SQLITE_LOCKED', 'SQLITE_BUSY_RECOVERY', 'SQLITE_ERROR_RETRY', 'SQLITE_INTERRUPT'];
        return retriables.includes(error.code);
    }

    getPool(tenant) {
        if (!this.pools[tenant]) {
            const dbName = "../data/" + tenant + ".sqlite";

            // Create database and set Write-Ahead Log for more performance. See https://sqlite.org/wal.html.
            const db = new sqlite(dbName);
            db.pragma('journal_mode = WAL');
            db.close();    
            
            // Create connection pool for it.
            this.pools[tenant] = new Pool(dbName, { max: this.config.connectionLimit, timeout: this.config.connectTimeout });
        }

        return this.pools[tenant];
    }

    createConnection(tenant, taskName, resolve, reject, cleanup) {
        const self = this;

        return {
            name:    taskName,
            tenant:  tenant,
            resolve: resolve,
            reject:  reject,
            cleanup: cleanup,
            state:   "Queued",
            initialize: function (conn) {
                this.state = "Connected";
                this.conn  = conn;
            },
            begin: function () {
                return new Promise ((resolve, reject) => {
                    try {
                        let result = this.conn.prepare("BEGIN TRANSACTION").run();
                        resolve(result);
                    }
                    catch (err){
                        reject(err);
                    }
                })
            },
            query: function (statement, values) {
                statement = statement.replace(self.table, "Entities");
                
                return new Promise ((resolve, reject) => {
                    let result;
                    try {
                        if (statement.startsWith("SELECT ")) {
                            result = this.conn.prepare(statement).all();
                        }
                        else if (values === undefined) {
                            result = this.conn.prepare(statement).run();
                            result.affectedRows = result.changes;
                        }
                        else {
                            result = this.conn.prepare(statement).run(values);
                            result.affectedRows = result.changes;
                        }
                        resolve(result);
                    }
                    catch (err) {
                        console.log(err);
                        reject(err);
                    }
                })
            },
            stream: function (statement) {
                // Create an interceptor through which we can simulate the events.
                const interceptor = {
                    events: {},
                    on: function (eventName, handler) {
                        // Add the subscription (event handler) for this event.
                        this.events[eventName] = this.events[eventName] || [];
                        this.events[eventName].push(handler);
                        return this;	// For chaining.
                    },
                    emit: function (eventName, data) {
                        // Call the subscribed event handlers.
                        const event = this.events[eventName] || [];
                        event.forEach(handler => { handler.call(null, data); });
                    }
                }

                this.query(statement)
                .then(rows => {
                    // We don't have metadata, sadly.
                    // interceptor.emit("fields", meta);
                    // TODO: Doc says its possibleuse an 'iterate'.
                    for (let i=0; i < rows.length; i++) {
                        interceptor.emit("data", rows[i]);
                    }
                    interceptor.emit("end");
                })
                .catch (err => {
                    interceptor.emit("error", err);
                });

                return interceptor;                
            },
            commit: function (onDone, onError) {
                if (this.conn) {
                    try {
                        this.conn.prepare("COMMIT").run();
                        this.release(() => onDone && onDone());
                    }
                    catch (err){
                        console.log(err)
                        this.release(() => onError && onError(err));
                    }
                }
                else {
                    onDone && onDone();
                }
            },
            rollback: function (onDone, onError) {
                if (this.conn)
                    this.conn.prepare("ROLLBACK").run()
                    .then(() => {
                        this.release(() => onDone && onDone());
                    })
                    .catch(error => {
                        this.release(() => onError && onError(error));
                    });
                else
                    onDone && onDone();
            },
            release: function (onReleased, noCleanup) {
                const self = this;
                const finish = function () {
                    if (!noCleanup) {
                        self.cleanup(self);
                    }
                    if (onReleased) {
                        setTimeout(onReleased, 1);
                    }
                    onReleased = null;
                    self.conn = null;
                };
                
                if (this.conn) {
                    this.conn.release();
                }

                finish();   // Yes, always.
            }
        };
    }

    acquire(connection, onAcquired) {
        connection.state = "Connecting";

        this.getPool(connection.tenant)
        .acquire()
        .then(conn => {
            connection.initialize(conn);
            connection.resolve(connection);
            if (onAcquired) onAcquired();
        })
        .catch(err => {
            this.logger.error("Could not obtain a database connection for " + connection.name + ": " + err);
            connection.cleanup(connection);
            connection.reject(err);
        });
    }

    get activeConnections() {
        return this.pool && this.pool.connections ? this.pool.connections.length : 0;
    }

    get totalConnections() {
        return this.pool ? this.pool.max : 0;
    }

    close(resolve) {
        try {
            if (this.pool)
                this.pool.close();
        }
        catch (err) {
            this.logger.error("Could not close connection:" + err);
        }
        resolve && resolve();
    }
}

module.exports = SQLiteAdaptor;