const { BlobServiceClient } = require("@azure/storage-blob");
const { DefaultAzureCredential } = require("@azure/identity");

// Load the .env file if it exists
require("dotenv").config();

const addPrefixToProps = (obj, prefix) =>
    Object.keys(obj).reduce(
        (acc, key) => ({
            ...acc,
            ...{ [prefix + key]: obj[key] }
        }),
        {}
    );

const AzureBlobUploader = function (config, tenants) {
    const self = this;

    this.events = {};

    var emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const event = self.events[eventName] || [];
        event.forEach(fn => { fn.call(null, data); });
    };

    var subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);
    };

    setTimeout(emit, 1, "ready");

    const instance = {
        on: subscribe,
        produce: config.produce,
        formatter: config.formatter,
        open: function (job, fieldGetter, onConnectionOpen) {
            const name = job.name;
            const cycle = job.cycle;
            const tenant = job.source;
            const tenantConfig = tenants[tenant];

            setTimeout(() => {
                onConnectionOpen(null, {
                    send: async function (messages, onComplete) {

                        if (instance.produce === "linkages") {
                            // Merge the source and target into one object, whith prefixed property names.
                            messages = messages.map(message => ({
                                ...addPrefixToProps(message.source, "source"),
                                ...addPrefixToProps(message.target, "target")
                            }));
                        }

                        try {
                            // Add metadata to JSON header
                            messages.jobName = job.name; 
                            messages.endTime = job.endTime;
                            messages.environment = tenantConfig.environment;

                            // Job may have been split. The name would look like: My_Job_Name.0.1, etc.
                            let subJobName = "";
                            const subJobSepIdx = job.name.indexOf(".");
                            if (subJobSepIdx > 0) {
                                messages.jobName = job.name.substring(0, subJobSepIdx);
                                subJobName = "_" + job.name.substring(subJobSepIdx + 1).replaceAll(".", "-");
                            }                            

                            // Connect to Azure storage account that hosts the blob container.
                            // Storage account name and shared key comes from env. variable.
                            const account = process.env.AZURE_STORAGE_ACCOUNT_NAME || "";

                            const blobServiceClient = new BlobServiceClient(
                                // When using AnonymousCredential, following url should include a valid SAS or support public access.
                                `https://${account}.blob.core.windows.net`,
                                new DefaultAzureCredential()
                            );

                            const blobName        = tenantConfig.environment + "_" + job.endTime + subJobName + ".json";
                            const containerName   = tenantConfig.storagefolder + "/" + config.storagefolder;
                            const containerClient = blobServiceClient.getContainerClient(containerName);
                            const blockBlobClient = containerClient.getBlockBlobClient(blobName);

                            // Upload data to Azure blob container as file.
                            const hostContent = JSON.stringify(messages);
                            blockBlobClient
                                .upload(hostContent, Buffer.byteLength(hostContent))
                                .then(data => onComplete(null, data))
                                .catch(onComplete);
                        }
                        catch (ex) {
                            console.log("Blob Upload Failed: " + ex.message);
                            setTimeout(onComplete, 10, ex.message);
                        }
                    },
                    close: function () {
                        return true;
                    }
                });
            }, 1);
        }
    };
    return instance;
};

module.exports = AzureBlobUploader;
