const fs = require('fs');

const CSVRendererAPI = function (config) {
    const self = this;

    config = config || {};
    this.config = {
        // If not specified or empty: use ../data
        // If specified and relative: under ../data
        // If specified and absolute: use as-as
        name:      config.address && config.address.charAt(0) != '/' ? "../data/" + config.address : (config.address || "../data/$jobName"),
        produce:   config.produce === "graphs" ?  "linkages" : config.produce,
        formatter: config.formatter,
        delim:     config.delim || ","
    };

    this.events = {};

    var emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const event = self.events[eventName] || [];
        event.forEach(fn => { fn.call(null, data); });
    };

    var subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);
    };

    setTimeout(emit, 1, "ready");

    const instance = {
        on: subscribe,
        produce:   self.config.produce,
        formatter: self.config.formatter,
        open: function (job, fieldGetter, onConnectionOpen) {
            const name = job.name;
            const cycle = job.cycle;
            const CSV = {
                headers: null,
                columns: null,
                getHeaders: function (prefixSet) {
                    if (this.headers) return null;  // Return headers only once.

                    const fields = fieldGetter(prefixSet);
                    this.headers = fields.headers;
                    this.columns = fields.columns;

                    return this.headers.join(self.config.delim);
                },
                stringify: function (message) {
                    if (!message)
                        return "";

                    const columns = [];
                    for (let i=0; i < this.columns.length; i++) {
                        columns.push(message[this.columns[i]]);
                    }
                    return columns.join(self.config.delim);
                },
                clearHeaders: function () {
                    this.headers = null;
                }
            };

            let timeStamp = (new Date(cycle)).toUTCString()
                                             .replace(/ /g, "_")
                                             .replace(/,/g, "")
                                             .replace(/\//g, "_")
                                             .replace(/:/g, "_");
            let epochTime = cycle;
            let fileName  = self.config.name
                .replace("$timeStamp", timeStamp)
                .replace("$epochTime", epochTime)
                .replace("$jobName",   name);

            // Create the directory if it does not exist.
            let fileParts = fileName.split("/");
            fileParts.pop();  // Take the filename part off.
            let filePath = fileParts.join("/");
            try { fs.statSync(filePath); }  // Does it exist?
            catch (ex) { fs.mkdirSync(filePath, { recursive: true }); }

            // Now we can open/create the file (hopefully).
            let file = fs.createWriteStream(fileName);
            if (!file) {
                setTimeout(onConnectionOpen, 1, "File " + fileName + " could not be opened");
                return;
            }
            file.on('error', err => {
                onConnectionOpen(err);
            });
            file.on('ready', () => {
                onConnectionOpen(null, {
                    send: function (messages, onComplete) {
                        if (!file)
                            return setTimeout(onComplete, 1, "File is not open");

                        // If 'produce' = "linkages", we will get an array to source/target nodes
                        if (instance.produce === "linkages") {
                            let headers = CSV.getHeaders(["source", "target"]);
                            if (headers) {
                                file.write(headers + "\n");
                            }

                            for (let i=0; i < messages.length; i++) {
                                file.write(
                                    CSV.stringify(messages[i].source) +
                                    self.config.delim +
                                    CSV.stringify(messages[i].target) +
                                    "\n"
                                );
                            };
                        }
                        else {
                            let headers = CSV.getHeaders();
                            if (headers) {
                                file.write(headers + "\n");
                            }

                            for (let i=0; i < messages.length; i++) {
                                file.write(CSV.stringify(messages[i]) + "\n");
                            };
                        }

                        setTimeout(onComplete, 1);
                    },
                    close: function () {
                        if (file) {
                            file.end();
                            CSV.clearHeaders();
                        }
                        file = null;
                        return true;
                    }
                });
            });
        }
    };
    return instance;
};

module.exports = CSVRendererAPI;
