'use strict';
/**
 * This is the interface that destination adaptors must implement.
 * @interface
 */
class DestinationAPI {
    /**
     * Create an instance of the DestinationAPI. During the contruction the instance
     * may prepare to connect to its output (if appropriate). The instance has to
     * emit a 'ready' event (asynchronously) once it has fully initialized and is ready
     * to accept calls to {@link DestinationAPI#open}.
     * @param {object} config - The configuration object as extracted for this adaptor
     * from config.json. Note that the config object will also include a 'formatter'
     * property. This property needs to be exposed at a public getter of this instance.
     */
    constructor(config) {}

    /**
     * Subscribe to an event.
     * @param {string} event - Valid values are 'error' and 'ready'.
     * @param {function} listener - The listener to this event.
     */
    on(event, listener) {}

    /**
     * Holds the strategy used by this adaptor when a job contains related entities.
     * Some implementations of this class are able to handle graphs (nodes and edges),
     * while others can only deal with arrays of tuples containing 'source' and 'target'
     * properties, and yet others only support arrays of messages.
     * The preferred value comes from the 'config' object passed to the constructor,
     * but it is up to the adaptor to set this property to the best supported value.
     * Valid values are "entities" (no relationships accepted), "linkages" (tuples only)
     * and "graphs" (all nodes and edges).
     */
    get produce() {}

    /**
     * The formatter as provided in the config object used to construct this instance.
     */
    get formatter() {}

    /**
     * Opens a connection to whatever destination the subclass supports. Upon completion
     * this method must call 'onConnectionOpen(error, destConn)' (asynchronously) with a
     * {@link DestinationConnection} object at which messages can be sent.
     * @param {object} job - The job. Notable properties in the job are 'name', 'cycle' and 'tenant'.
     * @param {function} fieldGetter - A function that will return an object with a 'headers'
     * property and a 'columns' property. This is useful for tabular outputs where headers need
     * to be written first, or for any output where the names of the available fields need to be
     * known in advance. The headers constitute all headers that need to be written to the output
     * in scenarios where entity relationships are involved. The columns represent just the names
     * of the fields found/created for entities while generating messages.
     * generating the messages.
     * @param {function} onConnectionOpen - The function that needs to be called upon completion.
     * It must pass a {@link DestinationConnection} object at which messages can be sent.
     * @async
     */
    open(job, fieldGetter, onConnectionOpen) {}
}

 /**
  * This is the interface that destination connectors must implement.
  * @interface
  */
class DestinationConnection {
    /**
     * Send an array of messages (or graphs, or tuples). Once done, the implementation must
     * (asynchronously) call 'onComplete()'.
     * @param {array} messages - The messages to be sent.
     * @param {function} onComplete - The function to be called upon completion.
     * @async
     */
    send(messages, onComplete) {}

    /**
     * Close the connection to the destination.
     */
    close() {}
}
