const fs = require('fs');

const addPrefixToProps = (obj, prefix) =>
    Object.keys(obj).reduce(
        (acc, key) => ({
            ...acc,
            ...{ [prefix + key]: obj[key] }
        }),
        {}
    );

const JSONRendererAPI = function (config) {
    const self = this;

    config = config || {};
    this.config = {
        // If not specified or empty: use ../data
        // If specified and relative: under ../data
        // If specified and absolute: use as-as
        name:      config.address && config.address.charAt(0) != '/' ? "../data/" + config.address : (config.address || "../data/$jobName"),
        produce:   config.produce,
        formatter: config.formatter
    };

    this.events = {};

    var emit = function (eventName, data) {
        // Call the subscribed event handlers.
        const event = self.events[eventName] || [];
        event.forEach(fn => { fn.call(null, data); });
    };

    var subscribe = function (eventName, fn) {
        // Add the subscription (event handler) for this event.
        self.events[eventName] = self.events[eventName] || [];
        self.events[eventName].push(fn);
    };

    setTimeout(emit, 1, "ready");

    const instance = {
        on: subscribe,
        produce:   self.config.produce,
        formatter: self.config.formatter,
        open: function (job, fieldGetter, onConnectionOpen) {
            const name = job.name;
            const cycle = job.cycle;

            let timeStamp = (new Date(cycle)).toUTCString()
                                             .replace(/ /g, "_")
                                             .replace(/,/g, "")
                                             .replace(/\//g, "_")
                                             .replace(/:/g, "_");
            let epochTime = cycle;
            let fileName  = self.config.name
                .replace("$timeStamp", timeStamp)
                .replace("$epochTime", epochTime)
                .replace("$jobName",   name);

            // Create the directory if it does not exist.
            let fileParts = fileName.split("/");
            fileParts.pop();  // Take the filename part off.
            let filePath = fileParts.join("/");
            try { fs.statSync(filePath); }  // Does it exist?
            catch (ex) { fs.mkdirSync(filePath, { recursive: true }); }

            // Now we can open/create the file (hopefully).
            let file = fs.createWriteStream(fileName);
            if (!file) {
                setTimeout(onConnectionOpen, 1, "File " + fileName + " could not be opened");
                return;
            }
            file.on('error', err => {
                onConnectionOpen(err);
            });
            file.on('ready', () => {
                onConnectionOpen(null, {
                    send: function (messages, onComplete) {
                        if (!file)
                            return setTimeout(onComplete, 1, "File is not open");

                        if (instance.produce === "linkages") {
                            // Merge the source and target into one object, whith prefixed property names.
                            messages = messages.map(message => ({
                                ...addPrefixToProps(message.source, "source"),
                                ...addPrefixToProps(message.target, "target")
                            }));
                        }

                        // Write the whole array at once, including '[', ',' and ']' characters.
                        file.write(JSON.stringify(messages, null, 4));

                        setTimeout(onComplete, 1);
                    },
                    close:  function () {
                        if (file) {
                            file.end();
                        }
                        file = null;
                        return true;
                    }
                });
            });
        }
    };
    return instance;
};

module.exports = JSONRendererAPI;
