// To log, we need to configure the logger provider before we 'require' "kafka-node".
function consoleLoggerProvider (name) {
    return {
      debug: console.debug.bind(console),
      info:  console.info.bind(console),
      warn:  console.warn.bind(console),
      error: console.error.bind(console)
    };
}
// If we want logging:
// const kafkaLogging = require('kafka-node/logging');
// kafkaLogging.setLoggerProvider(consoleLoggerProvider);

const kafka = require('kafka-node');

const KafkaBrokerAPI = function (config) {
    config = config || {};

    this.config = {
        hosts:     Array.isArray(config.host) ? config.host : config.host ? [config.host] : [],
        port:      config.port    || 9092,
        connect:   config.connect || {},
        topic:     config.address || config.topic || "No topic specified",
        produce:   "entities",
        formatter: config.formatter
    };

    /*  THIS ERROR MESSAGE HAVE BEEN OBSERVED IN KAFKA CONSUMERS AT OCCASION

        Commit cannot be completed since the group has already rebalanced and assigned the
        partitions to another member. This means that the time between subsequent calls to
        poll() was longer than the configured max.poll.interval.ms, which typically implies
        that the poll loop is spending too much time message processing. You can address
        this either by increasing the session timeout or by reducing the maximum size of
        batches returned in poll() with max.poll.records.
        (org.apache.kafka.clients.consumer.internals.ConsumerCoordinator)

        We support both (see options below). The Controller takes care of constraining the
        maximum batch size already for its own reasons, but we handle it here as well, by
        splitting message arrays that are rejected as too big, until all data is accepted.

        Conect Options:
        - kafkaHost: (string) A comma-separated list of kafka broker/host combinations. Example:
            "kafka-1.us-east.myapp.com:9093,kafka.us-east-1.myapp.com:9093,kafka-3.us-east.myapp.com:9093"
            Default: "localhost:9092". Note that we also support "host" and "port" in config.json.
        - connectTimeout: (ms) Time it takes to wait for a successful connection before moving to
            the next host. Default: 10000.
        - requestTimeout: (ms) Time for a kafka request to timeout. Default: 30000.
        - connectRetryOptions: (object) Options that apply to the initial connection. Sub-options:
            - retries: The maximum amount of times to retry the operation. Default: 10.
            - factor: The exponential factor to use. Default: 2.
            - minTimeout: The number of milliseconds before starting the first retry. Default: 1000.
            - maxTimeout: The maximum number of milliseconds between two retries. Default: Infinity.
            - randomize: Randomizes the timeouts by multiplying with a factor between 1 to 2. Default: false.
        - idleConnection: (ms) Allows the broker to disconnect an idle connection from a client
            (otherwise the clients continues to O after being disconnected). The value is elapsed
            time in ms without any data written to the TCP socket. Default: 5 minutes.
        - maxAsyncRequests: (count) Maximum async operations at a time toward the Kafka cluster. Default: 10.
        - sslOptions: (object) Options to be passed to the TLS broker sockets.
		    Example: { rejectUnauthorized: false }
        - sasl: (object) SASL authentication configuration (only SASL/PLAIN is currently supported).
		    Example: { mechanism: 'plain', username: 'foo', password: 'bar' }
     */
    const client = new kafka.KafkaClient(Object.assign({
        kafkaHost: this.config.hosts.map(host => host + ":" + this.config.port).join(",")
    },  this.config.connect));

    const producer = new kafka.HighLevelProducer(client);
    const self = this;

    return {
        on: function (event, handler) {
            /*  The following events are relevant:
                - ready: This event is emitted when producer is ready to send messages.
                - error: This is the error event propagated from the internal client that
                         the producer should always handle.
             */
            producer.on(event, handler);
        },
        produce:   self.config.produce,
        formatter: self.config.formatter,
        open:  function (job, fieldGetter, onConnectionOpen) {
            const name = job.name;
            const cycle = job.cycle;
            const tenant = job.source;

            setTimeout(onConnectionOpen, 1, null, {
                send: function (messages, onComplete) {
                    const me = this;

                    producer.send([{
                        topic: self.config.topic,
                        messages: messages.map(message => JSON.stringify(message))
                    }], error => {
                        if (error && error.message == "MessageSizeTooLarge") {
                            // If the producer API cannot even accept less than 100, we give up.
                            if (messages.length < 100)
                                return onComplete("KafkaBroker says: 'MessageSizeTooLarge'. Please make your messages smaller");

                            const firstHalf  = messages.splice(0, Math.ceil(messages.length / 2));
                            const secondHalf = messages;
                            me.sendAll(firstHalf, firstError => {
                                me.sendAll(secondHalf, secondError => {
                                    onComplete(firstError || secondError);
                                });
                            });
                        }
                        else {
                            onComplete(error ? error.message : null);
                        }
                    });
                },
                close:  function () {
                    return true;
                }
            });
        }
    };
};

module.exports = KafkaBrokerAPI;
