import { SharedSessionStore } from "@dt-esa/level-cluster";
import { AxiosInstance } from "axios";
import { ConfigParams } from "express-openid-connect";
import { ScopeMap } from "./dynatrace-scopes";
export interface PermissionMap {
    [key: string]: Array<string>;
}
export declare type AuthenticationOptions = {
    authorizations?: PermissionMap;
    scopeMappings?: ScopeMap;
    saml?: any;
    openid?: ConfigParams;
    dynatrace?: {
        endpoint: string;
        customAxios?: AxiosInstance;
    };
    client?: {
        port: number;
        timeout: number;
    };
};
/**
 * Connect-style middleware that sets up SSO authorization via passport.
 *
 *
 * @param options
 * ```ts
 * type AuthorizatonHandlerOptions = {
 * 	   mode:
 * 			 "client"    | // Use a seperate webserver running this middleware to authenticate & authorize transactions.
 * 						   // If `authorizations` is specified in both the client and server, the resulting request will have the permissions from both.
 * 			 "dynatrace" | // Use a Dynatrace instance to authenticate transactions and provide authorization.
 * 			 "azure"     | // Use Azure App Registration for Authentication. Authorization specified in `authorizations`.
 * 			 "saml",       // Use a generic SAML configuration for Authentication. Authorization specified in `authorizations`.
 * 	   authorizations?: Map<string, Array<string>>, // For `azure` and `saml` modes. An object containing Authorizations to apply.
 * 			// e.g.
 * 			// {
 * 			//     "grace.hopper@example.com": ["ReadConfig", "WriteConfig", "logs.read"],
 * 			//     "ryan.dahl@example.com":    ["WriteConfig", "logs.read"],
 * 			//     "brendan.eich@example.com": ["ReadConfig", "logs.read"],
 * 			// ...
 * 			// }
 *     saml?: Object,      // SAML configuration object provided to "passport-saml"
 *     azure?: Object,     // Azure configuration object provided to "passport-azure-ad"
 *     clientConnectionPort?: number, // When mode is `client`, the port that the seperate webserver authorizing transactions is running on.
 *     dynatraceEndpoint?:    string  // The URL that the Authorizer will check against for Dynatrace Authorization.
 *     customAxios?:   	      AxiosStatic // An optional axios client.
 *	   scopeMappings?: Map<string, Array<string>>, // A mapping of scope ID from the Dynatrace API to your router. This is only used in "dynatrace" mode.
 *
 * }
 * ```
 *
 *
 * This will export an express router with several route handlers:
 *  - `login`         (redirect to login provider)
 *  - `code`          (recieve the SAML/OpenID/Oauth response)
 *  - `logout`        (log user out & redirect back to root)
 *  - `authorization` (endpoint to check authorization)
 *
 * All subsequent requests in the chain will be decorated with the following properties:
 *  - `_username`:         string
 *  - `_scopeMapping`:     Map<string, Array<string>>
 *  - `_authorizedScopes`: Array<string>
 */
export declare const authentication: (options: any) => any;
/**
 * Connect-style middleware that asserts all permissions are satisfied by the
 * authenticated user. If a permission is missing, it will reject the transaction.
 * @param permissions A list of required Permissions or Scopes an authorized part must have.
 *
 * - Invoking this method with an empty array or no permissions specified will simply ensure
 * that the user is authenticated.
 */
export declare const authorize: (permissions?: any[]) => (req: any, res: any, next: any) => any;
/**
 * Method to initialize the shared user cache.
 * This is present as an example.
 *
 * Alternatively you can create a SharedSessionStore on the target port
 * with the instanceId "@dynatrace-esa/authorizer" manually.
 *
 * @param port port on which the shared memory store communicates.
 * @param options
 */
export declare const SharedUserCache: (port: any, options: any) => SharedSessionStore<unknown>;
