"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CacheClusterWorker = exports.CacheClusterPrimaryPM2 = exports.CacheClusterPrimary = void 0;
// @ts-nocheck
const cluster_1 = __importDefault(require("cluster"));
const channel = '@dt-esa/authorizer';
class CacheClusterPrimary {
    // Handle messages is false for running in pm2 mode.
    // In that case, the extending class handles messages.
    constructor(cacheId, handleMessages = true) {
        this.cacheId = cacheId;
        this.cache = [];
        console.log("Starting Cache Cluster Primary.");
        // Do not run when PM2 is used.
        if (handleMessages) {
            cluster_1.default.setMaxListeners(0);
            cluster_1.default.on('message', (worker, msg) => {
                // We handle the first message from a worker as an initialization point.
                if (msg && msg.channel == channel && msg.type == 'init') {
                    console.log(`Worker ${worker.pid} established connection.`);
                    worker.send({
                        channel,
                        type: 'init',
                        cacheId: msg.cacheId,
                    });
                }
                else {
                    this.processMsg(worker, msg);
                }
            });
        }
    }
    processMsg(worker, msg) {
        if (!msg || msg.channel !== channel || !msg.data) {
            return false;
        }
        try {
            const key = msg.data.key;
            const value = msg.data.value;
            switch (msg.func) {
                case 'fetch':
                    let result = this.cache[key];
                    return this.sendToWorker(worker, msg, 'resolve', result);
                case 'store':
                    this.cache[key] = value;
                    return this.sendToWorker(worker, msg, 'resolve', true);
                case 'delete':
                    let res = delete this.cache[key];
                    return this.sendToWorker(worker, msg, 'resolve', res);
                default:
                    return this.sendToWorker(worker, msg, 'reject', new Error("Function not specified."));
            }
        }
        catch (ex) {
            return this.sendToWorker(worker, msg, 'reject', ex);
        }
    }
    sendToWorker(worker, msg, type, data) {
        worker.send({
            channel,
            cacheId: msg.cacheId,
            promiseId: msg.promiseId,
            type,
            data,
        });
    }
}
exports.CacheClusterPrimary = CacheClusterPrimary;
class CacheClusterPrimaryPM2 extends CacheClusterPrimary {
    constructor(pm2, cacheId) {
        super(cacheId, false);
        pm2.launchBus((err, pm2Bus) => {
            console.log("pm2 bus launched");
            pm2Bus.on('process:msg', (packet) => {
                const msg = packet.raw;
                if (msg && msg.channel == channel && msg.type == 'init') {
                    console.log(`Worker ${packet.process.pm_id} established connection.`);
                    pm2.sendDataToProcessId(packet.process.pm_id, {
                        data: {},
                        topic: channel,
                        channel,
                        type: 'init',
                        cacheId: msg.cacheId,
                    }, (sendErr, res) => {
                        if (sendErr) {
                            console.log(sendErr, res);
                        }
                    });
                }
                else {
                    const worker = {
                        send: (msgData) => {
                            const pm2Message = msgData;
                            pm2Message.topic = channel;
                            if (typeof pm2Message.data === 'undefined') {
                                pm2Message.data = {};
                            }
                            pm2.sendDataToProcessId(packet.process.pm_id, pm2Message, (sendErr, res) => {
                                if (sendErr) {
                                    console.log(sendErr, res);
                                }
                            });
                        },
                    };
                    this.processMsg(worker, msg);
                }
            });
        });
    }
}
exports.CacheClusterPrimaryPM2 = CacheClusterPrimaryPM2;
class CacheClusterWorker {
    constructor(cacheId) {
        this.cacheId = cacheId;
        this.timeoutMs = 1000;
        // Map of active promises that we are waiting for a response on.
        this.promises = {};
        // Have we successfully connected to the primary process?
        this.initiated = false;
        /**
         *
         * @param key Key at which item is added into cache
         * @param value Value of item that gets added into cache
         * @returns
         */
        this.store = (key, value) => this.sendToPrimary("store", { key, value });
        /**
         *
         * @param key Key for item to retrieve from cache
         * @returns
         */
        this.fetch = (key) => this.sendToPrimary("fetch", { key });
        /**
         *
         * @param key Key for item to delete from cache
         * @returns
         */
        this.delete = (key) => this.sendToPrimary("delete", { key });
        this.promiseId = 0;
        console.log("Starting Cache Cluster Worker.");
        process.setMaxListeners(0);
        process.on('message', (msg) => {
            if (msg && msg.channel == channel && msg.type == 'init' && msg.cacheId == cacheId) {
                console.log("Worker has been connected to Primary.");
                this.initiated = true;
            }
            else {
                this.processMsg(msg);
            }
        });
        // Establish a connection to the primary process.
        process.send({
            channel,
            cacheId: this.cacheId,
            type: 'init'
        });
    }
    savePromise(resolve, reject) {
        const hrtime = process.hrtime();
        let promiseId = this.promiseId++ + hrtime[0].toString() + hrtime[1].toString();
        console.log("Saving promise", promiseId);
        if (typeof this.promises[promiseId] !== 'undefined') {
            console.log("duplicate promsie id");
            debugger;
            // promiseId += crypto.randomBytes(12).toString('base64');
        }
        this.promises[promiseId] = {
            resolve,
            reject,
            timeoutId: setTimeout(() => {
                delete this.promises[promiseId];
                console.log(this.promises);
                reject(new Error('IPC timeout: no answer from primary in time ' + promiseId));
            }, this.timeoutMs),
        };
        return promiseId;
    }
    sendToPrimary(func, data) {
        return new Promise((resolve, reject) => {
            const promiseId = this.savePromise(resolve, reject);
            // this.sendToPrimary('store', promiseId, { key, value });
            const payload = {
                channel,
                cacheId: this.cacheId,
                func,
                promiseId,
                data,
            };
            if (!this.initiated) {
                console.log("Worker not initiated");
                // Wait init before sending messages to master
                this.waitInit(payload);
            }
            else {
                console.log("Worker sending message");
                process.send(payload);
            }
        });
    }
    waitInit(payload) {
        setTimeout(() => {
            if (this.initiated) {
                process.send(payload);
                // Promise will be removed by timeout if too long
            }
            else if (typeof this.promises[payload.promiseId] !== 'undefined') {
                this.waitInit(payload);
            }
        }, 30);
    }
    processMsg(msg) {
        if (!msg || msg.channel !== channel || msg.cacheId !== this.cacheId) {
            return false;
        }
        if (this.promises[msg.promiseId]) {
            clearTimeout(this.promises[msg.promiseId].timeoutId);
            switch (msg.type) {
                case 'resolve':
                    console.log("received response from primary", msg);
                    this.promises[msg.promiseId].resolve(msg.data);
                    break;
                case 'reject':
                    this.promises[msg.promiseId].reject(msg.data);
                    break;
                default:
                    throw new Error(`Unknown message type '${msg.type}'`);
            }
            delete this.promises[msg.promiseId];
        }
    }
    ;
}
exports.CacheClusterWorker = CacheClusterWorker;
